REM	   ***********************************
REM		   Universal Plotter
REM	   ***********************************
REM
REM		Function FNP(X$,Y$,T$,P7,P8)
REM
REM
REM **	This function takes the following arguments:
REM
REM	   X$ - label for the X axis
REM	   Y$ - label for the Y axis
REM	   T$ - label for the graph
REM	   P7 - graphic mode:
REM		0 : normal rectangular plot mode
REM		1 : semi-log plot mode
REM		2 : log-log plot mode
REM		3 : polar plot mode
REM		   (array X is "r", array Y is "theta")
REM		   (in this mode, the axes are not labelled)
REM		4 : use previous axes, assume mode 0
REM		   (this can only be used on the video)
REM		5 : bar graph in rectangular plot mode
REM	   P8 - output device:
REM		0 : video screen
REM		>0: the printer.  In this case, P8 is equal
REM		    to the number of points to be plotted.
REM
REM  ********  The printer must be on file channel 2  *******
REM
REM	NOTE: In using logarithmic plotting modes, the user
REM	      should exercise care that the elements of the
REM	      arrays are strictly positive.
REM
REM
REM **	The function assumes the data to be in the arrays
REM	X and Y.  The values of these arrays will be
REM	"destroyed" by function FNP.
REM
REM
REM **	The following "global" variables are required for the
REM	exclusive use of this function.  They must not be used
REM	by the calling program or unrecoverable errors may
REM	occur.  All other variables used by the function are
REM	preserved.
REM		M8$, M9$, P0$, M9 (array and scalar)
REM
REM
REM **	A function FNM is used by FNP for printing on the
REM	video, and a function FNP1$ is used for producing
REM	the axes values.  The user should take care that the
REM	calling program does not have functions with these
REM	names.
REM
REM
REM
60000 DEF FNP(X$,Y$,T$,P7,P8)
REM
REM	If this is the first pass, dimension the arrays.
REM
60010 M9=M9+1 \ IF M9=1 THEN DIM M9$(1:64),M8$(80:1),M9(20)
REM
REM	Preserve the "global" variables.
REM
60020 M9(1)=I \ M9(2)=J \ M9(3)=P0 \ M9(4)=P1 \ M9(5)=P2
60030 M9(6)=P3 \ M9(7)=P4 \ M9(8)=P5 \ M9(9)=P6
60040 M9(10)=P9 \ M9(11)=X0 \ M9(12)=X \ M9(13)=Y0 \ M9(14)=Y
REM
REM	Check for non-rectangular plotting modes,
REM	and revalue the arrays if necessary.
REM
60050 IF P7<1 OR P7>3 THEN 60120
60060 ON P7 GOTO 60110,60100,60070
REM	(polar conversion)
60070 P0=0 \ MAT IF X=X THEN P0=P0+1
60080 FOR I=1 TO P0 \ P9=X(I)
60090 X(I)=P9*COS(Y(I)) \ Y(I)=P9*SIN(Y(I)) \ NEXT \ GOTO 60120
REM	(log-log conversion)
60100 MAT X=LOG(X)
REM	(semi-log conversion)
60110 MAT Y=LOG(Y)
REM
REM	Skip over this stuff if using mode 4 on video
REM
60120 IF P7=4 AND P8=0 THEN 60180
REM
REM	Otherwise, determine the maxima and minima..
REM
60130 P1=MIN(X) \ P5=MAX(X) \ P3=MIN(Y) \ P6=MAX(Y)
60140 P2=30/(P5-P1) \ P4=10/(P6-P3)
REM
REM	and the graph origin.
REM
60150 X0=(P1<0)*INT(-P1*P2) \ Y0=(P3<0)*INT(-P3*P4)
60160 IF P8=0 THEN Y0=Y0+4 ELSE Y0=Y0*3
60170 IF P8=0 THEN X0=X0+20 ELSE X0=16+2*X0
REM
REM	Scale the data arrays.
REM
60180 MAT Y=INT(3*(Y-P3)*P4) \ MAT X=16+2*(X-P1)*P2
60190 IF P8=0 THEN MAT Y=Y+10 \ MAT X=X+24
REM
REM	If using mode 4, branch to video plotter.
REM
60200 IF P7=4 AND P8=0 THEN 60460
REM
REM	Send a FORM FEED to the output device.
REM
60210 IF P8=0 THEN PRINT CHR$(12) ELSE PRINT:2,CHR$(12)
60220 IF P8=0 THEN PLOT 0,44,0
REM
REM	Print the graph title (properly centered).
REM
60230 IF P8=0 THEN P0=64 ELSE P0=80
60235 P0=(P0-LEN(T$))/2
60240 IF P8=0 THEN P0=FNM(P0,16,T$) \ GOTO 60270
60250 FOR I=1 TO P0 \ PRINT:2," ", \ NEXT \ PRINT:2,T$
60260 PRINT:2," " \ PRINT:2," "
REM
REM	Branch to printer mode if P8>0
REM
60270 IF P8<>0 THEN 60680
REM
REM	Draw the axes on video.
REM
60280 P0=7168-64*Y0 \ FOR X=20 TO 50 \ POKE X+P0,223 \ NEXT
60290 P0=2*X0+1 \ FOR Y=10 TO 40 \ PLOT P0,Y,1 \ NEXT
REM
REM	If in polar mode, go do the plot now.
REM
60300 IF P7=3 THEN 60460
REM
REM	Put in the hash lines.
REM
60310 FOR Y=4 TO 14 \ FOR X=20 TO 50 STEP 5
60320 IF X=X0 OR Y=Y0 THEN 60330 ELSE POKE 7168+X-64*Y,174
60330 NEXT \ NEXT
REM
REM	Determine the values for the X-axis.
REM
60340 P0=(P5-P1)/3
60350 FOR I=0 TO 3 \ P5=P1+I*P0 \ P0$=FNP1$(P5)
60360 P5=FNM(20+10*I-LEN(P0$)/2,3,P0$) \ NEXT
REM
REM	Determine the values for the Y-axis.
REM
60370 P0=.2*(P6-P3)
60380 FOR I=0 TO 5 \ P5=P3+I*P0 \ P0$=FNP1$(P5)
60390 P5=FNM(19-LEN(P0$),4+I+I,P0$) \ NEXT
REM
REM	Print the X axis label (properly centered).
REM
60400 P0=20+(30-LEN(X$))/2 \ P0=FNM(P0,2,X$)
REM
REM	Print the Y axis label (properly centered).
REM
60410 P5=12 \ P0=1 \ P9=LEN(Y$) \ FOR I=2 TO P9 \ IF I=P9 THEN 60430
60420 P0$=MID$(Y$,I,I) \ IF P0$<>" " THEN 60450
60430 P0$=MID$(Y$,P0,I) \ P6=INT(LEN(P0$)/2) \ P5=P5-1
60440 P0=FNM(5-P6,P5,P0$) \ P0=I+1
60450 NEXT
REM
REM	Plot the graph to the video.
REM
60460 MAT PLOT X,Y,1 \ IF P7<>5 THEN 60520 ELSE Y0=Y0*3-2
REM
REM	Complete the bar graph on the video.
REM
60470 IF MAX(Y)=Y0 THEN 60520
60480 MAT Y=Y*(Y<=Y0)+(Y-1)*(Y>Y0) \ MAT PLOT X,Y,1 \ GOTO 60470
60520 PLOT 0,3,0 \ GOTO 60830
REM
REM	Subroutine for putting Y-axis labels to printer.
REM
60530 P9=LEN(Y$) \ IF P9=0 THEN RETURN ELSE M9$=""
60540 P0$=LEFT$(Y$,1) \ M9$=M9$+P0$ \ Y$=RIGHT$(Y$,LEN(Y$)-1)
60550 IF LEN(Y$)=0 THEN 60560 ELSE IF P0$<>" " THEN 60540
60560 P2=(10-LEN(M9$))/2 \ IF P2<0 THEN P2=0
60570 FOR P4=1 TO LEN(M9$)
60580 M8$(P2+P4)=MID$(M9$,P4,P4) \ NEXT \ RETURN
REM
REM	Subroutine for putting Y-axis values to printer.
REM
60590 P2=P3+I*(P6-P3)/30 \ P0$=FNP1$(P2)
60600 P2=14-LEN(P0$)
60610 FOR P4=1 TO LEN(P0$) \ M8$(P4+P2)=MID$(P0$,P4,P4) \ NEXT
60620 RETURN
REM
REM	Subroutine for putting X-axis values to printer.
REM
60630 FOR I=0 TO 60 STEP 10 \ P2=P1+I*(P5-P1)/60
60640 P0$=FNP1$(P2) \ P2=15+I-LEN(P0$)/2
60650 FOR J=1 TO LEN(P0$)
60660 M8$(J+P2)=MID$(P0$,J,J) \ NEXT \ NEXT
60670 MAT PRINT:2,M8$, \ PRINT:2," " \ PRINT:2," " \ RETURN
REM
REM	This is the loop for printing a graph.
REM
REM
60680 FOR I=30 TO 0 STEP -1 \ MAT M8$=" "
REM
REM
REM	Possibly part of the Y-axis label is required.
REM
60690 IF I<17 THEN GOSUB 60530
REM
REM	Possibly a horizontal hash line and Y-axis value
REM	is required.
REM
60700 IF MOD(ABS(I-Y0),5)>0 THEN 60720
60710 FOR J=16 TO 76 STEP 2 \ M8$(J)="." \ NEXT \ GOSUB 60590
REM
REM	Possibly we need the X-axis drawn now.
REM
60720 IF I<>Y0 THEN 60740
60730 FOR J=16 TO 76 \ M8$(J)="-" \ NEXT \ GOTO 60780
REM
REM	Collect the vertical hash points.
REM
60740 FOR J=16 TO 76
60750 IF MOD(ABS(J-X0),10)=0 THEN M8$(J)="."
60760 NEXT
REM
REM	Collect the Y-axis point.
REM
60770 M8$(X0)="l"
REM
REM	Collect all the data on this
REM	line and then print it.
REM
60780 FOR J=1 TO P8 \ IF Y(J)=I THEN M8$(X(J))="*"
REM
REM	(And adjust if this is a bar graph.)
REM
60790 IF P7=5 AND I>Y0 AND Y(J)=I THEN Y(J)=Y(J)-1
60800 NEXT \ MAT PRINT:2,M8$, \ PRINT:2," " \ NEXT
REM
REM	Get the X-axis values.
REM
60810 PRINT:2," " \ MAT M8$=" " \ P0=48-LEN(X$)/2 \ GOSUB 60630
REM
REM	Print the X-axis label.
REM
60820 FOR I=1 TO P0 \ PRINT:2," ", \ NEXT \ PRINT:2,X$
REM
REM	Restore the "global" variables and SCRAM!
REM
60830 I=M9(1) \ J=M9(2) \ P0=M9(3) \ P1=M9(4) \ P2=M9(5)
60840 P3=M9(6) \ P4=M9(7) \ P5=M9(8) \ P6=M9(9) \ P9=M9(10)
60850 X0=M9(11) \ X=M9(12) \ Y0=M9(13) \ Y=M9(14) \ RETURN 0
REM
60860 FNEND
REM
REM	Function FNP1$
REM
60870 DEF FNP1$(X)
60880 P9=ABS(X) \ IF P9=>1E3 OR P9<1E-3 THEN RETURN STR$(X,%#8E1)
60890 P0$=STR$(X)
60900 IF LEN(P0$)<6 THEN RETURN P0$ ELSE RETURN MID$(P0$,2,6)
60910 FNEND
REM
REM	Function FNM
REM
60920 DEF FNM(X,Y,M9$)
60930 Y=3*Y-3 \ X=2*X-2 \ PLOT X,Y,0 \ X=X+2*LEN(M9$)
60940 PRINT M9$, \ PLOT X,Y,0 \ RETURN X/2+2
60950 FNEND
REM
