	PAGE
	IF	MACROS
;********************************************************
;*		MACRO STACK FUNCTIONS			*
;*							*
;*	The macro evaluation stack is a 'String-List-	*
;*  Stack'.  It is described in a document with similar	*
;*  name.  These are a subset of the primitives for	*
;*  the SLS:  the ones found useful in this system.	*
;*  The SLS stores the activation records for each	*
;*  macro reference or 'call' as well as the local inf-	*
;*  ormation to each call, such as the parameters that	*
;*  were passed in on the invocation line.		*
;*							*
;********************************************************
;
;MACINIT() initializes the macro stack by nulling out the
;	appropriate lists.  It is not a primitive of the
;	SLS.  It is called once on each pass.
;
;MACPUSH( LISTP,MEMP,MEMLEN ) returns( LISTP), error code
;	if MSTACK overflows.  The text at MEMP with
;	length MEMLEN is copied into the MSTACK at MSTKFREE.
;	The length of the text is saved in the new entry,
;	and the new entry is inserted into the list structure
;	whose head pointer is LISTP.  On return, the new list
;	with the new entry at the top is passed back in LISTP.
;
;MACPEEK( LISTP,MEMP,MEMLEN ) returns( LISTP,ACTUALLEN ). If
;	MEMLEN is shorter than the element in the stack
;	pointed to by LISTP, then ACTUALLEN is equal to MEMLEN.
;	Otherwise, ACTUALLEN is the length of the stack
;	element.  LISTP on return is equal to MACPOP( LISTP ).
;
;MACPRUNE( LISTP ).  Points MSTKFREE at LISTP, thus 'pruning'
;	the pointed to element and anything pushed after it
;	off of the SLS.  The element that is pruned off is
;	conventionally a node element that shortens the other
;	lists that are affected by the pruning so that they
;	no longer have elements beyond the pruned node.  The
;	node is usually MACPEEKed into memory over the area
;	where the other list head pointers reside.
;
;MACPOINT( LISTP1,LISTP2 ).  The stack element at LISTP1 is
;	modified to have predecessor at LISTP2.
;
MACINIT PUSH	H
	LXI	H,MSTKBASE+3	;Clear the MSTACK.
	SHLD	MSTKFREE
	LXI	H,0
	SHLD	EXPANDCOUNT	;At the zeroth macro.
	SHLD	EXPANDP		;No macros stacked up.
	SHLD	MACKEY		;This declares no macros in
	POP	H		;progress.
	RET
;
;	LISTP is in HL, MEMP is in DE, and MEMLEN is in B.
; The error code is CARRY and DE=2001H.
;
MACPUSH	PUSH	B
	PUSH	D		;<MEMP<old BC<...
	XCHG
	LHLD	MSTKFREE	;HL=FREE pointer in stack.
	DCX	H		;Now construct pointer,len
	MOV	M,B		;part of a stack element
	DCX	H		;just before the FREE ptr.
	MOV	M,D
	DCX	H
	MOV	M,E
	PUSH	D		;<old LISTP<MEMP<old BC<
	MOV	C,B
	MVI	B,0
	LHLD	MSTKFREE
	DAD	B		;Is (FREE+MEMLEN<TOP)?
	LXI	D,MSTKTOP-3	;if so, then oflow.
	CALL	DCMP
	JC	MACPS1		;jmp/ No.
	POP	H		;<old LISTP<MEMP<old BC<...
	POP	D
	POP	B
	LXI	D,2001H		;** MSTACK overflow - Sorry **
	STC
	RET
MACPS1	XCHG			;Get back old FREE before
	LHLD	MSTKFREE	;moving it.
	XCHG
	INX	H		;Put new FREE at 3 bytes after
	INX	H		;last byte of this entry.
	INX	H
	SHLD	MSTKFREE
	POP	H		;<old LISTP<MEMP<old BC<...
	POP	H		;DE=FREE,HL=MEMP
	PUSH	H
	PUSH	D		;<FREE<MEMP<old BC<...
	XCHG			;DE=MEMP,HL=FREE=new LISTP.
	CALL	COPY		;Move BC bytes from *DE to *HL.
	POP	H		;HL=old FREE = new LISTP.
	POP	D		;DE,BC unchanged.
	POP	B
	XRA	A
	RET
;
;	In this implementation, DE=MEMP,B=MEMLEN,HL=LISTP.
; On return, HL=MACPOP(HL) or in other words, HL is the
; pointer to the previous element as extracted from the
; element itself.  DE unchanged, but C,A squashed.  B is
; given length of actual entry on stack.  Buffer overflow
; gives CARRY, DE=2002H, other regs as defined.
;
MACPEEK PUSH	D		;<MEMP<...
	DCX	H
	MOV	A,M
	INX	H		;A=LISTLEN (element length).
	CMP	B		;Which is longer?
	JC	MACPEK1		;jmp/ MEMLEN>=LISTLEN.
	MOV	A,B		;Use shorter len.
MACPEK1 PUSH	PSW		;Save the shortest len.
	PUSH	H		;<LISTP<PSW<MEMP<...
	MOV	C,B
	MVI	B,0
	XCHG
	CALL	COPY		;Copy BC bytes from *DE to *HL.
	POP	H
	DCX	H
	DCX	H		;Get new LISTP from old.
	MOV	A,M
	DCX	H
	MOV	L,M
	MOV	H,A		;HL= new LISTP.
	POP	PSW
	MOV	B,A		;Actual length used is in B.
	POP	D		;DE=MEMP as on call.
	CMC			;If MEM area was long enough,
	RNC			;then ret NO CARRY. Else:
	LX	1
POUNDP	DB	'#'
	;
	; Label field requested for a parm.  Form: #L.
	;
LAMBD1	LHLD	LABP
	CALL	LAMBD5
	JMP	LAMBD6
	;
	; Opcode field requested for a parm.  Form: #O.
	;
LAMBD2	LHLD	OPCP
	CALL	LAMBD5
	JMP	LAMBD6
	;
	; All operand field parms requested as a c, DE=2003H.
;
MACPOP	MOV	A,H
	ORA	L
	JNZ	MACPP1
	LXI	D,2003H	  ;** Can't pop a null list - MACPOP **
	STC
	RET
MACPP1	DCX	H
	DCX	H
	MOV	A,M
	DCX	H
	MOV	L,M
	MOV	H,A
	XRA	A
	RET
;
;	DE= new LISTP, HL->element to be changed.
;
MACPOINT DCX	H
	DCX	H
	MOV	M,D
	DCX	H
	MOV	M,E
	INX	H
	INX	H
	INX	H
	RET
;
;	HL=LISTP.  All unchanged on ret.  MSTKFREE=HL
;  Mainly used for clarity: an SLS primitive.
;
MACPRUNE SHLD	MSTKFREE
	RET
	ENDIF			;End MACROS.
;
