	PAGE
; Arith.TX
; Last updated:
;	12/29/79	RTM	Disaster recovery and two user
;	3/25/80		RTM	Random axe work. Crude.
;	2/3/81		DKM	Make source more readable

; 1. Argstack routines
;	TOPFP	PSHAS	POPAS	PSHCS	FLOAT
; 2. PFIX - FP -> 16-bit integer
; 3. MUL10 - int mul by 10
; 4. ITEST - int test
; 5. IMUL - int mul
; 6. INTEGER - string -> int
; 7. CNS - int -> string
; 8. FMUL - FP mul (*)
; 9. FP utilities
;	LNIB	MLRND	NNORM	ADDD	NSUB
; 10. FSUB	- FP sub (-)
; 11. FADD	- FP add (+)
; 12. FDIV - FP div (/)
; 13. FPOUT - FP to string
; 14. FPIN - string to FP
;
	PAGE
;  ArgStack routines.

; Load ptr to fp value on stack to BC and HL.
; Puts fpsiz in DE.

TOPFP	LHLD	TSTKA
	CALL	fpsizB
	DAD	B
	MOV	D,H
	MOV	E,L
	RET

; Push value addressed by HL onto arg stack.
; Sets argf, clears carry.

PSHAS	PUSH	H
	CALL	MfpsizD
	CALL	PSHCS
	POP	D
	XCHG
	CALL	VCOPY
	XRA	A
	INR	A		; clear carry and set acc to 1
	STA	ARGF
	RET

; Pop arg stack.
; HL contains address to put popped value at.

POPFP	LHLD	FPSINK
POPAS	XCHG
	LHLD	TSTKA
	CALL	fpsizB
	DAD	B
	SHLD	TSTKA	; increment stack pointer
	MOV	A,M		; if zero, dump a true zero
	ORA	A
	JNZ	VCOPY
	PUSH	B
	XCHG
	CALL	MfpsizD
	INR	E
	DAD	D
	POP	D
	JMP	CLEAR

; Push frame onto control stack.
; Takes minus amount to sub from tstka in DE.
; Does overflow test and returns old tstka.

PSHCS	LHLD	TSTKA
	PUSH	H
	DAD	D
	CALL	STOV
	SHLD	TSTKA
	POP	H
	RET

; Float the value in HL to top of stack

FLOAT	LXI	D,OBUF+10
	PUSH	D		; we will need obuf again
	CALL	CNS
	MVI	A,CR
	STAX	D
	POP	D
	LHLD	FPSINK
	PUSH	H		; we need fpsink again
	LXI	B,-1
	XCHG
	CALL	FPIN
	POP	H
	JMP	PSHAS	; put converted value on stack
	PAGE
; FLOATING POINT TO 16 BIT INTEGER  CONVERSION
;
;	The floating point number from the top of the
;	ArgStack is converted to a 16 bit integer  returned
;	in DE.   Z flag is set for zero results.
;

PFIXE	CALL	EXPRB	; get expression from text

PFIX	CALL	TOPFP
PFIX05	PUSH	H		; entry point if fp # in HL
	CALL	AINT
	CALL	POPFP	; pop top fp to fpsink
	POP	H
Pfix	MOV	A,M		; exponent and sign (MOD entry)
	CALL	MfpsizD
	DAD	D		; h now points to msb-1
	LXI	D,0		; in case of zero
	ORA	A
	RZ		; return if zero
	JM	OBERR	; negative no good
	XRI	0C0H
	MOV	C,A		; mov exponent to c
	DCR	C		; set up for loop close test
PFIX1	INX	H
	MOV	A,M
	RRC
	RRC
	RRC
	RRC
	CALL	MUL10
	JC	OBERR
	DCR	C
	JP	PFIX2
	MOV	A,M
	CALL	MUL10
	JC	OBERR
	DCR	C
	JM	PFIX1
PFIX2	MOV	A,D
	ORA	E
	RET
	PAGE
; INTEGER MULTIPLICATION BY 10
;
;	Multiply DE by 10 and add the digit in A.
;	preserves all but a, de

MUL10	PUSH	H
	MOV	H,D
	MOV	L,E
	DAD	H		; 2x
	JC	MUL1C
	DAD	H		; 4x
	JC	MUL1C
	DAD	D		; 5x
	JC	MUL1C
	DAD	H		; 10x
	JC	MUL1C
	XCHG
	ANI	0FH
	ADD	E
	MOV	E,A
	MOV	A,D
	ACI	0		; propogate carry
	MOV	D,A
MUL1C	POP	H
	RET
	PAGE
; INTEGER TEST
;
;	Is the FP number addressed by HL an integer?
;	Z flag is set if so.  HL preserved.

ITEST	PUSH	H
	PUSH	H
	XCHG
	LHLD	FPSINK
	XCHG
	PUSH	D
	PUSH	D
	CALL	VCOPY
	POP	D
	CALL	AINT	; convert fpsink to integer
	POP	B		; FPSINK
	POP	D		; original value
	CALL	REL	; returns correct Z setting
	POP	H
	RET
	PAGE
; 16 BIT INTEGER MULTIPLY
;
;	perform unsigned 16 bit integer multiply
;	bc and de are arguments to multiply, hl returns result
;	overflow causes oberr

IMUL	MOV	A,B
	ORA	A
	JZ	IMUL2
	MOV	A,D
	ORA	A
	JNZ	OBERR

; d is zero, b is non-zero

	MOV	A,E		; set-up multiplier
	MOV	E,C		; set-up first multiplicand
IMUL1	CALL	MUL8
	PUSH	H
	MOV	E,B
	CALL	MUL8
	MOV	A,H
	ORA	A
	JNZ	OBERR
	MOV	A,L
	POP	H
	ADD	H		; add partial products
	MOV	H,A
	RNC
	JMP	OBERR

; b is zero

IMUL2	MOV	B,D		; save second multiplicand
	MOV	A,C		; setup multiplier
	JMP	IMUL1

; perform 8 bit multiply of a times e, 16 bit result to hl

MUL8	LXI	H,0
	MOV	D,H
	MVI	C,8
MUL80	DAD	H		; left shift result
	ORA	A
	JP	MUL81	; test if add
	DAD	D		; add multiplicand
MUL81	RLC		; cycle multiplier
	DCR	C
	JNZ	MUL80
	RET

; get integer from txa, set carry if not found
; returns integer value in hl, returns terminator in acc

INTEGER	LXI	B,1
	JMP	INTEGER0+3
INTEGER0
	LXI	B,-1
	SHLD	STRPT
	CALL	GNC
	CMC
	RC
	LXI	D,0
INTEGER1
	CALL	MUL10
	JC	OBERR
	CALL	GNC
	JC	INTEGER1
	LHLD	STRPT
	PUSH	PSW
	MOV	A,B
	CMA
	MOV	B,A
	MOV	A,C
	CMA
	INR	A
	MOV	C,A
	DAD	B
	SHLD	TXA
	XCHG
	POP	PSW
	RET		; carry is clear
	PAGE
; CONVERT INTEGER TO STRING
;
; DE contains address of string, return updated value in DE
; HL contains value to convert


CNS	XRA	A		; set for no leading zeroes
	LXI	B,-10000
	CALL	RSUB
	LXI	B,-1000
	CALL	RSUB
	LXI	B,-100
	CALL	RSUB
CNS1	LXI	B,-10
	CALL	RSUB
	LXI	B,-1
	CALL	RSUB
	RNZ
	MVI	A,'0'
	STAX	D
	DCX	D
	RET

; Takes HL and sub minus number in BC the most possible times
; Put val on string at DE. If A=0 then don't put zero on string
; Return non-zero in A if put on string

RSUB	PUSH	D
	MVI	D,-1
RSUB1	SHLD	RTMP
	INR	D
	DAD	B
	JC	RSUB1
	LHLD	RTMP
	MOV	B,D
	POP	D
	ORA	B		; A=0 iff A=0 and B=0
	RZ
	MVI	A,'0'
	ADD	B
	STAX	D
	DCX	D
	RET
	PAGE
; FLOATING POINT MULTIPLICATION

FMUL	LDAX	B		; zero operands?
	ORA	A
	RZ
	LDAX	D
	ORA	A
	JNZ	FPM0
	STAX	B
	RET

FPM0	PUSH	B		; save arg addresses
	PUSH	D

; clear result buffer

	LHLD	FPBUF
	LDA	FPNIB
FPM1	MVI	M,0
	DCX	H
	DCR	A
	JNZ	FPM1

; compute b'

	XRA	A
	LHLD	BPRIM
	INX	H
	CALL	LNIB0	; preserves bc, de

; now set up for multiply

	LDA	FPBYT	; loop count
	LHLD	FPBUF	; right byte of result
	DCX	D		; right byte of multiplicand

; BC has multiplier current byte addr
; DE has multiplicand right byte addr
; HL has current result byte addr

FPM2	PUSH	PSW
	DCX	B		; multiplier current byte
	PUSH	B
	PUSH	D		; multiplicand
	LDAX	B
	PUSH	PSW	; save multiplier byte
	ANI	0FH
	JZ	FPM4
	CALL	fpbytB
	MOV	C,A
FPM3	PUSH	H
	CALL	ADDD
	MVI	A,0
	ADC	M
	MOV	M,A
	POP	H
	DCR	C
	JNZ	FPM3
FPM4	POP	PSW		; multiplier byte
	RAR
	RAR
	RAR
	RAR
	ANI	0FH
	JZ	FPM6
	CALL	fpbytB
	INR	B
	XCHG
	LHLD	BPRIM
	XCHG
	MOV	C,A
FPM5	PUSH	H
	CALL	ADDD
	POP	H
	DCR	C
	JNZ	FPM5
FPM6	POP	D
	POP	B
	POP	PSW
	DCX	H		; decrement result pointer
	DCR	A
	JNZ	FPM2

; coef multiply done, now round result

	CALL	buftpDE
	CALL	MLRND

; now normalize coef

	MVI	C,0
	CALL	buftpDE
	CALL	NNORM	; now c is 0 or -1

; calculate new exponent

	POP	D		; right arg
	POP	H		; left arg
	LDAX	D
	ANI	7FH
	MOV	B,A
	MOV	A,M
	ANI	7FH
	ADD	B		; add exponents of args
	ADD	C		; adjust for normalize
	SUI	40H	; adjust bias
	JZ	FPM8
	JP	FPM7	; jump if no over or under flow
	RAL
	ORA	A		; set sign bit if underflow
	MVI	A,0
	JM	FPM8	; if underflow, result zero

OVERR	MVI	C,23	; overflow error
	JMP	ERROR

; compute sign of result

FPM7	MOV	B,A		; save exponent of result
	LDAX	D
	XRA	M
	ANI	80H
	ORA	B
FPM8	PUSH	H
	LHLD	BUFTP
	CALL	fpsizD
	DCR	E
	DAD	D
	MOV	M,A
	POP	D
	JMP	VCOPY
	PAGE
; FLOATING POINT UTILITIES

; SHIFTER
;	shifts a fp coef left one nibble
;	lsb+1 of dest addr in HL
;	lsb of source addr in DE. clobbers A,HL only
;	if msb left nibble is non-zero, then shifts it to msb-1

LNIB	LDAX	D
LNIB0	PUSH	D
	PUSH	B
	CALL	fpbytB
LNIB1	ANI	0F0H
	MOV	C,A
	DCX	D
	LDAX	D
	PUSH	PSW
	ANI	0FH
	ORA	C
	RLC
	RLC
	RLC
	RLC
	DCX	H
	MOV	M,A
	POP	PSW
	DCR	B
	JNZ	LNIB1
	ANI	0F0H
	POP	B
	POP	D
	RZ
	RLC
	RLC
	RLC
	RLC
	DCX	H
	MOV	M,A
	RET

;	ROUNDER
;	rounds but does not normalize a coef in fpbuf
;	de has msb, preserves de
;	coef of form .0xxxxxxxxxx or .xxxxxxxxxxx

MLRND	LDAX	D
	ANI	0F0H
	CALL	fpbytHL
	DAD	D		; compute lsb
	MOV	A,M
	JZ	RND
	CPI	50H
	RC
FIRND	DCX	H
ASRND	CALL	fpbytC
	MVI	A,1
	JMP	BUMP

; this entry expects rounding nibble in acc and lsb+1 in hl

RND	ANI	0FH		; unnormalized case
	CPI	5
	RC
	CALL	fpbytC
	INR	C
	MVI	A,10H
BUMP	ADD	M
	DAA
	MOV	M,A
	RNC
	DCX	H
	DCR	C
	MVI	A,1
	JNZ	BUMP
	RET

;	NORMALIZER
;	takes msb of coef addr in de, takes exponent in c
;	normalizes one nibble if needed

NNORM	LDAX	D
	ANI	0F0H
	RNZ		; already normalized
	DCR	C
	CALL	fpbytHL
	DAD	D
	MOV	D,H
	MOV	E,L
	JMP	LNIB

;	REPEAT ADDER
;	takes left arg lsb addr in hl,
;	takes right arg lsb addr in de
;	number of bytes to add in b. preserves bc, de
;	hl points to byte preceeding msb of result on return

ADDD	PUSH	B
	PUSH	D
	ORA	A		; clear carry

ADD1	LDAX	D
	ADC	M
	DAA
	MOV	M,A
	DCX	H
	DCX	D
	DCR	B
	JNZ	ADD1
	POP	D
	POP	B
	RET

;	subtracts hl(lsb) from de(lsb) until de would go neg
;	subtracts fpbyt+1 bytes
;	returns # of times subtracted in a

NSUB	CALL	fpbytB
	INR	B
	MVI	C,0FFH
NSUB1	INR	C
	PUSH	B
	PUSH	D
	PUSH	H
	STC
	MVI	C,99H
NSUB2	MVI	A,0
	ADC	C
	SUB	M
	XCHG
	ADD	M
	XCHG
	DAA
	STAX	D
	DCX	H
	DCX	D
	DCR	B
	JNZ	NSUB2
	POP	H
	POP	D
	POP	B
	JC	NSUB1	; no borrow
	XCHG
	CALL	ADDD
	MOV	A,C
	RET
	PAGE
; FLOATING POINT SUBTRACTION

FSUB	LHLD	RBUF	; buffer for right arg
	LDAX	D		; sign of right arg
	ORA	A		; check for subtracting 0
	RZ		; don't take minus 0
	XRI	80H		; change sign
	PUSH	B		; save left arg address
	CALL	fpbytC	; copy right arg to RBUF
	INR	C
FSUB1	MOV	M,A
	DCX	H
	DCX	D
	LDAX	D
	DCR	C
	JNZ	FSUB1
	POP	B
	XCHG
	LHLD	RBUF
	XCHG		; fall thru to...
	PAGE
; FLOATING POINT ADDITION

FADD	LDAX	D		; sign and exp of right arg
	ORA	A
	RZ		; left arg is answer
	PUSH	B		; address of left arg (result)
	LDAX	B
	ORA	A
	JZ	FADD0	; jump if left arg is zero
	ANI	7FH		; clear off sign of left arg
	MOV	L,A
	LDAX	D		; sign of right arg
	ANI	7FH
	SUB	L		; right arg exp - left arg exp
	MOV	H,B
	MOV	L,C
	JP	FADD1	; right arg is exponent bigger
	XCHG
	CMA
	INR	A

; scale the value pointed to by hl

FADD1	PUSH	B
	MOV	B,A
	LDA	FPNIB
	INR	A
	CMP	B
	MOV	A,B
	POP	B
	JNC	FADD2

; here return the larger argument

FADD0	XCHG
	POP	D		; result address
	JMP	VCOPY	; may be copying self to self

FADD2	PUSH	D		; unscaled argument
	MOV	D,M		; sign and exp
	PUSH	D		; sign and exp of unscaled val
	ADI	1
	RAR		; divide by 2
 	PUSH	PSW	; save this condition for later
	XCHG
	LHLD	FPBUF
	XCHG
	CALL	fpbytC
	XRA	A
	STAX	D
FADD3	DCX	D		; copy coef to buffer
	DCX	H
	MOV	A,M
	STAX	D
	DCR	C
	JNZ	FADD3

; now put necessary number of zero bytes in front

	POP	PSW		; a has number of bytes needed
	PUSH	PSW
	LHLD	FPBUF
	ORA	A
	JZ	FADD5	; jump if no zero bytes needed
	MOV	C,A		; move the count to c
	XRA	A
FADD4	DCX	H		; now has lsb of scaled value
	DCX	D
	STAX	D
	DCR	C
	JNZ	FADD4
FADD5	POP	PSW
	JC	FADD6	; C set means even scaling
	INX	H		; new lsb
	MOV	D,H
	MOV	E,L
	CALL	LNIB0	; left shift one nibble
	DCX	D
	XCHG

; hl now has lsb+1 of scaled coef in buffer
; stack sign+exp of scaled, unscaled arg exp addr, result addr

FADD6	MOV	A,M		; prepare to round
	CPI	50H
	DCX	H
	PUSH	H		; lsb of scaled
	CNC	ASRND	; round if necessary
	POP	D		; scaled lsb
	POP	PSW		; sign and exp of scaled value
	POP	H		; unscaled exp addr
	XRA	M
	MOV	B,M
	DCX	H		; unscaled lsb
	PUSH	B		; sign and exp of unscaled
	JP	FAD16	; signs are the same

; subtract case (signs different)
; subtract scaled value from unscaled value
; and put result back over scaled
; if no borrow, then sign is same as sign of unscaled value
; stack sign+exp of unscaled, dest addr

	PUSH	D		; scaled lsb
	CALL	fpbytC
	STC
	MVI	B,99H
FADD7	MVI	A,0
	ADC	B
	XCHG
	SUB	M
	XCHG
	ADD	M
	DAA
	STAX	D
	DCX	H
	DCX	D
	DCR	C
	JNZ	FADD7
	POP	H		; lsb of scaled value
	JC	FADD9	; goto fadd9 if was no borrow

; had borrow, so comp result and toggle result sign

	POP	PSW		; sign and exp to a
	XRI	80H
	PUSH	PSW	; push psign and exp of result
	PUSH	H		; lsb of result in buffer
	CALL	fpbytC
	MVI	B,90H+10
FADD8	MOV	A,B
	SBB	M
	ADI	0
	DAA
	MOV	M,A
	DCX	H
	DCR	C
	CMC
	JNZ	FADD8
	XCHG		; msb-1 of result in buffer
	POP	H		; lsb of result in buffer

; subtract done, now normalize result
; hl has lsb of result in buffer, de has msb-1 of same
; stack sign and exp of result, result address

FADD9	PUSH	D		; msb-1 of result in buffer
	MOV	D,H
	MOV	E,L
	CALL	fpbytB
	DCR	B
	XRA	A
FAD10	INX	H		; this loop adds trailing 0's
	MOV	M,A
	DCR	B
	JNZ	FAD10
	POP	H		; msb-1 of result in buffer
	POP	B		; sign and exp of result
	CALL	fpnibC
FAD11	INX	H		; start look for 0's from msb
	MOV	A,M
	ORA	A
	JNZ	FAD12
	INX	D		; computing new lsb of result
	DCR	C
	DCR	C		; do twice to help fixup
	JNZ	FAD11

; came here if result is zero

	POP	B		; address of destination
	STAX	B		; acc had 0
	RET

FAD12	POP	H		; destination address
	ANI	0F0H		; see if result is normalized
	PUSH	PSW
	JNZ	FAD13	; skip if exponent ok
	DCR	C		; decrement exponent by one
FAD13	LDA	FPNIB
	CMA
	INR	A
	ADD	B		; Sign and exp of result.
	ADD	C		; Adjusted for normalization..
	MOV	M,A		; exponent value
	XRA	B		; check for underflow
	JP	FAD14	; jump if ok
	POP	PSW		; clean up stack
	MVI	M,0		; put zero in destination
	RET

; now check again if lnib needed, and if so do it

FAD14	POP	PSW		; z set if lnib needed
	JNZ	FAD15
	XRA	A
	INX	D		; lsb+1 of result in buffer
	JMP	LNIB0
FAD15	DCX	H
	XCHG
	CALL	fpbytC
	JMP	VCOP1

; add case-sign and exp of result in b
; de has scaled lsb address (in buffer)
; h has unscaled lsb stck has sign+exp, dest addr

FAD16	PUSH	D		; lsb of scaled value in buffer
	CALL	fpbytB
	ORA	A
FAD17	LDAX	D		; do the add
	ADC	M
	DAA
	STAX	D
	DCX	H
	DCX	D
	DCR	B
	JNZ	FAD17
	JC	FAD18	; go if there was carry
	POP	H		; lsb of result in buffer
	POP	PSW
	POP	D		; destination address
	STAX	D		; store exponent
	DCX	D
	CALL	fpbytC
	JMP	VCOP1

; add had carry, requiring lnib and round

FAD18	MVI	A,1		; put the carry into msb-1
	STAX	D
	POP	H		; lsb of result in buffer
	POP	B		; sign and exp to b
	MOV	A,M
	PUSH	H		; lsb of result in buffer
	CALL	RND
	POP	D		; lsb+1 of result in buffer
	POP	H		; destination address
	INR	B		; bump exponent
	MOV	A,B		; set up check for overflow
	ANI	7FH
	JZ	OVERR
	MOV	M,B		; put exponent at destination
	JMP	LNIB	; left shift, value to dest.
	PAGE
; FLOATING POINT DIVIDE

FDIV	LDAX	D		; zero args?
	ORA	A
	JZ	DZERR
	LDAX	B
	ORA	A
	RZ
	CALL	MfpbytH
	DAD	B
	SHLD	RESTA
	PUSH	B		; left arg
	PUSH	PSW	; sign+exp of left arg
	PUSH	D		; right arg
	PUSH	H		; msb of left arg

; compute barg (right arg with 0 byte preceding)

	PUSH	B
	PUSH	D
	LHLD	BARG
	CALL	fpbytB

FDV0	DCX	D
	LDAX	D
	MOV	M,A
	DCX	H
	DCR	B
	JNZ	FDV0
	MVI	M,0		; msb of barg
	POP	D
	POP	B

; compute bprim

	XRA	A
	LHLD	BPRIM
	INX	H
	CALL	LNIB0

; copy left arg to buffer

	LHLD	FPBUF
	INX	H
	XRA	A
	CALL	fpbytD
	INR	D

FDIV0	MOV	M,A		; zero second half+1 of buffer
	DCX	H
	DCR	D
	JNZ	FDIV0
	CALL	fpbytD
FDIV1	DCX	B		; copy to first half of buffer
	LDAX	B
	MOV	M,A
	DCX	H
	DCR	D
	JNZ	FDIV1
	MVI	M,0		; msb of dividend

; perform divide of coef's; subtracts DE gets DE - HL

	PUSH	H
	CALL	buftpDE
	DCX	D
	CALL	fpbytHL
	DAD	D
	XCHG
	POP	H
	CALL	fpbytC
	INR	C
FDIV2	PUSH	D		; current subtrahend lsb
	PUSH	B		; to preserve c
	LHLD	BARG
	CALL	NSUB
	POP	B
	RLC
	RLC
	RLC
	RLC
	MOV	B,A
	POP	D		; current subtrahend
	INX	D
	PUSH	D
	PUSH	B
	LHLD	BPRIM
	CALL	NSUB
	POP	B
	ORA	B
	LHLD	RESTA
	MOV	M,A
	INX	H
	SHLD	RESTA
	POP	D		; current subtrahend
	DCR	C
	JNZ	FDIV2

; round and normalize result

	POP	D		; msb of result
	CALL	MLRND	; preserves de
	MVI	C,0
	CALL	NNORM

; compute exponent of result

	POP	D		; right arg exp addr
	LDAX	D
	ANI	7FH
	SUB	C
	MOV	H,A
	POP	PSW		; sign+exp of left arg
	MOV	L,A		; save sign of left arg
	POP	B		; left arg exp addr
	ANI	7FH
	SUB	H
	ADI	41H
	JZ	FDIV4
	JP	FDIV3

; overflow or under flow

	ANI	40H
	JNZ	FDIV4	; underflow
	JMP	OVERR

; compute sign of result and store

FDIV3	XCHG		; right arg addr to hl
	MOV	D,A		; save exponent value
	MOV	A,E		; get left arg sign
	XRA	M
	ANI	80H
	ORA	D
FDIV4	STAX	B
	RET
	PAGE
; PUT NUMBER INTO OUTPUT FORMAT

;	special routine. we bypass the format error and return
;	"***" to indicate a no-no

FORMAT	LHLD	BOMBER	; reset stack
	SPHL
	LDA	CWIDTH	; width of output field
	MOV	B,A
	MOV	C,A	; put in B & C
	LHLD	EOBUF	; our buffer
	MVI	M,'"'	; termination char

FORMAT1	DCX	H	; dink pointer
	MVI	M,'*'	; stick in '*'
	DCR	C
	JNZ	FORMAT1	; spin till done
	RET		; HL -> last char, B: length

;	f format subroutine
;	expects width in acc, value ptr in HL
;	set carry on failure to fit into field
;	return ptr to first char of result in HL
;	also, epoint points to last non-blank digit
;	of result (for free format use)

FFORM	SHLD	VALUE
	MOV	B,A		; remember width

; init the output buffer

	LHLD	EOBUF
	MVI	M,'"'
	LDA	CFRAC
	ORA	A
FF1	JZ	FF2
	DCX	H
	MVI	M,'0'
	DCR	B
	JZ	FORMAT	; fraction was>=width error
	DCR	A
	JMP	FF1

FF2	DCX	H
	MVI	M,'.'
	SHLD	DPOINT	; used by E format
	SHLD	EPOINT
	SHLD	RESTP	; this value used by lfp
FF3	DCR	B
	JZ	FF4
	DCX	H
	MVI	M,' '
	JMP	FF3

FF4	SHLD	RESTA	; result address
	DCX	H
	MVI	M,'!'	; beginning marker
	LHLD	VALUE
	MOV	A,M
	ANI	7FH
	JZ	FF82	; go remove trailing zeroes
	SUI	41H		; unbias
	JM	FF9	; d-case (all right of dp)
	INR	A		; correct for bias glitch
	MOV	B,A		; remember exponent value
	MOV	C,A
	LDA	FPNIB
	SUB	C
	JNC	FF7	; c-(value overlaps the dp)

; zero fill between dp and lsb of value

	JZ	FF6
	CMA
	INR	A
	MOV	C,A
FF5	MVI	A,'0'
	CALL	LFPD
	DCR	C
	JNZ	FF5
FF6	CALL	fpnibC
	MOV	B,C

; b has index of next digit to be written to the left

FF7	INR	C
FF71	CALL	GFD	; get digit
	CALL	LFPD
	DCR	B
	JNZ	FF71

; c has digit index in value
; loop terminates when rfpd exits to ff81 or ff82

FF80	MOV	B,C
FF8	CALL	GFD	; get the digit
	CALL	RFPD
	INR	B
	JMP	FF8

; come here from rfpd if rounding required
; h has pointer to the '"' character

FF81	DCX	H
	MOV	A,M
	CPI	'.'
	JZ	FF81
	CPI	','
	JZ	FF81
	CPI	'!'+1
	JC	FF811

; must be a digit, so increment it

	INR	A
	MOV	M,A
	CPI	'9'+1
	JNZ	FF82	; carry not propagated
	MVI	M,'0'
	JMP	FF81

; come here if blank or '!' encountered when propogating carry

FF811	LXI	D,EXP	; 200q unless e format
	LDAX	D
	CPI	80H
	JZ	FF812	; jump if not e format

; convert to 1.00 and bump exponent

	INX	H		; point to char preceding dp
	MVI	M,'1'
	INR	A		; bump exponent
	JZ	FORMAT	; exponent overflow
	STAX	D
	JMP	FF82	; done rounding

; here it was not e format

FF812	MVI	A,'1'
	CALL	LFPD

; remove trailing zeroes if required

FF82	LHLD	EOBUF
	DCX	H
	LDA	CFORM	; trialing zeroes removed
	CPI	'#'
	JZ	FF83
	LDA	COPT
	ORA	A
	JP	FF84	; jump if want trailing zeroes
FF83	MOV	A,M
	CPI	'0'
	JNZ	FF831
	MVI	M,' '
	DCX	H
	JMP	FF83

FF831	LDA	CFORM	; if free format, move up the "
	CPI	'#'
	JNZ	FF84
	INX	H
	MVI	M,'"'
	DCX	H

; check for special case of zero value

FF84	MOV	A,M		; pointer to last non-blank
	SHLD	EPOINT	; used by free format
	CPI	'.'
	JNZ	FF85
	DCX	H
	MOV	A,M
	CPI	' '
	JZ	FF841
	CPI	'!'
	JNZ	FF85
	INX	H
	INX	H
	MVI	M,'0'
	JMP	FF85

FF841	MVI	A,'0'
	CALL	LFPD

; now put on '-' if needed

FF85	LHLD	VALUE
	MOV	A,M
	ORA	A
	MVI	A,'-'
	CM	LFP	; put on '-' if negative value
	LHLD	RESTA
	RET

; d case-write zeroes after dpoint

FF9	CMA
	ORA	A		; set z in case of zero
	MOV	B,A
	MVI	C,1		; needed at ff8

FF91	JZ	FF80
	MVI	A,'0'
	CALL	RFPD	; will exit if field fills up
	DCR	B
	JMP	FF91

; add character to left of integer (for fpout)
; set C if out of room, chr expected in acc, preserve bc

LFPD	PUSH	PSW	; enter here if digit
	LDA	COPT
	ANI	40H		; test if commas required
	JZ	LFP1	; not desired
	LXI	H,COMCNT
	DCR	M
	JNZ	LFP1

; comma must be written

	MVI	M,3
	MVI	A,','
	CALL	LFP
LFP1	POP	PSW
LFP	LHLD	RESTP
	DCX	H
	MOV	D,A
	MOV	A,M
	CPI	'!'
	JZ	FORMAT
	SHLD	RESTP
	MOV	M,D
	RET

; add digit to right end of fp string (fpout), character in acc
; if chr beyond field, to FF81 if rounding required, else FF83
; preserve bc

RFPD	LHLD	EPOINT
	INX	H
	MOV	D,A
	MOV	A,M
	CPI	'"'
	JZ	RFPD1
	SHLD	EPOINT
	MOV	M,D
	RET

; here encountered end of field

RFPD1	POP	PSW		; clean off return link
	MOV	A,D
	CPI	'5'
	JC	FF82
	JMP	FF81	; round

; get bth digit from fp val, preserve bc, return '0' if b>prec

GFD	LDA	FPNIB
	MOV	D,A		; save in d
	CMP	B		; check for index > prec
	MVI	A,'0'
	RC
	MOV	A,B
	DCR	A
	SUB	D		; acc has index-prec-1
	RAR		; divide by 2
	MOV	E,A
	MVI	D,255
	PUSH	PSW
	LHLD	VALUE
	DAD	D
	POP	PSW
	MOV	A,M		; byte with digit
	JC	GFD1
	RAR
	RAR
	RAR
	RAR
GFD1	ANI	0FH
	ADI	'0'
	RET

;	output floating point value pointed to by hl
;	return pointer to first char of result string in hl
;	and size of result string in b

FPOUT	XCHG
	LXI	H,0
	DAD	SP
	SHLD	BOMBER
	XCHG
	PUSH	H
	MVI	A,4
	STA	COMCNT	; in case commas are required
	MVI	A,80H
	STA	EXP	; not E format
	LDA	CFORM	; load floating point format
	CPI	'I'		; integer
	JZ	IFORM
	CPI	'#'
	JZ	FREE
	CPI	'E'
	JZ	EFORM

; must be 'f' format

FF	LDA	CWIDTH
	CALL	FFORM

; now see about optional '+' and '$'

FP1	POP	H		; pointer to value
	LDA	COPT
	PUSH	PSW	; option word
	ANI	20H
	JZ	FP2	; jump if '+' not desired
	MOV	A,M		; exponent
	ORA	A
	MVI	A,'+'
	CP	LFP	; put on plus sign
FP2	POP	PSW		; option word
	RAR		; mov right bit to carry
	MVI	A,'$'
	CC	LFP
	LHLD	RESTA

; now see if want to do special free format stuff
; (leading and trailing blank removal)

	LDA	CFORM
	CPI	'#'
	JNZ	FP4
FP3	INX	H
	MOV	A,M
	CPI	' '
	JZ	FP3
	DCX	H

; now count characters in result sring to b register

FP4	MVI	B,0
	MOV	D,H
	MOV	E,L
FP41	LDAX	D
	CPI	'"'
	RZ
	INR	B
	INX	D
	JMP	FP41

FREE	PUSH	H
	LDA	FPNIB
	MOV	H,A		; Set up CWIDTH.
	RLC
	INR	A
	MOV	L,A		; Set up CFRAC.
	SHLD	CWIDTH
	POP	H
	MOV	A,M
	ANI	7FH		; trim off coefficient sign bit
	JZ	IFORM
	PUSH	B
	MOV	B,A
	LDA	FPNIB
	ADI	41H
	CMP	B
	MOV	A,B
	POP	B
	JC	EFORM	; exponent too big

; now we know exponent is lee than 100q+fpnib+1, see if
; value fits in f format (after stripping trailing zeroes)

	PUSH	H
	MOV	B,A
FR1	INR	B		; bump count
	DCX	H		; last unchecked byte of coef
	MOV	A,M
	ANI	0FH		; mask down to right digit
	JNZ	FR2	; jump if no more zeroes
	INR	B
	MOV	A,M
	ORA	A
	JZ	FR1	; loop if left digit also zero

; have increased exp (in b) by number of trailing zeroes+1

FR2	POP	H
	MOV	A,B
	CPI	40H+1
	JC	EFORM	; jump if still won't fit
	CALL	ITEST
	JNZ	FF	; jump if not integer

; fall thru to iform

IFORM	CALL	ITEST	; preserves hl
	JNZ	FORMAT	; jump if was not integer
	LDA	CFRAC
	STA	EFORMXX
	XRA	A
	STA	CFRAC
	LDA	CWIDTH
	INR	A		; make room for the dpoint
	CALL	FFORM
	LDA	EFORMXX
	STA	CFRAC
	MVI	A,'"'
	PUSH	H
	LHLD	EOBUF
	DCX	H
	MOV	M,A
	POP	H
	JMP	FP1

EFORM	MOV	A,M		; exponent of value
	STA	EXP	; save it
	ORA	A		; if the number is zero..
	JZ	IFORM	; print it like an integer
	ANI	80H		; preserve sign
	ORI	41H	; set for scientific notation
	MOV	M,A		; fake value for fform
	LDA	CWIDTH
	SUI	4		; leave room for exponent
	CALL	FFORM
	LHLD	EPOINT
	MOV	A,M
	CPI	'.'
	JZ	EF0
	INX	H
EF0	MVI	M,'E'
	INX	H
	LDA	EXP	; the exponent
	ANI	7FH
	MVI	B,'+'
	JZ	EF1	; jump if zero case
	SUI	41H	; subtract off bias+1
	JP	EF1	; jump if positive

; here was negative exponent

	MVI	B,'-'
	CMA
	INR	A
EF1	MOV	M,B		; sign character
	INX	H
	PUSH	H		; save pointer
	MOV	E,A		; low order exponent value
	MVI	D,0		; high order exponent value
	LHLD	DPOINT	; pointer to decimal point
	DCX	H
	DCX	H		; decrement pointer two
	MOV	A,M		; we are checking for overflow
	CPI	'1'
	JNZ	EF2	; jump if no overflow

; here had overflow, so bump exponent and
; move the sign and '1' over

	DCX	H		; points to sign (if '-')
	MOV	A,M
	MVI	M,' '
	INX	H
	MOV	M,A		; mov over the sign (or ' ')
	INX	H
	SHLD	RESTP
	MVI	M,'1'	; overwrite the 0 with 1
	INX	D		; bump exponent
EF2	XCHG		; exponent to hl
	POP	D

; We have to do this differently now.  Is perhaps a little
; faster than before, but is ugly as sin!  (Previously, we
; used an entry to the CNS code, but bass-ackwards now.)

	MOV	A,L
Expone	CPI	10
	JC	expone
	INR	H
	SUI	10
	JMP	Expone

expone	MOV	B,A
	MOV	A,H
	ADI	30H
	STAX	D
	INX	D
	MOV	A,B
	ADI	30H
	STAX	D
	INX	D
	MVI	A,'"'	; string terminator
	STAX	D
	JMP	FP1
	PAGE
; Get a number from the input stream

; HL - character pointer to first character
; DE - result address
; BC - plus or minus 1, indicating text direction
; Returns updated pointer in HL
; (past terminator, with terminator in B)

FPINQ	DAD	B

FPIN	MOV	A,M
	CPI	' '
	JZ	FPINQ	; skip leading blanks!
	CPI	09
	JZ	FPINQ	; tab is same as blank
	XRA	A
	STA	CSIGN
	STA	ESIGN	; exponent sign
	STA	EXP	; exponent value
	LDA	FPNIB
	CMA		; We needed -FPNIB-1.
	STA	PUTFL	; flag for the put subroutine
	MOV	A,M		; first character
	CPI	'+'
	JZ	FPIN1
	CPI	PLSRW
	JZ	FPIN1
	CPI	MINRW
	JZ	FPIN0
	CPI	'-'
	JNZ	FPIN2
FPIN0	MVI	A,80H
	STA	CSIGN

FPIN1	DAD	B		; incr pointer past sign chr
FPIN2	XCHG
	SHLD	RESTA	; address of result
	MVI	M,0		; clear exponent
	LDA	FPBYT
FPIN3	DCX	H
	MVI	M,0
	DCR	A
	JNZ	FPIN3
	SHLD	RESTP
	XCHG
	SHLD	STRPT	; string pointer address
	CALL	GNC
	JC	FIS21	; was digit
	CPI	'.'
	JNZ	FIFAIL

; dp came first, make sure at least one digit follows

	CALL	GNC
	JC	FIS41

FIFAIL	STC
	RET

; have dp, no sd yet

FIS4	LXI	H,EXP
	DCR	M
FIS40	CALL	GNC
FIS41	JZ	FIS4	; was a zero
	JC	FIS60	; must be a sd
	CPI	'E'
	JZ	FIS7

; get here means we must have a zero result

FIZERO	XRA	A
	JMP	FIDN2

; eat leading zeros, if there are any

FIS2	CALL	GNC
FIS21	JZ	FIS2
	JC	FIS5
	CPI	'.'
	JZ	FIS40
	JMP	FIZERO

; have sd before dp

FIS5	PUSH	B
	CALL	PUTD
	POP	B
	LXI	H,EXP
	INR	M
	CALL	GNC
	JC	FIS5
	CPI	'.'
	JZ	FIS6
	CPI	'E'
	JZ	FIS7

; here we are done, so put the number together

FIDN	XRA	A		; exponent value if no 'e'
FIDN1	LXI	H,EXP
	ADD	M
	ADI	40H
	JM	FIFAIL	; exponent out of range
	LXI	H,CSIGN
	ORA	M
FIDN2	LHLD	RESTA
	MOV	M,A
	LHLD	STRPT	; updated string pointer
	LDA	TERMC
	MOV	B,A
	RET		; carry is clear

; here we have a sd and a dp

FIS6	CALL	GNC
	JNC	FIS61
FIS60	PUSH	B
	CALL	PUTD
	POP	B
	JMP	FIS6
FIS61	CPI	'E'
	JZ	FIS7
	JMP	FIDN

; here we have nz coef and 'e'

FIS7	CALL	GNC
	JC	FIS81
	CPI	'+'
	JZ	FIS8
	CPI	PLSRW
	JZ	FIS8
	CPI	'-'
	JZ	FIS71
	CPI	MINRW
	JNZ	FIFAIL
FIS71	STA	ESIGN	; puts non-zero value in esign

; now get the exponent magnitude

FIS8	CALL	GNC	; get first digit
	JNC	FIFAIL	; complain if no first digit
FIS81	MOV	E,A		; save in e in case of mul10
	MVI	D,0
	CALL	GNC	; see if another digit
	JNC	FIS82
	CALL	MUL10
	CALL	GNC	; eat it
FIS82	LDA	ESIGN
	ORA	A
	MOV	A,E
	JZ	FIDN1
	CMA
	INR	A		; twos complement of exponent
	JMP	FIDN1

; special get character routine for fpin
; expexts pointer in strpt
; if digit, sets carry and cleans to 4 bits (z set if 0)

GNC	LHLD	STRPT
	MOV	A,M
	STA	TERMC
	DAD	B
GNC00	SHLD	STRPT
	CPI	'9'+1
	RNC
	CPI	'0'
	CMC
	RNC
	ANI	0FH
	STC
	RET

; routine for fpin to put next dig in result
; scalar putfl has count of digits already put, so if too big
; then digit is not put (the prec + 1 digit is used to round

PUTD	MOV	B,A
	LXI	H,PUTFL
	INR	M
	MOV	A,M		; load count to acc
	LHLD	RESTP	; pointer to current byte
	JZ	PUTD1	; rounding digit
	RP		; we are beyond rounding digit
	RRC
	JC	PUTR

; here we want to put digit in left nibble

	MOV	A,B
	RLC
	RLC
	RLC
	RLC
	MOV	M,A
	RET

; here put digit into right nibble

PUTR	MOV	A,M
	ORA	B
	MOV	M,A
	INX	H
	SHLD	RESTP
	RET

; came here if rounding digit

PUTD1	MOV	A,B		; rounding digit to acc
	CPI	5
	RC
	CALL	FIRND
	RNC

; here must change to 1.0000000, hl has msb-1

	INX	H
	MVI	M,10H
	LXI	H,EXP
	INR	M		; bump exponent
	RET

