	PAGE
; Variables.TX
; Last Updated:
;	12/29/79	RTM	Disaster recovery, two user
;	3/25/80		RTM	Random axe work. Crude.
;	2/4/81		DKM	Improve readability

; ** 1. VAR - Find variable
; ** 2. STLK - Lookup variable
; ** 3. STASS - String assignment controller
	PAGE
; **		Find Variable.
; **
; **	Number or matrix element case:
; **	  HL points to value, Z cleared.
; **	String case:
; **	  String max size returned in STRMX global.
; **	  Logical length returned in STRLG and DE.
; **	  Base address of string returned in STRBA and HL.
; **	  Sym.tab. ptr to logical length field returned in BC.
; **	  Z set.

VAR	CALL	NAME	; get variable name
	JC	BSERR	; wasn't a variable!
VAR0	CALL	SNAME	; is it a string?
	JNC	STRINGVAR
	XRA	A
	STA	VTYPE	; 0 means numeric array
	LDA	MATFLG
	ORA	A		; non-zero indicates a MAT statement
	JNZ	ARRAYVAR
	CALL	GC
	MVI	A,LPARRW	; a left paren here
	CALL	SCANC	; means a subscript (an array!)
	JZ	ARRAYVAR

SCALAR			; scalar variable processing
	PUSH	B		; variable name
	CALL	STLK	; Returns entry address in HL.
	POP	B		; variable name
	JNC	FOUNDVAR
	CALL	fpsizH
	CALL	ASTAB	; clear space for new variable
	JMP	FOUNDVAR

ARRAYVAR
	MVI	A,MTYPE
	ORA	C
	MOV	C,A		; Set type to matrix
ARRAYVAR0
	PUSH	B		; variable name
	CALL	STLK	; look it up
	POP	B		; variable name
	JNC	ARRAYEXISTS
	LDA	MATFLG
	ORA	A
	JZ	DMERR	; how can I find an unDIM'd array?
	CPI	40H
	JZ	DIMMST	; it was special - a string 'scalar'

; Well, we couldn't find an array with the specified name.
; However, since we got this far it means we are doing a
; MAT statement.  So, we adjust the variable name to say
; 'I am a scalar' - and look THAT up.

	MVI	A,0FH	; clear all special flags..
	ANA	C
	MOV	C,A
	JMP	SCALAR	; and get variable as a scalar

ARRAYEXISTS
	LXI	D,0
	PUSH	D		; Initial locator.
	INX	H		; Increment symtab ptr to first
	INX	H		; dimension table entry.
	PUSH	H		; Dimension table ptr.

; Now we do the calculations to point us at the proper element
; of the array.  We have 3 cases:
;	case 1	-  a normal subscript calculation
;	case 2	-  the initial MAT scan, gets last element
;	case 3	-  MAT look-up, where the index is available
;		   to us under one of the MAT rocks.
; MATFLG has the indicator

GETDIMENSIONS
	LDA	MATFLG
	ORA	A
	JZ	NORMALSUBSCRIPT

; for cases 2 & 3, we pretend that it is a "normal"
; look-up of the last element, so we grab the number from
; the dimension table.

	CALL	DLOAD	; dimtab entry into DE
	DCX	D
	JMP	SUBSCRIPTS

NORMALSUBSCRIPT
	LDA	VTYPE	; these may be squatted
	PUSH	PSW	; so save'em on stack
	LHLD	GROKK
	PUSH	H
	CALL	PFIXE	; get subscript from text
	POP	H
	SHLD	GROKK
	POP	PSW
	STA	VTYPE
	LHLD	DIMORG	; check array origin..
	MOV	A,L
	ORA	A
	JZ	SUBSCRIPTS
	DCX	D		; for 1 based arrays!

SUBSCRIPTS
	POP	H		; Dimtab.
	CALL	DCMP	; compare index with dim size.
	JNC	SBERR	; too big!!
	XTHL		; Locator -> HL, dimtab -> stack.
	DAD	D		; Index plus locator to HL.
	XCHG		; Index plus locator to DE.
	POP	H		; Dimtab to HL.
	INX	H		; Advance to
	INX	H		; next dimension size
	PUSH	H		; Dimtab ptr to next entry
	MOV	C,M
	INX	H		; grab next dim
	MOV	B,M

; at this point
;	BC: next dimension from table
;	DE: current locator

	MOV	A,B		; 0 marks end of dimtab
	ORA	C
	JZ	LASTDIMENSION
	MOV	H,B
	MOV	L,C
	SHLD	NLN	; gets STRMX for strings
	XCHG
	SHLD	GROKK	; also used for strings
	XCHG
	CALL	IMUL	; Next dim  *  current locator.
	XTHL		; Locator - stack, dimtab  - HL
	PUSH	H
	LDA	MATFLG	; for MAT operations we got no worries
	ORA	A
	JNZ	GETDIMENSIONS
	CALL	SCOMA	; eat comma
	JNC	GETDIMENSIONS
	LDA	VTYPE
	ORA	A
	JZ	BSERR	; better not be numeric array!
	INR	A
	STA	VTYPE	; to indicate correct # of dims
	LHLD	NLN	; for strings we do it one more time
	XCHG
	DCX	D
	JMP	SUBSCRIPTS

LASTDIMENSION
	LDA	MATFLG
	ORA	A
	CZ	EATRP	; normal stuff has a right paren here
	POP	H		; Dimension table ptr.
	INX	H
	INX	H		; get past end marker
	PUSH	H		; save "base addr"
	XCHG		; locator in HL
	LDA	VTYPE
	ORA	A		; numeric type?
	JZ	NOTSTRING
	LHLD	GROKK	; use this locator for strings
NOTSTRING
	LDA	MATFLG
	CPI	80H		; iff is MAT scan do we do this
	JNZ	NOTSCAN
	SHLD	MATROK	; this is the number of array elements
NOTSCAN
	INR	A		; MAT lookup?
	JNZ	NOTLOOKUP

; If we were doing a MAT, here is where we peek under
; the rock MATROK to get our index.

	XCHG
	LHLD	MATcnt
	CALL	CMPDH	; compare index with locator
	JC	MSERR	; jmp/ too big: subscript error
NOTLOOKUP
	XCHG		; locator to DE
	POP	H		; base addr to HL
	LDA	VTYPE
	ORA	A
	RNZ		; string array - go back there now
	PUSH	B		; variable name
	CALL	fpsizB
	CALL	RADD	; twiddle so we point at right element
	POP	B		; variable name
FOUNDVAR
	CALL	fpsizD
	DCR	E
	DAD	D	; now we are pointing at the exponent
	MOV	A,B
	CPI	'#'	; variable # is treated specially
	JNZ	VARLAST

; here we handle the special case where the user
; has requested the "indexing" variable #.

	PUSH	H		; pointer to variable
	LHLD	DIMORG	; array origin is either 0 or 1
	XCHG
	LHLD	MATcnt ; current array element
	DAD	D
	CALL	FLOAT	; put it on the stack
	POP	H
	PUSH	H
	CALL	POPAS	; and pop it to the variable "#"
	POP	H

VARLAST	ORI	1		; carry is clear
	RET

STRINGVAR
	MVI	A,1
	STA	VTYPE	; 1 indicates a string array
	LDA	MATFLG
	ORA	A		; is an array if MATFLG is set
	JNZ	STRINGARRAYS
	MVI	A,LPARRW
	CALL	SCANC
	JNZ	STRINGSCALAR
STRINGARRAYS		; do dimension look-up
	CALL	ARRAYVAR0
	LDA	MATFLG	; if a MAT statement, we are OK
	ORA	A
	JNZ	STRINGARRAY2
	LDA	VTYPE	; if not, better have this a 2
	CPI	2
	JNZ	SBERR	; OOPS!!
STRINGARRAY2
	PUSH	H		; store the string pointer
	LHLD	NLN	; last dimension (length)
	DCX	H
	DCX	H		; HL now has real maximum length
	SHLD	STRMX
	CPI	40H		; MATFLG = string scalar
	JNZ	STRINGVAR1
	XRA	A		; reset MATFLG
	STA	MATFLG

;	DE: locator
;	HL: true string length
;
; now we must multiply our locator by the length
; to point at the proper element

STRINGVAR1
	INX	H		; element length
	INX	H
	MOV	B,H		; into BC
	MOV	C,L
	CALL	IMUL	; HL=locator*length

	POP	D		; base addr
	DAD	D		; plus HL
	PUSH	H
	CALL	LHLI
	SHLD	STRLG	; current string length
	XCHG		; also in DE
	POP	H
	PUSH	H
	INX	H
	INX	H
	SHLD	STRBA	; string base address
	POP	B		; pointer to string length
	XRA	A		; carry clear, Z set
	RET

; We do some fancy footwork to allow string 'scalars'.

STRINGSCALAR
	MVI	A,40H	; special setting
	STA	MATFLG
	JMP	STRINGARRAYS
	PAGE
; **	Called after stlk and dimension table set up,
; **	the end of dimension table marker
; **	and rest of matrix sym.tab. entry set up.
; **	Number of element in DE, sym.tab. ptr to max in HL.

DIMM	PUSH	H
	CALL	fpsizB
	CALL	IMUL
	XCHG
	POP	H
DIMMX	CALL	DSTOR
	XCHG
	INX	H		; for dim table end marker.
	INX	H
	JMP	ASTAB	; The marker is a zero.

DIMMST	LXI	H,20	; 12 + 4*2
	CALL	ASTAB	; gimme some room
	LXI	D,12	; total space req for "arrray"
	CALL	DSTOR
	LXI	D,1		; 1-dimensional
	CALL	DSTOR
	LXI	D,12	; 10 chars (+2 for STRLG)
	CALL	DSTOR
	XCHG
	DCX	H
	DCX	H
	SHLD	STRMX	; real maximum length
	LXI	H,0
	SHLD	STRLG	; current length
	XCHG
	INX	H
	INX	H
	PUSH	H
	CALL	DSTOR
	MOV	B,H
	MOV	C,L
	SHLD	STRBA
	POP	B
	POP	PSW		; This gets rid of extra RET.
	XRA	A
	STA	MATFLG
	RET
	PAGE
; **  Symbol table lookup.
; **
; **	BC contains the variable name and type
; **	If not found then create entry and set carry
; **	HL has address after pointer on return
; **	If entry created then STNBA gets base addr of new entry
; **	and STNPTR gets address of pointer to new entry

STLK	MOV	A,B		; Compute chain pointer addr.
	CPI	'#'
	JNZ	STLKX
	MVI	A,52	; specially done for # indexing var
	JMP	STLKY

STLKX	SUI	'A'
	ADD	A
STLKY	LHLD	NOSE
	INX	H
	CALL	ADR	; HL contains chain pointer

VARIABLELOOKUP
	MOV	D,H
	MOV	E,L
	CALL	LHLI	; address of next var in chain
	MOV	A,H
	ORA	L
	JZ	NEWENTRY
	MOV	A,C
	ANI	del		; are we deleting?
	MOV	A,M		; get symbol tag...
	JNZ	DelENTRY
	CMP	C		; is this the one?
	JZ	SUCCESSFUL
	XRI	del		; don't care if it's deleted
	CMP	C		; check again.
	INX	H
	JNZ	VARIABLELOOKUP

; undelete a entry

	DCX	H
	MOV	M,C		; by putting in new tag
	ORA	A		; clear carry, nz says undelete

; we are entered with flags set:
;	carry is clear, plus z set for normal successful
;	lookup else it's an undelete

SUCCESSFUL
	INX	H
	INX	H		; successful lookup
	INX	H
	RET		; scram

DelENTRY
	ORI	del		; mark as deleted for compare
	ANA	C		; strip off other bits
	CMP	C		; is it the right type
	JNZ	DelENTRY2
	MOV	A,M
	ORI	del		; OR in deleted bit
	MOV	M,A		; resave
DelENTRY2
	INX	H
	JMP	VARIABLELOOKUP

NEWENTRY
	MOV	A,C		; symbol type
	ANI	del		; are we deleting
	RNZ		; if so, just scram
	LDA	DIMFLG	; create an entry if...
	ORA	A		;	dimensioning
	JNZ	CREATENTRY
	MOV	A,C		;	or variable type
	ANI	MTYPE	;	is not a matrix
	JZ	CREATENTRY
	LDA	MATFLG
	INR	A		; MAT lookup?
	JNZ	DMERR	; jmp/ nope, oops!!
	STC		; tell'em no such animal
	RET		; and leave

CREATENTRY
	PUSH	D
	LXI	H,3
	CALL	ASTAB	; allocate space in the chain
	SHLD	STNBA
	XCHG
	POP	H
	SHLD	STNPTR
	CALL	DSTOR
	XCHG
	MOV	M,C
	INX	H
	INX	H
	INX	H
	MVI	A,1
	STA	UNDEF	; in case of error
	STC
	RET

DelFNDEF
	MVI	C,del+FTYPE
	MVI	B,'A'	; start with A
DelFN2
	CALL	STLK	; delete this one
	INR	B		; next letter
	MOV	A,B
	CPI	'Z'+1
	JC	DelFN2	; spin thru thru Z inclusive
	RET
	PAGE
; **	String assignment controller.

STASS	PUSH	H		; Base address.
	PUSH	PSW	; Termination chr
	PUSH	B		; Symtab ptr to lg.
	LHLD	STRMX
	PUSH	H
	CALL	SEXPR	; A still has termination flag.
	XCHG		; Source length to DE.
	LHLD	TSTKA
	DAD	D
	SHLD	TSTKA
	LXI	H,0
	DAD	SP		; Puts address of STRMX in HL.
	CALL	DCMP
	POP	H		; Remove STRMX from stack.
	JC	STAS2
	XCHG

; The smaller of source length and
; destination max is in DE now.

STAS2	POP	H		; Ptr to log size in dest.
	POP	PSW		; Get c condition 
	CNC	DSTOR	; Modify logical length
	POP	B		; Base address of result.
	LHLD	TSTKA

; HL has first char on arg stack (if any).
; DE has count (possibly 0).
; BC has base address of dest.

STAS4	MOV	A,D
	ORA	E
	RZ
	MOV	A,M
	STAX	B
	DCX	H
	INX	B
	DCX	D
	JMP	STAS4

;**********
