;
; An authorization processor for the System 88.
; When hooked up as Auth.OV on the system disk, it asks
; the user for his name.  This name must appear on the
; authorization list kept in the program.  If the name is
; authorized, the user is asked for the password, which
; must match that kept in the authorization list.  If that
; matches, we return.  If we don't know this user,
; we keep giving him the oportunity to guess names.
;
; This Logon is similar to good old LOGON on CP-V.
;
; Logon for the System 88 designed and written by R. T. Martin
;
; Fixed several years later by Brian Smith.
;
; 06/10/81  FEA  Fixed bug which didn't search past CR for name
;
	REFS	SYSTEM
	REF
	IF	USERS=1
giveup	MACRO
#L	EI
	HLT
	ENDM
	ENDIF
;
;
; character codes.
;
CR	EQU	0DH
DEL	EQU	7FH
;
	ORG	OVRLY
	IDNT	$,$

	DB	'Auth'	; our name.
;
Tail	EQU	OVRLY+7FFH
;
; The start vector takes us to Logon, the reenter point
; takes us to the authorization processor
;
	PUSH	PSW
Aok	CALL	Killi	; kill command files and type-ahead
	POP	PSW
	ORA	A
	JZ	Logon	; 0 -> Logon authorization
	CPI	1
	JZ	Auth	; 1 -> go authorize users
	RET		; else split.
;
;	Clear the access we got on entry.
;
EXIT
	IF	USERS=2
	PUSH	PSW
	LDA	SYSRES
	MOV	C,A
	clrt	wrt
	JC	Err
	POP	PSW
	ENDIF
	RET
;
; Various messages we use
;
Greet	db	'Who are you? ',0
Noname	db	'I don''t know you.',0
Pass	db	'What is your password? ',0
Nope	db	'No user with that name.',0
Gotone	db	'That user is already authorized; use D or C',0
Bmsg	db	'Incorrect password: command ignored.',0
;
; Authorization menu/prompt
;
Amsg	DB	'Commands are:',CR,'	A - Add user',CR
	DB	'	D - Delete user',CR
	DB	'	L - List users',CR
	DB	'	X - eXit',CR
	DB	'	C - Change password',CR,0
Nmsg	DB	'User name:',0
Psmsg	DB	'Password:',0
Npmsg	DB	'New password:',0
Me	DB	4,'AuthOV'	; ourselves as a file
;
; Random number generator
;
Rnd	PUSH	D
	PUSH	B
	LHLD	Seed
	MVI	B,17
	MOV	A,L
Rnd1	ANI	2DH
	JPO	Rnd2
	CMC
Rnd2	MOV	A,H
	RAR
	MOV	H,A
	MOV	A,L
	RAR
	MOV	L,A
	DCR	B
	JNZ	Rnd1
	SHLD	Seed
	INX	H
	POP	B
	POP	D
	RET
;
; Logon.  Make sure the system bail-outs are nailed shut.
;
Logon	DI
	LXI	H,Ioret
	SHLD	UVEC
	SHLD	PVEC	; disable control-Y and user int.
	XRA	A
	STA	SCHR	; disable front panel entry
	EI
	LXI	H,Greet	; greeting for the user
	CALL	Msg	; greet and prompt
	LXI	D,Zero	; no prompt for Rlwe
	LXI	H,CMND	; read into here
	LXI	B,64	; length of buffer
	CALL	Rlwe	; read from the user
;
; Search Ulist for the name now in CMND.  Each entry in Ulist
; consists of a variable-length name terminated with a CR,
; followed by 16 bytes of encoded password.  The list is
; terminated by a 00 byte.
;
	CALL	Find	; go find it
	JNC	Getpw	; jmp/we found it.
;
; Wrong.  You lose.  Not on Ulist, or wrong password
;
Wrong	LXI	H,Noname
	CALL	Msg	; nope. You can't do it.
	JMP	Logon	; Give another chance.
;
; Get the password from the user.
;
Getpw	PUSH	H	; stash on the stack
	LXI	H,Pass
	CALL	Msg	; first part
	LXI	D,Zero	; no prompt, already printed it
	POP	H	; name pointer
	CALL	Valid	; go look and see if same
	JNZ	Wrong	; jmp/nope.
;
; You win.
;
	LXI	H,Seed
	MVI	A,Tail/256+1
Getp3	MVI	M,0
	INX	H
	CMP	H
	JNZ	Getp3
	MVI	A,0CH
	JMP	WH1	; split thru the wormhole.
;
; Seed holder
;
Seed	DW	0	; plotz.
Owl	JMP	Owl	; for contents of WH1
;
; Authorization code.  Responsible for building and modifying
; Ulist.
;
Auth
	IF	USERS=2
	LDA	SYSRES
	MOV	C,A
	sett	wrt	; Get write access so we can modify.
	JC	Err
	LXI	H,EXIT	; Set up to clear access we got.
	PUSH	H
	ENDIF

	LDA	SCHR
	ORA	A
	RZ		; you must be enabled.
	LXI	H,Ioret
	SHLD	PVEC	; croak control-Y interrupt!
Auth0	LXI	H,Amsg
	CALL	Msg	; print the initial message
Auth1	MVI	A,':'
	CALL	WH1
	CALL	WH0	; get a chr
	CALL	Fold	; fold LC to UC
	CPI	'X'
	RZ		; split on X.
	CPI	'A'
	JZ	Add
	CPI	'D'
	JZ	Del
	CPI	'L'
	JZ	List
	CPI	'C'
	JZ	Change
;
; Command we didn't understand
;
	JMP	Auth0
;
; List the authorized users.  Note that we DON'T list the
; passwords, which are scrambled, anyway.
;
Ltxt	DB	'Authorized users:',CR,0
;
; Lmsg prints out stuff till a CR is hit, does pagination
;
Linc	INX	H
Lmsg	MOV	A,M
	CALL	WH1
	CPI	CR
	JNZ	Linc	; do next chr
	DCR	C
	RNZ		; return on cr if page not full
	MVI	C,14	; eset page count
	MVI	A,'.'
	CALL	WH1
	CALL	WH0
	MVI	A,7FH
	JMP	WH1	; return deleting pagination '.'
;
List	LXI	H,Ltxt
	MVI	C,14	; pagination counter
	CALL	Lmsg
	LXI	H,Ulist
List1	MOV	A,M
	ORA	A
	JZ	Auth1	; split if end of list hit
	CALL	Lmsg	; else print entry
	LXI	D,17
	DAD	D
	JMP	List1	; point to next one and list.
;
; Subroutine to search for the name in CMND in Ulist.  If
; found, return HL pointing at the name, and C clear.  If
; not found, return C set and HL pointing to 00 terminator.
;
Find	LXI	H,Ulist
Find1	PUSH	H
	LXI	D,CMND	; look at this thing
Find2	LDAX	D
	CMP	M
	JNZ	Find3	; jmp/not it.
	CPI	CR
	JZ	Find5	; jmp/this is it!
	INX	H
	INX	D
	JMP	Find2
Find3	POP	H
	MOV	A,M	; did we hit end of list?
	ORA	A
	JNZ	Find41	; jmp/continue search
	STC		; set carry-not on list
	RET
Find4	INX	H
Find41	MOV	A,M
	CPI	CR
	JNZ	Find4	; skip to end of this entry
	LXI	D,17
	DAD	D
	JMP	Find1	; look at next one
Find5	POP	H	; this is the one
	XRA	A	; clear carry
	RET		; and split.
;
; Scramble the password in CMND based on the name pointed to
; by HL.  Don't change HL.
;
Scram	PUSH	H
	XCHG
	LXI	H,0	; seed we build
	MVI	B,0	; adder
Scram1	LDAX	D
	MOV	C,A
	DAD	B	; add in character
	INX	D
	CPI	CR
	JNZ	Scram1	; through the thing.
	MOV	A,L
	ORI	1	; insure odd
	MOV	L,A
	SHLD	Seed	; stuff in as the seed
;
; Now scramble the contents of CMND
;
	LXI	H,CMND
	MVI	C,16
Scram2	PUSH	H
	CALL	Rnd
	MOV	A,L
	POP	H
	XRA	M
	MOV	M,A
	INX	H
	DCR	C
	JNZ	Scram2
	POP	H	; restore name pointer
	RET		; and split.
;
; Validate the password for the name pointed to by HL.  Use
; the prompt pointed to by DE, and read into CMND.  Return
; with Z flag set if a match, NZ set of no match.
;
Znargh	CPI	CR	; echo CRs
	JZ	Owl
	CPI	DEL	; and DELETES
	JZ	Owl
	MVI	A,'?'	; fake WH1 routine!
	JMP	Owl	; just put out crud.
;
Valid	PUSH	H	; save name pointer into Ulist
	XCHG
	CALL	Msg	; print the prompt!
	DI		; we steal old wormhole for password
	LXI	D,Zero	; now no prompt for Rlwe, and also no
	LHLD	WH1+1	; echoing to the screen...
	SHLD	Owl+1
	LXI	H,Znargh
	SHLD	WH1+1
	LXI	H,CMND
	CALL	Rlwe	; read in password
	PUSH	H
	LHLD	Owl+1
	SHLD	WH1+1	; put back old wormhole
	MVI	A,CMND+64 AND 0FFH
	POP	H
Valid1	MVI	M,0
	INX	H
	CMP	L
	JNZ	Valid1	; zap rest of buffer
	POP	H	; get name pointer back
	PUSH	H	; stash gain
	MVI	A,7FH
	CALL	WH1
	MVI	A,18H
	CALL	WH1
	CALL	Scram	; scramble the entry
Valid2	MOV	A,M
	INX	H
	CPI	CR
	JNZ	Valid2	; point past the CR
	LXI	D,CMND	; now we want to compare.
	MVI	C,16	; length to compare
Valid3	LDAX	D
	CMP	M
	JNZ	Valid4	; split if no good
	INX	D
	INX	H
	DCR	C
	JNZ	Valid3	; look through whole thing
Valid4	POP	H	; restore name pointer!
	RET		; return with Z if ok.
;
; Delu is called to delete the entry pointed to by HL
; from the authorization list.
;
Delu	PUSH	H	; save this for later on!
Delu1	MOV	A,M
	INX	H
	CPI	CR
	JNZ	Delu1	; find end of name
	LXI	D,16
	DAD	D	; this is copy from
	XCHG		; in DE
	POP	H
	PUSH	H	; copy to in HL
;
; Copy an entry from DE to HL
;
Delu2	LDAX	D
	ORA	A	; end of list hit?
	JNZ	Delu4	; jmp/nope
;
; Zero to Tail hit.
;
	MVI	A,Tail/256
Delu3	MVI	M,0
	INX	H
	CMP	H	; see if we hit it yet.
	JNZ	Delu3	; jmp/nope
	POP	H
	RET		; return to caller
Delu4	LDAX	D
	MOV	M,A
	INX	H
	INX	D
	CPI	CR
	JNZ	Delu4	; copy name
	MVI	C,16
Delu5	LDAX	D
	MOV	M,A
	INX	H
	INX	D
	DCR	C
	JNZ	Delu5	; copy 16 byte password
	JMP	Delu2	; and then look at next entry
;
; Get name and do a Find on it.
;
Gnf	LXI	H,CMND
	LXI	D,Nmsg
	LXI	B,64
	CALL	Rlwe	; read in name
	JMP	Find	; return through Find
;
; Add user to Ulist
;
Add	CALL	Gnf	; go get name & look it up
	JC	Add1	; jmp/good, not there yet.
	LXI	H,Gotone
	CALL	Msg
	MVI	A,CR
	CALL	WH1
	JMP	Auth1
Add1	LXI	D,CMND
	PUSH	H	; save beginning of name pointer
Add2	LDAX	D
	MOV	M,A
	INX	D
	INX	H
	CPI	CR
	JNZ	Add2	; copy name into Ulist
	POP	D	; beginning of name pointer
	PUSH	H	; save end of name pointer!
	XCHG		; put in HL for Valid
	LXI	D,Psmsg	; prompt
	CALL	Valid	; go try to validate it and fail!
	POP	H	; copy to for password
Add2a	LXI	D,CMND
	MVI	C,16
Add3	LDAX	D
	MOV	M,A
	INX	D
	INX	H
	DCR	C
	JNZ	Add3	; copy password into Ulist
;
; Now punt us out again.
;
Punt	LXI	H,Me	; ourselves on the disk
	LDA	SYSRES	; on this drive
	CALL	Look	; see if we exist as INITIAL
	JC	Crot	; jmp/no, well, scram then.
	LXI	H,7	; point to FDA in the FDE
	DAD	D	; FDE start passed to us in DE
	MOV	E,M
	INX	H
	MOV	D,M	; that's FDA for the write
	INX	H
	LDA	SYSRES
	MOV	C,A	; this drive
	MOV	A,M	; # of sectors to write out.
	XCHG		; put disk addr in HL
	LXI	D,OVRLY	; memory address in DE
	MVI	B,0	; write
	CALL	Dio	; go do it
	JNC	Auth1	; jmp/ok, we're updated.
;
; Crot is jumped to to print out an error message, either on
; an I/O error, or an internally declared error
;
Crot	CALL	Ovrto
	DB	'Emsg'	; else gripe out error message
	MVI	A,CR
	CALL	WH1	; and a cr
	JMP	Auth1	; and back to the loop.
;
; Delete a user
;
Del	CALL	Gnf	; get name and find
	JNC	Del1	; jmp/found it
Nogot	LXI	H,Nope
	CALL	Msg
	MVI	A,CR
	CALL	WH1
Del1	LXI	D,Psmsg
	CALL	Valid	; go get password!
	JNZ	Del2	; jmp/junk
	CALL	Delu	; delete this entry
	JMP	Punt	; punt us out to disk and split
Del2	LXI	H,Bmsg
	CALL	Msg
	MVI	A,CR
	CALL	WH1
	JMP	Auth1
;
; Change the password for a user.  Note that the OLD password
; is required....
;
Change	CALL	Gnf	; get name and find it
	JC	Nogot	; jmp/no such user, scram.
	LXI	D,Psmsg
	CALL	Valid	; go get old password!
	JNZ	Del2	; jmp/wrong, I won't do it.
	LXI	D,Npmsg	; new password message
	CALL	Valid	; get it, scramble it.
Change1	MOV	A,M
	INX	H
	CPI	CR	; skip to end of name
	JNZ	Change1
	JMP	Add2a	; then go copy in the new password
;
; Here lies the User list.
;
Zero	DB	0
Ulist	EQU	$
	DS	Tail-Ulist	; reserve us some space
;
; The End.
;
	END
