;********************************************************
;*							*
;*    Emedit - an editor for error message overlays	*
;*							*
;*	Version 1.8			R. T. Martin	*
;*	03/03/81					*
;*							*
;*	Copyright (C) 1981, Interactive Products	*
;*  Corporation, dba PolyMorphic Systems, 460 Ward	*
;*  Drive, Santa Barbara, CA  93111			*
;********************************************************
;
; Last updated:
;	12/4/79		RTM	Two user system
;	1/3/80		RTM	Disaster recovery
;	03/03/81	BFS	Combine for Single and Twin
;				and Clean up.  Removed
;				self-modifying code.
;
; System references

	REFS	SYSTEM
	REF

FALSE	EQU	0
TRUE	EQU	NOT FALSE
CR	EQU	0DH		; carriage return
TAB	EQU	9		; tab chr

	MACLIST	2
	PAGE
	ORG	USER
	IDNT	USER,USER	; and away we go.

; The first thing we do is see if we are Enabled

	LDA	SCHR
	ORA	A	; enabled?
	RZ		; split if not!
	LXI	H,0
	DAD	SP
	SHLD	ESP	; save entry stack pointer for errors
	XRA	A
	STA	CHANGED	; mark nothing changed
	STA	FILE	; and no file open.
	LXI	D,VMSG	; version message and PROMPT text
	JMP	Loop1	; leap into loop
;
;	Loop is the command loop.  We put up a menu and get a
;  command from the user.  Loop1 is an internal entry used the
;  first time through for displaying the version text as well.
;
Loop	LXI	D,PROMPT
Loop1	XCHG
	CALL	Msg	; print stuff
	CALL	WH0	; get an input character
	;
	; Pick up first chr from buffer and scan command table
	;
	CALL	Fold	; fold to UC
	LXI	H,Table
	MOV	B,A	; put over here
	;
	; Scan Table until match or zero found
	;
Loop2	MOV	A,M	; get entry
	ORA	A
	JZ	Loop4	; jmp/not quite found.
	CMP	B
	JZ	Loop3	; jmp/found it
	INX	H
	INX	H
	INX	H	; point to nxt one
	JMP	Loop2
	;
	; We found it, call the thing attached.
	;
Loop3	LXI	D,Loop
	PUSH	D	; return point
	INX	H
	MOV	E,M
	INX	H
	MOV	D,M
	XCHG
	PCHL		; Call it.
;
; If not found, re-display commands.
;
Loop4	LXI	D,CMMSG
	JMP	Loop1
	PAGE
; The command table
; A character followed by the routine address, terminated by
; a zero byte.

Cmd	MACRO
#L	DB	#1
	DW	#2
	ENDM

Table	Cmd	'V',View	; V -> View error messages
	Cmd	'R',Repl	; R -> Replace error message
	Cmd	'D',Del		; D -> Delete error message
	Cmd	'A',Add		; A -> Add error message
	Cmd	'X',Exit	; X -> eXit
	Cmd	'E',Edit	; E -> Edit file name:
	Cmd	'S',Sort	; S -> Sort entries in the file
	Cmd	'L',List	; L -> List messages to printer

; End of table

	DB	0,0
	PAGE
;	Oops is internal error recovery for us.  Punt is
;  internal point for restoring stack and bailing back into
;  Loop.
;
Poke	XRA	A
	STA	FILE	; so no file marked open!
Oops	CALL	Gover
	DB	'Emsg'	; spit out error message
	MVI	A,CR
	CALL	WH1	; put out a CR
Punt	LHLD	ESP
	SPHL		; recover stack
	JMP	Loop	; and back to Loop
;
;	Check is called to see if we have a file open for
;  editing, as noted by FILE being zero or no-zero.  If no
;  file is open, we bail out with an error, else we return.
;
Check	LDA	FILE	; see if open
	ORA	A
	RNZ		; split/yup
	LXI	H,NMSG
	CALL	Msg
	JMP	Punt	; print message and scram.
	PAGE
;	Getn is called to get a hex number for an error
;  message.  It prompts the user and gets a line input,
;  crunches into a hex number returned in DE and HL pointing
;  to the message in buffer, if C clear. If C set, we're
;  pointing to the end.
;
Getn	CALL	Msg
	LXI	H,BUF	; input buffer
	LXI	D,GMSG	; prompt
	LXI	B,10	; 10 chrs max
	CALL	Rlwe	; read it in.
	LXI	H,0	; zero number built
	LXI	B,BUF	; buffer pointer
Getn1	LDAX	B
	CALL	Fold	; insure upper case
	SUI	'0'
	JC	Getne	; jmp/done.
	CPI	10
	JC	Getn2
	SUI	7
	CPI	10
	JC	Getne
	CPI	16
	JNC	Getne
Getn2	DAD	H
	DAD	H
	DAD	H
	DAD	H	; shift over 4 bits for new digit
	ORA	L
	MOV	L,A	; put in new digit
	INX	B
	JMP	Getn1	; and loop
Getne	XCHG		; copy to DE
;
;	Now traipse through the text in the buffer looking for
;  this thing.
;
Find	LHLD	OBUF+7	; text pointer
	MVI	A,(OBUF-OVRLY)/256	; delta
	ADD	H
	MOV	H,A	; now points to where text really is.
	;
	; See if this msg is the one, or is the end.
	;
Getn4	MOV	A,M
	CMP	D
	INX	H
	JNZ	Getn5	; nope, that not it.
	MOV	A,M
	CMP	E
	JZ	Getnf	; jmp/found it!
Getn5	DCX	H	; now pointing to start
	MOV	A,M
	INR	A
	JNZ	Getn5a
	STC		; set carry and
	RET		; split if end.
Getn5a	INX	H
	INX	H
Getn6	MOV	A,M
	INX	H
	ORA	A
	JZ	Getn4
	JMP	Getn6	; search for the end.
Getnf	DCX	H
	XRA	A	; clear carry,
	RET		; split pointing to message
	PAGE
;	D -> Delete a message from the file
;
Dtxt	db	'Delete ',0
Del	CALL	Check	; see if editing
	LXI	H,Dtxt
	CALL	Getn	; go get code to delete
	JNC	Del2	; jmp/found it.
;
; Huh is for reporting that code not found.
;
Huh	LXI	H,NOGOT
	JMP	Msg	; return thru Msg
;
; Croak the string
;
Del2	MVI	A,4
	STA	CHANGED	; mark changed.
	PUSH	H	; start of the message
	INX	H
	INX	H	; point past code fields
Del3	MOV	A,M
	INX	H
	ORA	A
	JNZ	Del3	; search for 00 terminator
	XCHG		; end point to DE
	POP	H	; start/copy to

; Copy (DE) to (HL) until FF end of message marker hit.

Del4	LDAX	D
	MOV	M,A	; first byte of code marker
	INR	A
	JZ	Dele	; jmp/looks like the end.
	INX	H
	INX	D
	LDAX	D
	MOV	M,A	; second code byte
	INX	H
	INX	D
Del5	LDAX	D
	MOV	M,A
	INX	D
	INX	H
	ORA	A
	JNZ	Del5	; copy whole message
	JMP	Del4	; back to top again
Dele	INX	H
	MVI	M,TRUE	; tromp terminator in
Dele1	INX	H
	MOV	A,H
	CPI	BUF/256
	RZ
	MVI	M,FALSE
	JMP	Dele1	; croak rest of stuff out.
	PAGE
; A -> Add a message. If it already exists, we'll delete it.

Atxt	db	'Add ',0
Add	CALL	Check
	LXI	H,Atxt
	CALL	Getn	; go get number
Add0	JC	Addnew	; go add new message

; The message already exists. Call Del2 to croak the old
; message.  Note tacit assumption of only one with that code
; exists in the buffer!

Adel	PUSH	D	; save error code
	CALL	Del2	; go delete old message
	POP	D	; get code back
	CALL	Find	; try and find message
	JNC	Adel	; we found it?!? Go croak same!

; Add new message to end, now pointed to by HL. Don't let
; the message overflow past BUF.

Addnew	MOV	M,D
	INX	H
	MOV	M,E
	INX	H	; store away code stuff
	SHLD	TAIL	; use for butt pointer
	MVI	A,3
	STA	CHANGED	; mark changed.
	LXI	H,ATERM
	CALL	Msg	; tell 'em how to terminate it.
Aread	LXI	D,APR	; add prompt text
	LHLD	TAIL	; read pointer
	LXI	B,120	; arbitrary upper limit
	CALL	Rlwe	; read me in a line
	INX	H
	SHLD	TAIL	; new end point
	DCX	H	; chr before CR
	DCX	H
	MOV	A,M
	CPI	1BH	; is it an ESC?
	JZ	Aend	; jmp/yup, go tromp down end.
	MOV	A,H
	CPI	BUF/256	; we blown it yet?
	JNZ	Aread	; jmp/nope

; Trunk!

	LXI	H,TRUNK	; oops, things getting too big.
	CALL	Msg	; tell 'em the news
	LXI	H,BUF-3	; to tromp down end
Aend	MVI	M,FALSE	; terminator
	INX	H
	MVI	M,TRUE	; end of list markers
	INX	H
	MVI	M,TRUE	; stomp down end marks
	RET		; and split
	PAGE
; R -> Repace text

Rtxt	db	'Replace ',0
Repl	CALL	Check
	LXI	H,Rtxt
	CALL	Getn	; go get a number
	JC	Huh	; jmp/not found?
	JMP	Add0	; jmp/join common path!
	PAGE
; V -> View the file.  Toss crud to the screen
; L -> List the file to the printer

Ltxt	db	'List',CR,0
List	CALL	Check
	LXI	H,Ltxt
	CALL	Msg
	MVI	A,TRUE
	STA	PFLAG	; Printer flag set.
	JMP	Viewa	; go to main line

; Vout is called rather than WH1, so we can swap things.

Vout	PUSH	B
	MOV	B,A
	LDA	PFLAG
	ORA	A
	MOV	A,B
	POP	B
	JZ	WH1
	JMP	WH7

Vmsg	MOV	A,M	; like Msg, but counts CR's
	ORA	A
	RZ		; return if end of msg hit.
	CALL	Vout
	INX	H
Vpow2	PUSH	B
	MOV	B,A
	LDA	PFLAG
	ORA	A
	MOV	A,B
	POP	B
	JNZ	Vmsg	; If printing, don't pause.
	CPI	CR
	JNZ	Vmsg
	DCR	B	; B is our CR counter
	JNZ	Vmsg
	MVI	A,'.'
	CALL	WH1
	CALL	WH0	; get a character to page display
	CALL	Fold
	CPI	'X'	; see if they want to stop
	JNZ	Vmsg2	; jmp/nope
	LHLD	ESP
	SPHL
	MVI	A,7FH
	CALL	WH1
	JMP	Loop	; blow out the side door.
Vmsg2	MVI	A,7FH
	CALL	WH1	; croak that
	MVI	B,14	; reset count
	JMP	Vmsg	; continue with display.
VHDR	db	'Error code ',0
VCR	db	CR,0

Vtxt	db	'View',CR,0
View	CALL	Check	; see if this is reasonable
	MVI	A,FALSE
	STA	PFLAG	; Not printing.
	LXI	H,Vtxt
	CALL	Vmsg	; print initial stuff
Viewa	LHLD	OBUF+7
	MVI	A,(OBUF-OVRLY)/256
	ADD	H
	MOV	H,A	; offset the stuff
	MVI	B,14	; set initial pagination counter
View1	MOV	D,M
	MOV	A,D
	INR	A
	RZ		; split-end of list.
	INX	H
	MOV	E,M
	INX	H
	PUSH	H	; save ptr
	LXI	H,VHDR
	CALL	Vmsg	; print header followed by
	CALL	DEout
	LXI	H,VCR
	CALL	Vmsg	; a carriage return, and then
	POP	H
	CALL	Vmsg	; print the text.
	INX	H
	PUSH	H
	LXI	H,VCR
	CALL	Vmsg
	POP	H
	JMP	View1	; print next one
;
;	Number in DE printed through Vout.
;
DEout	MOV	A,D
	CALL	De1
	MOV	A,E
De1	PUSH	PSW
	RRC
	RRC
	RRC
	RRC
	CALL	De2
	POP	PSW
De2	ANI	15
	ADI	90H
	DAA
	ACI	40H
	DAA
	JMP	Vout
	PAGE
; E -> Edit file.  See if we have a changed file, and if so
; call Cleanup to take care of that.

Edit	LDA	CHANGED
	ORA	A	; well?
	CNZ	Cleanup	; call Cleanup to clean it up.

; Goose Gfid to get us a file to edit.

	LXI	H,FMSG	; primpt string
	LXI	D,FILE	; place to buld FDE
	LXI	B,'OV'	; default extension
	MVI	A,0C0H	; read, look, ext.
	gfid
	JC	Oops	; oops... trouble...

; Now validate the file.  It must be .OV, and have LA=SA=2000H
; or we compain.

	LXI	H,FILE+1
	MOV	A,M
	ANI	31	; name length
	CPI	4
	CNZ	Wrong	; jmp/must be 4, turkey!
	INR	A
	MOV	E,A
	MVI	D,0
	DAD	D	; point to ext
	MOV	A,M
	CPI	'O'
	CNZ	Wrong	; ext not 'OV'
	INX	H
	MOV	A,M
	CPI	'V'
	CNZ	Wrong	; ext not 'OV'
	LXI	D,5
	DAD	D	; point to LA,SA
	MOV	A,M
	ORA	A
	CNZ	Wrong	; first byte should be 00
	INX	H
	MOV	A,M
	CPI	OVRLY/256
	CNZ	Wrong	; should be OVRLY
	INX	H
	MOV	A,M
	ORA	A
	CNZ	Wrong
	INX	H
	MOV	A,M
	CPI	OVRLY/256
	CNZ	Wrong	; jmp/you loose.

; Well, name, SA, and LA look ok. Let's read it in and check
; the pointer to the text.

	LXI	H,OBUF
	MVI	A,BUF/256	; end of overlay buffer
Ez	MVI	M,0
	INX	H
	CMP	H
	JNZ	Ez	; pre-zero the area.
	LHLD	FILE+8	; FDA from FILE 'cause we know length!
	LXI	D,OBUF	; read into here.
	LDA	FILE
	ANI	7FH	; Clear extesnion presence.
	MOV	C,A
	MVI	B,1	; read stuff
	LDA	FILE+0AH	; DNS in FDE
	CALL	Dio	; read it in!
	JC	Oops	; nope, that won't do.
	LHLD	OBUF+7	; text pointer
	MOV	A,H
	CPI	(OVRLY+2048)/256
	JNC	Wrong	; must be in 2K range
	CPI	OVRLY/256
	RNC		; split-looks OK.

; Wrong.  Not an error message overlay

WMSG	db 'That''s not an error message overlay!',CR,0
Wrong	LXI	H,WMSG
	CALL	Msg	; give 'em the bird.
	XRA	A
	STA	FILE	; mark nothing in for editing.
	JMP	Punt	; scram.
	PAGE
; X -> eXit the program.
;
Extx	db	'Exit',CR,0
Exit	LXI	H,Extx
	CALL	Msg
	LDA	CHANGED
	ORA	A
	CNZ	Cleanup	; clean up if changed
	LHLD	ESP
	SPHL
	RET		; else just split.
;
; Cleanup is called to clean up and write out the updated
; error message overlay.
;
Cleanup	LXI	H,CMSG
	CALL	Msg	; say cleaning up
	DI
	LXI	H,BUF-1	; start scan for tail here
Cl1	MOV	A,M
	ORA	A
	DCX	H
	JZ	Cl1	; search for the end of the critter
	INX	H	; point at last non-zero byte
	SHLD	TAIL	; save as tail address
	INR	H
	MOV	A,H
	SUI	OBUF/256	; calc # sectors
	STA	SIZE	; save as size
	MOV	B,A
	LDA	FILE+0AH	; size of current file
	CMP	B
	JC	Cl2	; jmp/got bigger

; The file is same size or smaller...  We'll punt the new
; contents to the same place.

	LHLD	FILE+8	; FDA from FDE
	LXI	D,OBUF	; buffer address
	LDA	FILE
	ANI	7FH	; Clear flag.
	MOV	C,A	; device
	MVI	B,0	; write
	LDA	SIZE	; # sectors
	CALL	Dio
	JC	Oops	; gronk...trouble!
Clx	XRA	A
	STA	FILE
	STA	CHANGED	; all done, mark closed.
	RET		; and split.

; File size has changed.  We must delete the old one and
; create a new one

Cl2	LXI	H,FILE+1
	MOV	A,M
	ORI	80H	; set deleted
	MOV	M,A	; in the FDE
	DCX	H	; point to FILE
	MVI	A,81H	; replace this FDE
	PUSH	H
	gfid
	JC	Oops	; oops....big trouble	
	DI
	POP	H	; pointer to FILE back again
	INX	H
	MOV	A,M
	ANI	7FH	; clear deleted bit for enter!
	MOV	M,A
	DCX	H
	MOV	A,M
	ANI	7FH
	MOV	M,A	; and on EX presence as well
	LDA	NFDIR	; see if directory is in
	CMP	M
	JZ	Cl3	; jmp/its in all right	
	MOV	A,M
	INX	H
	CALL	Look	; to force in directory
	JNC	Cl2	; still there? croak the damn thing!

; Pick up place to put this thing from directory and put in
; FDE for Gfid.  Write out the data to that address

Cl3	LHLD	SBUF1+0DH	; first free sector
	SHLD	FILE+8		; put away over here
	LDA	SIZE
	STA	FILE+0AH	; new size to write
	LXI	D,OBUF	; memory address
	MVI	B,0	; write
	LDA	FILE
	ANI	7FH	; Clear flag.
	MOV	C,A	; device
	LDA	SIZE	; # sectors
	CALL	Dio	; write out
	JC	Oops	; bad news...
	LXI	H,FILE
	MVI	A,1
	gfid		; call Gfid to do enter
	JC	Oops	; nope, guess it wont.
	JMP	Clx	; split.
	PAGE
; S -> Sort the entries in the file.  Of marginal utility,
; but I'll do it anyway

Stxt	db	'Sort!',CR,0
Sort	CALL	Check
	LXI	H,Stxt
	CALL	Msg
	LXI	H,BUF
	SHLD	SEND	; set up sort end pointer
;
; Do an n-pass insertion sort on the stuff
;
Sort1	LXI	D,0FEFEH	; starting compare code
	LXI	H,0
	SHLD	TAIL	; lowest msg found so far
	LHLD	OBUF+7
	MVI	A,(OBUF-OVRLY)/256
	ADD	H
	MOV	H,A	; point into the real text
Sort2	MOV	A,M
	CPI	0FFH	; hit the end yet?
	JZ	Sort6	; jmp/yes, start pass 2 of sort.
	PUSH	H	; point to start of thing
	INX	H
	CMP	D	; see if this one is lower
	JC	Sort5	; jmp/yes it is, eat same	
	JNZ	Sort3	; else just skip this one
	MOV	A,M
	CMP	E
	JC	Sort5	; jmp/eat this one
;
; Skip over this one
;
Sort3	POP	H
	INX	H
	INX	H
Sort4	MOV	A,M
	INX	H
	ORA	A
	JZ	Sort2	; this one skipped
	JMP	Sort4
;
; This one is lower, remember that.
;
Sort5	POP	H
	PUSH	H
	SHLD	TAIL	; remember pointer over here.
	MOV	D,M
	INX	H
	MOV	E,M	; and copy its code in here
	JMP	Sort3	; then skip it.
;
; End of one pass thru stuff.  See what we found as low msg.
;
Sort6	LHLD	TAIL
	MOV	A,L
	ORA	H
	JZ	Sort8	; jmp/start pass 2, out of stuff.
	XCHG		; put pointer in DE
	LHLD	SEND	; end of sorted list right now
	LDAX	D
	MOV	M,A
	INX	H
	INX	D
	LDAX	D
	MOV	M,A
Sort7	INX	H
	INX	D
	LDAX	D
	MOV	M,A	; shuffle byte
	ORA	A
	JNZ	Sort7	; not the end yet
	INX	H
	SHLD	SEND	; new end
	MVI	M,0FFH
	INX	H
	MVI	M,0FFH	; so mark that.
	LHLD	TAIL
	CALL	Del2	; snuff this one out.
	JMP	Sort1	; do another pass.
;
; Sort is done. Shuffle stuff down from BUF,
;
Sort8	LHLD	SEND	; last thing copied
	INX	H
	INX	H	; to last FF byte
	INX	H
	PUSH	H
	POP	B	; put in BC for compare
	LHLD	OBUF+7
	MOV	A,H
	ADI	(OBUF-OVRLY)/256
	MOV	H,A	; make HL right for copy to
	LXI	D,BUF	; copy from
Sort9	LDAX	D
	MOV	M,A
	INX	H
	INX	D
	MOV	A,D
	CMP	B
	JNZ	Sort9
	MOV	A,E
	CMP	C
	JNZ	Sort9
	RET		; copy in and split.
	PAGE
; Text

FMSG	db	'Edit file name:',0
VMSG	db	0CH,'Emedit version 1.8',CR
CMMSG	db	'Commands are:',CR
	db	'	A	Add message',CR
	db	'	D	Delete message',CR
	db	'	E	Edit file',CR
	db	'	R	Replace message',CR
	db	'	V	View all messages',CR
	db	'	X	eXit program',CR
	db	'	S	Sort messages',CR
	db	'	L	List messages to printer',CR
	db	':',0
PROMPT	db	':',0
NMSG	db	'No file open for editing- use E first',CR,0
CMSG	db	'...Creating updated file...',CR,0
GMSG	db	'error code:',0
NOGOT	db	'I can''t find that message.',CR,0
ATERM	db	'Terminate new message with ESC-CR',CR,0
APR	db	'<',0
TRUNK	db	'Message truncated-Overlay is full!',CR,0

; Data areas

ESP	DW	0	; save entry stack pointer
PFLAG	DB	0	; Printer flag for Vout.
CHANGED DB	0	; data changed flag 0=not changed.
TAIL	DW	0	; end of beastie pointer
SIZE	DB	0	; size of beastie in sectors (<=8)
SEND	DW	0	; sort end pointer

; Buffers for the overlay

OBUF	EQU	($+256) AND 0FF00H	; force to page bound
BUF	EQU	OBUF+800H	; buffer for reading text.

; That's All, Folks!

	END
