;********************************************************
;*							*
;*		   Relocatable Debugger			*
;*							*
;*	Version 4.3			Stanley Reifel	*
;*	03/13/81			Lennie Araki	*
;*					Brian Smith	*
;*							*
;********************************************************
;
;	This is the last in the ndb, xdb, adb line of
;  debuggers.  The original ndb, xdb and adb were written by
;  R. T. Martin.  Albert's debugger was written by
;  Stanley Reifel as an improvement on the features of
;  ndb.  The relocating loader was written by
;  Lennie Araki and Brian Smith.  Brian Smith combined
;  Albert's debugger and the relocating loader to create RDB.
;
;		>>>>>>>> CHANGES <<<<<<<<<
;
; 7/21/80    REV 3.1A:
;		Memtop only lowed if greater than RamTop
;
; 2/10/81    REV 3.2A:
;		Added J - jump with temp break point
;		Made I skip instruction if in instruction mode
;		Made V add 8 or 64 to window address if no arg
;		Single step port not displayed in window P mode
;		Added string searching
;		Added T for setting memtop
;		Added W for warm starting Exec
;		Added Q for quit
;
; 2/11/81    REV 3.2A changed to REV 4.0A
;
; 2/11/81	Made it compatable with Single & Twin Users
;			(by Don Moe)
; 03/02/81   BFS 4.0	Relocatable!
; 03/02/81   BFS 4.1	Make UCHR null on QUIT.
; 03/03/81   BFS 4.2	Increase break points to 8 and change
;			display to handle them.
; 03/13/81   DKM 4.3	Properly display registers, even if
;			pointing to screen.
;
False	EQU	0
True	EQU	NOT False
;
	REFS	SYSTEM
	REF
	IF	USERS=2
MOVE	EQU	Moven
SCREEN	EQU	1800H
DEOUT	EQU	Deout
SHAREDMEM EQU	0E000H
	ELSE
MOVE	EQU	100H
	ENDIF
;
SingStp	EQU	0CH
BRKPNTS EQU	8	; 8 break points allowed.
;
ptrcont	MACRO		;moves four bytes from address pointed
			;to by named register to buffer area
			;for that register
	LHLD	#1	;get contents of register
	LXI	D,#1ptr	;point to buffer block
	MVI	B,4	;four bytes
#1loop	MOV	A,M	;copy them
	STAX	D
	INX	H
	INX	D
	DCR	B
	JNZ	#1loop	;until finished
	ENDM
;
	MACLIST	0
;
;
;	Ascii character codes.
;
TAB	EQU	9
CR	EQU	0DH
UP	EQU	11H
DOWN	EQU	12H
RIGHT	EQU	13H
LEFT	EQU	14H
CtrlU	EQU	15H
ESC	EQU	1BH
;
;
;
	ORG	USER		; USER program but assemble
	RELOC			; relocatably!
	IDNT	$,$
;
;	Start code.  Check to see if we have already been 
;  installed and if so just return.  Else copy code above
;  MEMTOP, relocate, and hookup into UVEC.
;
PROG	JMP	START		; Start and
	JMP	START		; Reenter vectors.
;
Prompt	db	'RDB 4.3 connected to Ctrl-U. Memtop at ',0
;
;============================================================
;		    set up routine
;
;
START
	IF	USERS=1
	LHLD	MEMTOP
	PUSH	H
	XCHG
	LHLD	UVEC		; (UVEC) < MEMTOP?
	ELSE
	LHLD	MEMTOP
	PUSH	H
	LXI	D,SHAREDMEM
	LHLD	UVEC
	CALL	DCMP
	JNC	ST1		; It's in shared, init it.
	XCHG			; UVEC to DE
	POP	H		; Wasn't in shared see if
				; above memtop.
	PUSH	H		; Get and save memtop
	XCHG			; UVEC to HL
	ENDIF
	CALL	DCMP		; See if installed.
	JC	ST1		; No, normal init.
	POP	H		; Get old memtop off stack
	LXI	H,0
	PUSH	H		; Put zero as flag to INIT.
	JMP	MOK
ST1	LHLD	MEMTOP		; Get memtop back in HL.
	LXI	B,Size		; Size of resident area.
	INR	B		; bump to boundary.
	MVI	C,0
	CALL	SubBC		; Subtract it.
	SHLD	MEMTOP		; Now move MEMTOP down.
	INX	H
	XCHG			; Destination is MEMTOP+1
	LXI	H,Code		; Source is code.
	LXI	B,-PROGLEN
	CALL	MOVE		; Copy code above MEMTOP.
	;
	;	Now relocate the code.
	;
	LHLD	MEMTOP		; Start of resident-1
	LXI	B,1-ProgLen	; (Length of non-resident)
	DAD	B		; Point to starting 
	XCHG			; Memory addr (DE)
	LXI	H,ProgLen+PROGLEN
	DAD	D
	SHLD	RELEND		; Set RELEND for RELOCS
	LHLD	MEMTOP
	INX	H		; Where code runs
	LXI	B,INIT		; (Assembly addr)
	CALL	SubBC		; Subtract the two.
	MOV	B,H
	MOV	C,L		; Relocation offset (BC)
	LXI	H,TAIL		; Relocation table (HL)
	CALL	RELOCS		; Relocate it!
;
MOK	LXI	H,Prompt	;tell'em were here
	CALL	Msg
;
	LHLD	MEMTOP		;tell'em the top of RAM
	XCHG
	CALL	DEOUT
;
	MVI	A,CR		;print a return
	CALL	WH1
	LHLD	MEMTOP
	INX	H
	XTHL			; Put address on stack
				; and get old memtop off
	RET			; Initialize and return.
;
;
;
;		MODIFIED Relocation Subroutine
; RELOCS
;	Relocate a file loaded into memory at DE
;  with HL pointing at the start of the relocation
;  block and RELEND pointing at the end of the program.
;  Offset is in BC.
;
RELOCS	SHLD	RELBLK	 	;Save HL in RELBLK.
	MOV	H,B
	MOV	L,C
	SHLD	RELOFF
	LHLD	RELBLK
	XCHG
	JMP	Reloc3		;Jmp get into the loop!
	;
Reloc1	INX	H		;Main loop
	INX	H		;Bump HL two.
	PUSH	D
	XCHG
	LHLD	RELEND		;Get end of code pointer.
	XCHG
	CALL	DCMP		;Present pointer - end.
	POP	D
	RNC			;Ret/ all done, no carry.
	;
Reloc2	DCR	C		;Dink bit count
	MOV	A,B		;Get bit pattern in A
	JNZ	Reloc4		;Jmp/ some bits left
Reloc3	LDAX	D		; else get a new byte
	INX	D
	MVI	C,8		; and reset bit count
Reloc4	RAL
	MOV	B,A		;Put bits back into B.
	JNC	Reloc1		;Jmp/ 0: just skip two
	;
	DCR	C		;Dink bit count
	MOV	A,B		;Get bit pattern in A
	JNZ	Reloc5		;Jmp/ some bits left
	LDAX	D		; else get a new byte
	INX	D
	MVI	C,8		; and reset bit count
Reloc5	RAL
	MOV	B,A		;Put bits back into B.
	JNC	Reloc6		;Jmp/ 0: relocate even
	INX	H		; else relocate odd.
Reloc6	PUSH	B
	PUSH	D
	MOV	C,M
	INX	H
	MOV	B,M		;Get old addr in BC
	XCHG
	LHLD	RELOFF		;Offset to add.
	DAD	B		;Add the two together.
	XCHG
	MOV	M,D
	DCX	H
	MOV	M,E		;Stick back in memory.
	POP	D
	POP	B
	JMP	Reloc1		;Back to main loop.
;
;	Data section
;
RELBLK	  DS	2
RELEND	  DS	2
RELOFF	  DS	2
;
;	***** UTILITIES *****
;
DCMP	MOV	A,H
	CMP	D
	RNZ
	MOV	A,L
	CMP	E
	RET
;
DSUB	MOV	A,L
	SUB	E
	MOV	L,A
	MOV	A,H
	SBB	D
	MOV	H,A
	RET
;
SubBC	MOV	A,L
	SUB	C
	MOV	L,A
	MOV	A,H
	SBB	B
	MOV	H,A
	RET
;
ProgLen EQU	$-PROG		; Length of non-resident code.
	PAGE
;
;============================================================
;		entry points into debugger
;
;
;	Initialize at load time, entered with old memtop in HL.
;
Code	EQU	$		; Start moving here.
;
INIT	MOV	A,H
	ORA	L
	JZ	INIT1		; If HL=0 leave OldMT alone.
	SHLD	OldMT		; Set old memtop for exit.
INIT1	LXI	H,DEBUG		;set UVEC to point to DEBUG
	SHLD	UVEC
	LXI	H,DEBUG1
	SHLD	SRA7		;also single & break point RST
;
	MVI	A,CtrlU		;set UCHR to a control U
	STA	UCHR
;
	XRA	A		;set window to no display
	STA	WindFlg
	STA	SingUse		;zap single step flag
	STA	TempBrk		;disable temp break points
;
	LXI	H,BrkBuf	;point break pointer to buffer
	SHLD	BrkPntr
	RET			;return to exec
;
;
;
DEBUG				;returns here after user rupt
	MVI	A,-1
	STA	SingUse
;
DEBUG1	MVI	A,-1		;disable UVEC
	STA	DONT
;
	LXI	H,10		;find the pos of the SP
	DAD	SP
;
	LXI	D,REGISTR	;save the registers in buffers
	MOV	A,L		;save the address of the SP
	STAX	D		;first the low byte
	INX	D
	MOV	A,H		;then the high byte
	STAX	D
	INX	D
;
	LXI	H,0		;get sp again
	DAD	SP
	MVI	C,10		;save all 10 other registors
SaveR	MOV	A,M		;get the byte on the stack
	STAX	D		;save the byte in a buffer
	INX	H		;point to next byte on stack
	INX	D		;point to next space in buffer
	DCR	C		;have all registers been saved?
	JNZ	SaveR		;nope, get the next byte
;
	LDA	SingUse		;we ret cause of SS or UVEC?
	ORA	A
	JNZ	WERT		;ya, no sweat
;
	LHLD	PC		;no, came back because of BP
	DCX	H		;dcr the program counter
	SHLD	PC	
;
WERT	EI
	XRA	A		;reset SingUse
	STA	SingUse
;
	LXI	H,SCREEN	;save the screen
	LXI	D,ScrnBuf	;store it here
	LXI	B,-400H		;move all 1K
	CALL	MOVE
;
	LHLD	POS		;save cursor pos
	SHLD	OldPos
;
	LXI	H,BrkBuf	;take out the break points
;
OutNext	LDA	BrkPntr		;any more break points
	CMP	L
	JZ	TempRst		;no, restore temp break points
;
	MOV	E,M		;get the break point address
	INX	H
	MOV	D,M
;
	INX	H		
	MOV	A,M		;get the data
	STAX	D		;save at break point
	INX	H		;point to next break point
	INX	H
	JMP	OutNext
;
;
TempRst	LDA	TempBrk		;check temp break point flag
	ORA	A
	JZ	HART		;not set so skip it
;
	LHLD	PCplus1		;get address of temp brk pnt
	LDA	TempIns		;get the instruction
	MOV	M,A		;restore it in memory
	XRA	A		;clear the temp break point flg
	STA	TempBrk
	JMP	HART		;all done
;
;============================================================
;		main command processor
;
HART	CALL	Flush		;kill the keyboard buffer
	CALL	REFRESH		;display registors & memory
	CALL	HEXC		;see what's on the user mind
;
	MOV	A,B		;get termination char
	CPI	'!'		;does user wanta set break pnt
	JZ	BrkPnt		;ya, go set a break point
;
	CPI	'%'		;does user want to clear BPs?
	JZ	ClrBP		;ya, go clear break points
;
	CPI	''''		;user wanta enter ASCII in mem
	JZ	ASCII 		;ya, go for it
;
	CPI	'.'		;do they want to modify memory?
	JZ	Mod		;ya, lets go modify RAM
;
	CPI	':'		;do they wanta set a register?
	JZ	RegSet		;ya, go find out which reg
;
	CPI	';'		;they what to set the window?
	JZ	WindSet		;ya, go set the window
;
	CPI	'G'		;do they wanta ret to program?
	JZ	Exit		;return to where ever
;
	CPI	'I'		;howbout indirect jump
	JZ	Indirct		;yes, point with at (window)
;
	CPI	'J'		;wanta  jump with break point?
	JZ	Jump		;ya, jump with break point
;
	CPI	'M'		;possibly a move
	JZ	Moveit		;yes sir, go move some stuff
;
	CPI	'O'		;maybe output to a port?
	JZ	Outer		;yo, get byte and port	
;
	CPI	'Q'		;could it be they want to quit?
	JZ	QUIT
;
	CPI	'R'		;do they wanta restore screen?
	JZ	Restore		;ya, go restore the screen
;
	CPI	'S'		;they wanta search for a byte?
	JZ	Search		;ya, go sniff it out
;
	CPI	'T'		;they wanta set MEMTOP?
	JZ	MemTop		;ya, ya, set it to whatever
;
	CPI	'V'		;they want to point the window
	JZ	VIEW
;
	CPI	'W'		;how about going to Exec?
	JZ	WrmExec		;ya, goto Exec
;
	CPI	'X'		;do they want to single step?
	JZ	Single		;ya, go run 1 instruction
;
	CPI	'Z'		;do they want to fill memory?
	JZ	Fill		;ya, go get the data
;
;
	LXI	H,HART		;return to HART
DIRECT	PUSH	H
	CPI	UP		;they what to move window up?
	JZ	Up		;ya, go do it
;
	CPI	DOWN		;move the window down?
	JZ	Down		;sure, go down
;
	CPI	RIGHT		;how about right?
	JZ	Right		;ok ,right it goes
;
	CPI	LEFT		;left?
	JZ	Left		;ya, move window left
	RET			;none of the above, try again
;
;=============================================================
;		     user commands
;
;the user wants to set a break point
BrkPnt	XCHG			;save the break point address
;
	MOV	A,E		;is the break point at zero?
	ORA	D
	JNZ	GetBP		;no, go get the break point
;
	LDA	WINDOW		;get address of the window
	MOV	E,A		;stuff it in break point buff
	LDA	WINDOW+1	;get addr high
	MOV	D,A
;
GetBP	LXI	H,BrkBuf
;
LokAgn	LDA	BrkPntr
	PUSH	H
	LXI	H,BrkBuf+(BRKPNTS*4)
	CMP	L
	POP	H
;
;	CPI	(BrkBuf+(4*5)) AND 0FFH ; replaced by above
; push lxi cmp pop.
;
	JZ	HART		;skip it since there is no room	;
	CMP	L
	JZ	GotSpce		;ok got space for it
;
	MOV	A,M
	CMP	E
	INX	H
	JNZ	Next
	MOV	A,M
	CMP	D
	JZ	HART		;skip it since it is a dup
Next	INX	H
	INX	H
	INX	H
	JMP	LokAgn
;
GotSpce	MOV	M,E		;ya, save BP address in buffer
	INX	H
	MOV	M,D
;
	INX	H		;update the buffer pointer
	LDAX	D		;save instruction code
	MOV	M,A		;in breakpoint buffer
	INX	H
	INX	H
	SHLD	BrkPntr
	JMP	HART
;
;
;clear the break points
ClrBP	LXI	H,BrkBuf	;point break pointer to buffer
	SHLD	BrkPntr
	JMP	HART
;
;
;the user wants to enter ASCII characters into RAM
ASCII	LDA	WindFlg		;get current window mode
	STA	OldWind		;save it for later
;
	MVI	A,2		;set the window for number mode
	STA	WindFlg
;
ASCIIpl	CALL	REFRESH		;udate screen with new mode
;
	LXI	H,MemMod	;tell we're in mem mod mode
	CALL	Msg
;
	LXI	H,ASCIIpl	;if goto DIRECT return here
	CALL	Flush		;kill the keyboard buffer
	CALL	WH0		;get the char from user
;
	CPI	ESC		;are they through?
	JZ	LdoneO		;ya, return to the hart
;
	CPI	UP		;is char less than an UP
	JC	STORE		;ya, not a cursor control
	CPI	LEFT+1		;is it less than a left
	JC	DIRECT		;ya, its a cursor control char
;
STORE	LHLD	WINDOW		;get address of the window
	MOV	M,A		;write char into memory
;
	LXI	H,ASCIIpl	;return to here
	PUSH	H
	JMP	Left		;go move cursor left
;
;
;user wants to enter the memory modify mode
Mod	LDA	WindFlg		;get current window mode
	STA	OldWind		;save it for later
;
	MVI	A,2		;set the window for number mode
	STA	WindFlg
;
ModPoll	CALL	REFRESH		;udate screen with new mode
;
	LXI	H,MemMod	;tell we're in mem mod mode
	CALL	Msg
;
	CALL	Flush		;kill the keyboard buffer
	CALL	HEXC		;see what's on the user mind
;
	MOV	A,B		;get termination char
	CPI	' '		;do they wanta set a register?
	JZ	STUFF		;stick L in memory
;
	CPI	ESC		;user finished?
	JZ	LdoneO		;ya, return to the hart
;
	CPI	'V'		;user wanta change window add?
	JZ	Viewer		;ya, set the window
;
	LXI	H,ModPoll	;return to Mode Poll
	JMP	DIRECT		;see if they hit an arrow
;
STUFF	MOV	A,C		;did they hit any numbers?
	ORA	A		;check the digit count
	JZ	GoLeft		;no, move left
;
	MOV	A,L		;get the byte from user
	LHLD	WINDOW		;stick it in at the window
	MOV	M,A
;
GoLeft	LXI	H,ModPoll	;return to Mode Poll
	PUSH	H
	JMP	Left		;move window left
;
LdoneO	LDA	OldWind		;restore window mode
	STA	WindFlg
	JMP	HART		;return to the hart
;
Viewer	MOV	A,H		;any argument given
	ORA	L
	JNZ	Viewer1		;ya, set window to that address
;
	LHLD	WINDOW		;get address of the window
	LXI	B,64		;add 64 for number mode
	DAD	B
;
Viewer1	SHLD	WINDOW		;get user address and store  
	JMP	ModPoll
;
;	RegSet sets a register pair.
;
RegSet	LXI	D,HL		;point to the reg in memory
	XCHG
;
	CALL	LCFLD		;get the register to be set
	CPI	'H'		;how about the HL?
	JZ	SetReg		;ya, set HL
;
	INX	H		;no, point to DE reg in memory
	INX	H
	CPI	'D'		;could it be the DE pair?
	JZ	SetReg		;ya, set DE
;
	INX	H		;no, point to BC reg in memory
	INX	H
	CPI	'B'		;what about the BC registers?
	JZ	SetReg		;ya, set the BC
;
	INX	H		;no, point to AF reg in memory
	INX	H
	CPI	'A'		;is it the A and flag byte?
	JZ	SetReg		;ya, set the A and flag byte
;
	INX	H		;no, point to PC reg in memory
	INX	H
	CPI	'P'		;do they want to set the PC
	JZ	SetReg		;ya, go set the PC
;
	JMP	HART		;none of above, so forget it
;
SetReg	MOV	M,E		;stick low user byte in reg mem
	INX	H
	MOV	M,D		;now the high user byte in reg
	JMP	HART		;all done
;
;	WindSet set the window mode and address.
;
WindSet CALL	LCFLD		;how do they want to set it
	CPI	'Z'		;do they want to blank the wind
	JZ	ZAP		;ok set the window to 0
;
	CPI	'I'		;do they want the instruction?
	JZ	DisWind		;ya, go set for disasm 
;
	CPI	'N'		;do they want the number mode?
	JZ	NumWind		;ya, go set for display numbers
;
	CPI	'P'		;they wanta disply input ports
	JZ	InWind		;ya, go change the window
	JMP	HART		;none of the above, so skip it
;
ZAP	XRA	A		;turn off the window
	STA	WindFlg
	JMP	HART
;
DisWind	MVI	A,1		;set the window for disasm mode
	STA	WindFlg
	JMP	NewWind
;
NumWind	MVI	A,2		;set the window for number mode
	STA	WindFlg
	JMP	NewWind
;
InWind	MVI	A,3		;set the window for input mode
	STA	WindFlg
;
NewWind	MOV	A,H		;if no was given don't change
	ORA	L		;both high and low zero?
	JZ	HART		;ya so skip it
	JMP	VIEW		;go save new address
;
;	Exit causes the debugger to return to the program
;  at (PC).
;
Exit	LXI	H,EXIT		;return to exit after SS
	SHLD	SRA7
	JMP	Single1		;go run 1 instruction
;
;	After running 1 instruction return to here.
;
EXIT	LXI	H,DEBUG1	;set single step back
	SHLD	SRA7
;
	LXI	H,BrkBuf	;insert the break points
;
GetNext	LDA	BrkPntr		;any more break points
	CMP	L
	JZ	BPin		;no
;
	MOV	E,M		;get the break point address
	INX	H
	MOV	D,M
;
	INX	H
	LDAX	D		;get instruct at BP for latter
	MOV	M,A		;save it in memory
;
	MVI	A,0FFH		;store a RST7 at BP
	STAX	D
	INX	H
	INX	H
	JMP	GetNext
;
BPin	XRA	A		;enable UVEC
	STA	DONT
	JMP	Ioret
;
;	User wants to point window at what the window is
;  pointing at.
;
Indirct LHLD	WINDOW		;get address of the window
	LDA	WindFlg		;are we in instruction mode?
	CPI	1
	JNZ	NotInst
	INX	H		;ya, skip the instruction
NotInst MOV	E,M		;get low byte pointed by window
	INX	H
	MOV	D,M		;get hing byte pointed by wind
	XCHG
	SHLD	WINDOW		;point the window here
	JMP	HART
;
;	User wants to stick break point at instruction
;  following PC.
;
Jump	LHLD	PCplus1		;get addr of instruct after PC
	MOV	A,M		;save old instruction
	STA	TempIns		;save break point instruction
	MVI	M,0FFH		;store a RST7 at BP
	MVI	A,-1		;set temp break point flag
	STA	TempBrk	
	JMP	Exit		;go run the code
;
;	User wants to move some data.
;
Moveit	MVI	A,18H		;clear the bottom line
	CALL	WH1
	LXI	H,Strting	;prompt'em for starting addr
	CALL	Msg
	CALL	HEXC		;get address
	MOV	A,C		;if no # entered return to HART
	ORA	A
	JZ	HART
	PUSH	H		;save starting address
	LXI	H,Ending	;prompt for ending address
	CALL	Msg
	CALL	HEXC		;get the ending address
	XCHG
	LXI	H,To		;prompt for the move to addr
	CALL	Msg
	CALL	HEXC		;get the move to address
	MOV	B,H		;save it in BC
	MOV	C,L
	POP	H		;get starting address back
MOVEM	MOV	A,M		;get byte
	STAX	B		;save at (BC)
	CALL	Dcmp		;done?
	INX	H		;but first inc the pointers
	INX	B
	JNZ	MOVEM		;no, not done, keep moving
	JMP	HART		;ya, all finished
;
;	User wants to stuff a byte to a port.
;
Outer	LXI	H,BITE		;prompt for byte
	CALL	Msg
	CALL	HEXC		;get the byte
	MOV	E,L		;save the byte
	MOV	A,C		;if no # entered return to HART
	ORA	A
	JZ	HART
;	
	LXI	H,Port		;prompt for the port
	CALL	Msg
	CALL	HEXC		;get the port
;
	MOV	A,L		;stuff port address instruction
	STA	Outer1+1	;* * *  YUCK * * *
	MOV	A,E		;get the byte
Outer1	OUT	0		;output it
	JMP	HART		;return to HART
;
;	Quit.  Restore MEMTOP, set UVEC to Ioret and UCHR to 0.
;
QUIT	LHLD	OldMT		;get old memtop
	SHLD	MEMTOP		;restore memtop
	LXI	H,Ioret		;point UVEC to Ioret
	SHLD	UVEC
	XRA	A
	STA	UCHR		; Make UCHR null.
	JMP	WrmExec		; Warm start exec.
;
;	Restore the screen then wait for the user.
;
Restore	LXI	H,ScrnBuf	;restore the screen
	LXI	D,SCREEN	;put data here
	LXI	B,-400H		;move all 1K
	CALL	MOVE
	CALL	WH0		;wait for the user
	JMP	HART
;
;	The operator wishes to search for a string in RAM.
;
Search	LXI	H,String	;prompt'em for the string
	CALL	Msg
;
	LXI	D,StrgBuf	;point to the string buffer
GetStrg	CALL	HEXC		;get the byte in L
	PUSH	H		;save the byte
;
	MVI	A,' '		;print a space
	CALL	WH1
;
	LXI	H,StrgBuf + StrSiz  ;buffer full?
	CALL	Dcmp
	POP	H		;get the byte back
	JZ	Search1		;ya, skip it
;
	MOV	A,L		;no, get the byte
	STAX	D		;save it in the buffer
	INX	D		;inc the buffer
;
	MOV	A,B		;get the termination char
	CPI	CR		;is it a turn?
	JNZ	GetStrg		;nope, get more string
;
Search1	XCHG			;save end of string address
	DCX	H		;dcr end string address
	SHLD	StrgEnd
;
	MVI	A,18H		;clear line
	CALL	WH1
;
	LXI	H,Strting	;prompt'em for starting addr
	CALL	Msg
	CALL	HEXC		;get address
	DCX	H		;move behind it
;
	SHLD	SchMem		;save search address
;
	LDA	WindFlg		;is the window turned off?
	ORA	A
	JNZ	Checker		;no, its ok
	MVI	A,2		;ya, set it to mode 2, numbers
	STA	WindFlg
;
Checker	LXI	H,Srching	;tell'em we're searching
	CALL	Msg
;
NoMatch	LHLD	SchMem		;get the search address
	INX	H		;inc it
	SHLD	SchMem		;update it
;
	MOV	A,L		;wrapped around to zero?
	ANA	H
	INR	A
	JZ	NoMtch		;ya, string not here
;
	LXI	D,StrgBuf	;point to the string
Search2	LDAX	D		;get from the string
	CMP	M		;does it match memory?
	JNZ	NoMatch		;nope, no cigar
;
	LDA	StrgEnd		;end of string?
	CMP	E		;1 byte comp since string small
	JZ	MtchUp		;ya, we got a winner
;
	INX	H		;inc pointer
	INX	D
	JMP	Search2		;look some more
;
NoMtch	LXI	H,0		;no match so point to 0
	JMP	Search4
;
MtchUp	LHLD	SchMem		;HL -> beginning of string
Search4	SHLD	WINDOW		;got a match
	PUSH	H
	PUSH	D
	CALL	REFRESH		;point window to match
;
	LXI	H,Contin	;tell'em to hit C to continue
	CALL	Msg
	POP	D
	POP	H		;restore search address
	CALL	WH0		;wait for the user
	CALL	Fold
	CPI	'C'		;is it a C?
	JZ	Checker		;ya, keep looking
	JMP	HART		;nope, return to the HART
;
;
;call to set memtop
MemTop	LXI	H,Address	;ask for the address
	CALL	Msg
;
	CALL	HEXC		;get the address
	MOV	A,H		;if null skip it
	ORA	L
	JZ	HART
	SHLD	MEMTOP		;save it
	JMP	HART
;
;
;VIEW set the window address
VIEW	MOV	A,H		;any argument given
	ORA	L
	JNZ	VIEW1		;ya, set window to that address
;
	LHLD	WINDOW		;get address of the window
;
	LDA	WindFlg		;check window mode
	DCR	A		;is it instruction mode?
	JZ	VIEW2		;ya, add for last instruction
;
	DCR	A		;is in number mode?
	JZ	VIEW3		;ya, add 64 to window
;
	LXI	B,8		;add 8 to window for ports
	DAD	B
	JMP	VIEW1
;
VIEW2	LHLD	NxtWind		;get last instruction address
	JMP	VIEW1
;
VIEW3	LXI	B,64		;add 64 for number mode
	DAD	B
;
VIEW1	SHLD	WINDOW		;get user address and store  
	JMP	HART
;
;
;warm start to Exec
WrmExec	LXI	H,Warm		;get warm start address
	SHLD	PC		;save at PC
	JMP	Exit		;goto execute
;
;
;Single single steps on instruction at (PC)
Single	MVI	A,-1		;set single step flag
	STA	SingUse

Single1	LXI	H,ScrnBuf	;restore the screen
	LXI	D,SCREEN	;put data here
	LXI	B,-400H		;move all 1K
	CALL	MOVE
;
	LHLD	OldPos		;restore cursor pos
	SHLD	POS
;
	DI			;don't bother me
	LXI	H,0		;get the current stack pointer
	DAD	SP
;
	LXI	D,HL		;stick reg from RAM on stack
	MVI	C,10
;
RestR	LDAX	D		;get byte form reg RAM
	MOV	M,A		;stick it on the stack
	INX	H		;inc the stacker pointer
	INX	D		;inc the reg RAM pointer
	DCR	C		;all registers been restored?
	JNZ	RestR		;no, keep going
;
	POP	H		;restore the registors
	POP	D
	POP	B
	POP	PSW
	EI
	OUT	SingStp		;start the interrupt circuit
	RET			;go to program
;
;
;Fill fills memory with a byte
Fill	LXI	H,BITE		;prompt for byte
	CALL	Msg
	CALL	HEXC		;get the byte
;
	MOV	A,C		;if no # entered return to HART
	ORA	A
	JZ	HART
;
	MOV	C,L		;save it
	PUSH	B
;
	LXI	H,Startng	;prompt for starting address
	CALL	Msg
	CALL	HEXC		;get starting address
	XCHG			;save it in DE
;
	LXI	H,Ending	;prompt for ending address
	CALL	Msg
	CALL	HEXC		;get the ending address
	XCHG
;
	POP	B		;get the byte back
	MOV	M,C		;stick in the first byte
;
Fill1	MOV	A,H		;all done?
	CMP	D
	JNZ	Fill2		;nope keep stuffing
;
	MOV	A,L		;maybe
	CMP	E		;done now?
	JZ	HART		;ya, return to the HART
;
Fill2	INX	H		;inc the pointer
	MOV	M,C		;stick byte into memory
	JMP	Fill1
;
;
;Up moves the window up
Up	LDA	WindFlg		;what mode is the window in
	CPI	1		;is it in the instruction mode?
	JZ	InstUp		;ya, go move window 9 instruct
;
	LHLD	WINDOW		;no, must be in # mode
	LXI	B,8		;move window 8 bytes
UpDate	DAD	B		;add to current address
Update	SHLD	WINDOW		;update the window
	RET			;all done
;
InstUp	LHLD	NxtWind		;get addr of word after window
	JMP	Update		;go save it
;
;
;Down moves the window down
Down	LHLD	WINDOW		;move the window 8 bytes down
	LXI	B,-8
	JMP	UpDate
;
;
;Right move the old window right
Right	LHLD	WINDOW		;move the window right 1
	LXI	B,-1
	JMP	UpDate		;stick it in the window pointer
;
;
;Left moves the window to the left
Left	LDA	WindFlg		;what mode is the window in
	CPI	1		;is it in the instruction mode?
	JZ	InstLft		;ya, go move window 1 instruct
;
	LHLD	WINDOW		;no, must be in # mode
	LXI	B,1		;move window left one
	JMP	UpDate
;
InstLft	LHLD	ScndLin		;get address of second line
	JMP	Update		;make that the new window
;
;
;============================================================
;		    screen driver
;	Refresh the screen, display registers and windows.
;
REFRESH	CALL	RegPtr		;save what regs point to
	CALL	CLEAR		;go clear the screen
RegBp	CALL	DumpR		;dump one line of registers
	CALL	DumpBp		;disply 1 line of break points
	CALL	CROUT		;stuff out a CR
;
	LDA	LineCnt		;first 4 lines dumped?
	CPI	4
	JNZ	RegBp		;no, keep dumping
;
	CALL	DumpAF		;ya, now dump A & F
	CALL	CROUT2		;disply 2 returns
;
	LHLD	WINDOW		;point NxtWind to begin of Wind
	LDA	WindFlg		;check the window type flag
	CPI	2		;is it in number mode?
	JNZ	WndLoad		;no, set NxtWind to WINDOW
;
	LXI	B,-32		;ya, set NxtWind to WINDOW - 32
	DAD	B
WndLoad	SHLD	NxtWind	
;
	LHLD	PC		;get the address the PC
	CALL	DISASS		;disasm instruction at PC
	SHLD	PCplus1		;save second address after PC
	JMP	DispPC
;
DumpPW	LHLD	NxtPc		;get the next PC address
	CALL	DISASS		;disasm data following the PC
DispPC	SHLD	NxtPc		;save next address
;
	CALL	Tabber2		;tab over twice
;
	CALL	WindDsp		;now display the window
	CALL	CROUT
;
	LDA	LineCnt		;have we dump the last line?
	CPI	0FH
	JNZ	DumpPW		;no keep dumping
	RET			;ya, refresh all done
;
;	DumpR display the register lables and their contents.
;
DumpR	LXI	H,Lable		;display the lables
	CALL	AddLin4		;add LineCnt*4 to HL
	CALL	Msg		;show the lable
;
	LXI	H,REGISTR	;disply registor content
	CALL	AddLin2		;add LineCnt*2 to HL
	MOV	E,M
	INX	H
	MOV	D,M
	CALL	DEOUT		;display'em
;
	LDA	LineCnt		;get line counter back
	LXI	H,spptr		;point to register buffer
	CALL	AddLin4		;add LineCnt*4 to HL
	XCHG			;put into DE
	MVI	C,4		;dump 4 bytes in hex at DE
	CALL	BLK		;start off with a space
;
ForBite	CALL	BLK		;move over to the next field
	LDAX	D		;get byte
	CALL	BYTE		;display it in hex
	INX	D		;inc the memory pointer
	DCR	C		;have all bytes been dumped?
	JNZ	ForBite		;no, go dump the next one
	RET			;all done
;
RegPtr	PUSH	H		;save regs
	PUSH	D
	PUSH	B
	ptrcont	sp		;move stack pointer data
	ptrcont	HL		;move HL data
	ptrcont	DE		;move DE data
	ptrcont	BC		;move BC data
	POP	B
	POP	D
	POP	H
	RET
;
;	DumpBp display the break points.
;
DumpBp	CALL	Tabber		;move over again
	LXI	H,BrkBuf	;there a break point to show?
	CALL	AddLin8		;add LineCnt*8 to HL
;
	XCHG			;get the break point pointer
	LHLD	BrkPntr
	XCHG
	CALL	Dcmp		;if HL > BrkPntr no show
	RNC			;ya, nothing to show
;
	CALL	Tabber		;ya, tab over
;
	MOV	E,M		;get the address of the BP
	INX	H
	MOV	D,M
	XCHG
	CALL	DISASS		;display the address & instruct
	LXI	H,BrkBuf+4	; Now the next one.
	CALL	AddLin8		;add LineCnt*8 to HL
;
	XCHG			;get the break point pointer
	LHLD	BrkPntr
	XCHG
	CALL	Dcmp		;if HL > BrkPntr no show
	RNC			;ya, nothing to show
;
	CALL	Tabber		;ya, tab over
;
	MOV	E,M		;get the address of the BP
	INX	H
	MOV	D,M
	XCHG
	CALL	DISASS		;display the address & instruct
	RET
;
;	DumpAF displays the A & F bytes and flag symbols.
;
DumpAF	LXI	H,AFlable	;display 'AF '
	CALL	Msg
;
	LHLD	AF		;now print the content of A & F
	XCHG
	CALL	DEOUT		;display it
;
	CALL	BLK2		;print 2 spaces
;
	MOV	A,E		;see if the carry flag is set
	RRC			;roll the carry into the carry
	MVI	A,'C'		;display a C if carry is set
	CC	WH1
;
	MOV	A,E		;check the minus flag
	RAL
	MVI	A,'M'		;print M if minus
	CC	WH1
;
	MOV	A,E		;last the zero flag
	RAL
	RAL
	MVI	A,'Z'		;print Z for zero if set
	CC	WH1
;
	MOV	A,E
	RRC
	RRC
	RRC
	JNC	DuAF1
	LXI	H,PEmsg		;display Parity Even message
	CALL	Msg
;
DuAF1	MOV	A,E
	RAL
	RAL
	RAL
	RAL
	RNC
	LXI	H,ACmsg		;display Aux Carry message
	JMP	Msg
;
;
;display the window in hex-ascii, instuctions, input port
WindDsp	LDA	WindFlg		;check the window type flag
	ORA	A		;is it a 0?
	RZ			;ya, don't do anything
;
	DCR	A		;is it a 1?
	JZ	WindDis		;ya, disasm byte in NextWind
;
	DCR	A		;is it a 2
	JZ	WindNum		;ya, jmp to wind number mode
;
	CALL	Tabber
;
	LXI	H,NxtWind	;get the port address
	MOV	A,M
	INR	M		;inc port addr for next time
;
	MOV	B,A		;save port address
	ANI	0FCH		;mask off the single step ports
	CPI	18H
	JNZ	Inner1		;not single step port
;
	MVI	B,1CH		;next valid port
	MVI	M,1DH		;update memory	
;
Inner1	MOV	A,B		;get the port address
	STA	Inner+1		;peee uuuu  wish I had a Z-80
	CALL	BYTE		;display the port address
;
	CALL	Tabber		;one more time
;	
Inner	IN	0		;read the port
	CALL	BYTE
	RET
;
;
;display window in hex and ASCII
WindNum	LHLD	NxtWind		;get window address
	XCHG			;display address
	CALL	DEOUT
	XCHG
	CALL	BLK		;space over once
;
	LDA	LineCnt		;are we down to the 10 line?
	CPI	10
	MVI	A,' '		;if yes stick in a arrow
	JNZ	AteByt0		;no, stick in a space
;
	MVI	A,9AH		;ya, stick in the arrow
AteByt0	CALL	WH1
;
	MVI	C,8		;display 8 bytes
AteByt1	MOV	A,M		;get the byte
	INX	H		;inc the pointer
	CALL	BYTE		;display the byte in hex
	CALL	BLK		;stick in a space
;
	DCR	C		;have all 8 byte been dumped?
	JNZ	AteByt1		;no, dump the next one
;
	CALL	BLK		;ya, stick in a space
	LHLD	NxtWind		;now dump the 8 char in ASCII
	MVI	C,8
;
AteByt2	MOV	A,M		;get the byte
	INX	H		;inc the pointer
;
	CPI	20H		;is it less than 20
	JC	PAD		;ya, not ASCII, insert space
;
	CPI	7FH		;is it greater than 7F
	JC	AteByt3		;ya, not ASCII, insert space
;
PAD	MVI	A,' '		;since not ASCII stuff in space
AteByt3	CALL	WH1		;print the char
	DCR	C
	JNZ	AteByt2
;
	SHLD	NxtWind		;point window to next byte
	RET
;
;
;WindDis displays the instruction dissassembled at the window
WindDis	CALL	Tabber		;stick in another tab
;
	LHLD	NxtWind		;disasm instruction in window
	CALL	DISASS
	SHLD	NxtWind		;update the next window pos
;
	LDA	LineCnt		;is this the first line?
	CPI	6
	RNZ			;no, all done
;
	SHLD	ScndLin		;ya, save addr of second line
	RET
;
;============================================================
;
;CLEAR clears screen and zeros line counter
CLEAR	MVI	A,0CH		;clear the screen
	CALL	WH1
;
	XRA	A		;0 the line counter
	STA	LineCnt
	RET
;
;
;CROUT outputs a CR and updates the screen line counter
CROUT2	CALL	CROUT		;display 2 CR
CROUT	MVI	A,CR		;send a return
	CALL	WH1
	LDA	LineCnt		;inc the line counter
	INR	A
	STA	LineCnt
	RET
;
;
;Tabber displys a tab or two
Tabber2	CALL	Tabber		;print 2 tabs
Tabber	MVI	A,TAB		;stick a tab to WH1
	CALL	WH1
	RET
;
;
;BLK displys a space or two
BLK2	CALL	BLK		;print 2 spaces
BLK	MVI	A,' '		;stick a space to WH1
	CALL	WH1
	RET
;
;
;AddLin adds LineCnt * 4 or 2 to HL
AddLin2	LDA	LineCnt
	JMP	AddLine
;
AddLin8	LDA	LineCnt
	RLC
	RLC
	JMP	AddLine
;
AddLin4	LDA	LineCnt
	RLC
AddLine	RLC
	MOV	C,A
	MVI	B,0
	DAD	B
	RET
;
;
;16 bit compare, Z set if DE = HL, no C set if DE < HL 
Dcmp	PUSH	B
	MOV	B,A
	MOV	A,H
	CMP	D
	JNZ	DCMP1
	MOV	A,L
	SBB	E
DCMP1	MOV	A,B
	POP	B
	RET
;
;============================================================
;	disassemble one instruction pointed to by (HL)
;		Written By Lennie Araki
;
DISASS	PUSH	PSW		;save all the reg but HL
	PUSH	B
	PUSH	D
	PUSH	H
	LXI	H,RETURN	;return via RETURN
	XTHL
;
	MOV	D,H		;display the address of word
	MOV	E,L
	CALL	DEOUT
;
	CALL	BLK		;move over to the next field
;
	MOV	D,M		;get the instruction to disasm
	LXI	B,INSTR		;start looking @ top of  table
;
MASK	LDAX	B		;get the mask from the table
	INX	B		;inr the table pointer
;
	ANA	D		;mask D with data from table
	MOV	E,A		;save it
;
	LDAX	B		;get instruction from table
	INX	B		;inr the table pointer
	CMP	E		;does mask byte = instruction?
	JZ	FOUND		;ya, found the correct data
;
	CALL	SKIP		;no, get next entry in table
	INX	B
;
	JMP	MASK		;look one more time
;
FOUND	CALL	PRM		;print the text in the table
	LDAX	B		;get the instruction type
;
	CPI	TB		;is it less than TB?
	CC	BLK		;ya, so move over to next field
;
	PUSH	H		;save addr of instruct in RAM
;
	LXI	H,RTN		;jump to routine type indirect
	LDAX	B		;get type displacement
	CALL	ADDA		;add on displacement
	XTHL			;stick it on the stack
	MOV	A,D
	RET			;goto the routine 
;
ADDA	ADD	L		;add displacement to pointer
	MOV	L,A		;add on the low byte
	RNC			;did it carry?
	INR	H		;ya, inc the H too
	RET
;
;
RTN	EQU	$		;indirect routines ref here
;
CMP1	ANA	E		;mask for necessary bits
	MOV	E,A		;save the byte
CMP2	LDAX	B		;get the symbol from the table
	INX	B		;inc the table pointer
	CMP	E		;does symbol match masked byte?
	JZ	PRM		;ya, print the text in table
	CALL	SKIP		;no, try next table entry
	JMP	CMP2
;
;
;SKIP inc the table pointer until the end of entry is found
SKIP	LDAX	B		;get next char
	INX	B		;inc the pointer
	ORA	A		;is the high bit set?
	JP	SKIP		;no, keep looking
	RET			;ya, all done
;
;
;PRM print the ascii byte @ (BC)
PRM	LDAX	B		;get the byte from table
	INX	B		;inc the table pointer
	XRI	80H		;is this the end of entry?
	JP	WH1		;ya, print byte & ret via WH1
;
	ANI	7FH		;no, mask off high bit just set
	CALL	WH1		;display it
	JMP	PRM		;go print the next char
;
;
COMMA	MVI	A,','		;print a comma
	JMP	WH1		;return through WH1
;
;
RETURN	INX	H		;inr the memory address
	POP	D		;restore the registers
	POP	B
	POP	PSW
TYPO	RET			;return to call of DISASS
;
;
REGM	RRC			;display a reg, B,C,D,E,H,L,A,M
	RRC			;role register to low bits
	RRC
REG	LXI	B,RGT		;point to lables
	MVI	E,7		;set the mask
	JMP	CMP1		;go show the correct registor
;
;
RP1	LXI	B,RPT		;display reg pair, HL,DE,BC
	JMP	SR1
;
;
SR	LXI	B,SRT		;show reg pair, SP,HL,DE,BC
SR1	MVI	E,30H		;set the mask
	JMP	CMP1		;go print the correct reg pair
;
;
TYP5	CALL	REGM		;print register lable
	CALL	COMMA		;show a comma
	MOV	A,D		;print other register lable
	JMP	REG
;
;
TYP6	CALL	REGM		;print register lable
	CALL	COMMA		;now print a comma
;
;
TYP7	INX	H		;get operand from (HL)
	MOV	A,M
	JMP	BYTE		;display it in hex
;
;
TYP8	CALL	RP1		;print a register pair lable
	CALL	COMMA		;display a comma
;
;
TYP9	INX	H		;now show 2 byte operand in hex
	MOV	E,M		;get lower byte
	INX	H
	MOV	D,M		;get upper byte
	JMP	DEOUT		;print it
;
;
TYPA	RRC			;print # from 0-7 for RST
	RRC			;role for number in low bits
	RRC
	ANI	7		;mask off trask
	ADI	'0'		;make it ascii
	JMP	WH1		;print it
;
;
TYPB	CALL	COND		;print the conditional
	CALL	BLK		;space to next field
	JMP	TYP9		;now print the address
;
;
TYPC	CALL	COND		;print the conditional
	JMP	BLK		;space for tabs
;
;
COND	LXI	B,CCT		;display conditional, Z,C,M
	MVI	E,38H		;set the mask
	JMP	CMP1		;go print it
;
;
DATA	CALL	BLK		;undefined data
	DCX	H		;display byte
	JMP	TYP7
;
;
HEXC	LXI	H,0		;zero conversion buffer
	MOV	C,L		;and counter
NXNYB				;get character
	CALL	LCFLD		;fold it
	MOV	B,A		;save it
	CPI	'0'		;return if less than '0'
	RC
	CALL	WH1		;echo each char
	SUI	'0'		;change to binary
	CPI	10
	JC	NXNB1
	SUI	7
	CPI	10
	RC			;return if not hex
	CPI	16
	RNC			;return if not hex
NXNB1	INR	C		;count # of hex chars
	DAD	H		;shift HL
	DAD	H		;over for next digit
	DAD	H
	DAD	H
	ORA	L		;Or in next digit
	MOV	L,A
	JMP	NXNYB		;go get the next one.
;
;
BYTE	PUSH	PSW
	RRC
	RRC
	RRC
	RRC
	CALL	HEXO
	POP	PSW
HEXO	ANI	15
	ADI	90H
	DAA
	ACI	40H
	DAA
	JMP	WH1			;display hex digit
;
;
LCFLD	CALL	WH0
	JMP	Fold
;
;============================================================
;		disassembler data tables
;
;indirect jump table
T0	EQU	TYPO-RTN
T1	EQU	REG-RTN
T2	EQU	REGM-RTN
T3	EQU	RP1-RTN
T4	EQU	SR-RTN
T5	EQU	TYP5-RTN
T6	EQU	TYP6-RTN
T7	EQU	TYP7-RTN
T8	EQU	TYP8-RTN
T9	EQU	TYP9-RTN
TA	EQU	TYPA-RTN
TB	EQU	TYPB-RTN
TC	EQU	TYPC-RTN
TD	EQU	DATA-RTN
;
;
Z	EQU	128
;
;CPU registers
RGT	db	0,('B'+Z),1,('C'+Z)
	db	2,('D'+Z),3,('E'+Z)
	db	4,('H'+Z),5,('L'+Z)
	db	6,('M'+Z),7,('A'+Z)
;
;condition codes
CCT	db	0H,'N',('Z'+Z)
	db	8H,('Z'+Z)
	db	10H,'N',('C'+Z)
	db	18H,('C'+Z)
	db	20H,'P',('O'+Z)
	db	28H,'P',('E'+Z)
	db	30H,('P'+Z)
	db	38H,('M'+Z)
;
;register pair & stack registers
SRT	db	30H,'PS',('W'+Z)
RPT	db	00H,('B'+Z)
	db	10H,('D'+Z)
	db	20H,('H'+Z)
	db	30H,('S'+Z)
;
;instruction decode
;format : mask,result,msg,type
INSTR	db	-1,76H,'HL',('T'+Z),T0
	db	0C0H,40H,'MO',('V'+Z),T5
	db	0C7H,06H,'MV',('I'+Z),T6
	db	0C7H,4H,'IN',('R'+Z),T2
	db	0C7H,5H,'DC',('R'+Z),T2
	db	0F8H,80H,'AD',('D'+Z),T1
	db	0F8H,88H,'AD',('C'+Z),T1
	db	0F8H,90H,'SU',('B'+Z),T1
	db	0F8H,98H,'SB',('B'+Z),T1
	db	0F8H,0A0H,'AN',('A'+Z),T1
	db	0F8H,0A8H,'XR',('A'+Z),T1
	db	0F8H,0B0H,'OR',('A'+Z),T1
	db	0F8H,0B8H,'CM',('P'+Z),T1
	db	-1,0C6H,'AD',('I'+Z),T7
	db	-1,0CEH,'AC',('I'+Z),T7
	db	-1,0D6H,'SU',('I'+Z),T7
	db	-1,0DEH,'SB',('I'+Z),T7
	db	-1,0E6H,'AN',('I'+Z),T7
	db	-1,0EEH,'XR',('I'+Z),T7
	db	-1,0F6H,'OR',('I'+Z),T7
	db	-1,0FEH,'CP',('I'+Z),T7
	db	-1,07H,'RL',('C'+Z),T0
	db	-1,0FH,'RR',('C'+Z),T0
	db	-1,17H,'RA',('L'+Z),T0
	db	-1,1FH,'RA',('R'+Z),T0
	db	-1,0C3H,'JM',('P'+Z),T9
	db	0C7H,0C2H,'',('J'+Z),TB
	db	-1,0CDH,'CAL',('L'+Z),T9
	db	0C7H,0C4H,'',('C'+Z),TB
	db	-1,0C9H,'RE',('T'+Z),T0
	db	0C7H,0C0H,'',('R'+Z),TC
	db	0C7H,0C7H,'RS',('T'+Z),TA
	db	-1,0DBH,'I',('N'+Z),T7
	db	-1,0D3H,'OU',('T'+Z),T7
	db	0CFH,01H,'LX',('I'+Z),T8
	db	0CFH,0C5H,'PUS',('H'+Z),T4
	db	0CFH,0C1H,'PO',('P'+Z),T4
	db	-1,32H,'ST',('A'+Z),T9
	db	-1,3AH,'LD',('A'+Z),T9
	db	-1,22H,'SHL',('D'+Z),T9
	db	-1,2AH,'LHL',('D'+Z),T9
	db	-1,0EBH,'XCH',('G'+Z),T0
	db	-1,0E3H,'XTH',('L'+Z),T0
	db	-1,0E9H,'PCH',('L'+Z),T0
	db	-1,0F9H,'SPH',('L'+Z),T0
	db	0CFH,9H,'DA',('D'+Z),T3
	db	0EFH,2H,'STA',('X'+Z),T3
	db	0EFH,0AH,'LDA',('X'+Z),T3
	db	0CFH,3H,'IN',('X'+Z),T3
	db	0CFH,0BH,'DC',('X'+Z),T3
	db	-1,27H,'DA',('A'+Z),T0
	db	-1,2FH,'CM',('A'+Z),T0
	db	-1,37H,'ST',('C'+Z),T0
	db	-1,3FH,'CM',('C'+Z),T0
	db	-1,0FBH,'EI',(' '+Z),T0
	db	-1,0F3H,'DI',(' '+Z),T0
	db	-1,0H,'NO',('P'+Z),T0
	db	0,0,'--',('-'+Z),TD
;
;
Lable	db	'SP ',0,'HL ',0,'DE ',0,'BC ',0
AFlable	db	'AF ',0
;
;=============================================================
;			messages
;
BITE	db	18H,'Byte? ',0
String	db	18H,'String? ',0
Srching	db	18H,'   Searching',0
Startng	db	'   '
Strting	db	'Starting address? ',0
Ending	db	'   Ending address? ',0
Contin	db	'Hit C to continue ',0
Address	db	18H,'Memtop address? ',0
MemMod	db	'Memory Modify Mode ',0
Port	db	'   Port address? ',0
To	db	'   To? ',0
PEmsg	db	' (PE)',0
ACmsg	db	' (AC)',0
;
;=============================================================
;		flags and buffers
;
REGISTR	EQU	$	;registor buffer
sp	DS	2	;save stack pointer here
HL	DS	2	;save HL here
DE	DS	2	;contains DE
BC	DS	2	;contains BC
AF	DS	2	;low byte flags, high byte A
PC	DS	2	;contain program counter
NxtPc	DS	2	;points to the next byte in PC
PCplus1	DS	2	;points to second instruction at PC
WINDOW	DS	2	;contain address of the window
NxtWind	DS	2	;points to next byte in window
WindFlg	DS	1	;0=no wind,1=Instruct,2=#,3=input
OldWind	DS	1	;window flag when enter mem mod mode
ScndLin	DS	2	;address of the second line of window
LineCnt	DS	1	;video line number
SingUse	DS	1	;single step or user interrupt flag
OldMT	DS	2	;value of memtop when debugger entered
TempBrk	DS	1	;temp break point flag, -1 if set
TempIns	DS	1	;temp break point instruction
BrkPntr	DS	2	;break point buffer pointer 
BrkBuf	DS	BRKPNTS*4 ;break point buffer
spptr	DS	4	;save buffer for what SP points to
HLptr	DS	4	; "      "    "   "   HL   "     "
DEptr	DS	4	; "      "    "   "   DE   "     "
BCptr	DS	4	; "      "    "   "   BC   "     "
OldPos	DS	2	;cursor pos when enter bugger
StrSiz	EQU	10	;ten byte string buffer
StrgBuf	DS	StrSiz	;string search buffer
StrgEnd	DS	2	;end of search string address
SchMem	DS	2	;search memory address
ScrnBuf	DS	1	;screen buffer 1K
Scrend	EQU	ScrnBuf+1024
PROGLEN	EQU	Scrend-Code
Size	EQU	PROGLEN+1024
TAIL	EQU	$
;
;
	END
