;********************************************************
;*							*
;*			Szap				*
;*							*
;*	Version 3.0			R. T. Martin	*
;*	02/24/81					*
;*							*
;*	Copyright (c) 1981, Interactive Products	*
;*  Corporation dba PolyMorphic Systems 460 Ward Dr.	*
;*  Santa Barbara, CA  93111				*
;*							*
;********************************************************
;
;
; ***** SuperZap *****
;	written by R. T. Martin, June 1977
;
; Last updated:
;	12/26/78 	FEA
;	12/28/79	RTM	Disaster recovery
;	02/24/81	BFS	Support Units > 7.
;				Rip out overwriting code areas!
;
; Thanks to Joe Young for showing me what a SuperZap is.
;
; Symbols in the Disk system and such.
;
	REFS	SYSTEM
	REF

; Chr codes we use

UP	EQU	17	; up cursor control
DOWN	EQU	18	; down "      "
RIGHT	EQU	19	; right "     "
LEFT	EQU	20	; left  "     "
ESC	EQU	27	; escape
LF	EQU	10	; line feed, up to last line start
VT	EQU	11	; vert tab, home-up
FF	EQU	12	; form feed, clear screen + home-up
CR	EQU	0DH	; carriage return, start of next line
;
; Code
;
	ORG	USER
	IDNT	$,$
;
	JMP	START
	JMP	Warm	; no re-entry....
;
; Data
;
ESP	DW	0	; entry stack pointer
FMOD	DB	0	; frame modified flag
SMOD	DB	0	; screen modified
FRAME	DW	0	; frame address we're displaying
FDN	DB	0	; disk #
FDA	DW	0	; disk address
FFLAG	DB	0	; file or hard address mode
MFLAG	DB	0	; mode flag
CPOS	DW	0	; cursor position
CCHR	DB	0	; chr underneath cursor
CADR	DW	0	; memory address under cursor
CFLAG	DB	0	; chr (text) display flag
EFLAG	DB	0	; ignore errors flag!
;
; Routines
;
; Do I/O thru Dio and crab on errors.  Preserves all.
;
Doio	PUSH	H
	PUSH	D
	PUSH	B
	PUSH	PSW
	CALL	Dio
	CC	OOPS
	POP	PSW
	POP	B
	POP	D
	POP	H
	RET		; bye y'all.
;
; WHEX takes the address in FRAME and displays the hex
; representation of that on the video board by building
; the image into SCREAN. Somebody else has to copy that
; to VIDEO to force the display...
;
WHEX	PUSH	H
	PUSH	D
	PUSH	B	; might just as well.
	LHLD	FRAME	; frame address
	XCHG
	LXI	H,SCREAN ; screen model addr
;
; We puts out 2 characters and a space 16 by 16.
;
	MVI	B,16	; # lines
WHEX1	MVI	C,16	; # bytes per line
WHEX2	LDAX	D	; get byte
	INX	D
	CALL	WHX	; poke
	MVI	M,0A0H	; poke in a space
	INX	H
	DCR	C
	JNZ	WHEX2	; more on this line
WHEX3	MOV	A,L
	ANI	63
	JZ	WHEX4
	MVI	M,0A0H
	INX	H
	JMP	WHEX3
WHEX4	DCR	B
	JNZ	WHEX1
	POP	B
	POP	D
	POP	H
	RET		; bye....
WHX	PUSH	PSW
	RRC
	RRC
	RRC
	RRC
	CALL	WHX1
	POP	PSW
WHX1	ANI	15
	ADI	90H
	DAA
	ACI	40H
	DAA
	ORI	80H
	MOV	M,A
	INX	H
	RET
;
; WTXT takes the stuff in FRAME and builds text to 
; the screen model SCREAN (in the right spot)
;
WTXT	PUSH	H
	PUSH	D
	PUSH	B
	LHLD	FRAME
	LXI	D,SCREAN+48
	XCHG		; from is DE, to is HL
	MVI	B,16	; 16 lines
WTXT1	MVI	C,16	; 16 characters
WTXT2	LDAX	D
	ANI	80H
	LDAX	D
	JZ	WTXT3
	MVI	A,20H	; wutzit nurgh blank!
WTXT3	ORI	80H
	INX	D
	MOV	M,A
	INX	H
	DCR	C
	JNZ	WTXT2
	MOV	A,L
	MOV	L,A
	PUSH	D
	LXI	D,48
	DAD	D
	POP	D
	DCR	B
	JNZ	WTXT1
	POP	H
	POP	D
	POP	B
	RET		; scram.
;
; WHAM copies the SCREAN buffer to VIDEO.
;
WHAM	PUSH	H
	PUSH	D
	MVI	L,0
	LDA	SCRHM
	MOV	H,A	; video board address
	LXI	D,SCREAN
WHA1	LDAX	D
	MOV	M,A
	INX	H
	INX	D
	MVI	A,SCREAN/256+4
	CMP	D
	JNZ	WHA1
	POP	D
	POP	H
	RET
;
; Junk Hex input routine. Nothing fancy. Gobbles chrs
; building hex # in HL until crapper found, then
; splits with crapper in A, value in HL. DOES NOT ECHO
; CHARACTERS TO SCREAN. (Dat's you problem.)
;
HEXI	PUSH	D
	LXI	H,0	; clear it.
	MOV	C,L	; digit counter
HEX1	CALL	WH0	; gimmie some stuff
HEXA	CALL	Fold	; make it upper case!
	MOV	B,A
	CPI	'0'
	JC	HEXX
	SUI	'0'
	CPI	10
	JC	HEX2
	SUI	7
	CPI	10
	JC	HEXX
	CPI	16
	JNC	HEXX
HEX2	DAD	H
	DAD	H
	DAD	H
	DAD	H
	ORA	L
	MOV	L,A
	INR	C	; bump digit count accepted
	JMP	HEX1
HEXX	MOV	A,B	; term chr
	POP	D
	RET		; bye....
;
; Some actual driver code of sorts....
;
LOOP	LHLD	ESP
	SPHL		; recover SP!
	LHLD	CPOS
	MOV	A,M
	STA	CCHR
	MVI	M,0FFH	; put up a cursor
	CALL	WH0	; get a keyboard character
	MOV	B,A	; stash
	LDA	CCHR
	MOV	M,A	; put chr back
	LDA	MFLAG
	ORA	A	; check for text entry mode
	JNZ	ETEXT	; jmp/that's it, go do it.
	MOV	A,B	; chr again...
	CALL	Fold	; make it upper case
	MOV	B,A
	CPI	5	; Control E?
	JZ	SCRAM	; scram- bug out to Exec
	CPI	RIGHT
	JZ	CRIGHT
	CPI	LEFT
	JZ	CLEFT	; right or left move commands
	CPI	'.'
	JZ	NEWD	; just renew the display
	CPI	LF
	JZ	CUP	; LF- up a frame.
	CPI	DOWN
	JZ	CCR	; up arrow- move down a line
	CPI	UP
	JZ	CLF	; down arrow- move up a line
	CPI	CR
	JZ	CDOWN	; CR- down a frame
	CPI	ESC
	JZ	CESC
	CPI	'/'
	JZ	CSET
	CPI	'I'
	JZ	CIND	; go indirect
	CPI	':'
	JZ	CDEV	; set device #
	CPI	'!'
	JZ	CERR
	CPI	'Z'
	JZ	CZAP	; clear to end of page...
	CPI	3
	JZ	CHKSM	; Checksum 4 blocks
	CPI	27H	; "'"
	JZ	CQUOTE	; enter a character....
	LXI	H,L2
	PUSH	H
	PUSH	D
	LXI	H,0
	MOV	C,L
	JMP	HEXA
L2	MOV	A,C
	ORA	A	; any digits?
	JZ	LOOP	; jmp/nope
	STA	FMOD	; mark diddled
	STA	SMOD	; and screen to be updated.
	MOV	A,L
	LHLD	CADR
	MOV	M,A	; diddle it
	JMP	CRIGHT	; and go do right.
;
; Some semblance of error recovery......
; The error code/subcode expected in DE
;
OOPS	LDA	EFLAG
	ORA	A	; do we ignore 'em??
	RNZ		; return to caller if so.
	XTHL
	POP	H	; junk return addr pushed...
OOPS1	CALL	WHEX
	LXI	H,SCREAN+52
	LDA	FDN
	CALL	WHX
	INX	H
	LDA	FDA+1
	CALL	WHX
	LDA	FDA
	CALL	WHX
	INX	H
	MOV	A,D
	CALL	WHX
	MOV	A,E
	CALL	WHX
	JMP	NEWQ
;
; the E command turns off aborts on disk errors...
;
CERR	LDA	EFLAG
	CMA
	STA	EFLAG
	MOV	E,A
	MOV	D,A
	JMP	OOPS1	; display state of marker....
;
; Enter text goodie. Terminate on esc.
;
ETEXT	MOV	A,B	; the chr
	CPI	ESC
	JNZ	CQ1	; jmp/enter it
	XRA	A
	STA	MFLAG	; else clear flag byte
	JMP	LOOP	; and scram
;
; Single quote says enter a character
;
CQUOTE	CALL	WH0	; get a character
CQ1	LHLD	CADR
	MOV	M,A	; poke into memory
CQ2	MVI	A,1
	STA	MFLAG	; set text entry flag
	STA	SMOD
	STA	FMOD	; set modified
	JMP	CRIGHT	; and zoop off...
;
; Go indirect on current thing
;
CIND	LHLD	CADR
	MOV	E,M
	INX	H
	MOV	D,M
	XCHG
	LDA	FDN
	ORA	A
	JNZ	NEWF	; 16 bits ok
	MVI	L,0
	JMP	NEWF	; else only top 8 bits used
;
; zap to end of page
;
CZAP	LHLD	CADR
CZ1	MVI	M,0
	INR	L
	JNZ	CZ1
	MVI	A,1
	STA	FMOD
	JMP	NEWD	; go do new display.
;
; The guy wants a checksum of junk so let's give it to him
;
CHKSM	LHLD	FDA		; First grab 3 more sectors
				; 1 sector
	INX	H		; at a time.
	LDA	FDN		; Which disk #
	MOV	C,A
	MVI	B,1		; Read command
	LXI	D,BUF+100H	; Where to put stuff
	MVI	A,1		; Do 1 sector
	CALL	Doio		; Grab it
	INX	H		; Point to next sector
	LXI	D,BUF+200H	; Next sector goes here
	CALL	Doio
	INX	H		; Last one
	LXI	D,BUF+300H
	CALL	Doio
	;
	; Sectors to be checksummed are now in RAM af BUF
	;
	LXI	H,BUF+1		; Start checksumming here
	XRA	A		; Initialize checksum
CHKSM1	ADD	M
	INX	H
	MOV	B,A
	MOV	A,H
	CPI	(BUF+400H)/256
	MOV	A,B
	JNZ	CHKSM1
	;
	; When we fall through here we have the checksum so
	; let's stick it where it will do us some good.
	;
	STA	BUF
	MVI	A,1
	STA	FMOD	; Tell machine we modified the frame
	LHLD	FDA	; Tell NEWF what sector to punt
	JMP	NEWF	; Update everything and split
;
;	Control Y exit. No update of current frame, tell user.
;
EXIT	POP	H
	POP	H
	POP	H
	POP	H
	POP	H	; Interrupt level reggies and ret add
	EI		; Re-enable interrupts.
	LXI	H,Bailout ; Tell 'em what happened.
	CALL	Msg
	JMP	EXIT1	; and leave.
;
; The user said control-e: we scram back to Exec.
;
SCRAM	LDA	FMOD
	ORA	A
	CNZ	FPUNT	; punt modified frame out.
	MVI	A,FF
	CALL	WH1	; Clear the screen.
EXIT1
	IF	USERS=2
	LDA	FDN
	ORA	A
	JZ	Warm
	MOV	C,A
	clrt	excl
	ENDIF

	JMP	Warm	; and warm-start.
;
; FPUNT is called if the frame mod flag is set. We 
; have a modified disk sector to punt out. Note that we
; know about disk directories, so if you say to punt
; a modified sector 0-3, we do the right thing.
;
FPUNT	LDA	FDN
	ORA	A
	JZ	FPUNX	; jmp/not disk!
	LHLD	FDA	; current thing.
	MOV	A,H
	ORA	A
	JNZ	FPUN1	; jmp/not directory
	MOV	A,L
	ORA	A
	JZ	FPUND	; jmp/directory!
	CPI	4
	JC	FPUND	; jmp/directory sector!
FPUN1	LXI	D,BUF	; memory address
	MVI	B,0	; write
	LDA	FDN
	MOV	C,A	; device #
	MVI	A,1
	CALL	Doio	; do it...
FPUNX	XRA	A
	STA	FMOD	; clear flag
	RET		; and split.
FPUND	LXI	D,SBUF1
	LXI	H,0
	MVI	B,1
	LDA	FDN
	MOV	C,A
	MVI	A,1
	CALL	Doio	; read in directory
	INX	H
	LXI	D,SBUF1+100H
	CALL	Doio
	INX	H
	LXI	D,SBUF1+200H
	CALL	Doio
	LXI	D,SBUF1+300H
	INX	H
	CALL	Doio
;
; We read in the directory one sector at a time so
; that  we can recover as much as possible of
; it in case of an error in one sector.....
;
	LXI	H,SBUF1
	LDA	FDA	; low byte of thing
	ADD	H
	MOV	H,A	; copy to address
	LXI	D,BUF	; copy from
FPD1	LDAX	D
	MOV	M,A
	INX	D
	INR	L
	JNZ	FPD1	; copy over.
	CALL	Ckdr	; checksum
	STA	SBUF1
	STA	NFCK	; stash appropriately
	LDA	FDN
	STA	NFDIR	; stash that also
	MOV	C,A
	LXI	D,SBUF1
	LXI	H,0
	MVI	B,0
	MVI	A,4
	CALL	Doio	; write it out
	JMP	FPUNX	; split.
;
; Up
;
CUP	LHLD	FDA
	LDA	FDN	; which one? disk or ram?
	ORA	A
	JZ	UPM
	DCX	H	; for disk, inx
	JMP	NEWF
UPM	DCR	H
NEWF	PUSH	H
	LDA	FMOD	; modified?
	ORA	A
	CNZ	FPUNT	; punt it out if so.
	POP	H
	SHLD	FDA	; new frame address
	SHLD	CADR	; set temp CADR
	PUSH	H
	MVI	L,0
	LDA	SCRHM
	MOV	H,A	; video board address
	SHLD	CPOS	; store that
	POP	H
	LDA	FDN	; see if disk or memory....
	ORA	A
	JZ	NEWM	; jmp/it's memory
;
; Get new disk sector
;
	MOV	C,A	; devn
	LHLD	FDA	; here is address
	LXI	D,BUF
	XCHG
	SHLD	CADR	; update these two....
	SHLD	FRAME
	XCHG
	MVI	B,1	; read
	MVI	A,1	; 1 sector
	CALL	Doio
	JMP	NEWD	; display.
NEWM	SHLD	FRAME	; keep as same address
NEWD	CALL	WHEX	; fill next part of display
	LXI	H,SCREAN+52
	LDA	FDN
	CALL	WHX
	INX	H	; space....
	LDA	FDA+1
	CALL	WHX
	LDA	FDA
	CALL	WHX
	LDA	CFLAG	; chr display?
	ORA	A
	CNZ	WTXT	; do it/don't do it, whatever.
NEWQ	CALL	WHAM	; flash on screen.
	JMP	LOOP	; and loop.
;
; Down
;
CDOWN	LHLD	FDA
	LDA	FDN
	ORA	A
	JZ	CDM
	INX	H
	JMP	NEWF
CDM	INR	H
	JMP	NEWF	; go to it, kid.
;
; Tab moves us to start of next line on screen.
;
CCR	LHLD	CADR
	MOV	A,L
	INR	A
	JZ	CDOWN	; move to next frame
	INX	H
	SHLD	CADR
	XCHG		; save in DE for a bit!
	LHLD	CPOS
	LXI	B,3
	DAD	B	; move over
	MOV	A,L
	ANI	30H
	CPI	30H
	JNZ	Ccr1
	LXI	B,16
	DAD	B
Ccr1	SHLD	CPOS
	MOV	A,E
	ANI	15
	JNZ	CCR	; move to line start
	JMP	CRTH	; clean up.
;
; LF - Move up a line
;
CLF	LHLD	CADR
	MOV	A,L
	ORA	A
	JZ	CUP	; jmp/go up if at top
Clf1	LHLD	CADR
	DCX	H
	SHLD	CADR
	XCHG		; save in DE
	LHLD	CPOS
	LXI	B,-3
	DAD	B
	MOV	A,L
	ANI	30H
	CPI	30H
	JNZ	Clf2
	LXI	B,-16
	DAD	B
Clf2	SHLD	CPOS
	MOV	A,E	; to see if at line start
	ANI	0FH
	JNZ	Clf1	; jmp/not yet, keep going
	JMP	CRTH	; cut out when done.
;
; Escape toggles chr display
;
CESC	LDA	CFLAG
	CMA
	STA	CFLAG
	JMP	NEWD	; display
;
; Set new frame address
;
CSET	CALL	HEXI
	LDA	FDN
	ORA	A	; memory or disk?
	JNZ	NEWF	; disk/leave it alone.
	MVI	L,0	; haha.
	JMP	NEWF
;
; D sets device #
;
CDEV	LDA	FMOD
	ORA	A
	CNZ	FPUNT	; clean up BEFORE changing...

	IF	USERS=2
	LDA	FDN
	ORA	A
	JZ	CDEV1
	MOV	C,A
	clrt	excl
	CC	OOPS
	ENDIF

CDEV1	CALL	HEXI
	MOV	A,L
	STA	FDN

	IF	USERS=2
	ORA	A
	JZ	CDEV2
	MOV	C,A
	sett	excl
	CC	OOPS
	ENDIF

CDEV2	LDA	FDN
	LXI	H,0	; you get sector 0 no charge...
	JMP	NEWF
;
; Right move stuff
;
CRIGHT	LHLD	CADR	; target block address
	MOV	A,L
	INR	A	; if zero, don't move it,
	JZ	CDOWN	; but move down to next frame.
	INX	H
	SHLD	CADR	; well, that's the easy part.
	LHLD	CPOS	; cursor position
	LXI	D,3
	DAD	D	; just grind it out, kid...
	MOV	A,L
	ANI	30H
	CPI	30H	; see if we went too far...
	JNZ	CRTH	; jmp/nope, ok.
	LXI	D,16
CRTG	DAD	D
CRTH	SHLD	CPOS	; store new position
	LDA	SMOD
	ORA	A
	MVI	A,0
	STA	SMOD
	JNZ	NEWD	; new display if mod set...
	JMP	LOOP	; and to top again.
;
; Left move it he wants to.
;
CLEFT	LHLD	CADR	; memory block address
	MOV	A,L
	ORA	A
	JZ	CUP	; move to previous block...
	DCX	H
	SHLD	CADR
	LHLD	CPOS
	LXI	D,-3
	DAD	D
	MOV	A,L
	ANI	30H	; bits should be off...
	CPI	30H
	JNZ	CRTH	; jmp/looks cool to me...
	LXI	D,-16
	JMP	CRTG
;
; START code
;
START	LDA	SCHR	; see if we are enabled
	ORA	A
	RZ		; and split if not!
	LXI	H,0
	SHLD	FRAME
	SHLD	CADR
	SHLD	FDA	; set to zero
	DAD	SP
	SHLD	ESP	; save for error recovery
	XRA	A
	STA	FMOD
	STA	MFLAG
	STA	SMOD
	STA	CFLAG
	STA	FDN	; zero flags too.
	MVI	L,0
	LDA	SCRHM	; screen home
	MOV	H,A
	SHLD	CPOS	; cursor position
	LXI	H,EXIT	; Control Y goes here.
	SHLD	PVEC
	LXI	H,Hello	; text
	CALL	Msg	; display it
	CALL	Killi	; flush input
	MVI	A,0C9H
	STA	USER	; stomp on entry point
	JMP	LOOP	; and go process commands
;
	MACLIST	0
;
Hello
 db FF,CR,'SuperZap version 3.0 02/24/81- '
 db 'Commands are:',CR,CR
 db '^E - Exit',9,'. - Display',9,'^C - Checksum next '
 db '4 sectors',CR
 db '!  - Toggle error flag',9,9,'Z - Zero from cursor',CR
 db 'I - Display indirect',9,9,':n - Use device n',CR
 db '/n - Display frame n',9,9,27H,' - Enter text',CR
 db 'LF - Display prev. frame',9,'CR - Display next frame',CR
 db 9AH,' - Move cursor right',9,9,9BH,' - move cursor left',CR
 db 9CH,' - Move cursor up',9,9,9DH,' - Move cursor down',CR
 db 'nn - Enter data at cursor and move right',CR
 db 'ESC - Toggle text display, terminate text entry',CR
 db 'nn = Hexadecimal number',CR
 db CR,'If you don''t know what you''re doing, '
 db 'don''t do anything!',VT,0

Bailout db FF,'Control Y - exiting without writing current'
	db ' display page!',CR,0
;
	ORG	($+255) AND 0FF00H
SCREAN	EQU	$		; 1K screen buffer
BUF	EQU	SCREAN+400H	; sector buffer
;
	END
