	PAGE
;********************************************************
;*		    ASSEMBLE				*
;*							*
;*	We talk to the user about assembly options,	*
;*  open some disk files, initialize some globals, and	*
;*  call PASS twice, with PASSFLAG=1 and then 2.  Then	*
;*  we print summaries, symbol table lists, cross ref	*
;*  tables etc.  					*
;*							*
;********************************************************
;
; Last updated:
;	2/8/80		RTM JDW	clock date for >= <= fix
;	2/27/80		RTM	Clock date for devlocks
;
	ORG	USER		;Assembly loc for assmblr.
	IDNT	$,$		;LA,SA for disk sys...

	JMP	START		;The usual cold starts
	RET			;No warm start.

START	LXI	H,HELLOMSG	;Say hello.
	CALL	Msg
	LXI	D,-(SYMLLEN+50)	;Is there enough room to run?
	LHLD	MEMTOP		;MEMTOP must be > (SYMLLEN+50)
	DAD	D		;Add and expect negative.     
	LXI	D,206H		;I don't have enough memory
	JNC	Err		; to do that

ASMB	LXI	H,CMND		;Invocation text.
	CALL	OPEN		;OPEN DISK  FILES.
	XRA	A
	STA	MLTFLG		;Init multiple file flag.

	LXI	H,HRDMSG	;Does he want hard copy?
	CALL	YORN
	STA	HARDLIST

	LXI	H,EROMSG	;Does he want errors only?
	CALL	YORN
	CMA
	STA	ERRSONLY

	LXI	H,STLMSG	;Does he want symtab listing?
	CALL	YORN
	STA	STLFLAG

;  Done with the listing conversation.

	CALL	SCNINIT		;Initialize listing driver.
	CALL	SYMINIT		; symbol table,
	CALL	RELINIT 	; and relocation systems.

	MVI	A,1
	STA	PASSFLAG	;Indicate pass 1.
	LXI	H,P1MSG
	CALL	Msg		;Tell user we're doing it.

;  Do pass one.

	CALL	PASS
	RC			;Ret/ fatal error

	MVI	A,2
	STA	PASSFLAG	;Ready for pass two.
	LXI	H,P2MSG
	CALL	Msg		;Tell user we got to pass two.

	LDA	MLTFLG		;If multiple files used,
	ORA	A		;don't rewind. User will have
	JNZ	ASMB2		;to respecify first file.
	CALL	REWIND		;Rewind input file.
	JMP	ASMB21
ASMB2	LXI	D,REOMSG	;Re-open message.
	LXI	H,LINEBUF
	LXI	B,LINEBFLEN
	CALL	Rlwe
	LXI	H,LINEBUF
	CALL	OPIN		;Glenn's canned opener.
	JNC	ASMB21
	CALL	Ovrto
	DB	'Emsg'
	LXI	H,CRMSG
	CALL	Msg
	JMP	ASMB2		;Ask again.
ASMB21	
	;  Do pass two.

	CALL	PASS
	RC			;Ret/ fatal error
	LDA	HARDLIST
	ORA	A
	JZ	ASMB22
	LXI	H,FFBUF 	;Send formfeed to printer.
	MVI	C,1
	CALL	PUTLINE

;  Output the relocation block to the disk file.

ASMB22	CALL	RELDUMP

;  Print error totals on scrn (and possibly paper).

	LXI	H,ETMSG		;'Error total = '
	CALL	Msg
	LDA	HARDLIST	;If printing onto display
	ORA	A		;screen, don't duplicate
	JZ	ASMB23		;the error msg.
	LXI	H,ETMSG
	MVI	C,ETMSGLEN
	CALL	PUTLINE		;Print them too.
ASMB23	LHLD	ERRCOUNT	;Convert error count into a
	XCHG			;decimal number in LISTBUF.
	LXI	H,LISTBUF	;Conversion area.
	MVI	C,5		;Max 5 digits.
	MVI	B,10		;Base 10
	MVI	A,' '		;Blank padding.
	CALL	@NUM
	MVI	A,CR
	STA	LISTBUF+5	;A CR for PUTLINE.
	XRA	A
	STA	LISTBUF+6	;Zero after number for MSG.
	LXI	H,LISTBUF
	CALL	Msg		;Print number on scrn.
	LDA	HARDLIST	;Do we want a hard copy?
	ORA	A
	JZ	ASMB24
	LXI	H,LISTBUF
	MVI	C,6
	CALL	PUTLINE		;Put count on paper (decimal).

;  Close the output file, if one was specified.

ASMB24	LDA	OTFLG
	ORA	A
	JZ	ASMB25
	CALL	SHUT
ASMB25	;
;  Close the DEFS file, if one was used.

	IF	REFSDEFS
	CALL	DEFCLOSE	; Close it.
	ENDIF

;  Print the symbol table listing, if the user wants it.

	LDA	STLFLAG		;Does user want one?
	ORA	A
	JNZ	PUTSYMS 	;Print symbol table listing.
	RET			;Return to supervisor.

;  This is an optimization.

YORN	CALL	Msg		;User's prompt then '(Y or N)'
YORN2	LXI	H,LINEBUF	;Read into text buffer.
	LXI	B,5		;Max chars.
	LXI	D,YRNMSG
	CALL	Rlwe
	LDA	LINEBUF		;Get the Y or N.
	CALL	Fold		;Make into UC.
	CPI	'Y'
	JZ	YORN1
	CPI	CR
	JZ	YORN3
	CPI	'N'
	JNZ	YORN2
YORN3	XRA	A
	RET
YORN1	MVI	A,0FFH
	RET

	 IF	M88ASM
HELLOMSG DB	'Macro-88 version 5.1: 06/10/81',CR,0
	 ELSEIF M86ASM
HELLOMSG DB	'Macro-86 version 5.1: 06/10/81',CR,0
	 ELSEIF M65ASM
HELLOMSG DB	'Macro-65 version 5.1: 06/10/81',CR,0
	 ELSEIF M89ASM
HELLOMSG DB	'Macro-89 version 5.1: 06/10/81',CR,0
	 ENDIF
;
HRDMSG    DB	'Hardcopy? (else video display)',0
EROMSG	  DB	'Full listing? (else errors only)',0
STLMSG	  DB	'Symbol table printout?',0
YRNMSG    DB	' (Y or N):',0
P1MSG	  DB	'Pass one.',CR,0
P2MSG	  DB	'Pass two.',CR,0
REOMSG	  DB	'Re-specify original input file please:',0
ETMSG	  DB	CR,'Error total = ',0
ETMSGLEN  EQU	$-ETMSG
	PAGE
;********************************************************
;*			PASS				*
;*							*
;*	Here we go reading source and writing object	*
;*  code and printing listings. There are some		*
;*  globals that control PASS, such as PASSFLAG,	*
;*  HARDLIST, LISTFLAG, etc.  All are set by operator	*
;*  or something before PASS is called.  Between	*
;*  passes (between calls to PASS) the input file	*
;*  should be rewound.  Before the first call, of 	*
;*  course, the input file must be opened, as		*
;*  well as the output file.  The output file need 	*
;*  not be changed between passes, as it will simply	*
;*  be left alone in pass one.   			*
;*	LISTFLAG is set by STATEMENT when the 'LIST'	*
;*  pop is found, and when macro expansions start it	*
;*  is set by MACREF.  The printing of the current	*
;*  line can be controlled by setting VISIBLE or	*
;*  INVISIBLE to non-zero.  These are normally 		*
;*  mutually exclusive.  They override LISTFLAG temp-	*
;*  orarily.
;*	Statements that wish to control whether they	*
;*  themselves are listed (PAGE, LIST, MACLIST...) can	*
;*  do so by controlling the INVISIBLE flag.  It is	*
;*  reset automatically to 0 (False or 'visible').	*
;*							*
;********************************************************
;
; Initialize globals.
;
PASS	XRA	A
	;
	IF	REFSDEFS
	STA	REFYET		;No REFS pop seen yet.
	STA	DEFYET		;No DEFS pop seen yet.
	ENDIF
	;
	STA	GENYET		;No code produced yet.
	STA	ENDFLAG		;Not end of text yet.
	STA	IFNEST
	STA	IFFAILS		;'IF' nest level 0/0
	STA	ELSEFLG		;Haven't seen any ELSEs yet.
	STA	MACDEF		;Not defining MACRO.
	MVI	A,16		;Default hex listing
	STA	PTBASE
	MVI	A,0FFH
	STA	LISTFLAG	;As if a 'LIST 1' was seen.
	STA	MLISTFLAG	;We are listing macro expands.
	LXI	H,0
	SHLD	ERRCOUNT	;No errors at all yet.
	SHLD	LOCATION	;As if an 'ORG 0' was seen
	SHLD	CHECKSUM
	SHLD	LASTREL
	SHLD	OLDA		;As if an 'IDNT 0,0' was seen.
	SHLD	OSTA		;(Output file LA & SA)
	;
	CALL	MACINIT		;Init Macro String-List-Stack.
	;
	CALL	STKINIT		;Init Arithmetic Value Stack.
	;
	CALL	BITINIT 	;Init bit output variables.
	;
	; Top of main loop.
	;
PASS1	CALL	GETLINE		;From disk or MACRO...
	JNC	PASS11
PASS12	MVI	A,0FFH		;Set mult. file flag.
	STA	MLTFLG
	LXI	H,LINEBUF
	LXI	B,LINEBFLEN
	LXI	D,NOENDMSG	;No 'END' pop found.
	CALL	Rlwe
	LXI	H,LINEBUF
	CALL	OPIN		;Glenn's canned opener
	JNC	PASS1		;Good - get the line.
	LDA	IFDE+1
	ANI	1FH		;How  long  was name?
	RZ			;If zero, done.
	CALL	Ovrto
	DB	'Emsg'		;Else, bad error.
	MVI	A,CR
	CALL	WH1
	JMP	PASS12		;But keep trying.
	;
	; Now HL points at buffer and C=# chars read.
	;
PASS11	PUSH	H
	PUSH	B
	CALL	STATEMENT	;Assemble it.
	POP	B
	POP	H
	JNC	PASS2		;Jmp/ no error.
	;
	;	Fatal error.  Print the current line (pointed
	; at by HL, length in C), print error (code in DE),
	; and return CARRY.
	;
	PUSH	D
	CALL	PUTLIST 	;Show current line
	POP	D
	CALL	PUTERR		; then spit out error.
	STC
	RET
	;
	;  Print the assembly listing.
	;
PASS2	LDA	PASSFLAG	;Skip listing altogether if
	CPI	2		;not pass 2.
	JNZ	PASS4
	LDA	ERRSONLY	;If listing only errors,
	ORA	A
	JNZ	PASS3		;jmp to errors only.
	LDA	INVISIBLE	;If that statement wants to
	ORA	A		;be invisible, print it only
	JNZ	PASS3		;if it is an error.
	LDA	VISIBLE		;If that statement wants to be
	ORA	A		;visible, print it.
	JNZ	PASS31
	LDA	LISTFLAG	;Else test the listing flag,
	ORA	A
	JZ	PASS3		;If false, print only errors.
	;
	;  Full listing (LIST on).  Print the source line
	; modified according to the LISTFLAG.
	;
	CALL	PUTMLIST 	;Write out listing from (HL,C).
	JMP	PASS32
	;
	;  Errors only listing. (LIST off)
	;
PASS3	PUSH	H		;then check for errors,
	LHLD	WARNING		;and if any present,
	MOV	A,H		;then print BOTH the
	ORA	L		;listing AND the error
	LHLD	ERR		;printout.
	ORA	H
	ORA	L
	POP	H		;(HL points at list buff.)
	JZ	PASS4		;jmp/ skip both list & error.
	;
	;  Print the original source line, followed by 
	; possible errors.
	;
PASS31	CALL	PUTLIST
	;
	;  Print and count any errors or warnings.
	;
PASS32	LHLD	ERR
	XCHG
	CALL	PUTERR		;Put possible ERRor.
	LHLD	WARNING
	XCHG
	CALL	PUTERR		; and possible WARNING.
	;
	;  Output object code to disk file.
	;
PASS4	CALL	DUMP
	;
	; Have we done the last statement?
	;
	LDA	ENDFLAG	
	ORA	A
	JZ	PASS1		;No - keep assembling.
	CALL	GETLINE 	;Found the 'END', but is
	JC	PASS42		;there still text?
	LXI	D,1601H    ;** STATEMENTS FOUND AFTER 'END' **
	CALL	PUTERR
PASS41	CALL	GETLINE		;Be sure input file is
	JNC	PASS41		;really exausted.
PASS42	XRA	A		;Successful termination.
	RET
;
NOENDMSG DB	'No END found.  Continuation file = ',0
