;********************************************************
;*		ASSEMBLER MESSAGE OVERLAY		*
;*							*
;*	This code is an overlay associated with the	*
;*  MACRO-88 assembler.  It is called to generate a	*
;*  piece of text from an error code passed in DE.  	*
;*  buffer pointed to by HL is used for storing the	*
;*  resulting text, and the maximum length is passed	*
;*  in in C.  It can be used, of course, for other 	*
;*  than strict error message generation:  any kind	*
;*  of text can be referred to by a code, so the 	*
;*  whole assembler need not be re-assembled just	*
;*  to change the text of a message.			*
;*	The message storage format is compatible with	*
;*  the System 88 error overlay editor 'Emedit'.	*
;*							*
;********************************************************

	MACLIST 0

	ORG	2000H
	IDNT	$,$		;Weird stuff for overlays.

	DB	'Amsg'		;Assembler message overlay.
	JMP	GO
	DW	ETXT		;Pointer for error editor.

GO	PUSH	D
	PUSH	B
	PUSH	H		;Save buffer pointer
	LXI	H,ETXT
	EI			; let in interrupts!
	JMP	LOOP1		;Start in middle of loop.

DEFAULT LXI	D,0		;Can't find desired message,
	LXI	H,ETXT		;so use the default one.
	JMP	LOOP1

; Search for next message.  HL-> text of last message.

FINDNEXT MOV	A,M		;First, find the zero at the
	ORA	A		;end.
	INX	H
	JNZ	FINDNEXT

LOOP1	MOV	A,M		;Are we looking at right one?
	CPI	0FFH		;None left if code is 0FFH.
	JZ	DEFAULT
	CMP	D
	INX	H
	INX	H		;(leave HL->text in this entry)
	JNZ	FINDNEXT	;No - go find next entry.
	DCX	H
	MOV	A,M
	CMP	E
	INX	H		;(leave HL->text in this entry)
	JNZ	FINDNEXT
	XCHG			;DE -> message.
	POP	H		;Get back buffer ptr.
	POP	B
	PUSH	B
	PUSH	H
LOOP2	MOV	A,C		;Is print buffer empty yet?
	ORA	A
	JZ	AMSG5		;Yes - go return.
	LDAX	D		;Get message char.
	ORA	A		;Terminated by zero.
	JZ	AMSG5		;If no more message, ret.
	MOV	M,A		;If good, move to buff.
	INX	H
	INX	D
	DCR	C
	JMP	LOOP2	

AMSG5	MOV	A,C		;Leave HL->message buffer,
	POP	H
	POP	B		;C=message length.
	SUB	C
	CMA			;length = -(residual count -
	INR	A		;             max length)
	MOV	C,A
	POP	D
	RET
	PAGE
;  The messages themselves.  Terminated by zero.
;  Messages with exclamation points in them are
;  not supposed to appear to the user:  they are
;  system failure messages or internal error codes.
;	The 0000H error code message is also used
;  for defaulting messages not found in the rest
;  of the table.  When the end of the table is reached,
;  the 0FFxxH error code is found, and the table is
;  re-searched for the 0000H message.
;	System message codes are defined here for convenience
;  but since they make the overlay longer, are not normally
;  included.  They may be included by removing the semicolons
;  before them.

err	MACRO
	DB	#1H/256,#1H AND 0FFH,#2,0
	ENDM

ETXT				;For Emedit.

	err  0,'No message for this error!!'

	err  0101,'Division by zero'

	err  0201,'Expr has unbalanced parenthesis'
	err  0203,'String is too long for a double-word'
	err  0204,'Invalid operand'
	err  0205,'Illegal expression'

	err  0301,'Opcode symbol undefined as a macro.'
	err  0302,'OPCODE FLD HAS-NON OPC.!!'
	err  0303,'Opcode field symbol name illegal'
	err  0304,'BAD OPCODE SUBTYPE!!'
	err  0305,'Opcode symbol type illegal'
	err  0306,'Long jump substituted'

	err  0401,'String too long!!'

	err  0501,'Expression too complex !'
	err  0502,'VSTACK UNDERFLOW!!'
	err  0503,'VSTACK GROUP UNDERFLOW!!'

	err  0701,'Bad base specifier'
	err  0702,'NUMBER HAS NON-NUMERIC FIRST CHAR!!'

	err  0801,'Expression nested too deep!!'

	err  0B01,'Missing comma'
	err  0B02,'Invalid register'
	err  0B03,'Register not allowed here'
	err  0B04,'Need Reg8'
	err  0B05,'Need Reg16'
	err  0B06,'Need Sgreg'
	err  0B07,'Suffix/ reg mismatch'

	err  0C01,'Symbol table full!!'
	err  0C02,'SYM PASSED NON-ID!!'
	err  0C03,'BAD INDEX INTO SYM FIELD!!'
	err  0C04,'STINSERT CONFUSED!!'
	err  0C05,'SYMFIND CAN''T FIND SYMBOL!!'
	err  0C06,'Too many symbols to sort !!'
	err  0C07,'BIT BUFFER FULL!!'
	err  0C08,'STFREE MOVED IN PASS 2!!'

	err  0D01,'Number has illegal digit'
	err  0D02,'Base must be 2 -> 36'
	err  0D03,'PRINT FIELD ERR IN @NUM!!'
	err  0D04,'Number has value > 65535'

	err  0E01,'LABEL HAS NO SYMTAB ENTRY!!'

	err  0F01,'ENDM without previous MACRO'
	err  0F02,'BAD POP CODE SUBTYPE!!'

	err  1001,'Backwards ORG/ late RELOC ignored'

	err  1101,'Undefined symbol reference'
	err  1102,'Illegal symbol type referenced'

	err  1203,'Byte value truncated'
	err  1204,'Signed value truncated'
	err  1205,'Value out of range'

	err  1301,'Can''t redefine asmb symbol'
	err  1302,'Doubly defined symbol'
	err  1303,'Cannot SET a defined LABEL'
	err  1304,'Cannot use a SET VAR as a LABEL'
	err  1306,'Phase error - value changed'
	err  1308,'LABDEF LOST A VAR!!'

	err  1501,'ENDIF/ELSE/ELSEIF without IF'

	err  1601,'Statements after END'

	err  1701,'Definition error by REF'
	err  1702,'REF without REFS'
	err  1703,'REF LOST DEF BETWN PASSES!!'
	err  1704,'Symbol not in REF file'

	err  1801,'DEF CAN''T FIND SYMBOL!!'
	err  1802,'DEF given an undefined symbol'
	err  1803,'DEF given a doubly defined symbol'
	err  1804,'DEF without DEFS'
	err  1805,'DEF without symbol name'

	err  1901,'Unmatched brackets'
	err  1902,'Unmatched quotes'

	err  2001,'MSTACK overflow!!'
	err  2002,'MACPEEK buffer too short!!'
	err  2003,'MACPOP on DE=0!!'

	err  2101,'Line too long !'

	err  2201,'Illegal char after ''#'''

	DW	0FFFFH		;End of messages.

	END
