	PAGE
;********************************************************
;*		    OBJECT CODE OUTPUT			*
;*							*
;*	This presently a simple dumper for the data	*
;*  generated - it is written into consecutive bytes	*
;*  of a relocatable load module type of file with	*
;*  an extension of '.RL'.				*
;*	The data is read from the Value Stack and then	*
;*  SKIP number of zeroes are written out to the file.	*
;*  This allows generation of more zeroes than could be *
;*  held in the Value Stack, as when a long distance 	*
;*  forward ORG pop is seen.				*
;*							*
;********************************************************
;
; DUMP
;	Puts the contents of the VSTACK and SKIP number
;  of zeroes out to disk.
;
DUMP	LDA	PASSFLAG	;Only output code in pass 2
	CPI	2
	JNZ	SUCCEED
	CALL	STKPOP
	JC	DUMP3		;Jmp/ shouldn't ever jump.
	CALL	STKPUSH
	JC	DUMP3		;neither should this one.
	;
	LHLD	LOCATION	;Calculate location of the
	CALL	DSUB		; first byte in the inst.
	SHLD	DUMPLOC
	CALL	STKRDINIT	;Prepare for reading VSTACK.
	;
DUMP1	CALL	STKREAD
	JC	DUMP3		;Done when get CARRY.
	CALL	DUMP5		;Stick on disk.
	LDA	RELFLAG
	ORA	A		;Is it ABS?
	JZ	DUMP2		;Jmp/ yup.
	LHLD	DUMPLOC
	CALL	RELADDR 	;Relocate this address.
	RC
DUMP2	LHLD	DUMPLOC
	INX	H		;Bump pointer.
	SHLD	DUMPLOC
	JMP	DUMP1		;Spin for next byte.
	;
DUMP3	LHLD	SKIP		;Now write SKIP # of zeroes.
	MVI	E,0
DUMP4	MOV	A,H
	ORA	L		;If count=zero, stop.
	RZ
	PUSH	H
	CALL	DUMP5		;Write a zero.
	POP	H
	DCX	H
	JMP	DUMP4
;
; DUMP5
;	Dumps the byte in E out to the disk file.
;
DUMP5	LHLD	CHECKSUM	; Update 16 bit running
	MVI	D,0		; object code checksum.
	DAD	D
	SHLD	CHECKSUM
	MOV	A,E		;Output byte in E
	JMP	PUTC		; to the output file.
	PAGE
;********************************************************
;*		RELOCATION SYSTEM			*
;*							*
;*	The relocation scheme uses a sequence of bits 	*
;*  to inidicate which words(addresses)	need to be 	*
;*  relocated.  A 0 bit means skip two bytes.  A 10	*
;*  bit pattern means relocate the word pointed at and	*
;*  then advance two bytes.  A 11 bit pattern means 	*
;*  skip a byte relocate that word and then advance	*
;*  two bytes.  Note that this scheme guarantees that	*
;*  the relocation dictionary is between 1/16 and 1/8	*
;*  the size of the actual program.			*
;*							*
;********************************************************
;
; RELINIT
;	Initialize the relocation system.
;
RELINIT XRA	A		;Clear relocation flag.
	STA	RELOCATE
	RET
;
; RELADDR
;	Relocate the address pointed at by HL.
;
RELADDR LDA	RELOCATE
	ORA	A
	RZ			;Ret/ no relocation
	XCHG
	LHLD	LASTREL		;Last relocation
	XCHG
	INX	H		; point past the two bytes
	INX	H		; to be relocated.
	SHLD	LASTREL 	;Set new last relocation.
	DCX	H
	DCX	H
	CALL	DSUB		;Calculate difference
	XCHG			; move to DE.
RELADR1 MOV	A,D
	ORA	E		;Is DE=0?
	JZ	RELADR2 	;Jmp/ yup, relocate even.
	DCX	D		;Dink count by one.
	MOV	A,D
	ORA	E		;Is DE=0 now?
	JZ	RELADR3 	;Jmp/ yup, relocate odd.
	XRA	A
	CALL	PUTBIT		;Dump a zero bit.
	RC
	DCX	D		;Dink count again.
	JMP	RELADR1
	;
RELADR2 MVI	A,TRUE
	CALL	PUTBIT		;Dump a 1
	RC
	XRA	A
	JMP	PUTBIT		;Followed by a 0.
	;
RELADR3 MVI	A,TRUE
	CALL	PUTBIT		;Dump a 1
	RC
	MVI	A,TRUE		;Followed by another 1.
	JMP	PUTBIT
;
; RELDUMP
;	Output the relocation block to the disk file.
;  But first we must output enough zero bits to force us past
;  the last byte of the program (LOCATION).
;
RELDUMP LDA	RELOCATE
	ORA	A
	RZ
	LHLD	LASTREL		;Last relocated thing
	XCHG
	LHLD	LOCATION
	SHLD	OLDA		;Set file's load address.
	CALL	DSUB		;Calculate difference
	XRA	A		; into HL.
	MOV	A,H
	RAR
	MOV	D,A
	MOV	A,L
	RAR
	MOV	E,A		;Want half the difference
	INX	D		; plus one.
RELDMP1 MOV	A,D
	ORA	E
	JZ	DUMPBITS	;Jmp/ done, dump to disk.
	XRA	A
	CALL	PUTBIT		;Dump a zero bit.
	RC
	DCX	D		;Dink count.
	JMP	RELDMP1
	PAGE
;********************************************************
;*		BIT OUTPUT SYSTEM			*
;*							*
;*	The bit output system uses the space directly	*
;*  after the symbol table free pointer for a buffer	*
;*  space and assumes the symbol table will not grow.	*
;*  If it does we return error 0C07H when the bits are	*
;*  dumped through DUMPBITS.				*
;*							*
;********************************************************
;
;	Initialize the bit output system.
;
BITINIT PUSH	H
	LHLD	STFREE		;Grab ST pointer. UGGGH!!!
	INX	H		; plus four for fake entry
	INX	H		; build by SYMFIND.
	INX	H
	INX	H
	SHLD	BSTART		;Save start of bit map
	SHLD	BITPTR		; and end.
	LXI	H,9
	SHLD	BITWORD 	;Pattern=0, count=8+1.
	XRA	A
	STA	BITOVFLW 	;Clear overflow flag.
	POP	H
	ORA	A
	RET
;
; DUMPBITS
;	Ouput the bits accumulated by successsive calls to
;  PUTBIT to the currently open disk file.
;
DUMPBITS LDA	BITOVFLW
	ORA	A
	JZ	DUMPBT1
	LXI	D,0C07H 	;## BIT BUFFER OVERFLOW ##
	JMP	PUTERR
DUMPBT1 LHLD	BSTART		;BSTART should be exactly
	DCX	H		;Four bytes past STFREE.
	DCX	H
	DCX	H
	DCX	H
	XCHG
	LHLD	STFREE		;Compare to make sure STFREE
	CALL	DCMP		;didn't change in pass 2.
	JZ	DUMPBT2
	LXI	D,0C08H 	;## STFREE MOVED IN PASS 2 ##
	JMP	PUTERR
DUMPBT2 XRA	A
	CALL	PUTBIT		;Output zero bits
	RC
	LHLD	BITWORD
	MOV	A,L
	CPI	8		;To assure that the last bits
	JNZ	DUMPBT2 	; are placed into memory.
	;
	LHLD	BSTART
	XCHG
	LHLD	BITPTR
	CALL	DSUB		;Calculate number of bytes
	XCHG			;buffered.
	LHLD	BSTART		;Start of our buffer.
DUMPBT3 MOV	A,D
	ORA	E
	RZ			;Ret/ done.
	MOV	A,M		;Get byte.
	INX	H		;Bump pointer.
	CALL	PUTC		;Output to disk file.
	RC
	DCX	D		;Dink pointer.
	JMP	DUMPBT3 	;Spin.
;
; PUTBIT
;	Dumps a bit.
;  A<>0 = dump a 1 bit, else dump a 0 bit.
;
PUTBIT	PUSH	B
	PUSH	H
	CALL	PUTBT1
	POP	H
	POP	B
	RET
	;
PUTBT1	PUSH	D
	MOV	B,A		;Save Acc.
	LHLD	BITWORD
	DCR	L		;Dink bit count.
	JNZ	PUTBT3		;Jmp/ haven't shifted 8 times
	MOV	A,H		;Bit pattern in A.
	LHLD	MEMTOP
	XCHG
	LHLD	BITPTR		;Get buffer pointer.
	MOV	M,A		;Stuff bit pattern in buffer.
	INX	H
	CALL	DCMP		;MEMTOP>BITPTR+1?
	JC	PUTBT2		;Jmp/ yup, have space.
	POP	D
	MVI	A,TRUE
	STA	BITOVFLW
	RET
	;
PUTBT2	SHLD	BITPTR		;Update pointer.
	LXI	H,8		;Bit pattern=0, count=7+1.
	;
PUTBT3	POP	D
	MOV	A,B		;Get the 0 or NZ
	ADI	0FFH		;Set CARRY unless 0.
	MOV	A,H
	RAL			;Rotate into bit pattern
	MOV	H,A		; and put back.
	SHLD	BITWORD
	ORA	A
	RET
