	PAGE
;********************************************************
;*		EXPRLIST PARSER 			*
;*							*
;*	This parser uses the EXPR parser and the string *
;* recognizer to return an indefinite length sequence	*
;* of 16 bit integer expression values.  They are	*
;* all returned on the VSTACK, with the number of 	*
;* them in TOS.  It uses the STRING parser to get	*
;* characters from a string, if one is found.  STRING	*
;* is also used by EXPR for string expressions that are *
;* less than 3 characters long.  So, in order to avoid	*
;* ambiguity with 2 character strings (they mean diff-	*
;* erent things to EXPR and EXPRLIST) we rule that 2	*
;* character strings will NOT be submitted to EXPR:	*
;* this means both characters will be used.		*
;*							*
;********************************************************
;
;	The syntax of an expression list is as follows:
;
;<expr.list> ::= { <expr> | <str.expr> | <null> } , <expr.list>
;
;	A wierdism of the parser is that a one character
;  string must be taken as an <expr>, so that math may be
;  performed on it.  A two character string, however, must
;  not be given to expr, because it will throw away the first
;  (high order) character - but we want both characters to
;  assemble.  Hence we must try for a string first, and then
;  backup (!) if it is only one char long and pass it to EXPR.
;
EXPRLIST LXI	D,0
EXPL1	PUSH	H		;Save buffer ptrs for backup.
	PUSH	B
	PUSH	D		;Save value count.
	CALL	STKMARK		;Save VSTK extension for bckup
	PUSH	D
	CALL	STRING
	JC	EXPL2
	MOV	A,E
	XRI	1		;If DE=1, then string is
	ORA	D		;one char long, and should be
	JZ	EXPL2		;parsed by EXPR.
	POP	D
	CALL	STKUNMARK	;Unmark, save VSTK contents.
	POP	D
	POP	PSW		;Throw backup info.
	POP	PSW
	CALL	STKPUSH		;VSTACK=EXPRCOUNT<STRINGLEN<..
	RC
	CALL	STKADD		;VSTACK=NEWEXPRCOUNT<...
	RC
	CALL	STKPOP		;VSTACK=<... , DE=NEWEXPRCOUNT
	RC
	JMP	EXPL4		;Go scan comma, and continue.
;
EXPL2	CALL	STKBACK		;Throw VSTK contents.
	POP	D
	CALL	STKUNMARK
	POP	D
	POP	B		;Backup buffer pointers.
	POP	H
	PUSH	D		;Save value count.
	CALL	BYTE		;Get a byte valued thing.
	JC	EXPL3		;Jmp/ bomb out.
	POP	D
	INX	D		;Count the valid EXPR.
	JMP	EXPL4		;Go scan comma, and continue.
;
EXPL3	XTHL
	XCHG
	CALL	STKPUSH		;Put value count on TOS.
	XCHG
	POP	H
	STC			;Return error as generated by
	RET			;whatever got us here.
;
EXPL4	CALL	BREAK
	CALL	COMMA
	JNC	EXPL1		;Found comma - keep going.
	JMP	STKPUSH		;Put total EXPR count on TOS.
				;Success depends on STKPUSH.
