	PAGE
;********************************************************
;*		INSTRUCTION PARSER			*
;*							*
;*	The OPCODE and OPERAND fields of an assembly	*
;* source line are both handled by this parser.      	*
;* The resulting value of the instruction is returned	*
;* on the VSTACK, 					*
;* with the number of total bytes for the instruction	*
;* on TOS.  						*
;*							*
;********************************************************
;
INSTR	CALL	BREAK
	CALL	VAR		;Get OPCODE.
	JNC	INST1
	LXI	D,0303H    ;** OPCODE FLD. NOT VALID SYMBOL **
	STC
	RET
INST1	CALL	SYMFIND
	JNC	INST2
	LXI	D,0301H		;** OPCODE SYMBOL UNDEFINED **
	STC
	RET
INST2	LXI	D,0		;Zeroth byte is type.
	CALL	SYMGET		;Get value from entry.
	CPI	1		;Must have an OPCODE.
	JZ	INST3	
	LXI	D,0302H		;$$ OPCODE FLD HAS NON-OPC $$
	STC
	RET
INST3	LXI	D,1		;We do have an OPCODE.
	CALL	SYMGET
	RC
	MOV	E,A		;Save OPCODE value on VSTACK.
	MVI	D,0
	CALL	STKPUSH
	RC
	LXI	D,3		;Fourth byte is sub-type.
	CALL	SYMGET
	RC
	CPI	1
	JZ	PAIRTYPE	;INX,DAD,PUSH...
	CPI	2
	JZ	ADDRTYPE	;JMP,CALL,LHLD,LDA...
	CPI	3
	JZ	RRTYPE		;MOV
	CPI	4
	JZ	R2TYPE		;ADD,SUB...
	CPI	5
	JZ	R1TYPE		;DCR,INR,RST
	CPI	6
	JZ	R1D8TYPE	;MVI
	CPI	7
	JZ	D8TYPE		;ADI...
	CPI	8
	JZ	NOPTYPE		;NOP,RET...
	CPI	9
	JZ	BDTYPE		;LDAX,STAX.
	CPI	10
	JZ	PRADTYPE	;PAIR & ADDRESS (eg. LXI ).
	LXI	D,0304H		;## BAD OPCODE SUBTYPE ##
	STC
	RET
;
PAIRTYPE CALL	PAIR?		;Get a pair number.
	LXI	D,3
	CALL	STKPUSH
	RC
	CALL	STKSHL
	RC
	CALL	STKOR		;PAIRCK guarantees no mask
	RC			;needed.
	LXI	D,1		;Length of a PAIRTYPE
	JMP	STKPUSH		;OPCODE.
;
ADDRTYPE CALL	EXPR?
	CALL	ABSADDR 	;Expand top VSTACK element
	RC			;into two one-byte elements.
	LXI	D,3
	JMP	STKPUSH 	;3 byte instruction.
;
RRTYPE	CALL	REG?
	LXI	D,3		;Shift over for R1 field.
	CALL	STKPUSH
	RC
	CALL	STKSHL
	RC
	CALL	COMMA?
	CALL	REG?		;Get second register.
	CALL	STKOR		;Combine them.
	RC
	CALL	STKOR		;Combine with OPCODE.
	RC
	LXI	D,1		;RRTYPE has length one.
	JMP	STKPUSH
;
R2TYPE	CALL	REG?
	CALL	STKOR
	RC
	LXI	D,1
	JMP	STKPUSH
;
R1TYPE	CALL	REG?
	LXI	D,3		;Shift left 3 for R1 field.
	CALL	STKPUSH
	RC
	CALL	STKSHL
	RC
	CALL	STKOR
	RC
	LXI	D,1
	JMP	STKPUSH
;
R1D8TYPE CALL	R1TYPE		;Use stuff from R1TYPE, but
	RC
	CALL	STKPOP		;throw away the 1 on TOS.
	RC
	CALL	COMMA?		;Must have comma.
	CALL	BYTE?		;Get a byte valued expression.
	LXI	D,2		;Real length is 2.
	JMP	STKPUSH
;
D8TYPE	CALL	BYTE?
	LXI	D,2		;Length is 2.
	JMP	STKPUSH
;
NOPTYPE LXI	D,1
	JMP	STKPUSH		;Length is 1.
;
BDTYPE	CALL	BORD?
	LXI	D,3		;Goes in R1 field.
	CALL	STKPUSH
	RC
	CALL	STKSHL
	RC
	CALL	STKOR		;B=0,D=2, so no need to mask.
	RC
	LXI	D,1		;Length is 1.
	JMP	STKPUSH
;
PRADTYPE CALL	PAIRTYPE	;Use code in PAIRTYPE, but
	CALL	STKPOP		;Throw away the length on TOS.
	RC
	CALL	COMMA?		;Must have comma or error.
	JMP	ADDRTYPE
;
; Register operand parsers.
; Return the value of the register on VSTACK.
; The names with '?' will call WARN if can't parse correctly.
; Beware! This parser is susceptible to prefix ambiguity!
;
REG?	CALL	REG		;Do the parse,
	JMP	???		;then go handle possible err.
PAIR?	CALL	PAIR
	JMP	???
BORD?	CALL	BORD
	JMP	???
;
REG	CALL	BREAK
	LXI	D,7		;Accumulator is a 7 code.
	MVI	A,'A'		;(Make D=0 with LXI...)
	CALL	TESTC
	JNC	STKPUSH
	MVI	E,1
	MVI	A,'C'		;C reg is a 1 code.
	CALL	TESTC
	JNC	STKPUSH
	MVI	E,3		;E reg is a 3 code.
	MVI	A,'E'
	CALL	TESTC
	JNC	STKPUSH
	MVI	E,5		;L reg is a 5 code.
	MVI	A,'L'
	CALL	TESTC
	JNC	STKPUSH
	MVI	E,6		;Memory is a 6 code.
	MVI	A,'M'
	CALL	TESTC
	JNC	STKPUSH
	;
PAIR	CALL	BREAK
	LXI	D,6		;Register pairs are a proper
	CALL	TEST		;subset of registers.
	DB	'PSW',0		;PSW is a 6 code.
	JNC	STKPUSH
	CALL	TEST
	DB	'SP',0		;as is the SP.
	JNC	STKPUSH
	MVI	E,4		;H reg is a 4 code.
	MVI	A,'H'
	CALL	TESTC
	JNC	STKPUSH
	;
BORD	CALL	BREAK
	LXI	D,0		;B reg is a 0 code.
	MVI	A,'B'		;B or D spec is proper subset
	CALL	TESTC		;of PAIRs.
	JNC	STKPUSH
	MVI	E,2		;D reg is a 2 code.
	MVI	A,'D'
	CALL	TESTC
	JNC	STKPUSH
	;
	; Not recognized as a valid register.
	;
	LXI	D,0		;Else push something,
	CALL	STKPUSH		;and give error:
	LXI	D,0B03H		;** Invalid register spec. **
	STC
	RET
