	PAGE
;********************************************************
;*		    I/O					*
;*							*
;*	Some of Glenn's original code to do I/O with	*
;*  the crutched-up micro-tec assembler.  The screen	*
;*  driver has been completely rewritten to handle	*
;*  multiple line transfers, and transfers with no	*
;*  carriage returns in them at all.			*
;*	REWIND has been added for the input file.	*
;*							*
;********************************************************
;
;***** PRINTER DRIVER *****
;
; Initialization of WH7 driver.
;
SCNINIT XRA	A
	STA	SPOS		;Print head at left.
	MVI	A,14		;init line count
	STA	SCNT
	RET
;
; Entry point to print a line pointed to by HL.
; Length limit is in C (used if no CR found).
;
SCNPR	MOV	A,C
	STA	BSAVE		;save max length.
	LDA	HARDLIST
	ORA	A
	JNZ	SCNPT		;jmp/ printer on.
;
; Here we put the buffer chars on the screen, with 
; proper spacing for TABs.  If the right edge of the
; screen is found, we eat chars until a CR, and then
; we continue printing chars after that.  We always
; look at all the chars given us, not halting at CR.
;
SCRN	LDA	BSAVE		;BSAVE has the buffer
	ORA	A		;length that was passed us.
	RZ			;We count it down to zero.
	DCR	A
	STA	BSAVE
	LDA	SPOS		;Room to print?
	CPI	64		;Not if at or beyond 64.
	MOV	A,M		;Get char to be printed.
	JNC	SCRN1		;If no room, just skip
	CALL	WH1		;the printing.
SCRN1	CPI	CR
	JZ	SCRN2
	CPI	TAB
	JZ	SCRN3
	LDA	SPOS		;Found a non-CR, non-TAB.
	INR	A
	STA	SPOS
	INX	H
	JMP	SCRN
;
SCRN2	XRA	A		;Found a CR.  zero print
	STA	SPOS		;position, and decrement
	LDA	SCNT		;line count.  If no more
	ORA	A		;lines available on this
	JNZ	SCRN21		;display page, 
	MVI	A,'.'		;Then print a period,
	CALL	WH1
	CALL	WH0		;wait for operator response,
	MVI	A,RB		;then wipe out the dot and
	CALL	WH1		;continue dumping.
	MVI	A,14		;We leave line count at 13.
SCRN21	DCR	A
	STA	SCNT
	INX	H
	JMP	SCRN
;
SCRN3	LDA	SPOS		;Found a TAB.
	ADI	8
	ANI	0F8H		;This does what a TAB does
	STA	SPOS		;to the position counter.
	INX	H
	JMP	SCRN
;
; We have to print the line, rather than display on CRT.
;
SCNPT	LDA	BSAVE		;Is there any more text?
	ORA	A
	RZ
	DCR	A		;Yes - count it.
	STA	BSAVE
	MOV	A,M		;Get char.
	INX	H
	CALL	WH7		;Print anything else now.
	JMP	SCNPT
;
;
; ***** DISK I/O *****
;
;	Read from HL buffer to get file names to open for
;  both input and output (respectively).
;
OPEN	CALL	NXTC
	JNZ	OPEN		;Scan to delimeter
	XRA	A
	STA	INFLG
	STA	OTFLG
	CALL	OPIN		;Open input file
	JC	Err		;Ret/ error - mail out.
	MOV	A,M
	CPI	CR
	RZ			;Output file is optional.
	LXI	B,'GO'		;Default extension.
	CALL	OPOUT		;Open output file.
	JC	Err
	RET
;
;	Open an input file.  Read from buffer at HL to get
;  filename.
;
OPIN	SHLD	FSTPTR		;Save ptr
	MVI	A,CR
	DCX	H
	CMP	M		;We looking at CR yet?
	JZ	ER0703		;Yes - error.
	INX	H
	LXI	D,IFDE		;Setup
	MVI	A,40H
	gfid
	JC	ER0705
	MVI	A,1
	STA	INFLG
	PUSH	H		;Save ptr to input
	LXI	H,IFDE+1	;--> input FDE
	LXI	D,3
	MOV	A,M
	ANI	1FH
	ADD	E
	MOV	E,A
	DAD	D		;--> FDA slot in FDE
	SHLD	IFDP		;Save pointer.
	LXI	D,IFDA
	MVI	B,4
OPLP	MOV	A,M
	STAX	D
	INX	H
	INX	D
	DCR	B
	JNZ	OPLP		;Copy FDE good stuff.
	LHLD	IFDA		;Initialize REWIND variables.
	SHLD	REW1
	LHLD	INBS
	SHLD	REW2
	LXI	H,INBUF+0FFH
	SHLD	IPTR
	POP	H		;Restore H to ptr
	RET
;
;	Open an output file.  BC=Default Extension.
;  Skip the first char, assuming it is a delimeter.
;
OPOUT	MVI	A,CR
	CMP	M
	INX	H		;HL-> 2nd FD if there
	JNZ	TWOFDS		;jmp/ two FD'S given
	LHLD	FSTPTR		;--> First FD in buffer.
TWOFDS	LXI	D,OFDE		;Setup for Gfid
	MVI	A,60H		;Lookup function.
	gfid
	JNC	ER0702		;If already there, report.
	PUSH	H
	LXI	H,0300H		;Must get a 300H error
	CALL	DCMP		;normally.
	POP	H
	JNZ	FAIL		;If not 300H, ret CARRY.

; For two user systems only, get MUNG access to device.

	IF	USERS=2
	LDA	OFDE
	ANI	DEVMASK		; device #
	MOV	C,A
	sett	mung		; see if we can MUNG it
	JC	FAIL		; crap out if we can't
	ENDIF

	MVI	A,1
	STA	OTFLG
	LXI	H,OFDE+1	;--> output FDE
	LXI	D,3
	MOV	A,M
	ANI	1FH
	ADD	E
	MOV	E,A
	DAD	D		;--> FDA in putput FDE
	SHLD	OFDP		;Save pointer
	LXI	D,OFDA
	MOV	A,M
	STAX	D
	INX	H
	INX	D
	MOV	A,M
	STAX	D		;copy FDA
	LXI	H,0
	SHLD	ONBS		;Clear output number sectors.
	LXI	H,OTBUF
	SHLD	OPTR		;Clear output buffer pointer.
	RET
;
; Get a character from disk input file.
;
GETC	CALL	GETBYTE		;Get a binary byte.
	RC			;If empty file, ret CARRY.
	ORA	A		;Skip zeroes.
	JZ	GETC
	STC
	CMC			;Ret NO CARRY.
	RET
;
GETBYTE	LDA	INFLG
	ORA	A
	JZ	ER0703		;No input file.
	PUSH	B
	PUSH	D
	PUSH	H
	LHLD	IPTR
	INR	L		;Bump bottom byte (!!!!RLD).
	JZ	GTBUF		;jmp/ fill tank
	SHLD	IPTR
GTGNG	MOV	A,M
GOAWY	POP	H
	POP	D
	POP	B
	RET
;
GTBUF	LHLD	INBS		;Sector counter.
	MOV	A,H
	ORA	L
	JNZ	GTBF1		;jmp/ more out there.
	STC
	JMP	GOAWY		;Ret CARRY on EOF.
;
GTBF1	DCX	H
	SHLD	INBS		;bump sector ctr.
	LHLD	IFDA
	INX	H
	SHLD	IFDA		;Inc disk address.
	DCX	H
	LDA	IFDE
	ANI	DEVMASK		;Unit #
	MOV	C,A
	MVI	A,1		;1 sector
	LXI	D,INBUF
	MOV	B,A		;Read
	CALL	Dio
	JC	Err		;Oops.
	LXI	H,INBUF
	SHLD	IPTR		;Reset ptr
	JMP	GTGNG
;
;  Put a character to the output file.
;
PUTC	PUSH	B
	PUSH	D
	PUSH	H
	MOV	B,A
	PUSH	B
	LDA	OTFLG
	ORA	A		;Is there an output file?
	JZ	PTAWY		;Jmp/ nope.
	LHLD	OPTR
	MOV	M,B
	INR	L
	SHLD	OPTR
	JNZ	PTAWY		;jmp/ not full
	LHLD	OFDA		;output disk address
	LDA	OFDE
	ANI	DEVMASK		;Unit number
	MOV	C,A		;in C.
	LXI	D,OTBUF		;Flush out buff.
	MVI	A,1		;Tell Dio to move 1 sector.
	MVI	B,0		;Write command.
	CALL	Dio
	JC	Err		;Oops!
	LHLD	OFDA
	INX	H
	SHLD	OFDA		;Inc output disk address.
	LHLD	ONBS
	INX	H
	SHLD	ONBS		;and output file sector ctr.
	XRA	A		;Set zero flag.
	;
PTAWY	POP	B
	MOV	A,B		;Save A but not flags.
	POP	H
	POP	D
	POP	B
	RET
;
; Rewind the input file.
;
REWIND	PUSH	H
	LHLD	REW1
	SHLD	IFDA
	LHLD	REW2
	SHLD	INBS
	LXI	H,INBUF+0FFH
	SHLD	IPTR
	POP	H
	RET
;
; Close output file.
;
SHUT	LDA	OTFLG
	ORA	A
	RZ			;Ignore if no output file.
	LHLD	OPTR		;Clear rest of buffer.
	DCR	L
	INR	L		;Set Z according to L.
	JZ	SHUT3		;If already full, quit.
SHUT2	XRA	A
	CALL	PUTC		;Put zeroes until full.
	JNZ	SHUT2
SHUT3	LDA	OFDE+1		;Get name length.
	ANI	31		;Leave only name length.
	ORI	20H		;Set the 'new' bit.
	STA	OFDE+1
	LHLD	OFDP		;-->FDA slot
	INX	H
	INX	H		;--> NBS slot
	LXI	D,ONBS		;--> source block
	MVI	B,6
SHTLP	LDAX	D		;Move NBS, LA & SA to FDE.
	MOV	M,A
	INX	H
	INX	D
	DCR	B
	JNZ	SHTLP
;
	LXI	H,OFDE
	MVI	A,1
	gfid
	JC	Err

	IF	USERS=2		; Clear allocation
	LDA	OFDE
	ANI	DEVMASK
	MOV	C,A
	clrt	mung
	JC	Err
	ENDIF

	RET
;
; Disk I/O utilities.
;
NXTC	MOV	A,M
	INX	H
	CPI	CR
	RZ
	CPI	TAB
	RZ
	CPI	' '
	RZ
	CPI	','
	RET
;
; Disk I/O errors.
;
ER0702	MVI	E,2	;** Output file already exists **
	DB	21H
ER0703	MVI	E,3	;** No input file specified **
	DB	21H
ER0705	MVI	E,5	;** Input file does not exist **
	MVI	D,07H
	STC		;Return waving the flag.
	RET	
