	PAGE
;********************************************************
;*		LAMBDA CONVERSION			*
;*							*
;*	Substitute the actual parameters for the	*
;*  formal parameters in a line of text.  The formals	*
;*  are represented by pound signs ('#') followed by	*
;*  specifiers, such as a digit, an 'L', an 'O', or a	*
;*  pound sign '#'.  LAMBDA actually writes the text of *
;*  the actual parm into the area passed in HL,C.  	*
;*  There are supposed to be no CR's in a parameter's	*
;*  text.  						*
;*							*
;********************************************************
;
	IF	MACROS
LAMBDA	SHLD	LAMBDP		;Save pointer.
	MOV	A,C		;and length.
	STA	LAMBDLEN
	CALL	GETTEXT		;Get char following '#'.
	JC	LAMBD6		;Ret on EOF.
	CPI	'#'		;Two pounds are a quoted
	JZ	LAMBD0		;pound.
	CPI	'L'		;If 'L', get label field.
	JZ	LAMBD1
	CPI	'O'		;If 'O', get Opcode field.
	JZ	LAMBD2
	CPI	'A'		;If 'A', entire
	JZ	LAMBD3		;parameter list.
	CPI	'0'+1		;If digit, select the parm.
	JC	LAMBD01		;Don't accept a '0'.
	CPI	'9'+1		;But do accept digits 1-9.
	JC	LAMBD4
LAMBD01 LXI	D,2001H		;** Illegal char after '#' **.
	CALL	WARN
	XRA	A
	RET
	;
	; Two pounds request a single, quoted pound.
	;
LAMBD0	LXI	H,POUNDP	;A fake SLS entry with just
	CALL	LAMBD5		;a pound sign in it.
	JMP	LAMBD6
	DW	0
	DB	1
POUNDP	DB	'#'
	;
	; Label field requested for a parm.  Form: #L.
	;
LAMBD1	LHLD	LABP
	CALL	LAMBD5
	JMP	LAMBD6
	;
	; Opcode field requested for a parm.  Form: #O.
	;
LAMBD2	LHLD	OPCP
	CALL	LAMBD5
	JMP	LAMBD6
	;
	; All operand field parms requested as a concatenated
	; parameter.  Form: #A.
	;
LAMBD3	LHLD	PARMP		;Put together the parms.
LAMBD31 PUSH	H
	CALL	LAMBD5		;Put first one in.
	POP	H
	RC
	CALL	MACPOP		;CDR down the list.
	JC	LAMBD6		;If list was null go away,
	MOV	A,H		;or if it became null, also
	ORA	L		;go away.
	JZ	LAMBD6
	PUSH	H
	LXI	H,COMMAP	;Points at a dummy comma entry.
	CALL	LAMBD5
	POP	H
	RC
	JMP	LAMBD31	
	DW	0		;This is a fake SLS entry,
	DB	1		;containing only a comma.
COMMAP	DB	','
	;
	; A selected operand field parm requested.
	; Form:  #1,#2,#3,#4...
	;
LAMBD4	SBI	'0'		;Convert digit to binary.
	MOV	B,A
	LHLD	PARMP		;CDR down parameter list.
LAMBD41 MOV	A,B
	ORA	A
	JZ	LAMBD42		;If Done, go copy the parm.
	DCR	B
	CALL	MACPOP		;Else CDR one more element.
	JNC	LAMBD41
LAMBD42 CALL	LAMBD5		;Copy that one, fall out.
	;
	; Go back.  Leave HL,C->remaining text.
	;
LAMBD6	LHLD	LAMBDP
	LDA	LAMBDLEN
	MOV	C,A		;Leave HL,C->remaining text.
	RET
	;
	; HL has a pointer to an element on the SLS.
	; copy it into the text at LAMBDP,LAMBDLEN
	; Move LAMBDP and shorten LAMBDLEN.
	;
LAMBD5	MOV	A,H
	ORA	L
	RZ			;If list is null, forget it.
	XCHG
	LHLD	LAMBDP
	XCHG
	LDA	LAMBDLEN
	MOV	B,A		;DE,B->text area.
	CALL	MACPEEK		;Fill in the text.
	RC			;Give up if buffer too short.
	MOV	L,B
	MVI	H,0
	DAD	D
	SHLD	LAMBDP		;Move text pointer by act.len.
	LDA	LAMBDLEN
	SUB	B		;Shorten by actual parm len.
	STA	LAMBDLEN
	XRA	A
	RET
	ENDIF
;
; GETLINE reads an input line from either the input file
; or the text in the macro whose symbol table key is in
; MACKEY.  The index into the MACKEY element of the symtab
; is in MACINDEX.  If MACKEY is zero, there is no macro, and
; we read from disk.  If we find a pound sign in reading
; the line, some text is to be substituted for it and we
; call LAMBDA to do this.  LAMBDA does a 'lambda conversion'
; by substituting actual parameters in for the '#' symbols, 
; which represent formal parameters.
;
GETLINE LXI	H,LINEBUF	;New line goes into linebuf.
	MVI	C,LINEBFLEN	;Length of line buffer in C.
;
GETLN1	MOV	A,C		;If no more room, truncate
	ORA	A		;by just returning.
	JNZ	GETLN11
	LXI	D,2101H		;** Line too long - Sorry **
	CALL	WARN
	JMP	GETLN10		;Go away.
GETLN11 CALL	GETTEXT		;Get a char.
	RC			;On EOF, ret CARRY.
	MOV	M,A
	;
	IF	MACROS
	CPI	'#'		;Pounds are parameter refer-
	JNZ	GETLN12		;ences in the macro version.
	PUSH	H
	LHLD	MACKEY		;Doing an expansion?
	MOV	A,H
	ORA	L
	POP	H
	MOV	A,M
	JZ	GETLN12		;Only do Lambda conversion on
	CALL	LAMBDA		;expansions:  not definitions
	JMP	GETLN1		;or main text.
	ENDIF
	;
GETLN12 DCR	C
	INX	H
	CPI	CR		;Done if it was a CR.
	JNZ	GETLN1
	;
	; Normal return path.
	;
GETLN10 LXI	H,LINEBUF	;HL-> text we read.
	MVI	A,LINEBFLEN	;C = remaining line buff.
	SUB	C
	MOV	C,A		;C = length of text we read.
	XRA	A		;ret NO CARRY.
	RET
;
; Get a character from disk.
;
	IF	NOT MACROS
GETTEXT	PUSH	H
	PUSH	B
	CALL	GETC
	POP	B
	POP	H
	RET
	ENDIF
;
; Get a character - from macro or disk.
; Even '#' characters will be fetched: LAMBDA works on
; the next higher level than GETTEXT (it calls GETTEXT).
;
	IF	MACROS
GETTEXT PUSH	H
	PUSH	B
	CALL	GETTXT1
	POP	B
	POP	H
	RET
GETTXT1	LHLD	MACKEY		;Is this a macro expansion?
	MOV	A,H
	ORA	L
	JZ	GETC		;jmp/ no.
	;
	; Macro expansion - get from macro prototype.
	; (HL=MACKEY)
	;
	XCHG
	CALL	SYMUNKEY	;Symtab points at macro entry.
	LHLD	MACINDEX
	INX	H		;Bump it (skips type code on
	SHLD	MACINDEX	;first call).
	XCHG
	CALL	SYMGET		;Get byte from text.
	RNC			;Return it if found.
	;
	; Macro out of text.  Pop expansion level and
	; try again for a character.
	;
	LHLD	EXPANDP		;Head of expansion list.
	CALL	MACPRUNE	;Set free pointer there.
	LXI	D,MSTG		;Macro ptr. storage area.
	MVI	B,MSTGEN-MSTG	;Length of storage area.
	CALL	MACPEEK		;Read last act. rec.
	SHLD	EXPANDP		;Pops EXPANDP. (Ignore CARRY)
	JMP	GETTXT1		;Try again.
	ENDIF
