	IF	MACROS
	PAGE
;********************************************************
;*		PARAMETER PARSER			*
;*							*
;*	Macro calls can have a list of text elements	*
;*  called 'parms' which are substituted for occurr-	*
;*  rances of '#1','#2'... in the macro text before	*
;*  the text is assembled.  The parms have a special	*
;*  form, similar to the elements of an 'expression-	*
;*  list, but which are not evaluated when parsed.	*
;*	A parm is a sequence of characters terminated	*
;*  by a comma, or it is a '[' followed by a list of	*
;*  comma separated parms followed by a ']'.  In the	*
;*  first case, the blanks are stripped off the ends	*
;*  end of the parm.  In the latter case, only the	*
;*  surrounding brackets are stripped off, leaving any	*
;*  blanks between the  brackets intact, as well as	*
;*  the internal commas between internal parms.  Also,	*
;*  a bracket parm may contain other nested internal	*
;*  bracket pairs, as long as they are matched.  If	*
;*  a bracket is found inside an apostrophe-quoted	*
;*  string, it does not count in the matching.  In	*
;*  short, the brackets are a kind of 'quote'.		*
;*	A bracketed parm need not be delimited from	*
;*  its neighbors by the usual comma.  The brackets	*
;*  serve as delimiters as well as a comma.  A comma	*
;*  may be used before or after a bracketed parm with	*
;*  no effect other than increased clarity.		*
;*							*
;********************************************************
;
;	Augmented transition network description of this
; parser can be found in 'Deeper into the Gutz of MACRO-88'.
;	The parm that is located from the text passed in
; HL,C is returned in DE,B, and HL,C is moved to immediately
; after the parm.  A comma, if present after the parm, is
; left as the first char in HL,C.
;	Labels in the parser:  PARM21 is a label to jump to
; when going from state 1 to state 2.  If many such exist,
; they are given extra letters on the end.  The state itself
; gets just PARM2 label.  Before each state are the entry
; points; after the state is the decision tree for exiting.
;
; Errors:   0H  null parm list - simple mismatch.
;	   1901H ** Parm list - unmatched brackets **
;	   1902H ** Parm list - unmatched quotes **
;
;
; State 1 is eating initial spaces.
;
PARM	XRA	A
	STA	DEPTH
	MOV	D,H
	MOV	E,L
	MVI	B,0
	JMP	PARM1
PARM11	INX	H
	DCR	C
	INX	D
PARM1	MOV	A,C
	ORA	A
	JZ	PARM2		;If C==0 then parm is null.
	MOV	A,M		;Check for a space char.
	CPI	' '+1		;All control chars and space
	JC	PARM11		;are accepted as spaces.
	CPI	'['		;Type two parm? Go to 21.
	JNZ	PARM2		;Any other char goes to 2.
;
;  State 2 is reading chars from the parm.
;
PARM21	INX	D
	MVI	A,1
	STA	DEPTH
	INX	H
	DCR	C
	JMP	PARM2
PARM22B LDA	DEPTH
	DCR	A
	STA	DEPTH
	JMP	PARM23		;Share code.
PARM22A	LDA	DEPTH
	INR	A
	STA	DEPTH		;Fall into shared code.
PARM23	INR	B		;From string state.
	DCR	C
	INX	H
PARM2	MOV	A,C
	ORA	A
	JZ	PARM2A
	MOV	A,M
	CPI	''''
	JZ	PARM32	
	LDA	DEPTH		;C!=0.
	ORA	A
	JZ	PARM2C
	CPI	1		;Depth!=0.
	JNZ	PARM2D
	MOV	A,M		;Depth=1.
	CPI	']'		;Last right bracket is the one
	JZ	PARMOK2		;at depth one.
PARM2D	MOV	A,M		;Depth!=1 && Depth!=0.
	CPI	'['
	JZ	PARM22A		;Nest one level.
	CPI	']'
	JZ	PARM22B		;Unnest one level.
	JMP	PARM23
	;
PARM2A	LDA	DEPTH		;C=0.
	ORA	A
	JZ	PARM4		;Ok only if at zero depth.
	JMP	PARMNE		;Nest error.
	;
PARM2C	MOV	A,M		;Depth=0.
	CPI	'['		;At zero depth, [ is term.
	JZ	PARM4
	CPI	','		;As well as comma.
	JZ	PARM4
	CPI	']'		;and a right bracket,
	JZ	PARM4
	CPI	';'		;or the comment delimiter.
	JZ	PARM4
	JMP	PARM23		;Any other char just accepted.
;
; State 3 is in the middle of a string.
;
PARM33
PARM32	INX	H
	DCR	C
	INR	B
PARM3	MOV	A,C
	ORA	A
	JZ	PARMSE		;If out of text, string error.
	MOV	A,M
	CPI	''''
	JZ	PARM23
	JMP	PARM33
;
; State 4 is a transition state for pushing HL.
; It's kind of an optimization.
;
PARM4	PUSH	H
	DCX	H
	JMP	PARM5
;
; State 5 is eating spaces at the end of the parm.
;
PARM55	DCX	H		;Space found, so backup.
	DCR	B
PARM5	MOV	A,B
	ORA	A
	JZ	PARMOK5		;If parm null, done.
	MOV	A,M		;A space is backed up over.
	CPI	' '+1		;Any ctl char or ' ' is space.
	JC	PARM55
	JMP	PARMOK5
;
PARMSE	LXI	D,1902H  	;** Unmatched quotes **
	STC
	RET
;
PARMNE	LXI	D,1901H 	;** Unmatched brackets **
	STC
	RET
;
PARMOK2 INX	H
	DCR	C
	XRA	A
	RET
PARMOK5 POP	H
PARMOK	XRA	A
	RET
	ENDIF			;End of MACROS.
