	PAGE
;********************************************************
;*		LISTING GENERATORS			*
;*							*
;*	All these routines format things for the 	*
;*  assembly listing.  Their names start with PUT.	*
;*							*
;********************************************************
;
;  PUTSYMS prints the sorted table of all the symbols
;  that were defined as labels in this assembly.  We
;  do a selection sort by calling SYMPRESORT to get to
;  the first lexicographical element in the symtab and
;  then SYMSORTNEXT to get to successive elements.  These
;  functions are part of the symbol table system, and
;  will do the searching for us, without us having to
;  violate the imbedded data structure.  SYMNAME gets
;  the name of the symbol currently being examined.
;
; The size of the name subfield of a symbol printout:
;
SLSIZE	EQU	11		;16-4(hex digits)-1=11
;
PUTSYMS
	IF	MACROS		;First, the macros.
	LXI	H,MACHEADER
	MVI	C,MACHDLEN
	MVI	A,8 OR 80H	;List them with no value
	CALL	PUTSYM1 	;field printout (the OR 80H
	ENDIF			;does it).
	;
	LXI	H,UDLHEADER
	MVI	C,UDLHDLEN
	MVI	A,6 OR 80H	;List UNDEF LABELs first.
	CALL	PUTSYM1 	;The OR 80H puts in the 'no
	LXI	H,DDLHEADER	;value' bit which suppresses
	MVI	C,DDLHDLEN	;the value printout.
	MVI	A,7		;Then DOUBLE DEF LABELs.
	CALL	PUTSYM1
	LXI	H,SVRHEADER
	MVI	C,SVRHDLEN
	MVI	A,5
	CALL	PUTSYM1
	LXI	H,LABHEADER
	MVI	C,LABHDLEN
	MVI	A,4		;Finally, LABELs.
	CALL	PUTSYM1
	LXI	H,CRMSG 	;Always leave carriage
	MVI	C,1		;at left.
	JMP	PUTLINE
;
;  Here we print all the symbols in the symbol table
;  of the type passed in in A in sorted (alphabetic)
;  order by name.  If there is no symbol of that type,
;  we do nothing, but if there are some, we print the
;  'header' pointed to by HL, with length in C.
;	If bit 8 of A was set on entry, we print only the
;  name of the thing.  Otherwise, we take the value of the
;  thing to be the SYMGET(1) and SYMGET(2) bytes (Gotten by
;  SYMVAL).
;
PUTSYM1 STA	PUTSMTYPE	;Save symbol type, leave in A.
	ANI	NOT 80H		;All but the 'novalue' bit.
	MOV	D,A		;SYMPRESORT & SYMSORTNEXT use
	MOV	E,A		;A,D and E for alternate types.
	CALL	SYMPRESORT	;Setup for the sort.
	JC	PUTSYM5 	;Jmp/ something special.
	CALL	PUTLINE		;Print header for table.
PUTSYM2 MVI	A,4		;Start new line of listing.
	STA	PUTSMCNT
	LXI	H,CRMSG		;Print CR to left zero print
	MVI	C,1		;head.
	CALL	PUTLINE
PUTSYM3 CALL	PUTSYMN 	;Print the symbol name.
	LDA	PUTSMTYPE
	ANI	80H		;Get the bit that says we put
	JNZ	PUTSYM4 	;print numbers. jmp/ don't.
	CALL	SYMVAL		;Get the value.
	LDA	PTBASE		;What is print base?
	CALL	PFCALC		;and print field for that
	MOV	C,A		;base? (for 16 bit #)
	LDA	PTBASE		;Leave base in B.
	MOV	B,A
	PUSH	B
	MVI	A,'0'		;Pad with ASCII zeroes.
	LXI	H,LISTBUF
	CALL	@NUM		;Generate number.
	POP	B		;Get back print field.
	RC	
	LXI	H,LISTBUF
	CALL	PUTLINE		;Print number.
	;
	;  Done printing number.  Print space, get next thing
	;  to print or return if none.
	;
PUTSYM4 LXI	H,SPMSG 	;A buffer with just ' '.
	MVI	C,1
	CALL	PUTLINE
	LDA	PUTSMTYPE	;Pass desired type in A.
	ANI	NOT 80H		;(All but top bit.)
	MOV	D,A		;and in alternate type
	MOV	E,A		;parameter cells.
	CALL	SYMSORTNEXT	;Skip to next higher name.
	JC	SUCCEED
	LDA	PUTSMCNT	;Count down symbols/line.
	DCR	A
	STA	PUTSMCNT
	JNZ	PUTSYM3 	;Keep going on this line.
	JMP	PUTSYM2 	;Start new line.
	;
PUTSYM5 MOV	A,D		;Either have no symbols of
	ORA	E		; given type/ true error
	RZ			;Ret/ it's the former.
	PUSH	D
	CALL	PUTLINE 	;Put up header
	POP	D
	JMP	PUTERR		; and spew error.
;
;
THREESP   DB	' ',' ' 	;Buffer with three spaces.
SPMSG	  DB	' ',0		;Buffer with just one.
;
UDLHEADER DB	CR,CR,TAB,'Undefined symbols:'
	  DB	CR,0
UDLHDLEN  EQU	$-UDLHEADER
;
DDLHEADER DB	CR,CR,TAB,'Doubly defined symbols:'
	  DB	CR,0
DDLHDLEN  EQU	$-DDLHEADER
;
SVRHEADER DB	CR,CR,TAB,'Set variables used:'
	  DB	CR,0
SVRHDLEN  EQU	$-SVRHEADER
;
LABHEADER DB	CR,CR,TAB,'Labels defined in this assembly:'
	  DB	CR,0
LABHDLEN  EQU	$-LABHEADER
	  ;
	  IF	MACROS
MACHEADER DB	CR,CR,TAB,'Macros defined in this assembly:'
	  DB	CR,0
MACHDLEN  EQU	$-MACHEADER
	  ENDIF
;
;
;	Print the currently referenced symbol in a
; field SLSIZE chars wide padded with spaces.
;
PUTSYMN LXI	D,LISTBUF	;Blank SLSIZE bytes of
	MOV	H,D		;the LISTBUF.
	MOV	L,E
	MVI	B,SLSIZE	;Size of print field in
	MOV	C,B		;listing for one symbol
PUTSYN1 MVI	M,' '		;name + max number len.
	INX	H
	DCR	C
	JNZ	PUTSYN1
	CALL	SYMNAME 	;Get name of current symbol
	LXI	H,LISTBUF	;into LISTBUF over blanks.
	MVI	C,SLSIZE
	JMP	PUTLINE		;Print name.
;
;
; PUTLIST puts out the listing text, which includes the
; object code and source text, followed by any extra
; lines necessary for the rest of the object code.
;
OBJSIZE EQU	16		;Two TABs worth.
AFSIZE	EQU	5		;4 hex digits + ' '.
;
PUTLIST PUSH	H
	PUSH	B
	CALL	STKRDINIT	;Prepare to dump VSTACK
	CALL	PUTLS4		;Generate object code part.
	CALL	PUTLINE		;Print it.
	POP	B
	POP	H
	CALL	PUTLINE		;Print source line.
PUTLS9	CALL	PUTLS4		;Generate possible extension.
	RC			;If no bytes printed, ret.
	CALL	PUTLINE
	LXI	H,CRMSG		;A buffer with just a CR.
	MVI	C,1
	CALL	PUTLINE
	JMP	PUTLS9
;
;
; PUTMLIST is just like PUTLIST except it modifies the
; listing depending on the value of LISTFLAG, defined
; as follows:
;	2: "code lines" only, object field listed
;	3: "code lines" only, object & source fields listed
;	default: full listing
;
PUTMLIST PUSH	H
	PUSH	B
	CALL	STKRDINIT	;Prepare to dump VSTACK
	CALL	PUTLS4		;Generate object code part.
	JNC	PUTMLS1
	LDA	LISTFLAG	;No object code was generated
	CPI	2		; so if either case 2 or 3
	JZ	PUTMLS4 	; just return.
	CPI	3
	JZ	PUTMLS4
PUTMLS1 CALL	PUTLINE		;Print it.
	POP	B
	POP	H
	LDA	LISTFLAG
	CPI	2
	JZ	PUTMLS3 	;Jmp/ object code only.
	CALL	PUTLINE 	;Print source line.
PUTMLS2 CALL	PUTLS4		;Generate possible extension.
	RC			;If no bytes printed, ret.
	CALL	PUTLINE
PUTMLS3 LXI	H,CRMSG 	;A buffer with just a CR.
	MVI	C,1
	CALL	PUTLINE
	JMP	PUTMLS2
PUTMLS4 POP	B
	POP	H		;No CARRY (by JZ)
	RET
;
;  We are supposed to generate the object code part of the
;  listing line and return a pointer to it in HL for printing.
;  If no data bytes were printed, ret CARRY.  We inc AFIELD
;  by the number of bytes dumped.
;
PUTLS4	LXI	H,LISTBUF	;Convert into list buffer.
	MVI	C,OBJSIZE	;Only use part of it.
	LDA	AFLDBLK		;Do we print address field?
	ORA	A
	JNZ	PUTLS1		;If nonzero, we don't.
	XCHG
	LHLD	AFIELD		;We do.
	XCHG
	LDA	PTBASE
	CPI	4		;If base < 4, no AFIELD on
	JNC	PUTLS8		;next time through.
	MVI	A,0FFH		;(The AFIELD takes up the
	STA	AFLDBLK		;whole object code area.)
	LDA	PTBASE
PUTLS8	MOV	B,A		;What base we printing in?
	PUSH	B
	CALL	PFCALC		;A=16 bit pfld.
	POP	B
	PUSH	PSW		;Save pfld.
	MOV	C,A
	MVI	A,'0'		;Pad with zeroes.
	CALL	@NUM		;Print DE, bump HL.
	JC	PUTLS71 	;If no good, system error.
	POP	PSW		;Apfld.
	MOV	B,A
	MVI	A,OBJSIZE
	SUB	B
	MOV	C,A		;C=remaining buffer.
	JZ	PUTLS10		;Jmp/ handle zero case.
PUTLS1	MVI	M,' '		;Space to AFIELD SIZE
	INX	H
	DCR	C
	MOV	A,C
	CPI	OBJSIZE-AFSIZE
	JNC	PUTLS1		;Go till C is less.
PUTLS10 XRA	A
	STA	PUTLCNT		;No bytes dumped yet.
PUTLS2	LDA	PTBASE
	MOV	B,A
	PUSH	B
	CALL	PFCALC		;What is print field?
	MOV	A,B		;8 bit pfld in B.
	POP	B		;A=pfld,B=base,C=buf size.
	INR	A		;Allow for space after num.
	CMP	C		;Should be less than buf siz.
	JNC	PUTLS6		;No - Go fill rest with ' 's.
	DCR	A		;(Correction for INR above)
	CALL	STKREAD		;Got space - get value to
	JC	PUTLS6		;Print unless no more values.
	PUSH	PSW
	PUSH	B
	MOV	C,A		;pfld <C, so we print byte.
	MVI	A,'0'		;B=base,C=print field,A=padding
	MVI	D,0		;Printing only bottom 8 bits.
	CALL	@NUM		;Print 8 bit value.
	JC	PUTLS7		;If no good, system error.
	POP	B
	POP	PSW
	MOV	B,A
	MOV	A,C
	SUB	B
	MOV	C,A		;Dec C by pfld used.
	LDA	PUTLCNT
	INR	A		;Count printed byte.
	STA	PUTLCNT
	PUSH	H
	LHLD	AFIELD		;For each printed byte,
	INX	H		;we bump the address field
	SHLD	AFIELD		;display.
	POP	H
	JMP	PUTLS2		;Keep going.
;
PUTLS6	MOV	A,C
	ORA	A		;Handle zero case.
	JZ	PUTLS61
PUTLS62 MVI	M,' '		;Fill rest of buff with ' 's.
	INX	H		;Length of buffer was exactly
	DCR	C		;the length of the desired
	JNZ	PUTLS62 	;object code area.
	PUSH	H
	LHLD	MACKEY
	MOV	A,H
	ORA	L
	POP	H
	JZ	PUTLS61 	;Jmp/ not in macro expansion.
	DCX	H		; else we must replace
	MVI	M,'+'		; the last space with a '+'
	;
PUTLS61 LXI	H,LISTBUF	;Return pointing at finished
	MVI	C,OBJSIZE	;product.
	LDA	PUTLCNT
	ORA	A
	JZ	FAIL
	JMP	SUCCEED
;
PUTLS7	POP	B
PUTLS71 POP	PSW		;@NUM blew up !!!!!!
	JMP	PUTERR
;
CRMSG	DB	CR,0		;For printing a CR.
;
; PUTERR puts out error messages.  An error message
; looks like: 
;  ** ERROR 0B01H: Missing operand - comma assumed **
;
PUTERR	PUSH	B
	PUSH	D
	PUSH	H
	CALL	PUTER1
	POP	H
	POP	D
	POP	B
	XRA	A
	RET
	;
PUTER1	MOV	A,D
	ORA	E
	RZ			;Just return if zero.
	;
	PUSH	D		;Save error code.
	LHLD	ERRCOUNT
	INX	H		;Count number of errors.
	SHLD	ERRCOUNT
	MVI	C,ERRMS1LEN
	LXI	H,ERRMSG1	;Print '	** ERROR '
	CALL	PUTLINE
	CALL	PUTHEXN 	;Print error number in hex.
	LXI	H,ERRMSG2
	MVI	C,ERRMS2LEN
	CALL	PUTLINE		;Print ': '
	;
	;  Generate message text through Amsg.OV
	;
	POP	D		;Get error code back.
	LXI	H,LISTBUF	;Put message in LISTBUF.
	MVI	C,LISTBFLEN
	CALL	Gover
	DB	'Amsg'
	CALL	PUTLINE		;Print message
	;
	LXI	H,ERRMSG3	;Print ' **',CR
	MVI	C,ERRMS3LEN
	CALL	PUTLINE
	RET
;
;  Print a hex number.  We use LISTBUF for conversion.
;
PUTHEXN LXI	H,LISTBUF
	MVI	B,16		;Base 16 (hex)
	MVI	C,4		;Print field.
	MVI	A,'0'		;Pad with zeroes.
	CALL	@NUM		;This bumps HL by 4, prints DE
	LXI	H,LISTBUF
	MVI	C,4
	JMP	PUTLINE		;Print results.
;		
;  Error message templates.
;
ERRMSG1   DB	'	** ERROR '
ERRMS1LEN EQU	$-ERRMSG1
;
ERRMSG2   DB	': '
ERRMS2LEN EQU	$-ERRMSG2
;
ERRMSG3   DB	' **',CR
ERRMS3LEN EQU	$-ERRMSG3
;
; PUTLINE gets rid of a line pointed to by HL with length
; in C. SCNPR is a routine that gets rid of one char
; by putting it on the screen or out to WH7, which contains
; a system line printer driver.
;
PUTLINE PUSH	B		;Print that line.
	PUSH	H
	CALL	SCNPR		;See code in Glenn's drivers.
	POP	H
	POP	B
	XRA	A
	RET
