	PAGE
;********************************************************
;*		STRING PARSER				*
;*							*
;*	This parser puts all the ASCII code values	*
;*  of the characters in a quoted string onto the	*
;*  VSTACK and leaves the total number of chars as	*
;*  the top element of VSTACK.  It recognizes single	*
;*  quotes as end-delimiters, and an occurrence of two	*
;*  single quotes inside the string is taken as a	*
;*  literal single quote.  The null string, or just	*
;*  quotes back to back, has a value given by a         *
;*  single zero value (count) on the VSTACK.		*
;*							*
;********************************************************
;
;  This is a simple three state finite state machine.
; The three states are: (1) Looking at a possible starting
; quote, (2) Eating characters between quotes, looking for
; the final quote mark, and (3) Looking after the final
; quote for an immediate double quote, which indicates
; a quoted single quote.
;
; (1) Looking for possible starting quote.
;
STRING	LXI	D,0		;Start char count at zero.
	PUSH	D		;Count stays on 8080 stack.
	CALL	BREAK
	MVI	A,027H		;Single quote?
	CALL	TESTC
	JNC	STRG1		;Yes - got a string.
	POP	D
	LXI	D,0		;Simple mismatch code.
	STC
	RET
;
;  (2) Eating characters between quotes.
;
STRG1	MVI	A,027H		;Have we found balancing quote?
	CALL	TESTC
	JNC	STRG3		;Yes - goto state (3).
	MOV	A,C
	ORA	A		;Is there anything left?
	JZ	STRG3		;No - act like balancing quote
	POP	D		;Count the char.
	INX	D
	PUSH	D
	MOV	E,M		;Get the char in the STRING.
	MVI	D,0		;Make a double-word.
	INX	H
	DCR	C
	CALL	STKPUSH		;Put it on VSTACK.
	JNC	STRG1		;Keep loading if no overflow.
;
STRG6	CALL	STKPOP		;Overflow.
	POP	D		;Put addmittedly short count
	CALL	STKPUSH		;on TOS.
	LXI	D,0401H		;** STRING TOO LONG **
	STC
	RET
;
;  (3) Looking after possible termination quote for an
; immediate quote that would indicate a quoted quote.
;
STRG3	MVI	A,027H
	CALL	TESTC		;Is this an imbedded quote?
	JNC	STRG5		;Yes - go load a quote & cont.
	POP	D		;No - done with string.
	PUSH	D
	CALL	STKPUSH		;Save string length on TOS.
	JC	STRG6		;If no room, string too long.
	POP	D		;DE=string len.
	RET			;Else success.
STRG5	LXI	D,027H		;Load quote literally.
	CALL	STKPUSH
	JC	STRG6		;If no room, string too long.
	POP	D
	INX	D		;Count that as a char.
	PUSH	D
	JMP	STRG1		;Keep loading chars.
