	PAGE
;********************************************************
;*		SYMBOL TABLE FUNCTIONS			*
;*							*
;*	Rules:  A variable name is indicated by a 	*
;* pointer (to a buffer area containing the name)	*
;* and a length of the name.  These utilities		*
;* conventionally pass the length in B, and the pointer *
;* in DE.  This makes them compatible with the parser	*
;* utilities, which produce these kinds of descriptors. *
;* An element in the symbol table has a name field, 	*
;* and a value field, or value area.  The name field	*
;* is used only by the symbol table functions 		*
;* themselves, but the value area is available,         *
;* via the symbol table functions, for any use		*
;* desired.						*
;*							*
;********************************************************
;
;	**** SYMBOL TABLE INITIALIZATION ****
;
;SYMINIT.  Initialize the symbol table functions.  Note:
;we add two to SYMLIST and SYMLLAST in order to point past
;the linkage bytes in front of the symbols.
;
SYMINIT LXI	H,SYMLIST+2
	SHLD	STBASE
	SHLD	STREF
	LXI	H,SYMLLAST+2
	SHLD	STLAST
	LXI	H,SYMLLEN
	SHLD	STFREE
	;
	;	Now we must initialize the original symbol
	;  list by putting the symbols onto the appropriate
	;  linked list in the hash table.
	;
	LXI	H,HASHTAB	;Zap hash table
	LXI	B,HASHSIZ*2
	XRA	A
	CALL	FILL
	LXI	H,SYMLIST+2
SYMIN1	CALL	SYMLINK 	;Put on appropriate chain.
	RC
	MOV	E,M
	INX	H
	MOV	D,M		;Get total length.
	DCX	H
	DAD	D		;Point at next symbol.
	INX	H
	INX	H
	LXI	D,SYMLLAST
	CALL	DCMP
	JC	SYMIN1		;Jmp/ not done yet.
	RET			;(NO CARRY by JC).
;
;	**** SYMBOL DEFINITION FUNCTIONS ****
;
;SYMENTER puts the given name into symtab, creating
;a new symbol element.  It leaves nothing in the value
;area of the new element.  The resulting entry is completely
;self-consistent, although undefined.
;If the symbol table overflows while entering the new 
;element, we return CARRY and error 0C01H.
;The name of the new symbol is pointed to by DE, with
;length in B.  The name is checked to be sure it follows
;the rules for a valid name:  It must start with a letter
;and contain only letters, digits, or the characters
;at sign (@), question mark (?) or underscore (_).
;If it does not, internal error code 0C02H is returned,
;with CARRY.
;
SYMENTER PUSH	H
	PUSH	B
	MOV	H,A
	PUSH	H		;Save A without saving CARRY.
	CALL	SYME1
	POP	H
	MOV	A,H
	POP	B
	POP	H
	RET
SYME1	LHLD	STFREE		;Give SYMENTER a new pointer.
	INX	H
	INX	H		;Make space for linkage bytes.
	SHLD	STFREE
	SHLD	STLAST
	XCHG			;Put DE pointer into HL in 
	SHLD	DSAVE
	MOV	C,B		;order to use standard VAR fcn.
	CALL	VAR		;Did user give us garbage name?
	JNC	SYME2
	LXI	D,0C02H
	RET			;(CARRY by JNC)
SYME2	MOV	C,B
	MVI	B,0
	INX	B
	INX	B
	INX	B
	MOV	A,C		;First part of an entry is
	CALL	SYMFILL 	;the total length.  For now,
	RC
	MOV	A,B		;total length=name length+3.
	CALL	SYMFILL
	RC
	DCX	B
	DCX	B
	DCX	B		;Restore name length in C.
	MOV	A,C
	CALL	SYMFILL 	;Put in name length byte.
	RC
	LHLD	DSAVE
SYME3	MOV	A,C
	ORA	A
	JZ	SYME4
	DCR	C
	MOV	A,M
	INX	H
	CALL	SYMFILL 	;write in characters from name.
	RC
	JMP	SYME3
SYME4	LHLD	STLAST
	CALL	SYMLINK 	;Link symbol into hash table.
	ORA	A
	RET	
;
;SYMLINK.  Links the symbol pointed at by HL onto the
;appropriate linked list in the hash table.
;The symbol is added onto the end of the chain to guarantee
;that the system symbols (opcodes and popcodes) will remain
;near the beginning of the list and thus always be fast to
;find.
;
SYMLINK PUSH	D
	PUSH	B
	MOV	B,A
	PUSH	B
	CALL	SYMLNK1
	POP	B
	MOV	A,B
	POP	B
	POP	D
	RET
SYMLNK1 PUSH	H		;Need this later.
	INX	H
	INX	H
	MOV	B,M		;Name length in B.
	INX	H
	XCHG			;Pointer to name in DE.
	CALL	SYMHASH		;Get appropriate bucket addr.
SYMLNK2 MOV	E,M
	INX	H
	MOV	D,M		;Get pointer to next entry
	XCHG			; in HL (DE points at MSB).
	MOV	A,H
	ORA	L
	JNZ	SYMLNK2 	;Jmp/ not at end yet.
	;
	POP	H		;Old HL.
	DCX	H
	DCX	H		;Point at linkage bytes.
	MOV	A,H
	STAX	D
	DCX	D
	MOV	A,L		;Link new symbol into
	STAX	D		; the chain.
	MVI	M,0
	INX	H
	MVI	M,0		;Mark end of linked list.
	INX	H		;Restore HL.
	ORA	A		;NO CARRY.
	RET
;
;SYMHASH.  This is a simple hashing function which adds up
;the bytes in the name and modulos the resulting sum by a
;prime number (HASHSIZ).  The name pointer is in DE and the
;length in B.  It returns a pointer to the appropriate
;bucket in HL.
;
SYMHASH PUSH	D
	PUSH	B
	MOV	C,A
	PUSH	B
	CALL	SYMHSH1
	POP	B
	MOV	A,C
	POP	B
	POP	D
	RET
SYMHSH1 LXI	H,0		;Initialize sum to zero.
	MOV	A,B
	ORA	A		;Must handle zero case.
	JZ	SYMHSH2
SYMHSH3 DAD	H		;Multiply by 4 to introduce
	DAD	H		; character position.
	LDAX	D		;Get byte from name in A.
	INX	D		;Bump name pointer.
	ANI	3FH		;** "Dirty" the byte **
	MOV	C,A		;(C has 2 MSBits cleared)
	RRC
	RRC			;Rotate LSBits into MSBbits.
	ANI	NOT 3FH 	;Clear all but top 2 bits.
	ORA	C		;OR into C,
	ADD	L		; and finally add to HL.
	MOV	L,A
	ADC	H
	SUB	L
	MOV	H,A		;HL=HL+A.
	DCR	B		;Count down name length
	JNZ	SYMHSH3 	;Jmp/ keep summing til done.
SYMHSH2 MOV	B,H
	MOV	C,L		;Move sum to BC.
	LXI	D,HASHSIZ	;Size of our hash table.
	CALL	DDIV		;DE=BC MOD DE.
	LXI	H,HASHTAB
	DAD	D		;Point into hash table by
	DAD	D		; adding DE twice.
	ORA	A		;NO CARRY.
	RET
;
;SYMINSERT. Puts A into value area of the last symbol
;in the symbol table after any other data fields
;already there.  The last symbol in the table is the one
;currently being defined at any instant.  The
;result is a self consistent symbol, still under
;definition status.  Successive SYMINSERT calls can
;enter more and more bytes until the symbol table
;overflows or until SYMENTER is called again, in which
;case, the symbol is deemed completely defined and
;a new symbol is setup as the current-symbol-being-
;defined.
;If the symbol table overflows, return CARRY, with
;0C01H error code .
;
SYMINSERT PUSH	H
	PUSH	B
	STA	PSWSAVE		;Save A but not flags.
	CALL	SYMI1
	LDA	PSWSAVE
	POP	B
	POP	H
	RET
;
SYMI1	PUSH	D
	LHLD	STLAST
	XCHG
	LHLD	STFREE
	CALL	DSUB		;STFREE-STLAST=element length?
	LDAX	D		;DE=STLAST.
	INX	D
	MOV	B,A
	LDAX	D
	MOV	D,A
	MOV	E,B		;DE=element length.
	CALL	DCMP		;Double compare DE and HL.
	JZ	SYMI2
	POP	D
	LXI	D,0C04H		;## SYMINSERT CONFUSED ##
	STC
	RET
SYMI2	LDA	PSWSAVE		;What char supposed to insert?
	CALL	SYMFILL 	;Preserves DE.
	JNC	SYMI3
	POP	H		;Clean stack
	RET	
SYMI3	INX	D		;OK, bump total length.
	LHLD	STLAST		;Put back in last element.
	MOV	M,E
	INX	H
	MOV	M,D
	POP	D		;Must restore DE.
	RET			;Carry clear by JNC.
;
SYMFILL PUSH	H
	PUSH	D
	LHLD	MEMTOP		;Pretend top of memory is
	LXI	D,-4		;2 bytes lower for linkage
	DAD	D		;bytes inserted at SYMENTER.
	XCHG			;(plus 2 for protection).
	LHLD	STFREE
	PUSH	PSW
	CALL	DCMP		;(MEMTOP-4>STFREE)?
	JC	SYMF2		;If not, then error:
	POP	PSW
	POP	D
	POP	H
	LXI	D,0C01H 	;** SYMBOL TABLE FULL **
	STC
	RET
SYMF2	POP	PSW		;So insert the char already.
	MOV	M,A
	INX	H		;Bump STFREE.
	SHLD	STFREE		;Update the SYMTAB FREE ptr.
	POP	D
	POP	H
	STC
	CMC
	RET
;
;
;	**** SYMBOL REFERENCE FUNCTONS ****
;
;  SYMKEY and SYMUNKEY are for getting and using a 'Key' for
; a particular entry in the symtab.  The key is actually just
; a pointer to the element gotten (and replaced into) STREF.
; Used for high speed reference to a previously SYMFINDed
; element, such as a macro prototype.
;
SYMKEY	XCHG
	LHLD	STREF
	XCHG
	XRA	A
	RET
;
SYMUNKEY XCHG
	SHLD	STREF
	XCHG
	XRA	A
	RET
;
; SYMFIND sets up the 'currently-examined' pointer.
; The name of the symbol pointed to by DE and whose
; length is in B is used as a key for a search of the
; symbol table.  When a symbol with a matching name is
; found, we return NO CARRY.
; If it is not found, we return CARRY.
;
SYMFIND PUSH	B
	PUSH	H
	PUSH	D
	CALL	SYMTEST 	;Probably already pointing at
	JNC	SYMFD3		;thing we want.
	CALL	SYMHASH
SYMFD1	MOV	A,M
	INX	H
	MOV	H,M		;Get pointer to next symbol
	MOV	L,A		; in chain (in HL).
	MOV	A,H
	ORA	L		;At end of chain?
	JNZ	SYMFD2		;Jmp/ nope, keep testing.
	POP	D
	LXI	D,0C05H		;$$ SYMFIND CAN'T FIND IT $$
	POP	H
	POP	B
	STC			;Couldn't find symbol.
	RET			;Return CARRY and error code.
SYMFD2	PUSH	H		;Save link list chain pointer.
	INX	H
	INX	H		;Point at total length.
	SHLD	STREF		;Set current reference pointer.
	CALL	SYMTEST 	;Compare symbols.
	POP	H
	JC	SYMFD1		;Jmp/ not it, continue search.
SYMFD3	POP	D		;Restore old name pointer.
	POP	H
	POP	B		;Found the symbol.
	ORA	A		;Return (NO CARRY).
	RET
;
;SYMTEST checks to see whether we are pointing at the 
;element in the symbol table whose name is in the buffer.
;It will match a null name, if it finds a null name in the
;symbol table (useful for storing non-named data in ST).
;
SYMTEST PUSH	PSW
	PUSH	H		;Got to save buff for retries.
	PUSH	D
	PUSH	B
	LHLD	STREF		;Current reference pointer.
	INX	H
	INX	H		;Point at name length.
	MOV	A,B
	CMP	M		;If not equal, fails match.
	JNZ	SYMT3
	INX	H		;Bump to first name char.
SYMT1	MOV	A,B		;Is there any more name?
	ORA	A
	JZ	SYMT2		;No - succeeds match.
	DCR	B
	LDAX	D		;Get name char.
	CMP	M		;Does it match name in ST?
	INX	H		;skip along names.
	INX	D
	JZ	SYMT1		;Jmp/ matched...keep trying.
SYMT3	POP	B
	POP	D
	POP	H
	POP	PSW
	STC
	RET			;Fail - ret. CARRY.
SYMT2	POP	B
	POP	D
	POP	H
	POP	PSW
	STC
	CMC
	RET			;Succeed - ret. No CARRY.
;
;SYMGET is the method of referencing the value of a symbol
;in the table.  DE is passed in with an index into the 
;element's value area, and the value of the indexed byte
;is returned in A.  If DE is larger than the length of the 
;element's value area, error code 0C03H is returned.
;The length of the value area is given by the total length
;of the element (the first DW value in an element's cell),
;minus the name length, minus three for the total length
;DW and the name length byte themselves.  In this routine,
;DE is allowed to go negative, and it will then access
;the bytes in the name area.
;
SYMGET	PUSH	H
	PUSH	D
	LHLD	STREF		;Index down from STREF.
	MOV	E,M
	INX	H
	MOV	D,M
	INX	H
	MOV	A,M		;A = name len.
	XCHG			;Total element len -> HL
	POP	D		;Retrieve input DE index.
	PUSH	D
	INX	D
	INX	D
	INX	D
	ADD	E
	MOV	E,A
	MVI	A,0
	ADC	D
	MOV	D,A		;DE=input.index+name.len+3.
	XCHG
	CALL	DCMP		;(element length - index) > 0?
	JC	SYMG2		;Yes - index inside bounds.
	POP	D
	POP	H
	LXI	D,0C03H		;## BAD INDEX INTO SYMTAB ##
	STC
	RET
SYMG2	XCHG
	LHLD	STREF		;Inside bounds, so index it.
	DAD	D		;HL=STREF+index.
	MOV	A,M		;Get desired char finally.
	POP	D
	POP	H
	STC
	CMC
	RET
;
;SYMPUT is the routine that changes a byte in the definition 
;of a symbol table element.  The length of the element stays
;the same, but the byte indexed by DE is changed to the value
;passed in A.  If DE is larger than the length of the element, 
;CARRY is returned, with error code 0C03H.  
;The code here is almost identical to SYMGET, so look at it
;for details.
;
SYMPUT PUSH	H
	PUSH	PSW
	PUSH	D
	LHLD	STREF
	MOV	E,M
	INX	H
	MOV	D,M
	INX	H
	MOV	A,M
	XCHG
	POP	D
	PUSH	D
	INX	D
	INX	D
	INX	D
	ADD	E
	MOV	E,A
	MVI	A,0
	ADC	D
	MOV	D,A
	CALL	DCMP
	JNC	SYMP2
	POP	D
	POP	PSW
	POP	H
	LXI	D,0C03H		;## BAD INDEX INTO SYMTAB ##
	STC
	RET
SYMP2	LHLD	STREF
	DAD	D
	POP	D
	POP	PSW
	MOV	M,A
	POP	H
	STC
	CMC
	RET
;
;	**** SYMBOL TABLE DUMP ROUTINES ****
;
;	The following routines are for dumping the symbol
;table to various places, such as a disk file. 
;The internals of the symbol table design dictate the ordering
;scheme used for the symbols in the table.  All the symbols
;must occur at least but not more than once, somewhere in
;the sequence.
;
;SYMFIRST.  The first in a sequence of symbol table elements
;is located.
;On return, the buffer has the name, and B=actual 
;name length.  The usual routines can be used to get
;successive bytes from the data area of the
;symbol.
;
SYMFIRST PUSH	H
	LHLD	STBASE
	SHLD	STREF
	POP	H
	STC
	CMC
	RET
;
;SYMNEXT.  The next in the sequence of symbols is selected
;and we proceed as in SYMFIRST.  When the last symbol has
;been selected, return CARRY.
;
SYMNEXT PUSH	H
	PUSH	D
	CALL	SYMNE1
	POP	D
	POP	H
	RET
SYMNE1	LHLD	STREF		;STREF -> total length field
	MOV	E,M
	INX	H
	MOV	D,M
	DCX	H
	DAD	D		;HL=STREF+total length field.
	INX	H
	INX	H		;Past linkage bytes.
	XCHG			;Before updating STREF, got
	LHLD	STFREE		;to see if found last element.
	DCX	H		;(want CARRY on equal)
	CALL	DCMP
	RC			;If no more, ret CARRY.
	XCHG
	SHLD	STREF		;HL -> new total length field.
	RET			;(CARRY by RC)
;
;SYMNAME puts the name of the currently referenced symbol
;into the buffer pointed to by DE with max length in B.
;The DE pointer is incremented over the name and B is dec-
;remented by the name length.  A is returned = B.
;
SYMNAME	PUSH	H
	PUSH	B
	LHLD	STREF
	INX	H
	INX	H
	MOV	C,M		;Get name length.
	INX	H		;HL->name in symtab.
SYMN1	MOV	A,B		;Any name buffer left?
	STA	BSAVE+1
	ORA	A
	JZ	SYMN2		;jmp/ no - done.
	MOV	A,C		;Any name left?
	ORA	A
	JZ	SYMN2		;jmp/ no.
	MOV	A,M
	STAX	D
	INX	H
	INX	D
	DCR	C
	DCR	B
	JMP	SYMN1
SYMN2	LDA	BSAVE+1
	POP	B		;Leave DE->after name,
	MOV	B,A
	POP	H		;B = OLD B - name len.
	STC			;HL,C undisturbed.
	CMC			;A has garbage.
	RET
;
;SYMVAL is a sort of optimization in that it just gets the
;two-byte value field of a symbol by calling the already
;existing SYMGET function with the appropriate constant
;arguments.
;
SYMVAL	PUSH	PSW
	LXI	D,2
	CALL	SYMGET		;Get MSB byte first
	JC	SYMVL1
	PUSH	PSW
	DCX	D
	CALL	SYMGET		; and then the LSB.
	JC	SYMVL2
	MOV	E,A
	POP	PSW
	MOV	D,A		;Return value in DE.
	POP	PSW
	STC
	CMC
	RET
SYMVL2	POP	PSW
SYMVL1	POP	PSW
	STC
	RET
;
;SYMPRESORT moves the currently referenced symbol to the
;symbol which has the alphabetically first name in the
;symbol table and which has one of the type codes
;passed in A,D or E.
;  Error code 0 is returned if symbol table has no
;symbols of the given type.
;
SYMPRESORT
	PUSH	H
	PUSH	D
	PUSH	B
	STA	PSWSAVE		;Save all the alternate type
	XCHG			;codes passed in A,D and E.
	SHLD	DSAVE
	CALL	SYMFIRST	;Start at first symbol
	LHLD	STFREE
	LXI	B,5		;Five bytes past for weird
	DAD	B		;SYMFIND stuff.
	CALL	HEAPINIT	;Initialize heap.
	;
SYMPS1	CALL	SYMNEXT		;Get next symbol.
	JC	SYMPS4		;If none, go cleanup.
	LXI	D,0		;Get type code of symbol.
	CALL	SYMGET
	JC	SYMPS5		;If weird, ret CARRY & code.
	MOV	B,A
	LDA	PSWSAVE		;Compare against all.
	CMP	B		;of the alternate type
	JZ	SYMPS2		;codes until match, 
	LDA	DSAVE		;else we keep looking.
	CMP	B
	JZ	SYMPS2
	LDA	DSAVE+1
	CMP	B
	JNZ	SYMPS1		;Nope. Keep looking.
SYMPS2	LHLD	STREF		;Currently referenced symbol.
	XCHG
	INX	D
	INX	D		;Want symbol length.
	CALL	HEAPENTER	;Enter pointer onto heap
	LXI	D,0C06H 	;** Too many symbols **
	JC	SYMPS5		;Jmp/ report error
SYMPS3	CALL	SIFTUP		;Sift it up the heap.
	JMP	SYMPS1		;Spin for more.
;
;  Entry point for SYMHEAP.
;
SYMPS4	CALL	HEAPTOP
	JNC	SYMPS6		;Jmp/ heap has something.
	LXI	D,0
SYMPS5	POP	B
	POP	H		;Have error code in DE.
	POP	H
	STC
	RET
SYMPS6	DCX	D
	DCX	D		;Point at total length.
	XCHG
	SHLD	STREF		;Set as currently referenced
	XCHG			; symbol
	POP	B
	POP	D
	POP	H
	XRA	A
	RET
;
SYMSORTNEXT
	PUSH	H
	PUSH	D
	PUSH	B
	CALL	HEAPDELETE	;Delete one element and
	CALL	SIFTDOWN	; and sift it down the heap.
	JMP	SYMPS4		;Jmp/ return in STREF.
;
;********************************************************
;*		HEAP SORT FUNCTIONS			*
;*							*
;*	These utilities support a primitive heapsort	*
;* on elements arranged with a length byte followed	*
;* by the name itself.					*
;*							*
;********************************************************
;
;	CONVERT changes an element number into an absolute
; address.
;
CONVERT MACRO
	PUSH	D
	DAD	H
	XCHG
	LHLD	HEAPORG
	DAD	D
	POP	D
	ENDM
;
;	SAR shifs HL right (up the heap).
;
SAR	MACRO
	ORA	A
	MOV	A,H
	RAR
	MOV	H,A
	MOV	A,L
	RAR
	MOV	L,A
	ENDM
;
;	SAL shifts HL left (down the heap).
;
SAL	MACRO
	DAD	H
	ENDM
;
	MACLIST 0		;Trash expansions
;
;	Initialize the heap.
;	Simply set the base of the heap and zero HEAPEND.
;
;	In this implementation, the base or origin of the
;  heap is passed in HL.
;
HEAPINIT
	PUSH	H
	SHLD	HEAPORG 	;Set origin of heap.
	LXI	H,0
	SHLD	HEAPEND
	POP	H
	ORA	A
	RET
;
;	Enter an element onto the heap.
;	Place the element at HEAPEND+1 and increment
;	HEAPEND, thus expanding the heap.
;
; 	In this implementation, DE is the element.  Returns
;  CARRY if out of memory.
;
HEAPENTER
	PUSH	B
	PUSH	H
	CALL	HEAPEN1
	POP	H
	POP	B
	RET
	;
HEAPEN1 PUSH	D
	LHLD	HEAPEND		;Get end of heap
	INX	H
	CONVERT 		;Convert to absolute addr.
	XCHG
	LHLD	MEMTOP
	DCX	H		;Pretend it is two bytes
	DCX	H		; lower for space for entry.
	CALL	DCMP		;Check for end of memory.
	XCHG
	POP	D
	RC			;Ret/ CARRY=out of memory
	MOV	M,E
	INX	H
	MOV	M,D		;Build heap element.
	LHLD	HEAPEND
	INX	H
	SHLD	HEAPEND 	;Bump HEAPEND.
	ORA	A
	RET
;
;	Return the element at the top of the heap.
;
;	In this implementation, we return the pointer to the
;  item in DE or CARRY if the heap is empty.
;
HEAPTOP
	PUSH	H
	LHLD	HEAPEND
	MOV	A,H
	ORA	L
	STC			;Return CARRY
	JZ	HEAPTP1		;if heap is null
	LXI	H,1
	CONVERT
	MOV	E,M
	INX	H
	MOV	D,M		;DE returns --> ITEM
HEAPTP1 POP	H
	RET
;
;	Delete the top element from the heap.
;	Replace the top of the heap by the bottom and
;	decrement HEAPEND, thus shrinking the heap.
;
;	In this implementation, CARRY set indicates the
;  heap is null.
;
HEAPDELETE
	PUSH	B
	PUSH	D
	PUSH	H
	CALL	HEAPDL1
	POP	H
	POP	D
	POP	B
	RET
	;
HEAPDL1 LHLD	HEAPEND		;Get end of heap
	MOV	A,H
	ORA	L
	STC
	RZ			;Ret/ CARRY=heap is null
	CONVERT 		;Convert to absolute addr
	MOV	E,M
	INX	H
	MOV	D,M
	;
	LXI	H,1		;For heap element #1
	CONVERT 		;Convert to absolute addr
	MOV	M,E
	INX	H		;Replace first heap element
	MOV	M,D		; with last.
	LHLD	HEAPEND
	DCX	H
	SHLD	HEAPEND 	;Decrement HeapEnd.
	ORA	A
	RET
;
;	Sift the element at HEAPEND (bottom of heap) up.
;	Starting at the bottom of the heap, compare the item
;	the for that element with the item for the element
;	one level up and swap the items if they are out of
;	order.  If a swap occurs, move up the heap and repeat
;	the process.
;
SIFTUP
	PUSH	B
	PUSH	D
	PUSH	H
	LHLD	HEAPEND 	;Start at bottom of heap.
	CALL	SIFTUP1
	POP	H
	POP	D
	POP	B
	ORA	A
	RET
	;
SIFTUP1 MOV	E,L		;Save element in DE.
	MOV	D,H
	SAR			;Move up the heap (HL=HL/2)
	MOV	A,H
	ORA	L
	RZ			;Ret/ hit top of heap, done
	PUSH	H
	XCHG
	CALL	ITEMCOMP
	JNC	SIFTUP2		;Jmp/ no swap needed
	CALL	ITEMSWAP	; else...swap,
	POP	H		; move up the heap
	JMP	SIFTUP1 	; and continue process.
	;
SIFTUP2 POP	H
	RET
;
;	Sift element1 (top of the heap) down.
;	Starting at the top of the heap, compare the items
;	for the two elements below that node to find the
;	lesser of the two.  Then compare that item with the
;	item for the node and swap the items if they are out
;	of order.  If a swap occurs, repeat the process using
;	the swapped element as the new node.
;
SIFTDOWN
	PUSH	B
	PUSH	D
	PUSH	H
	LXI	H,1		;Top of heap is first node
	CALL	SIFTDW1
	POP	H
	POP	D
	POP	B
	ORA	A
	RET
	;
SIFTDW1 SHLD	HEAPNODE	;Save node.
	SAL			;Traverse down the heap.
	XCHG
	LHLD	HEAPEND
	CALL	DCMP
	RC			;Ret/ past HEAPEND
	PUSH	D		;DE: left branch
	INX	D
	LHLD	HEAPEND
	CALL	DCMP		;Past end of heap?
	XCHG			;HL: right branch
	POP	D		;DE: left branch
	JC	SIFTDW2 	;Jmp/ yup, left is only one.
	PUSH	D
	PUSH	H
	CALL	ITEMCOMP	;Compare the two.
	POP	H
	POP	D
	JNC	SIFTDW2
	XCHG			;DE: has lesser of the two.
	;
SIFTDW2 LHLD	HEAPNODE	;Compare node
	PUSH	D		; with the smaller branch.
	XCHG
	CALL	ITEMCOMP
	JNC	SIFTDW3 	;Jmp/ no swap needed, done.
	CALL	ITEMSWAP 	; else swap node with branch
	POP	H		; and continue process with
	JMP	SIFTDW1 	; the branch as the new node.
	;
SIFTDW3 POP	H
	ORA	A
	RET
;
;	Compare two items.  HL = element1, DE = element2
;  Sets CARRY if element1<element2.  Returns BC: --> ITEM1
;  and DE: --> ITEM2.
;
ITEMCOMP
	PUSH	D
	CONVERT 		;Convert to absolute addr.
	SHLD	HEAPEL1
	MOV	E,M
	INX	H
	MOV	D,M		;--> item1
	POP	H
	CONVERT 		;Convert to absolute addr.
	SHLD	HEAPEL2
	MOV	A,M
	INX	H
	MOV	H,M
	MOV	L,A		;--> item2
	XCHG
	PUSH	H
	PUSH	D
	MOV	C,M		;C: len[ITEM1] stringA
	INX	H		;HL: --> ITEM1
	LDAX	D
	MOV	B,A		;B: len[ITEM2] stringB
	INX	D		;DE: --> ITEM2
	CALL	COMPARE 	;Compare two ITEMS
	POP	D
	POP	B
	RET
;
;	Swap the two items last compared.
;
ITEMSWAP
	LHLD	HEAPEL1
	MOV	M,E
	INX	H
	MOV	M,D
	LHLD	HEAPEL2
	MOV	M,C
	INX	H
	MOV	M,B
	ORA	A
	RET
