	PAGE
;
;
; **** MACRO-88 SYMBOL TABLE STORAGE ****
;
;	The initial symbol table contents are the instruction 
; opcodes, the register operand values, and the pseudo-
; opcodes. Arranged in order of decreasing expected frequency.
;
; The type byte has following meanings:
;
;	1=OPCODE
;	2=POPCODE
;	3=REGISTER	(Defunct as of version 003)
;	4=LABEL
;	5=SET VARIABLE	(May change value during assembly)
;	6=UNDEFINED LABEL
;	7=DOUBLE DEFINED LABEL
;	8=MACRO NAME
;
;  Format for SYMLIST entries:
;	DW	<linkage bytes>
;	DW	<total length>
;	DB	<name length>,<ASCII name>,<type code>
;	data ...
;
;  where <linkage bytes> double word is used to link the
;  symbols in the hash table chain and <total length> is
;  used to find the next entry (total length does include
;  length of length word itself), and the <name length> is
;  as used by the ENTER function.
;
;	There are no LABELs in this initial list of course,
;  but for documentation purposes, the format of a LABEL
;  data field bytes are:
;
;	DW	<const val>
;	DB	<rel flag>
;
; where <const val> is the value of the label and <rel flag>
; tells whether or not this symbol is a relocatable value
; (non-zero) or not (0).
;
;
;***************
;
;	THE INITIAL SYMBOL TABLE LIST.
;
SYMLIST
;
;	A macro to create a symbol table entry which is called
;  by both the opc and popc macros.
;
symbol	MACRO	symname,type,value,subtype
	DW	0
	IF	NULL[#4]
	DW	LEN['#1']+6
	DB	LEN['#1'],'#1',#2
	DW	#3
	ELSE
	DW	LEN['#1']+7
	DB	LEN['#1'],'#1',#2
	DW	#3
	DB	#4
	ENDIF
	ENDM
;
;
;  OPCODES
;
; The OPCODE sub-types are specified by the sub-type field of
; opcode entries, which is after the value field.
;
code	MACRO
#L	EQU	codeval
codeval SET	codeval+1
	ENDM
	;
codeval SET	1
;
PAIRCODE  code			;Register pair
ADDRCODE  code			;Address type
RRCODE	  code			;Register to Register
R2CODE	  code			;Register 2
R1CODE	  code			;Register 1
R1D8CODE  code			;Register 1 w/ immed byte
D8CODE	  code			;Immediate byte
NOPCODE   code			;Nop/ single byte
BDCODE	  code			;B or D register
PRADCODE  code			;Pair w/ address
;
opc	MACRO	value,subtype
	symbol	#L,1,0#1,#2
	ENDM
;
;
MOV	opc	40H, RRCODE
PUSH	opc	C5H, PAIRCODE
POP	opc	C1H, PAIRCODE
LXI	opc	01H, PRADCODE
LHLD	opc	2AH, ADDRCODE
SHLD	opc	22H, ADDRCODE
LDA	opc	3AH, ADDRCODE
STA	opc	32H, ADDRCODE
JMP	opc	C3H, ADDRCODE
JNZ	opc	C2H, ADDRCODE
JZ	opc	CAH, ADDRCODE
JNC	opc	D2H, ADDRCODE
JC	opc	DAH, ADDRCODE
JPO	opc	E2H, ADDRCODE
JPE	opc	EAH, ADDRCODE
JP	opc	F2H, ADDRCODE
JM	opc	FAH, ADDRCODE
PCHL	opc	E9H, NOPCODE
CALL	opc	CDH, ADDRCODE
CNZ	opc	C4H, ADDRCODE
CZ	opc	CCH, ADDRCODE
CNC	opc	D4H, ADDRCODE
CC	opc	DCH, ADDRCODE
CPO	opc	E4H, ADDRCODE
CPE	opc	ECH, ADDRCODE
CP	opc	F4H, ADDRCODE
CM	opc	FCH, ADDRCODE
RET	opc	C9H, NOPCODE
RNZ	opc	C0H, NOPCODE
RZ	opc	C8H, NOPCODE
RNC	opc	D0H, NOPCODE
RC	opc	D8H, NOPCODE
RPO	opc	E0H, NOPCODE
RPE	opc	E8H, NOPCODE
RP	opc	F0H, NOPCODE
RM	opc	F8H, NOPCODE
INX	opc	03H, PAIRCODE
DCX	opc	0BH, PAIRCODE
DAD	opc	09H, PAIRCODE
XCHG	opc	EBH, NOPCODE
ADD	opc	80H, R2CODE
ADC	opc	88H, R2CODE
SUB	opc	90H, R2CODE
SBB	opc	98H, R2CODE
ANA	opc	A0H, R2CODE
XRA	opc	A8H, R2CODE
ORA	opc	B0H, R2CODE
CMP	opc	B8H, R2CODE
DCR	opc	05H, R1CODE
INR	opc	04H, R1CODE
MVI	opc	06H, R1D8CODE
ADI	opc	C6H, D8CODE
ACI	opc	CEH, D8CODE
SUI	opc	D6H, D8CODE
SBI	opc	DEH, D8CODE
ANI	opc	E6H, D8CODE
XRI	opc	EEH, D8CODE
ORI	opc	F6H, D8CODE
CPI	opc	FEH, D8CODE
LDAX	opc	0AH, BDCODE
STAX	opc	02H, BDCODE
IN	opc	DBH, D8CODE
OUT	opc	D3H, D8CODE
DAA	opc	27H, NOPCODE
CMA	opc	2FH, NOPCODE
STC	opc	37H, NOPCODE
CMC	opc	3FH, NOPCODE
XTHL	opc	E3H, NOPCODE
SPHL	opc	F9H, NOPCODE
HLT	opc	76H, NOPCODE
DI	opc	F3H, NOPCODE
EI	opc	FBH, NOPCODE
RLC	opc	07H, NOPCODE
RRC	opc	0FH, NOPCODE
RAL	opc	17H, NOPCODE
RAR	opc	1FH, NOPCODE
RST	opc	C7H, R1CODE
NOP	opc	00H, NOPCODE
