	LIST	ldio
	PAGE
; diskequates.TX
; Last updated:
;	12/28/79	RTM	disaster recovery
;	1/10/80		RTM	Don't stomp CPU ram...
;	1/9/81		FEA	Added equates for hard disk
;
; ***** Definitions for MS drives *****

; The following equates are the controller board's para-
; meter passing area.  This is where command and control
; information is passed to the MS.

PARAMTR	rorg	1FE0H	; Start of area

; DISKADD (0-133FH) is the first disk address where
; I/O is to occur.

DISKADD	rds	2

; MEMADD (1000-1700H) is the address in buffer RAM where the
; I/O operation will transfer data to or from.
;
MEMADD	rds	2

; UNIT (4-7) Contains the number of the drive

UNIT	rds	1

; SECTS (1-8) contains the number of sectors to be transfered

SECTS	rds	1

; COMMAND contains the operation:
;	0= Write
;	1= Read
;	2= Initialize
;	3= Cache Read
;	4= Execute 1000H
;	Commands 5 and 6 are done by 8080.  5 is made
;	identical to 2, 6 calculates disk size from double
;	sided flag.

COMMAND	rds	1

; STATUS contains the completion status of the I/O or diagnos-
; tics run on the system, or a command to start I/O using the
; parameters in the rest of the parameter passing area.
;	0= Normal I/O completion
;	1= Bad Parameters
;	2= Preamble bad error
;	3= CRC error
;	4= Verify error
;	5= Write protected
;	6= Disk not ready
;
;	Diagnostics completion codes
;
;	80H= Checksum error on root
;	82-89H= Memory error in RAM
;	81H= Self test failure on SDLC
;	8AH= Data loop back test failure
;
;	FFH= Command to start I/O
;
STATUS	rds	1
;
; The following parameters are set up by the system and should
; not be fiddled with except by wizards.
;
; TRIES is the number of times we will attempt an operation
; before declaring an error. This is initialized to 10.
;
	rds	1	; Was RAW.
TRIES	rds	1
;
; TRIES1 is the counter for TRIES.  The reason this is in the
; parameter passing area is to allow diagnostics to determine
; how many attempts were made at any operation.  This can be
; found by subtracting TRIES1 from TRIES.
;
TRIES1 rds	1
;
; The MSDio transfer buffer resides at 1000H and is 2K bytes
; long resulting in a maximum transfer of 8 sectors at a time.
; In a write operation, data to be written out to the disk
; will be shoveled into here after first making sure I/O is
; not in progress.  On a read or a cache read the data in this
; area will be valid when the STATUS byte goes to 0.
;
BUFFER	EQU	1000H
;
; MS Boot Time is 3 seconds however we will allow a 2
; second excess for memory errors.  To see if the board is
; there, we cannot test for memory at 1FEFH as in earlier
; schemes because the controller could be testing its memory
; in its boot process and we would give it an error.  Secondly,
; the only byte left untested (STATUS) is not owned by us
; during the boot process.  Time out of Boot Time will mean a
; No-Controller error.
;
MSBootTime EQU	(1000/88)*5*1000 ; 88 microseconds in MSDio1
				 ; loop

; ***** Definitions for double density 5" drives *****

;	Communications locations.  Only the 8080 plays with
;  them except Status, which is the shared state code.

	rorg	01C00H	; 3/4K shared memory area on card.
DiskAdd rds	2	; Logical disk address.
Unit	rds	1	; Drive number passed to Z-80.
Pages	rds	1	; Number of pages in a transfer.
Status	rds	1	; Shared status/error code.
PageSwitch rds	1	; 0 indicates Page0. 1: Page1.

;	Error reporting info in shared controller RAM.

Tries	rds	1	; 8080 Settable soft error # retries.
SeekTries rds	1	; # seek retries.
Badbyte	rds	1	; The XOR of good and bad RAM bytes.
Soft2	rds	1	; Soft error counters for preamble,
Soft3	rds	1	; checksum,
Soft4	rds	1	; verify,
SoftSeek rds	1	; seek.

	rorg	((rpc+256) AND 0FF00H) -2

Page0	rds	256+1	; Page plus checksum byte.
Page1	rds	256+1

NONEXIST EQU	0FFH	; A non existent controller card.
ERRACK	EQU	0FEH	; If 8080 sees error, this resets Z-80.
BOOTING EQU	0FDH	; While Z-80 self-tests, it gives this
READ	EQU	0FCH	; Read a sector command.
WRITE	EQU	0FBH	; Write a sector command.
DISKSIZE EQU	0FAH	; Measure diskette size command.
READSTART EQU	0F9H	; Begin a read sequence.
WRITESTART EQU	0F8H	; Begin a write sequence.
EXEC1D00  EQU	0F7H	; Dio doesn't do it, but here is code.

LOWESTCOMMAND EQU 0D0H	; All above this is a Z-80 state!
			; All below is owned by 8080!

WORKING EQU	0CFH	; After Z-80 accepts command.
DONE	EQU	0	; When Z-80 doing nothing.

DDBootTime EQU	0FFFFH	; A lot of cycles in boot timeout.
DDTimeout EQU	10000	; Cycles before board doesn't exist.

; ***** Definitions for dumb 5" drives *****

	rorg	BUSIES+16	; start after CPU RAM stuff!

TIMER	EQU	0C00H	; time counter

MTIME	EQU	50*5	; 5 seconds worth of spin.
MTO	rds	1	; motor time-out counter
CDRIV	rds	1	; current drive number

; TRK1,2,3 are current positions for drives 1, 2, and 3.

TRK1	rds	1
TRK2	rds	1
TRK3	rds	1

; MUNG1,2,3,4 are scratch pairs. Likely to be anything.

MUNG1	rds	2
MUNG2	rds	2
MUNG3	rds	2
MUNG4	rds	2
MUNG5	rds	2

; DTRIES is the number of retries we attempt on an operation
; before deciding to call it a hard error. It is set to 10.

DTRIES	rds	1

; SINT is the interrupt-level exit for sector pulses. When SSEC
; equals the sector number we're at, we bail out to contents of
; SINT, disabled, and at the  interrupt level.

SINT	rds	2
SSEC	rds	2	; bail out and sector number

; DSTAT, DSMA, and DSCT are for the device drivers, that work
; at the interrupt level performing disk i/o. DSTAT is the I/O
; status: FF while in progress, 0 for normal completion, and
; <>0 for error completion. DSMA is the memory address, and
; DSCT is the number of sectors.

DSTAT	rds	2	; I/O completion status
DTRK	rds	1	; track number we're to be on
DSMA	rds	2	; memory address
DSCT	rds	1	; number of sectors
SCUZZ	rds	2	; junk word.

; BUGS counts the number of 102, 103, and 104 errors

BUGS	rds	3	; error counters

MemAdd	rds	2	; DD memory address
Command	rds	1	; DD command
Pages1	rds	1	; DD # pages

; 5" SSSD Disk controller stuff

DCO	EQU	2BH	; commands out to controller
DCI	EQU	2AH	; command status back

; You send the command port (DCO) an 86H to reset it to
; eating command words. The commands are:

CDSI	EQU	4	; disable sector interrupt
CESI	EQU	5	; enable sector interrupt
CMH	EQU	6	; motor halt
CMG	EQU	7	; motor go
CWH	EQU	8	; write off (halt)
CWG	EQU	9	; write on (go)
CRH	EQU	0AH	; read halt (off)
CRG	EQU	0BH	; read go (on)
CSI	EQU	0CH	; step in
CSO	EQU	0DH	; step out
CPL	EQU	0EH	; step pulse low
CPH	EQU	0FH	; step pulse high

; The status from the command status port looks like:

CSF	EQU	1	; sector flag (found a hole)
CINT	EQU	4	; interrupt enable

; Drive select is port 28H, drive status is port 29H.
; the bits to those look like:

DST	EQU	29H	; drive status (read only)
DSEL	EQU	28H	; drive select (write only)

; Drive status back bits

DWP	EQU	20H	; write protected
DTR0	EQU	10H	; track 0 hit
DSM	EQU	0FH	; sector number mask

; Select bits out

DS0	EQU	0	; select no drives
DS1	EQU	1	; select drive 1
DS2	EQU	2	; select drive 2
DS3	EQU	3	; select drive 3

; The synch byte used on the disks

DSYN	EQU	0E6H	; That's it.

; Finally, the USART is ports 20 and 21H

UDAT	EQU	21H	; data port
USTAT	EQU	20H	; status port
RBF	EQU	1	; read buffer full
TBE	EQU	2	; xmit buffer empty
;
; PRIAM hard disk inteligent controller register equates
;
; Read registers on the disk controller
;
BASE	EQU	38H		; Where I'm addressed
;
HSTATUS	EQU	BASE		; Controller status
DATA	EQU	BASE+1		; Disk read / write data
TSTAT	EQU	BASE+2		; Transaction status
RSLT0	EQU	BASE+3		; Result 0 register
RSLT1	EQU	BASE+4		; Result 1 register
RSLT2	EQU	BASE+5		; Result 2 register
RSLT3	EQU	BASE+6		; Result 3 register
;
; Write registers on the disk controller
;
HCMND	EQU	BASE		; Controller command
PARM0	EQU	BASE+2		; Parameter 0
PARM1	EQU	BASE+3		; Parameter 1
PARM2	EQU	BASE+4		; Parameter 2
PARM3	EQU	BASE+5		; Parameter 3
PARM4	EQU	BASE+6		; Parameter 4
;
; Important status bits
;
BUSY	EQU	00001000B	; Bit 3 is controller busy flag
COMPLET	EQU	01000000B	; Bit 6 is Cont Completion Req
DXFER	EQU	00000100B	; Bit 2 is Data Transfer Req
REJECT	EQU	10000000B	; Bit 7 is command reject
