	PAGE
; doesio.TX
; Last updated:
;	12/28/79	RTM	Disaster recovery
;	1/10/80		RTM	sssd bug found!!!!
;	1/11/80		RTM	Rom support for Dio
;	1/28/80		RTM	Drop rom support.
;	2/15/80		RTM	Do better param checking
;	2/18/80		RTM	Transfer right to driver
;	2/22/80		RTM	Hacked out SSSD driver!
;	2/22/80		RTM	Introducing Devlock code...
;	3/20/80		RTM	Split up large transfers
;	7/8/80		RTM	SSSD eats it
;	9/23/80		Staff	The famous cache bug squashed
;				by changing disk address to
;				illegal number (1FEE) in MS
;				disk size.
;	10/31/80	BFS	Changed dio chunks to 40
;				sectors so initializing a
;				5" drive doesn't take four
;				days.
;	11/05/80	BFS	Find non-existent MS controller
;				within user's lifetime.  Remove
;				wait for boot code because by
;				the time this code is in boot
;				is done.
;	1/9/81		FEA	Add full support for volume
;				manager and put in hard disk
;				driver code.
;	1/29/81		BFS	Added per device sector
;				blocking.
;	02/04/81	BFS	fixed per device sector
;				blocking so that it works.
;				It was assuming unit was same
;				as device... mucho explodo.
;	02/10/81	BFS	Removed DONT twiddling in hard
;				disk driver, it's done by Dio
;				so we had a double unlock type
;				problem which lead to DEADLOCK!
;	02/18/81	FEA	Set up retry recovery for 
;				Data Timeout Errors.
;	04/15/81	BFS	Change chunk size on HD to 20H
;				to improve twin performance.
;	05/04/81	FEA	Fixed volume legal address
;				checking so that it adds the
;				number of sectors being xfered
;				to the base address before it
;				checks for address overflow.
;

Stupid	SET	0	; Don't assemble 5" SSSD driver at all


	LIST	ldio
; Doio is the head of the disk I/O code. It has in it the
; volume manager, and takes care of blocking on access, etc.

; For a Dio call, we are entered with:
;	HL	Disk address
;	DE	Memory address
;	C	Volume #, 1-9 inclusive
;	B	Command: 0=write, 1=read, 2=init, 3=read
;			 4=execute 1000H (ms only)
;			 5=return vol size in HL
;	A	# of sectors to transfer, 00 is illegal.
; Return with C clear for no errors, C set and error in DE.

; This process is driven from the volume manager's control
; blocks. Each volume has a control block, which contains:
;
;	Offset	Size	Use
;	  0	  2	"Busy" semaphore address
;	  2	  2	Device driver address
;	  4	  2	Parameter area address
;	  6	  2	Volume size
;	  8	  3	Offset from physical 00 on device
;	  B	  1	Translated drive #
;	  C	  1	Volume status

;
;	First, determine the chunk size for this unit and
;  stick it in CHUNK.  Then check to see if # sectors to do
;  is less than or equal to that and split things up if
;  necessary.
;
Doio	PUSH	B		; Save command and drive.
	MOV	B,A		; Save number of sectors.
	PUSH	H		; Save disk address.
	CALL	GetChunk	; Get the chunk size in A.
	STA	CHUNK		; Put it away.
	POP	H
	MOV	A,B
	POP	B
	RC			; Error.
	PUSH	H		; Save.
	LXI	H,CHUNK		; Point at chunk size.
	CMP	M		; # sectors <= CHUNK?
	POP	H		; Restore
	JC	Doesio		; less, jump.
	JZ	Doesio		; Equal, jump.


; Break it up into CHUNK sector chunks.

	PUSH	PSW
	PUSH	B
	PUSH	D
	PUSH	H
	LDA	CHUNK
	CALL	Doesio		; go do CHUNK sectors
	JC	LockErr		; oops... a bit of trouble.
	POP	H		; disk address
	LDA	CHUNK
	PUSH	PSW
	ADD	L
	MOV	L,A
	JNC	Doio10
	INR	H
Doio10	POP	PSW
	POP	D		; memory address
	ADD	D
	MOV	D,A
	POP	B		; command info
	POP	PSW		; sector count
	PUSH	H
	LXI	H,CHUNK
	SUB	M
	POP	H
	JMP	Doio		; look again.

;	Get chunk size.  Unit number is in C.  Get device
;  number for unit then get chunk size for that device.
;  Return size in A.

GetChunk
	PUSH	B
	PUSH	D
	MVI	A,10H		; Index command.
	CALL	Vmgr
	JC	GC2		; Report error and clean stack.
	LXI	B,0BH		; Device number offset.
	DAD	B		; Point at device number.
	POP	D
	POP	B
	MOV	C,M		; device number in C
	LXI	H,DrvAdTab	; Base of table.
	DCR	C		; index with device-1
	MOV	A,C
	ADD	A
	ADD	A
	ADD	A
	SUB	C		; (Device-1) * 7 in A.
	ADD	L		; Add offset to pointer.
	MOV	L,A
	MOV	A,H
	ACI	0
	MOV	H,A
	MOV	A,M		; Get chunk size.
	ORA	A		; Set flags
	RNZ			; Non-zero is ok.
	LXI	D,109H		; No drive.
	STC
	RET
GC2	POP	B		; Error code is in DE
	POP	B		; but we have D and B on stack.
	RET			; Doesn't matter if we trash B. 
;
;
;
Doesio	PUSH	PSW
	PUSH	B
	PUSH	D
	PUSH	H		; stash 'em all.
	MOV	E,A		; put # sectors in E

; Idiot check parameters!

	MOV	A,C		; check dev #
	ORA	A
	JZ	Er101		; jmp/won't do!
	CPI	10
	JNC	Er101		; neither will that!
	MOV	A,B		; now look at command
	CPI	10
	JNC	Er101		; jmp/crud

; Call Devlock to see if we can hack at this thing!
; We want to get a temporary allocation for the device,
; either read, write or mung for a write to sector 00

	PUSH	B		; stash for a bit...
	MOV	A,B		; pick up command
	MVI	B,0A2H		; assume it's a read
	ORA	A		; see if a write type op
	JNZ	Gosee		; jmp/yup, go see!

; We think it's a write type op. Check for sector 00 (in hl)
; and request MUNG access rather than just write...

	MVI	B,0A3H		; assume write
	MOV	A,L
	ORA	H
	JNZ	Gosee		; jmp/nonzero DA

Getmung	MVI	B,0A5H		; else request MUNG

Gosee	CALL	Devlock		; see what comes back
	POP	B
	JC	LockErr		; Sorry, error.

; We got access thru Devlock, go ahead...

Okbyme	PUSH	B		; stash BC on top for later
	MVI	A,10H		; Index Command to vol mgr
	CALL	Vmgr		; Index me
	POP	B		; get right BC back again

; See it there's anything connected to this volume

	PUSH	H		; Save pointer
	LXI	D,0CH		; Index to volume status
	DAD	D
	MOV	A,M		; Is it 0?
	ORA	A
	POP	H		; Restore pointer
	JZ	Er101		; Lurch off if non-existant

; See if we have a disk size command, and if we do, whether we
; can pick our disk size out of the volume manager's table.

	MOV	A,B
	CPI	5		; Dsize Command?
	JNZ	GetBusy		; Nope, skip this stuff
	PUSH	H		; Else save index
	LXI	D,6		; Offset for size
	DAD	D		; Index to it
	MOV	E,M		; Get the size in DE
	INX	H
	MOV	D,M
	MOV	A,E		; Check to see if size is 0?
	ORA	D
	POP	H		; Balance stack in case its not
	JZ	GetBusy		; If zero must ask driver
	XCHG			; Else put size into HL
	POP	D		; Balance stack
	POP	D
	POP	D
	POP	D
	RET			; And split

; Pick up the address of the busy semaphore and wait for device
; to go not busy. We store an address because the "busy" is on
; a per controller basis.

GetBusy	MOV	E,M
	INX	H
	MOV	D,M		; address of semaphore
	INX	H		; leave pointing at driver addr

; Wait for semaphore to indicate not busy (00)

Doio1	DI
	LDAX	D
	ORA	A		; well?
	JZ	Doio2		; jmp/not busy, we win.
	CALL	VGiveup	; wait for a while.
	JMP	Doio1

; Set the semaphore to mark device busy, with 16*drive+fn
; this 16 * drive + fn is not used for anything so I'm going
; to use 0FFH instead to save some memory which I desperately
; need. BFS

Doio2	MVI	A,0FFH
	STAX	D		; device is busy now.
	STA	DONT		; and so are we!
	XCHG
	SHLD	DioBsy		; save its address
	XCHG

; Pick up driver address and stash for a while.

	MOV	E,M
	INX	H
	MOV	D,M
	XCHG
	SHLD	DioDrv
	XCHG
	INX	H		; now pointing at param addr

; Now pick up translated drive # and put in DioDn

	LXI	D,7
	DAD	D
	MOV	A,M
	STA	DioDn		; new drive # for driver code

; Checks to see if Parameter area address is non-zero, if it
; is add the disk address to the Offset and put it in the
; parameter area.  If it's zero skip this mush.

	LXI	D,4-0BH		; Index back to Parm area addr.
	DAD	D
	MOV	E,M		; Get parameter area address
	INX	H
	MOV	D,M
	MOV	A,E		; Is it 0?
	ORA	D
	JZ	GoToIt		; If so just do it
	XCHG			; Put DE on stack under old adr
	XTHL
	PUSH	H
	XCHG
	INX	H		; Point to size
	MOV	E,M		; Size to DE
	INX	H
	MOV	D,M
	INX	H
	XTHL			; Get old addr saving pointer
	PUSH	H		; Put address back on top
	LXI	H,11		; Index up stack pointer to
	DAD	SP		; where # sectors is being kept
	MOV	A,M		; # sectors to Accum.
	DCR	A
	POP	H		; Get disk address
	PUSH	H		; and re-save
	ADD	L		; Add # of sectors to disk addr
	MOV	L,A
	MVI	A,0
	ADC	H
	MOV	H,A
	JC	TooBig		; If last addr >FFFFH, overflow
	CALL	DCMP		; Compare to DISK SIZE in DE
	JNC	TooBig		; If > DSIZE report error
	POP	H		; Address back into HL
	XTHL			; Address to stack, VCB pointer
				; back into HL
	MOV	E,M		; Put offset into CDE
	INX	H
	MOV	D,M
	INX	H
	MOV	C,M
	POP	H		; Get old address in HL
	XRA	A		; zero MSB of this address
	DAD	D		; AHL=AHL+CDE
	ADC	C
	XCHG			; ADE= physical address
	POP	H		; Get parameter address
	PUSH	H		; Save again
	MOV	M,E		; Put address in parameter area
	INX	H
	MOV	M,D
	INX	H
	MOV	M,A

; Pull regs from stack, and go to it.

GoToIt	POP	H		; disk addr (untouched!)
	POP	D		; memory address
	POP	B		; command and dev
	LDA	DioDn
	MOV	C,A		; Kapow! translated dev!
	POP	PSW		; # sectors to do
	SHLD	DioHL		; save for a bit.

; Build return crud onto the stack.

	LHLD	DioBsy
	PUSH	H		; semaphore address
	LXI	H,Doio3		; return point
	PUSH	H
	LHLD	DioDrv		; driver address
	PUSH	H
	LHLD	DioHL		; HL (disk address) back.

; Call the driver.

	EI			; driver access blocked
	RET			; by the VCB semaphore...

; The driver returns to here. We clear the semaphore, and
; DONT, and must not change any regs!

Doio3	DI
	XTHL			; semaphore addr to HL
	PUSH	PSW
	XRA	A
	MOV	M,A		; mark it cleared.
	STA	DONT		; let world in again.
	POP	PSW
	POP	H		; Restore H
	JMP	VGiveup		; exit giving up processor.

;
; TooBig is the exit point when the disk address is bigger
; than the tabulated disk size.
;
TooBig	LHLD	DioBsy		; Clear busy flag
	MVI	M,0
	POP	D		; Throw two stack levels away
	POP	D
	JMP	Er101		; Bail out through here

; First crack at VCB control tables. We set up defaults for
; things that will get mucked with later on.

vcb	MACRO	Flag,Driver,Params,Size,Offset,Drive,Stat
#L	DW	#1,#2,#3,#4
	DB	0,0,0		; #5 ignored for now!
	DB	#6
	DB	#7
	ENDM

; Busy semaphores for the various types of drives.

SAbusy	EQU	BUSIES+1	; SA400 (dumb 5")
DDbusy	EQU	BUSIES+2	; DDDS 5"
MSbusy	EQU	BUSIES+3	; MS (8" DD)
HDbusy	EQU	BUSIES+4	; hard disk (7710)

; VCB list. Starts with "drive" 1. Right now, all point to
; DIO code.
VCBS	vcb	DDbusy,DDDio,0,1400,0,1,-1	; DD drive 1
	vcb	DDbusy,DDDio,0,1400,0,2,0	; DD drive 2
	vcb	DDbusy,DDDio,0,1400,0,3,0	; DD drive 3

; Drives 4 thru 7 are the MS

	vcb	MSbusy,MSDio,0,77*64,0,4,-1	; MS drive 4
	vcb	MSbusy,MSDio,0,77*64,0,5,0	; MS drive 5
	vcb	MSbusy,MSDio,0,77*64,0,6,0	; MS drive 6
	vcb	MSbusy,MSDio,0,77*64,0,7,0	; MS drive 7

; Volumes 8 and 9 are small undefined

	vcb	0,0,0,0,0,0,0
	vcb	0,0,0,0,0,0,0

; Table of values for flags and addresses for the various 
;  physical device drivers.  The table is organized as follows:
;
;	Size	Definition
;	 1	Chunk size/ existance byte. (Zero = not here,
;		non-zero = chunk size.)
;	 2	Busy flag address
;	 2	Device driver address
;	 2 	Parameter area address

; We'll use a macro to make our lives easier.

drvdef	MACRO	Existance, Busy Addr, Drv Addr, Parm Addr
	DB	#1
	DW	#2
	DW	#3
	DW	#4
	ENDM	drvdef

; Here it is

DrvAdTab

; Physical devices 1-3 are the DD controller (driver present)
; Chunk size = 20.

DrDef1	drvdef	20,DDbusy,DDDio,0
DrDef2	drvdef	20,DDbusy,DDDio,0
DrDef3	drvdef	20,DDbusy,DDDio,0

; Physical devices 4-7 are the MS controller (driver present)
; Chunk size = 8.

DrDef4	drvdef	8,MSbusy,MSDio,0
DrDef5	drvdef	8,MSbusy,MSDio,0
DrDef6	drvdef	8,MSbusy,MSDio,0
DrDef7	drvdef	8,MSbusy,MSDio,0

; Physical devices 8-10 are the 5" SSSDs (no driver present)

DrDef8	drvdef	0,0,0,0
DrDef9	drvdef	0,0,0,0
DrDef10	drvdef	0,0,0,0

; Physical devices 11-14 are the PRIAM disk (driver present)
; Chunk size = 32.

DrDef11	drvdef	32,HDbusy,HardDio,HardAdr
DrDef12	drvdef	32,HDbusy,HardDio,HardAdr
DrDef13	drvdef	32,HDbusy,HardDio,HardAdr
DrDef14	drvdef	32,HDbusy,HardDio,HardAdr

; Physical devices 15-20 are undefined (no driver present)

DrDef15	drvdef	0,0,0,0
DrDef16	drvdef	0,0,0,0
DrDef17	drvdef	0,0,0,0
DrDef18	drvdef	0,0,0,0
DrDef19	drvdef	0,0,0,0
DrDef20	drvdef	0,0,0,0

	PAGE
;	**** MS Driver ****

; The MS interface consists of breaking the I/O down into 
; chunks that the DIO code on the controller board can handle.
; This is now done by the dio chunk size logic.
; The largest transfer the controller board is capable of is 8
; sectors.  The code will then pass parameters into the memory
; on the large controller card and wait for status to come back

; DD INIT is accomplished by doing a series of writes and
; reads.  The Exec has the code to do this.  MS INIT is
; accomplished by passing a command of 2 to the DIO code.
; This will automatically write a preamble on all sectors of
; the disk. It will also verify the medium by writing a
; pattern on all sectors on the disk and verifying it.  When
; this has been accomplished the disk will require a second
; write command to put the directory on the first 4 sectors.

;
;	This RAM version of the MS driver code assumes that
;  the boot process has finished because the code itself has
;  to come off the disk.  If it comes off of a 5" DD controller
;  by the time the DD has run its diagnostics and loaded this
;  code and then made a call to it, the MS controller power on
;  diagnostics are done testing low memory and check for
;  existence is valid.  We still wait for controller to go
;  non-busy in case high memory is still being tested.

MSDio	PUSH	PSW
	PUSH	H	; Save parameter.
	LXI	H,1FEFH	; Controller test byte.
	XRA	A	; Make a zero.
	MOV	M,A	; Put it in test location.
	CMP	M	; did it go?
	POP	H	; restore parameters.
	MVI	A,7	; no controller code in A.
	JNZ	ARGH	; jmp/ no controller.
MSDio00	LDA	STATUS	; Read status.
	ORA	A
	JZ	MSDio0	; It's OK go ahead.
	CPI	0FFH
	JZ	MSDio00	; It's not done.
ARGH	POP	H	; Garbage on stack.
ARGH1	MOV	E,A
	MVI	D,1
	STC
	RET
;
;	Now shovel a couple of things which will not change
; into their Parameter areas so we can free up the registers.
;
MSDio0	MOV	A,C
	STA	UNIT
	MOV	A,B
	STA	COMMAND
	CPI	2
	JZ	NOXFER		; INIT and EXECUTE require no
	CPI	4		; transfer of bytes, so
	JZ	NOXFER		; go to NOXFER.
	CPI	5
	JZ	MSDiskSize
	SHLD	DISKADD
	POP	PSW		; # sectors.
	CALL	XFER		; Do the transfer.
	JNZ	ARGH1		; Jmp/ Error.
	ORA	A		; NOT CARRY.
	RET
	PAGE
; Here we do the Init and Execute 1000H commands

NOXFER	CALL	DIOEX
	JNZ	ARGH		; bail out on errors
	POP	PSW
	ORA	A		; Clear carry
	RET

; Do I/O transfers for either read or write

XFER	STA	SECTS		; # to do
	LXI	H,BUFFER	; Where to put stuff
	SHLD	MEMADD

; Figure out how much memory to move around

	MOV	B,A
	XRA	A
	MOV	C,A
	LDA	COMMAND
	CPI	3
	JZ	RDXFR		; jmp/read op
	CPI	1
	JNZ	WRXFR		; jmp/must be write

; Do read transfer

RDXFR	CALL	DIOEX
	RNZ			; do it and split if bad
	LXI	H,BUFFER
	JMP	Move		; return thru move.

; Do write transfer

WRXFR	XCHG
	LXI	D,BUFFER
	CALL	Move
	XCHG

; DIOEX executes the transfer and reports errors

DIOEX	LXI	H,STATUS
	MOV	A,M
	ORA	A
	JNZ	DIOEX3		; jmp/nfg- back out.
	MVI	M,0FFH		; and turn controller loose.
DIOEX1	CALL	VGiveup
	MOV	A,M		; wait for completion
	CPI	0FFH
	JZ	DIOEX1
	ORA	A		; if status <>0, error
DIOEX3	MVI	M,0		; set to done (finally)
	RET			; on error gets NOT ZERO.
	PAGE
; DiskSize function returns size of hardware in pages.

MSDiskSize POP	H		; Toss number of pages.
	MVI	A,1		; READ command
	STA	COMMAND		; of one page from a guaranteed
	STA	SECTS		; -to-exist page: page zero.
	LXI	H,1FEEH		; Illegal disk address so
				; cache won't find it and
				; drive will be selected.
	SHLD	DISKADD		; UNIT is already set up to
	CALL	DIOEX		; users requested drive #.
	JZ	MSDS1
	CPI	6		; Only a disk not ready error
	JZ	ARGH1		; is bad. Report it.
MSDS1	LXI	H,1FEEH		; The disk descriptor byte.
	MOV	A,M
	ANI	040H		; If high, this bit says got
	LXI	H,77*32*1	; single sided.
	RNZ			; So return. NOT CARRY
	DAD	H		; Else got double: double it.
	RET			; NOT CARRY
	PAGE
;	**** Double Density Dio ****

DDDio	PUSH	PSW
	PUSH	B	; Another DIO param.
	LXI	B,DDBootTime ; Wait this long if booting.
DDDio2	LDA	Status
	ORA	A	; Test Status
	JZ	DDDio0	; If was DONE (zero), go do operation.
	CPI	BOOTING	; If is booting, wait it out.
	JZ	DDDio20	; Jmp/ booting
	CPI	NONEXIST
	JZ	DDDio21	; Jmp/ non-existent controller
	PUSH	D
	PUSH	H
	CALL	AErr	; Get an error acknowledge from Z-80
	POP	H
	MOV	A,B	; and stick the error code in DE.
	ORA	C	; Did AErr time out?
	JNZ	DDDio01 ; Jmp/ no - controller is ready to go.
	POP	H	; Get DE back, don't mess up error code
	POP	B	; Controller looks really sick.
	POP	PSW
	STC
	RET
DDDio20	DCX	B	; Do this BC times.
	MOV	A,B
	ORA	C	; If out of time, it's really broken
	JNZ	DDDio2	; so fall into below.
DDDio21	POP	B
	POP	PSW
NoCont	LXI	D,107H	; 'Controller does not exist'
	STC
	RET

; A controller has responded and has DONE status now.
	
DDDio01	POP	D
DDDio0	POP	B
	POP	PSW
	STA	Pages		; Save # of pages to xfer.
	LDA	Status		; Is the board out there?
	INR	A		; If got 0FFH, it is not, 
	JZ	NoCont		; because that is what non-
	XCHG			; memory looks like.
	SHLD	MemAdd		; Save memory addr pointer.
	LXI	H,DiskAdd 	; Pass disk addr to controller.
	MOV	M,E
	INX	H
	MOV	M,D
	MOV	A,B
	STA	Command		; Save Dio command.
	MOV	A,C
	STA	Unit		; Pass drive # to controller.

	EI
	CALL	DDDio00
	RET

DDDio00	LDA	Command
	ORA	A
	JZ	DDWrite		; 0: Write
	DCR	A
	JZ	DDRead		; 1: Read
	DCR	A		; 2: Initialize media. (None)
	DCR	A
	JZ	DDRead		; 3: Cache read. Same as plain.
	DCR	A		; 4: Execute (None)
	DCR	A
	JZ	DDSize		; 5: Disk size.
	LXI	D,101H		; 'DIO Says: Bad parameters'
	STC
	RET
	PAGE
; Write one or more sectors onto disk.

DDWrite LHLD	MemAdd
	XCHG			; DE is moving mem ptr.
	LDA	Pages		; Below we assume Pages<>0.
	STA	Pages1		; Initialize the 8080 side
	ORA	A		; page counter.
	RZ			; Zero pages is easy.
	XRA	A		; The double buffer
	STA	PageSwitch	; always starts with zero page.
	LXI	H,Status 	; Send DiskAdd, Unit, Pages by
	MVI	M,WRITESTART	; means of WRITESTART.
DDWrite0 CALL	VGiveup
	MOV	A,M
	CPI	WRITESTART	; Wait for DONE.
	JZ	DDWrite0
	ORA	A		; If it changes but not to
	JNZ	AErr		; DONE, it is an error.
	CALL	DDWrite3	; Buffer the first page.
	LXI	H,Status	; Send first page.
	MVI	M,WRITE

; Top of double buffered page at a time write ahead loop.
; (Pump is primed)

DDWrite1 EI
	LXI	H,Pages1	; Are we done?
	DCR	M		; If so, go wait out last page.
	JZ	DDWrite41	; Jmp/ no more to do.
	CALL	DDWrite2	; Wait for WORKING.
	RC
	LXI	H,PageSwitch	; We own PageSwitch now.
	MOV	A,M		; So we can toggle it
	CMA			; for next page.
	MOV	M,A
	CALL	DDWrite3	; Write to PageSwitch page.
	LXI	H,Status
	MVI	M,WRITE		; Send that page.
	JMP	DDWrite1	; Loop.

; Just sent last page: Wait for DONE ( WORKING not good enough)

DDwrite40 CALL	VGiveup	; wait for a bit.
DDWrite41 LDA	Status		; Wait for last page.
	CPI	WRITE		; The last page is special
	JZ	DDwrite40	; because we don't signal
	CPI	WORKING		; for the next page.
	JZ	DDwrite40
	ORA	A
	JNZ	AErr		; We must get DONE.
	RET			; NOT CARRY.

; Do actual write transfer of the page.

DDWrite3 LDA	PageSwitch	; We have control of PageSwitch
	ORA	A		; PageSwitch
	LXI	H,Page1		; Tells which page buffer to
	JNZ	DDWrite31	; use.
	LXI	H,Page0
DDWrite31 LXI	B,256		; One page = 256 bytes + cksum.
	XCHG
	CALL	Move		; Move [HL] to [DE], BC bytes.
	XCHG			; Leaves DE moved.
	MOV	A,B		; Get checksum.  NOTE: we DON'T
	CMA			; negate but only complement!
	MOV	M,A		; This means 257 zero bytes
	RET			; would be bad like it should.

; Wait for latest page done writing and signal for a new one,
; when the last is WORKING or DONE, we can change the
; PageSwitch.  We set WRITE and then must leave PageSwitch
; alone until the next cycle.

DDwrite2 CALL	VGiveup
DDWrite2 LDA	Status
	CPI	WRITE
	JZ	DDwrite2
	CPI	WORKING
	RZ			; WORKING or DONE are good.
	ORA	A
	RZ
	JMP	AErr		; If last page bad, ret err.
	PAGE
;	Read one or more pages from disk.  Start off by
; sending the Z-80 the number of pages to xfer in Pages
; and the starting 0 based logical page number in DiskAdd.
; Z-80 acknowledges the READSTART command code passed in
; Status by either an error code or DONE code.  8080
; starts the first page by setting DONE back to READ.
; On subsequent pages, the Z-80 sets READ to WORKING
; as soon as it finishes the previous page: the 8080
; then sets READ for the next page if there is one and
; transfers the current page.  On last page we have to
; remember not to start the next page before the xfer.
; One big worry is the 8080 side interruption (CTRL-Y, CTRL-Z
; etc) which should not leave Z-80 hanging forever.  The
; Z-80 will always finish by itself in error Status or
; DONE, and if all the pages are not yet moved, the
; next READSTART or WRITESTART will generate an error.

DDRead	LHLD	MemAdd
	XCHG			; DE is moving mem ptr.
	LDA	Pages		; Zero pages is easy.
	STA	Pages1		; Init counter.
	ORA	A		; Below, we assume Pages<>0.
	RZ
	XRA	A		; Always starts 0.
	STA	PageSwitch	; Start Z-80 reading.
	LXI	H,Status	; DDRead2 will wait out the
	MVI	M,READSTART	; READSTART before going on.
DDRead0	CALL	VGiveup
	MOV	A,M
	CPI	READSTART	; Wait for no more READSTART.
	JZ	DDRead0		; If it changes, but not to
	ORA	A		; DONE, it is an error.
	JNZ	AErr
	MVI	M,READ		; Then signal read first page.
	CALL	DDRead2		; Wait for page
	RC

; Top of read-ahead loop.

DDRead1 EI
	LXI	H,PageSwitch	; Only here can we change
	MOV	A,M
	CMA			; PageSwitch: we have owner-
	MOV	M,A		; ship while Status is DONE
	LXI	H,Pages1	; Is there only one left?
	DCR	M		; (Zero case done above)
	JZ	DDRead4		; jmp/ Yes.
	MVI	A,READ		; or WORKING.
	STA	Status		; Start next read!
	CALL	DDRead2		; Wait for WORKING.
	RC
	CALL	DDRead3		; Xfer a middle page.
	RC
	JMP	DDRead1		; Jmp/ got >2 pages.

; Transfer last page without signalling for new.
; Got to do more waiting, because WORKING is not
; good enough for last page: got to have DONE.

DDread4	CALL	VGiveup	; go away and wait for a bit
DDRead4	LDA	Status		; Wait for last page done.
	CPI	READ		; The last page is special
	JZ	DDread4		; because we don't signal
	CPI	WORKING		; for the next page.
	JZ	DDread4
	ORA	A
	JNZ	AErr		; We must get DONE.
				; Fall into xfer of last page.

; Do actual read transfer of a page.

DDRead3 LDA	PageSwitch
	ORA	A		; PageSwitch
	LXI	H,Page1		; Tells which page buffer to
	JZ	DDRead31	; use. NOTE: Opposite sense due
	LXI	H,Page0		; to read-ahead!
DDRead31 LXI	B,256		; One page.
	CALL	Move		; Move [HL] to [DE], BC bytes.
	MOV	A,M		; Get NOT checksum from buffer.
	CMA			; Convert to checksum.
	CMP	B		; Must equal checksum of data.
	RZ			; If so, is good. NOT CARRY.
ICkErr	LXI	D,108H		; 'Internal checksum error'
	STC
	RET

;  Wait for latest page done reading and signal for new one.

DDread2	CALL	VGiveup
DDRead2 LDA	Status
	CPI	READ
	JZ	DDread2
	CPI	WORKING
	RZ
	ORA	A
	JNZ	AErr		; If last page bad, ret err.
	RET
	PAGE
; Size function tells how big a given diskette in
; a given drive is. Ret size in HL.

DDSize	LXI	H,Status
	MVI	A,DISKSIZE
	MOV	M,A
DDSize1	CMP	M
	JNZ	DDSize2
	CALL	VGiveup
	JMP	DDSize1
DDSize2	MOV	A,M
	ORA	A
	JNZ	AErr
	LXI	H,DiskAdd	; Disk Size function gives
	MOV	E,M		; number of pages hardware is
	INX	H		; capable of in HL.
	MOV	D,M
	XCHG
	RET			; NOT CARRY (by JNZ)
	PAGE
; Error bailouts for the DD controller

AErr	MOV	E,A		; Code in A becomes 1xx
	MVI	D,1		; error code in DE.
	LXI	H,Status
	LXI	B,DDBootTime	; Wait for controller card
AErr1	MOV	A,M		; to give us Done.
	ORA	A		; If Done, don't give Z80
	STC			; another ERRACK - It takes up
	RZ			; to 1ms for Z80 to respond w/
	CPI	LOWESTCOMMAND	; DONE status.  If Status is
	JNC	AErr3		; one of our 8080 owned command
	MVI	M,ERRACK	; commands, then put in the
AErr3	NOP			; ERRACK.  If not we cannot
	NOP			; change Status! NOTE ERRACK 
	NOP			; is greater than
	DCX	B		; LOWESTCOMMAND, so we only
	MOV	A,B		; hit it once.
	ORA	C
	JNZ	AErr1
	STC
	RET
	PAGE
	LIST	0		; Don't even print this mush!
	IF	Stupid		; rom based code!
	LIST 1

; Dumbio is the dumb 5" disk I/O handler. The registers are:
;	A	number of sectors
;	B	command, 0=write, 1=read, 2=verify
;	C	unit number (1, 2, or 3)
;	DE	memory address
;	HL	disk address (will bust later)

; If an error is detected, return with C set and code in DE

;Dumbio				; For ROM based code
;	CALL	VLock		; just me for a while, please!
;	PUSH	PSW
;	PUSH	B
;	PUSH	D
;	PUSH	H		; stash all the regs
;	LDA	087BH		; check out the driver
;	CPI	0F5H		; should be PUSH PSW
;	JNZ	Dumbdum		; jmp/shit on you!
;	LDA	087CH
;	CPI	79H		; and a MOV A,C
;	JNZ	Dumbdum
;	CALL	Dxchg		; exchange I/O area of memory
;	POP	H
;	POP	D
;	POP	B
;	POP	PSW
;	CALL	087BH		; ROM sssd driver!
;	PUSH	PSW
;	PUSH	B
;	PUSH	D
;	PUSH	H
;	CALL	SUDhalt		; then halt 'em
;	CALL	Dxchg		; reswap stuff
;	JMP	Ioret		; and scram

Dumbdum	LXI	D,109H		; no controller erro!
	JMP	LockErr

;Dxchg	DI
;	LXI	H,0D00H
;	LXI	D,2DA0H
;	LXI	B,60H
;Dxchg1	MOV	B,M		; from HL
;	LDAX	D		; from DE
;	MOV	M,A		; swap this
;	MOV	A,B
;	STAX	D		; swap that
;	INX	H
;	INX	D
;	DCR	C
;	JNZ	Dxchg1
;	RET			; ugly!


Dumbio	PUSH	D
	PUSH	B
	PUSH	PSW
	PUSH	H		; stash on stack.
	ORA	A		; check for 0 blocks
	JZ	DERR1		; jmp/nuts to you.
	LXI	D,-350 		; ??huh?
	DAD	D		; check disk address, I guess.
	JC	DERR1		; jmp/DA stinks.
	MOV	E,A
	MVI	D,0
	DCX	D
	DAD	D		; ??
	JC	DERR1		; jmp/junk.
	MOV	A,C
	CPI	1
	JC	DERR1		; jmp/unit number no good.
	CPI	4
	JNC	DERR1		; jmp/unit number no good.
	MOV	A,B
	CPI	1
	JZ	DIO1		; jmp/read command
	CPI	2
	JZ	DIO1		; jmp/verify
	ORA	A
	JZ	DIO1		; jmp/a write
	CPI	5
	JNZ	DERR1		; jmp/disk size function
	LXI	H,350		; 350 for small disks
	POP	B		; trash entry HL value
	POP	PSW		; restore PSW
	POP	B		; BC
	POP	D		; and DE
	JMP	Diox		; split.


;DIO1	LXI	H,SPINT
;	SHLD	SRA2	; insure WE get the interrupts..

; Now we does a very dangerous thing. We saves SP
; under a rock for error 106 bail-out later...

	LXI	H,0
	DAD	SP
	SHLD	DSMA	; save over here.

; See if the motor is going. If it is, go make final
; check for write enable on write command. Motor on is
; indicated by MTO being non-zero. If it is zero, we 
; must wait the power-up and synch time, check for a 
; disk in that unit, etc.

DIO2	LXI	H,MTO
	MOV	A,M
	ORA	A
	MVI	M,MTIME
	JZ	DIO3	; jmp/motor not on.
	LDA	CDRIV	; well, is the drive we want
	CMP	C	; selected now?
	JZ	DIOG	; jmp/yup, scratch off.
	MOV	A,C
	STA	CDRIV	; we want this one.
	MVI	A,40	; wait 20 ticks for head load
	JMP	DIO3A	; go select that drive.

; Start up the motor, see if we have a disk by counting
; sector interrupts over a period of time. We set up the
; RTC to 1 in the lower byte, and wait for it to count
; around a while, then check to see if DTO has changed.
; If it hasn't, no media in the drive-gripe.

DIO3	MOV	A,C
	STA	CDRIV	; Mark this one as current
	MVI	A,CMG	; motor go command
	OUT	DCO	; huzzak!
	XRA	A
DIO3A	STA	TIMER	; set up timer
	MOV	D,M	; get mtime vale
	MOV	A,C	; drive number
	OUT	DSEL	; select a drive
	MVI	A,CESI
	OUT	DCO	; enable sector interrupts
	IN	DST	; flush crud interrupt

; Now wait for 60 ticks (1 second) for motor to speed up

DIOW	CALL	VGiveup
	LDA	TIMER
	CPI	60	; a second yet?
	JC	DIOW	; nope, wait

; A second has gone by. See if DTO has changed.
; Note still disabled, and (HL) -> MTO.

	MOV	A,M
	ADI	5	; want at least 5...
	CMP	D	; any interrupts?
	JC	DIOG	; jmp/yup, assume motor going
	CALL	DHALT	; else halt the disks
	JMP	DERRB	; and report error-no disk!

; Drive is going; see if write protected and write req.

DIOG	MVI	M,MTIME	; reset the counter.
	XRA	A
	CMP	B	; is this a write?
	JNZ	DIOA	; jmp/nope, do it
	IN	DST	; get status
	ANI	DWP	; see if write protected
	JZ	DIOA	; jmp/nope

; The start of the error chain for DIO. Error code in
; D, subcode in E. The error code is 01.

	MVI	E,5	; Disk is write protected
	DB	21H	; two byte skip
DERR1	MVI	E,1	; bad parameters. pfui!
	DB	2EH
DERR6	MOV	E,A	; hard disk error!
	DB	21H
DERRB	MVI	E,6	; no media in drive.

; Now stuff in code and pull registers for exit.

	MVI	D,1
	POP	H
	POP	PSW
	POP	B
	XCHG		; DE and HL
	XTHL		; EC on stack, old DE in HL
	POP	H	; We want to leave HL alone
	XCHG		; and error code in DE.
Diox	XRA	A
	STC		; set carry
	EI		; insure enabled on return!
	RET		; not sure that's right!
;
DIOA	POP	H
	POP	PSW
	LXI	D,0FFF6H
	MVI	B,-1
DIO4	INR	B
	DAD	D
	JC	DIO4
DIO5	CPI	11
	JNC	DIO5a	; jmp/more than a track
	ADD	L
	JM	DIO6
	JZ	DIO6
	JMP	DIO5b
DIO5a	ADD	L
DIO5b	MOV	H,A
	MOV	A,L
	ADI	10
	MOV	D,A
	SUI	11
	CMA
	STA	MUNG2
	MOV	L,B
	SHLD	MUNG3
	POP	H
	SHLD	MUNG4
	MOV	C,L
	MOV	E,H
	POP	H
	CALL	DOIO		; go do the shit.
	LDA	MUNG1+1		; number sectors we did
	ADD	H		; update memory address
	MOV	H,A		; for the next one....
	PUSH	H
	LHLD	MUNG4
	PUSH	H
	LHLD	MUNG3
	MOV	A,H
	MOV	B,L
	INR	B
	MVI	L,0F6H
	JMP	DIO5
DIO6	SUB	L
	PUSH	PSW
	MOV	A,L
	ADI	10
	MOV	D,A
	POP	PSW
	POP	H
	MOV	E,H
	MOV	C,L
	POP	H
;
DOIO	STA	MUNG1+1
	LDA	DTRIES		; number of retries
DOI1	DI
	PUSH	PSW
	PUSH	H
	PUSH	D
	PUSH	B
	LDA	MUNG1+1
	CALL	RDSEC		; do the operation
	MOV	H,A		; save possible error code!!!
	ORA	A		; set condition codes!
	MOV	A,B
	STA	MUNG5
	POP	B
	POP	D
	JZ	DOI2		; jmp/no errors, check for vfy.
	PUSH	D
	MOV	E,H		; error code- a 2, 3, or 4
	LXI	H,BUGS-2	; error counter cell
	MVI	D,0
	DAD	D		; point at counter
	INR	M		; bump.
	POP	D
	JNZ	DOI1a		; jmp/no overflow
	DCR	M		; make it stick at 255.
DOI1a	POP	H
	MVI	A,MTIME
	STA	MTO		; dont timeout retries!
	POP	PSW
	DCR	A
	JNZ	DOI1		; jmp/retry the operation
	LDA	DSTAT		; error code to report
	PUSH	H
	PUSH	H
	PUSH	H		; should be nuff...
	PUSH	H
	JMP	DERR6		; else hard disk error.

RHD	POP	H
	POP	D
	EI
	RET

DOI2	MOV	A,E
	ORA	A
	JNZ	RHD		; jmp/return popping H and D
	POP	H		; we ALWAYS verify
	POP	PSW
	MVI	E,2
	JMP	DOI1		; fake reentry for verify.
	PAGE

; WSEC is called to wait a number of sectors, passed in A. 

WSEC	PUSH	B
	PUSH	PSW		; save stuff.
	MOV	B,A
	IN	DST		; sector we're on now
	ANI	0FH		; just sector, pleeze.
	ADD	B
	DAA
	ANI	0FH
	MOV	B,A		; the one we wait for
	MVI	A,1
	STA	TIMER		; clock out this process...
WSECL	CALL	VGiveup
	DI
	IN	DST
	ANI	0FH		; just sector
	CMP	B		; right one yet?
	JZ	WSECX		; jmp/yup, scram.
	LDA	TIMER		; have we timedout?
	ORA	A
	JNZ	WSECL		; jmp/nope, keep going

; Deep shit. Sector didn't come around in 4 sec.

	LHLD	DSMA		; sp hid under rock
	SPHL			; restore this.
	JMP	DERRB		; report something wrong.
WSECX	POP	PSW
	POP	B
	RET			; vanish.
	PAGE
; TRKS is called to step in or out a number of tracks.
; Entry is with HL pointing to track counter for the
; drive, A contains number of tracks to move- +=in, -=out.

TRKS	ORA	A		; well, what do we do
	RZ			; ret/nothing.
	PUSH	B		; save.
	MVI	B,CSO		; set step in
	JP	TRKS1
	DCR	B
	CMA
	INR	A
TRKS1	MOV	C,A		; number steps
	MOV	A,B		; command
	OUT	DCO		; issue direction set command
	PUSH	H		; save.
	LXI	H,MTO		; point at motor time out

; We need to step the number of steps in C. After issuing
; the step pulse, we wait for 2 sectors to go by to
; let the step complete.

TRKS2	MVI	M,MTIME		; don't let motor stop
	MVI	A,CPH		; step ff high
	OUT	DCO		; raise line
	DCR	A
	OUT	DCO		; lower line
	MVI	A,2
	CALL	WSEC		; wait 2 sector times for that
	DCR	C		; step some more?
	JNZ	TRKS2		; jmp/yes
	POP	H
	POP	B
	RET			; split.
	PAGE

; RDSEC is the actual device control code. You enter with:
;	HL	memory address
;	DE	sector/command
;	BC	track/unit number
;	A	number of sectors

RDSEC	PUSH	H
	PUSH	D
	PUSH	B
	PUSH	PSW		; save everything.

; A small bitch in the two user system. Remember which user
; we're doing the I/O for, so we can put the data in the right
; place!!!

	LDA	BRG		; if RAM code, get BRG contents
	ANI	PHANTOM		; figure out which user
	STA	SSEC+1		; stuff that over here.

	MVI	A,9
	ADD	D		; we want to start on the prev.
	DAA			; sector (mod 10)
	ANI	0FH
	STA	SSEC		; starting sector.
	MOV	D,B		; trk desired
	MVI	B,0
	LXI	H,TRK1-1
	DAD	B		; point to right drive ctr
	MOV	A,M		; current track for drive
	CPI	0FFH		; is it set up?
	JNZ	RDS1		; jmp/yes

; If the track cell contins FF, it means we have not
; used this drive yet. As a precautionary measure, we
; go through the song-and-dance of stepping IN 20 steps,
; then stepping out to track 0. This is to make sure 
; the head positioning mechanism on the Shugart is in
; the cam groove. (Mickey Mouse is what it is)...

	MVI	A,20
	CALL	TRKS		; step in 20
MMSE	MVI	A,0FFH		; step out 1
	CALL	TRKS
	IN	DST
	ANI	DTR0		; does it say track 0 yet?
	JZ	MMSE		; jmp/mickey says no.
	MVI	M,0		; set holder to 0.

; Now figure out which way to step, and step there

RDS1	MOV	A,D		; track we want
	SUB	M		; minus where we are
	CALL	TRKS		; go there already.
	MOV	M,D		; say we are there
	MOV	A,D
	STA	DTRK		; remember track for I/O code.

; Now set up to do the operation.

	MOV	A,E
	ADD	A		; double
	MOV	E,A
	MVI	D,0
	LXI	H,IVEC
	DAD	D
	MOV	E,M
	INX	H
	MOV	D,M
	XCHG
	SHLD	SINT		; bail-out point
	POP	PSW
	STA	DSCT		; save number sectors to do
	POP	B
	POP	D
	POP	H		; unwind
	SHLD	DSMA		; memory address to use
	IN	DST		; clear any trouble pending.
	MVI	A,0FFH
	STA	DSTAT		; set status up

; Now wait for it to be done.
; Note that when DSTAT goes to other than FF, if
; it does not go to 0 we have an error, with error code
; in DSTAT, passed as non-zero condition code back to
; caller....

RDS2	CALL	VGiveup
	DI			; provide small window
	LDA	DSTAT		; see what status looks like.
	INR	A
	JZ	RDS2		; jmp/still working.
	XRA	A
	LDA	DSTAT		; get status
	ORA	A		; provide condition codes
	RET			; to caller.

; Vector of processing routines

IVEC	DW	CWRT	; 0=write
	DW	Read	; 1=read
	DW	Verify	; 2=verify

; CWSUB is called to wait for TBE to go high. Saves me
; memory, because 8+3n < 7n for n>2.

CWSUB	IN	USTAT
	ANI	TBE		; is it up yet?
	RNZ			; jmp/yes, split
	JMP	CWSUB

; CURST is called to reset the cursed USART.

CURST	LHLD	DSMA
	MOV	B,H
	MOV	C,L
CURS1	MVI	A,0CBH
	OUT	USTAT
	MVI	A,0C0H
	OUT	USTAT
	RET
	PAGE
; The Write routine is entered at the
; interrupt level. We have DSMA containing the memory
; address, DSCT conaining the number of sectors to
; write. When we finish, set DSTAT to 0 and Ioret.

CWRT	CALL	CURST
	MVI	A,CWG
	OUT	DCO		; set write on
CWRTA	MVI	L,5
CWRT0	CALL	CWSUB		; wait for TBE
	XRA	A
	OUT	UDAT
	OUT	UDAT
	DCR	L
	JNZ	CWRT0		; write 10 zero bytes
	MVI	H,0		; the sneeky way to do it...
CWRT1	CALL	CWSUB		; wait for TBE
	MVI	A,DSYN
	OUT	UDAT
	OUT	UDAT
	CALL	CWSUB
	IN	DST		; get sector number
	ANI	0FH		; mask to 0-9
	ORI	80H		; tag
	OUT	UDAT		; shovel out
	LDA	DTRK		; first shovel out sector+80H
	OUT	UDAT		; then output track number
	MVI	A,80H
CWRT2	PUSH	PSW		; byte count
	CALL	CWSUB		; wait for TBE and then shovel.
	LDAX	B
	INX	B
	OUT	UDAT
	MOV	E,A
	LDAX	B
	INX	B
	OUT	UDAT
	MOV	D,A
	DAD	D
	POP	PSW
	DCR	A
	JNZ	CWRT2		; still going in record
	CALL	CWSUB		; wait for TBE before cksum...
	MOV	A,L
	CMA
	OUT	UDAT
	MOV	A,H
	CMA
	OUT	UDAT		; send out comp checksum
	LXI	H,DSCT
	DCR	M
	MVI	A,CWH
	OUT	DCO		; start writing ones
	JZ	CWRTX		; Jmp/all done.
	MVI	A,CWG
	OUT	DCO		; start write again,
	IN	DST		; clear magic bit
CWRT3	IN	DCI		; now wait for sector to come
	ANI	CSF		; around, before we will
	JNZ	CWRTA		; start writing again.
	JMP	CWRT3		; loop waiting for hole.

; Read/Verify logic

CVERR	MVI	A,4
	DB	21H		; skip next two bytes.
CRDE1	MVI	A,2		; preamble error
	POP	B		; crud from stack.
	DB	21H
CRDE2	MVI	A,3
	PUSH	PSW
	MVI	A,CRH
	OUT	DCO		; turn off read
	POP	PSW
	JMP	CRDX		; and split reporting error.

; Combined Read-Verify routine

Read	MVI	L,77H		; MOV M,A routine
	DB	11H		; skip 2 bytes
Verify	MVI	L,0BEH		; CMP M routine
	MVI	H,0C9H
	SHLD	SCUZZ		; code to do for each byte
CRVA	LHLD	DSMA		; since this code is common,
	XCHG			; put MA into DE
CRVAA	CALL	CURS1
	MVI	A,CRG
	OUT	DCO		; read go
	MVI	A,80H		; number of bytes to read
	PUSH	PSW		; bury it
	LXI	H,0		; clear checksum counter
	MOV	B,H
	MOV	C,L		; copy for state flags

; The timeout code provided by John Stephenson

	IN	DST		; reset CSF bit in controller
Crva1	IN	DCI
	ANI	CSF		; seen a hole yet?
	JZ	Crva1		; wait for the hole
	MVI	C,40H		; timeout value
Crva2	DCR	C		; dcr timeout
	JZ	CRDE1		; jmp/no synch, crap out!
	IN	USTAT
	ANI	RBF
	JZ	Crva2		; wait for data
	IN	UDAT
	MOV	B,A		; junk or sector number
	IN	UDAT
	MOV	C,A		; track number
	LDA	DTRK		; we hope....
	CMP	C
	JNZ	CRDE1		; jmp/oops, wrong track?!?
	CMP	B
	JZ	Rv		; jmp/old disk has trk-trk on it
	IN	DST
	ANI	0FH		; new disk has sector+80H on it
	ORI	80H
	CMP	B
	JNZ	CRDE1		; jmp/wrong sector?!??!!!
Rv	IN	USTAT
	ANI	RBF
	JZ	Rv
	XCHG			; put MA into HL, cksm in DE
	IN	UDAT
	MOV	C,A
	CMP	A		; preset cc's for read case
	CALL	SCUZZ		; do store or compare
	JNZ	CVERR		; and barf if bad
	INX	H
	IN	UDAT
	MOV	B,A
	CMP	A
	CALL	SCUZZ
	JNZ	CVERR
	INX	H
	XCHG			; ma to DE, cksm to HL
	DAD	B		; update cksm
	POP	PSW
	DCR	A		; done yet?
	PUSH	PSW
	JNZ	Rv		; jmp/nope
	POP	PSW

CRVB	IN	USTAT
	ANI	RBF
	JZ	CRVB		; this is read cleanup-
	IN	UDAT		; we check checksum and such
	MOV	C,A
	IN	UDAT
	MOV	B,A		; remember, MA in DE...
	DAD	B
	INX	H
	MOV	A,L
	CMP	H
	JNZ	CRDE2		; oops....checksum stinks.
	ORA	A
	JNZ	CRDE2
	LXI	H,DSCT
	MVI	A,CRH
	OUT	DCO		; read stop
	DCR	M
	JNZ	CRVAA		; jmp/more to do, do setup

; All done. Split.

CWRTX	XRA	A
CRDX	STA	DSTAT		; store status

; For the RAM code, return to the proper user, restoring the
; correct stack pointer.

	LDA	BRG
	OUT	BRGEN		; put it back!
	LHLD	USP
	SPHL			; and roar off.

	MVI	A,0FFH
	STA	SSEC		; don't come around again!
	LXI	H,Ioret
	SHLD	SINT		; clear the bail-out
	PCHL			; and spit.
	PAGE

; SPINT is entered as a result of a sector pulse interrupt
; from the controller. If SSEC equals the sector we just hit,
; we bail out thru SINT to process. Otherwise, we go to Ioret,
; after checking to see if we time out motors by decrementing
; DTO. If 0, turn off motor and deselect drive.

SPINT	IN	DST		; drive stat
	ANI	DSM		; mask to sector number
	MOV	B,A
	LDA	SSEC
	CMP	B
	JNZ	SPIN1		; jmp/not there yet.

; For the two user system, let's make sure we do the transfer
; to the correct user...

	LXI	H,0
	DAD	SP		; save current SP
	SHLD	USP		; in case we park the clown.
	LDA	BRG		; see who we're running now
	ANI	PHANTOM
	MOV	B,A
	LDA	SSEC+1		; the guy that wants the I/O
	CMP	B
	JZ	SPINTA		; jmp/we win! the right one!
	LDA	BRG
	XRI	PHANTOM		; else flip to the other one
	OUT	BRGEN		; ***** SWAP *****
	LHLD	USP
	SPHL			; get that guy's stack.
SPINTA	LHLD	SINT
	PCHL			; zoom off if we hit it

SPIN1	LXI	H,MTO
	DCR	M
	CZ	DHALT		; halt drives if timed out.
	JMP	Ioret		; and split.

; DHALT called to stop things, and mark them stopped.

DHALT	PUSH	PSW		; does not change any regs.
	MVI	A,0FEH		; if any I/O is in progress, we
	STA	DSTAT		; will abort it code FE.
	MVI	A,DS0
	OUT	DSEL		; select drive 0 to raise heads
	STA	CDRIV		; mark nobody selected
	STA	MTO		; mark motors off
	MVI	A,CMH
	OUT	DCO		; and turn off motors
	MVI	A,CDSI
	OUT	DCO		; disable interrupt on sp
	POP	PSW
	RET			; and leave.
	PAGE
	ENDIF
	LIST	1

; Priam driver for the system 88 computers.
;
; Written by Frank Anderson 11/6/80
; Copyright 1980  Interactive Products Corperation
;
;
RETADDR	EI		; Guarentees that we will return
	RET		; from Hardio with interupts enabled

DTETRYS	EQU	2		; # Data Timeout retries we'll
				; do before declaring an error

HardDio	PUSH	H		; Save HL
	LXI	H,RETADDR	; Where to go to split
	XTHL			; Save so that's where we go
;
;  Actual driver code for the Priam hard disk.    
;    Registers:
;	ENTRY				EXIT
;	AF= # sectors (no flags)	AF= carry set if error 
;	B = command			B = junk
;	C = drive #			C = junk
;	DE= Memory address		DE= error code if error
;	HL-> Disk address (3 bytes)	HL-> Size if Dsize cmnd
;
	STA	NSECTS		; Save # sectors
	MVI	A,DTETRYS	; Set up Data Timeout Error
	STA	DTETrys		;  counter.
	SHLD	HDISKAD		; Save disk address pointer
	XCHG
	SHLD	HMEMADD		; Save memory address
	MOV	A,C		; Get drive #
	SUI	11		; Subtract Dev # offset
	STA	DRIVE		; Save it
	;
	; Find out if drive exists
	;
	CALL	INDXPMS		; Point at drive parm block
	MOV	A,M		; Is it here?
	ORA	A
	LXI	D,0109H		; No drive error (assume bad)
	STC			; Set error flag (assume bad)
	RZ			; If zero split, no drive
	;
	; Else check command and dispatch if valid.
	;
	LXI	D,0101H		; Bad parameters command
	MOV	A,B		; Get command
	STA	HCOMAND		; Save for later
	ORA	A		; Write command?
	JZ	DOIO		; jmp/ yup do it
	DCR	A		; Read command?
	JZ	DOIO		; jmp/ yup
	CPI	5-1		; Dsize command?
	JZ	DSIZE		; jmp/ yup
	CPI	3-1		; Cache read?
	STC			; Assume not
	RNZ			; Split if illegal command
	;
	; Else fall through
;
; Do a read or a write
;
DOIO	DI			; Don't bug me now!
	LDA	HCOMAND		; Was the command a write?
	ORA	A
	JZ	HWRITE		; jmp/ yup, go do it.
	;
	; Else read or cache read, so fall through
	;
	MVI	A,053H		; Command to A
	CALL	CMNDIT		; Mail command
	LHLD	HMEMADD		; Initialize pointers
READ1	CALL	WAIT
	JC	OOPS		; On error bail out
	JNZ	HARDIO3		; If completion do that
	MVI	B,0		; Initialize 256 byte counter
READ2	IN	DATA		; else geta a byte
	MOV	M,A		; save it
	INX	H		; bump pointer
	DCR	B		; Sector done?
	JNZ	READ2		; jmp/ nope, more to sector
	JMP	READ1		; loop, do next sector
;
; Do a write of NSECTS (less than CHUNK)
;
HWRITE	MVI	A,052H		; Command to A
	CALL	CMNDIT		; Mail command
	LHLD	HMEMADD		; Initialize pointers
WRITE1	CALL	WAIT		; Wait for something to happen
	JC	OOPS		; jmp/ error
	JNZ	HARDIO3		; jmp/ done
	MVI	B,0		; 256 byte counter initialized
WRITE2	MOV	A,M		; Get data
	OUT	DATA		; and mait it	
	INX	H		; Bump address
	DCR	B		; Done 1 sector?
	JNZ	WRITE2		; jmp/ nope continue
	JMP	WRITE1		; loop
	;
HARDIO3	ORA	A		; Transaction status 0?
	JNZ	OOPS		; jmp/ nope, report error
	RET			; Else split

; How big is the disk?
;
DSIZE	CALL	INDXPMS		; Index to correct parm block
	MOV	A,M		; get device existance flag
	INX	H		; Point to dsize entry
	INX	H
	INX	H
	INX	H
	ORA	A		; Does device exist?
	RNZ			; Split, yup (HL -> dsize)
	LXI	D,109H		; Else report a device
	STC			;  doesn't exist error
	RET

;-------------------------------------------------------------
;			UTILITIES
;	
; Handshake a command at the controller
;
CMNDIT	PUSH	PSW		; Save command
	CALL	INDXPMS
	INX	H		; Point to sects/trk
	MVI	D,0		; MSB of # sects/trk to D
	MOV	E,M		; # sects / track to DE
	INX	H
	PUSH	D		; Save # sects/trk on stack
	MOV	E,M		; # sects / cyl into DE
	INX	H
	MOV	D,M
	PUSH	D		; Save on stack
CMNDIT1	IN	HSTATUS
	ANI	BUSY
	JNZ	CMNDIT1		; Loop, device is busy
	;
	; Calculate from disk address where we want to go
	; as we need to specify cylinder and head addresses.
	;
	LDA	DRIVE		; Tell the controller which
	OUT	PARM0		;  drive number
	LHLD	HDISKAD		; Get pointer to disk address
	MOV	E,M		; Load disk address into HLDE
	INX	H
	MOV	D,M
	INX	H
	MOV	L,M
	MVI	H,0	
	POP	B		; Put # sects/cyl into BC
	CALL	DVID32		; DE = cyl, HL = remainder
	POP	B		; # sects / track to BC
	PUSH	D		; Save cyl #
	XCHG			; DE = LSW of remainder
	LXI	H,0		; HL = MSW of remainder
	CALL	DVID32		; DE = head, HL = sector
	MOV	A,L
	OUT	PARM3		; Sector # to PARM3 register
	MOV	A,E		; Get head #
	RLC			; Shift into most significant
	RLC			;    nibble for combination
	RLC			;    with MSB of cylinder #
	RLC
	POP	D		; Get back cylinder #
	ORA	D		; Combine head and MSB cyl.
	OUT	PARM1		; Stick it into register
	MOV	A,E		; Get LSB of cyl #.
	OUT	PARM2		; Put it into register.
	LDA	NSECTS		; Get # sectors
	OUT	PARM4		; Stick into its register.
	POP	PSW		; Get back command
	;
	; Entry point for no parameters passed
	;
CMNDITa	PUSH	PSW		; Save command
CMNDITb	IN	HSTATUS		; Wait till ready
	ANI	BUSY
	JNZ	CMNDITb
	POP	PSW		; Get back command
	OUT	HCMND		; Mail to command port
	RET			; Split
;
; INDXPMS points HL at the correct parameter block for the
; drive I'm working on (the one in DRIVE).  
;
INDXPMS	PUSH	B
	PUSH	PSW
	LDA	DRIVE
	LXI	H,PARMS		; Point at area where we store
				;  parameter for our drives
	MOV	C,A		; Multiply unit # by 7
	ADD	A
	MOV	B,A
	ADD	A	
	ADD	B
	ADD	C
	ADD	L		; Index into PARMS table
	MOV	L,A
	MOV	A,H
	ACI	0
	MOV	H,A		; HL -> drive parameters
	POP	PSW
	POP	B
	RET
;
; Wait for command completion,  returns with carry set if we
; get a command reject.  Returns with carry not set and zero
; not set for a controller completion request or with carry
; not set and zero set if we get a data transfer request.
; In case of a controller completion request it returns with
; the transaction status in the accumulator.

WAIT	IN	HSTATUS		; Get status from drive
	MOV	B,A		; Save it
	ANI	REJECT		; Was it a command reject?
	STC			; Assume yes and set flag
	JNZ	WAIT2		; Split, rejected
	MOV	A,B		; Get back status
	ANI	DXFER		; Data transfer request?
	JZ	WAIT1		; jmp/ nope try this
	XRA	A		; Else set zero flag
	RET			; And split
WAIT1	MOV	A,B		; Get back status
	ANI	COMPLET		; Command completion?
	JZ	WAIT		; jmp/ nope, try again
WAIT2	IN	TSTAT		; Get transaction status 
	PUSH	PSW		; Save transaction status
	XRA	A		; Acknoweledge completion
	OUT	HCMND
	POP	PSW		; Get back transaction status
	RET			; Split, flag set
	;
	; Translate error code
	;
OOPS	IN	TSTAT		; Get transaction status
	ANI	03FH		; Mask off error code
	CPI	33H		; Data Timeout Error?
	JNZ	OOPS1		; jmp/ retry transfer
	LXI	H,DTETrys	; Decrement retry counter
	DCR	M
	JNZ	DOIO		; jmp/ not out of tries yet
OOPS1	MOV	E,A		; LSB of offset to E
	MVI	D,0		; MSB of offset
	LXI	H,ErrTbl	; Point at base of table
	DAD	D		; Index into code
	MOV	A,M		; Get code
	ORA	A		; Soft error?
	JZ	SOFT
	CPI	40H		; No message?
	JNZ	HERROR		; jmp/ nope report it
	ORA	E		; Else report whatever it
				; gave me with a 40H tacked on
HERROR	MOV	E,A
	MVI	D,1		; Put code into D
	STC
EXIT	RET			; Split
;
SOFT	LHLD	SftCnt
	INX	H		; Bump count
	SHLD	SftCnt
	ORA	A		; Set carry off
	RET	

;
; IndxVol points HL at the head of the volume control block for
; the logical device specified in C.  All regs preserved except
; HL which contains information.
;
IndxVol	PUSH	PSW		; Save all
	PUSH	B
	PUSH	D
	MVI	A,10H		; Command to A
	CALL	Vmgr		; Do it
	POP	D		; Restore all
	POP	B
	POP	PSW
	RET			; Split

;	Compare HL - DE. Trash A.

DCMP	MOV	A,H
	CMP	D
	RNZ
	MOV	A,L
	CMP	E
	RET

;
; This is the integer math package lifted from FIG-FORTH v1.1
; Thank you very much!    9/26/80   Lifted by Frank E. Anderson
;	P.S.  The stuff is in the public domain.
;
;--------------------------------------------------------------
;	32 bit divide routine
;
;	Entry				Exit
;	------------------		--------------------
;	DE = Numerator(LW)		DE = Quotient
;	HL = Numerator(HW)		HL = Remainder
;	BC = Denominator
;
DVID32	; First check to see if the numerator is larger than
	; 256 * the denominator, if so it's overflow time.
	;

	MOV	A,L		; Double subtract BC from HL
	SUB	C
	MOV	A,H
	SBB	B
	JNC	OVRFLOW		; JMP/ too big for result regs.
	;
	; Else do the divide
	;
	MVI	A,16		; Initialize loop counter
DVID32a	DAD	H		; 32 bit shift of HLDE through 
	RAL			; the carry flag
	XCHG
	DAD	H
	JNC	DVID32b
	INX	D
	ANA	A
DVID32b	XCHG			; Shift done
	RAR			; Restore first carry
	PUSH	PSW		; Save counter
	JNC	DVID32c		; If carry do this subtract
	CALL	DSUB
	JMP	DVID32d
DVID32c	CALL	DSUB		; Else try HL = HL - BC
				; HL = partial remainder
	JNC	DVID32d		; Underflow?
	DAD	B		; Underflow, restore
	DCX	D
DVID32d	INX	D		; Increment quotient
	POP	PSW		; Get back counter
	DCR	A		; Decrement loop count
	JNZ	DVID32a		; If not done continue
	RET			; DONE!, Split
;	
; Subtract BC from HL
;
DSUB	MOV	A,L		; Start with LSB
	SUB	C		; Do the subtract
	MOV	L,A		; Put result back
	MOV	A,H		; Now do MSB
	SBB	B		; Do the subtract
	MOV	H,A		; Put result back
	RET			; Split
;
OVRFLOW	LXI	H,-1		; Set HL and DE to maximum
	LXI	D,-1		; Values
	STC			; Turn on error flag
	RET			; Split
;
; The error tabel is organized as follows:
; The poly code we will report is in the DB field.
; The code we got and its meaning are in the comment field
;
ErrTbl	DB	0	; 00 Good complete
	DB	0	; 01 No index signal
	DB	0	; 02 No seek complete
	DB	40H	; 03 Illegal
	DB	40H	; 04 Illegal
	DB	40H	; 05 Illegal
	DB	40H	; 06 Illegal
	DB	40H	; 07 Illegal
	DB	40H	; 08 Illegal
	DB	40H	; 09 Illegal
	DB	40H	; 0A Illegal
	DB	40H	; 0B Illegal
	DB	40H	; 0C Illegal
	DB	40H	; 0D Illegal
	DB	40H	; 0E Illegal
	DB	40H	; 0F Illegal
	DB	40H	; 10 Late data transfer
	DB	3	; 11 CRC error
	DB	0BH	; 12 Seek Fault, drive reported
	DB	09H	; 13 Drive fault
	DB	40H	; 14 Defect map error
	DB	0BH	; 15 Seek fault, cylinder mismatch
	DB	0	; 16 Initialization complete
	DB	80H	; 17 PDA Stack error
	DB	80H	; 18 PDA Hardware trap
	DB	40H	; 19 Illegal
	DB	40H	; 1A Illegal
	DB	40H	; 1B Illegal
	DB	40H	; 1C Illegal
	DB	40H	; 1D Illegal
	DB	40H	; 1E Illegal
	DB	40H	; 1F Illegal
	DB	6	; 20 Drive not ready
	DB	5	; 21 Write protected
	DB	9	; 22 Drive not present
	DB	40H	; 23 Illegal
	DB	40H	; 24 Illegal
	DB	40H	; 25 Illegal
	DB	40H	; 26 Illegal
	DB	40H	; 27 Illegal
	DB	40H	; 28 Illegal
	DB	40H	; 29 Illegal
	DB	40H	; 2A Illegal
	DB	40H	; 2B Illegal
	DB	40H	; 2C Illegal
	DB	40H	; 2D Illegal
	DB	40H	; 2E Illegal
	DB	40H	; 2F Illegal
	DB	2	; 30 Sector not found
	DB	1	; 31 SMART command reject
	DB	6	; 32 Drive busy command timeout
	DB	40H	; 33 Data transfer timeout
	DB	1	; 34 Illegal cylender/head
	DB	1	; 35 Invalid drive number
	DB	1	; 36 Illegal sector number
	DB	1	; 37 DCB busy
	DB	1	; 38 Command double write
	DB	1	; 39 Command reject
	DB	1	; 3A Multi sector operation error
	DB	40H	; 3B Illegal
	DB	40H	; 3C Illegal
	DB	40H	; 3D Illegal
	DB	40H	; 3E Illegal
	DB	40H	; 3F Illegal
	;
	; End of table
	;
;
; Data storage
;
DRIVE	DS	1		; Drive #
NSECTS	DS	1		; # of sectors
HDISKAD	DS	2		; Disk address pointer
HMEMADD	DS	2		; Memory address
HCOMAND	DS	1		; Command
SftCnt	DS	2		; How many soft errors
DTETrys	DS	1		; Data Timeout Retries
PARMS	DS	4*7		; Where we keep parameters
				;  for each drive.  Format:
				;   Do I exist?	= 1 byte
				;   # sects/trk	= 1 byte
				;   # sects/cyl	= 2 bytes
				;   # sects/drive = 3 bytes
HardAdr	DS	3		; disk address buffer

	PAGE
