	PAGE
;========================================================
;|							|
;|			 Efun.OV			|
;|		  Editor Function Overlay		|
;|							|
;|	Version 2.0.5		     Glenn A. McComb	|
;|	5 May 1980					|
;|							|
;========================================================
;
;	History
;	~~~~~~~
;	10/07/80 LSA	Added Reenter & PrtMsg Services
;	10/14/80 LSA	Reorganization
;	10/27/80 LSA	Added two Abort messages
;	03/20/81 DKM	Added language options
;	05/15/81 DKM	Add Danish option
;
	REFS	EditRefs
	REF

	ORG	OVRLY
	IDNT	$,$
Head	DB	'Efun'

Start	;
	IF	DEBUG
	DB	0FFH		; Insert a RST 7.
	ENDIF
	;
	CPI	NumFcn
	RNC			; ret/ bad function code
	PUSH	H		; don't alter BC, DE, or HL
	PUSH	D
	ADD	A		; x 2
	MOV	E,A
	MVI	D,0		; offset in DE
	LXI	H,jtab
	DAD	D		; --> slot in jump table
	MOV	E,M
	INX	H
	MOV	D,M
	XCHG			; routine address in HL
	POP	D
	XTHL			; restore DE, HL
	EI			; reenable interrupts
jRet	RET			; and jump to routine
	PAGE

;  This is the dispatch table for the Efun services.
;  Caveat Emptor...

jtab	DW	jRet		; 00  no operation
	DW	Startup		; 01  startup editor
	DW	Reenter 	; 02  Reenter checks
	DW	Anykey		; 03  put msg, wait for key
	DW	PrtMsg		; 04  Print message
	DW	Enters		; 05  enter string
	DW	Claim		; 06  claim buffer space
	DW	GnInit		; 07  initialize gnomus
	DW	Gtext		; 08  fill half of ram
	DW	Ptext		; 09  flush half of ram
	DW	FillIn		; 10  fill all of ram
	DW	FlshOut		; 11  flush all of ram
	DW	Txtin		; 12  get N characters
	DW	Txtout		; 13  put N characters
NumFcn	EQU	($-jtab)/2
	PAGE
;========================================================
;|		    Startup and Reenter 		|
;========================================================
;
;	First validate checksum of resident and overlay.
;
Startup LXI	H,USER		; from here
	LXI	D,TAIL		; to here
	CALL	Checkem 	; go to it.
	RC
	LXI	H,Head		; then from here
	LXI	D,Eftail	; to here
	CALL	Checkem 	; check that, too...
	RC
	;
	;	initialize the video screen and the bubble...
	; ...if we have a video screen, that is!
	;
	LHLD	SCRHM-1
	MOV	A,H
	CMP	L		; if SCRHM and SCEND are equal
	LXI	D,208H		; you can't use the editor!
	JZ	Ederr		; so we'll die...
	;
	MVI	L,0		; HL = video board address
	SHLD	rlscn		; call this 'real screen' addr
	LXI	H,vbuf		; --> video buffer
	SHLD	fkscn		; stash that.
	LXI	H,textb 	; --> start of text area
	SHLD	ustrt		; save as start of user ram
	SHLD	bblb		; and create bubble start.
	LHLD	MEMTOP		; get end of memory
	SHLD	utop		; save as end of user ram
	SHLD	bble		; and create bubble end.
	;
	;	input/output files already open, so finish up 
	;  the initialization process.
	;
	XRA	A
	STA	inflgs		; clear inhibit flags
	MVI	A,stx
	CALL	Edchr		; put stx in text buffer
	MVI	A,etx
	CALL	Edchr		; put etx in buffer
	CALL	Mvbl		; put bubble in between stx/etx
	CALL	Upint		; init display driver
	LHLD	rlscn		; put brick in vbuf for flash
	XCHG
	LHLD	POS
	CALL	Dsub
	SHLD	ptsav		; set the pointer
	XCHG
	LHLD	fkscn		; -> video buffer
	DAD	D
	MVI	M,del+80H  	; put brick @ spot in vbuf
	MVI	A,1
	STA	WMsgFlg		; default to give warning msgs
	;
	CALL	Gnomus		; hookup Gnomus
	;
	LDA	eoflg
	ORA	A
	CZ	Append		; grab some text if not eof
	CALL	Edtb		; go to the top of the buffer
	CALL	Anykey		; wait for user response...
	ORA	A		; make sure we clear the CARRY
	RET			; and return to caller
;
;  Checkem is only called as part of the startup sequence to
;  verify checksums of resident and overlay portion of the
;  editor. This catches folks that bugger the editor...
;
Checkem MVI	B,0		; checksum goes here.
	DCX	H
Check1	INX	H
	MOV	A,M
	ADD	B
	MOV	B,A		; add in a byte
	MOV	A,H
	CMP	D		; right page yet?
	JNZ	Check1		; keep going
	MOV	A,L
	CMP	E		; right byte?
	JNZ	Check1		; jmp/ no, keep going
	;
	;  Area added up now. See if we return, or die!
	;
	IF	DEBUG
	ORA	A
	RET			; just return NO CARRY.
	ELSE
	INR	B		; should go to zero
	RZ			; ret/ good checksum
	ENDIF
	;
	LXI	D,0DEADH	; else go
	JMP	Ederr		; DIE!
;
;****************
;
;	Perform the REEENTER checks before leaping back 
;  into the editor.  For the single-user system just return.
;  For the 2-User system we need to "remung" the output drive 
;  and check our disk addresses we've kept in memory.
;
Reenter ;
	IF	USERS=2
	LDA	odev
	MOV	C,A
	sett	mung		; Try to regain mung access
	JC	Reent1		; jmp/ didn't get it
	ENDIF
	;
	LXI	H,ofde
	MVI	A,2		; Look.
	gfid
	JNC	Reent1		; jmp/ output file appeared
	LXI	H,300H
	CALL	Cmpdh
	JNZ	Reent1		; jmp/ not 300 error
	;
	LHLD	NFA
	XCHG			; Now check to see if NFA
	LHLD	offda		; has moved by comparing it
	CALL	Cmpdh		; against offda.
	RZ			; ret/ NFA's still the same
	;
Reent1	LXI	H,ReentMsg
	CALL	Msg		; give message
	STC			; and return error.
	RET
	PAGE
;========================================================
;|			Anykey				|
;========================================================
;
;  We are putting this routine here because it helps to
;  keep the Efun overlay resident without disturbing the
;  editor i/o code interaction with Gfid (believe me!).
;
Anykey	LXI	H,Anymsg
	CALL	Msg		; pop the message
	CALL	Flush
	EI
	CALL	Upclr		; force screen update
	JMP	WH0
	PAGE
;========================================================
;|		    Print message			|
;========================================================
;
;	Print message service.  Print the message given by
;  the code in HL.  Changes only A,HL.
;
PrtMsg	PUSH	B		; Preserve these
	PUSH	D
	CALL	PrtMsg1 	; print message and
	CALL	Upclr		; force screen update.
	POP	D
	POP	B
	RET
PrtMsg1 XCHG
	LXI	H,LastMsg
	CALL	Cmpdh
	RNC			; Ret/ invalid code
	LXI	H,MsgTab
	DAD	D
	DAD	D		; Point into table
	MOV	E,M
	INX	H
	MOV	D,M
	XCHG			; Get pointer to text
	CALL	Msg		; and print thru WH1
	RET
	PAGE
;========================================================
;|		    Enter string service		|
;========================================================
;
; Enter will allow any processor to accept a string from
; the user. The only edit function available is the delete
; key. A single escape terminates the enter function. For now,
; all control keys will echo as greek characters.
;
; Entry:  HL:  buffer		Exit:  C set if no disp room
;	  DE:  Max # chars	       HL: last char in buffer
;
Enters	SHLD	enbst	; save buffer address
	XCHG
	DAD	D
	DCX	H
	SHLD	enbnd	; save buff+maxlen as buffer limit
	;
	CALL	bblsiz
	CMC
	RC		; ret/ no bubble, we no enter...
	XCHG
	SHLD	bbls	; save the original bblb
	MVI	M,mkr	; put up the marker
	INX	H	; bump bblb
	SHLD	bblb
	LHLD	enbst	; --> enter buffer
	SHLD	bptr	; save as index
	;
Entlp	CALL	Update	; update screen
	CALL	WH0	; get a char
	CPI	esc
	JZ	Entesc	; jmp/ escape... weez done!
	CPI	del
	JZ	Entdl	; jmp/ del... delete a char
	;
	ANI	7FH	; strip marker bit
	MOV	B,A	; save char
	CALL	bblsiz	; How much room do we have?
	JNC	Entlp	; jmp/ no room - spin.
	XCHG
	MOV	M,B	; put char in bubble
	INX	H
	PUSH	H	; save new bblb
	LHLD	enbnd	; --> enter buffer end addr
	XCHG
	LHLD	bptr
	CALL	Cmpdh	; check for room in buffer
	POP	D	; clean stack/ new bblb in DE
	JZ	Entlp	; jmp/ no room - spin.
	;
	MOV	M,B	; stuff char in enter buffer
	INX	H
	SHLD	bptr	; inc ptr
	XCHG		; get new bblb
	SHLD	bblb	; set it
	JMP	Entlp	; and spin till esc char.
	;
Entdl	LHLD	bblb	; compare bblb & bbls+1
	XCHG
	LHLD	bbls
	INX	H
	CALL	Cmpdh
	JZ	Entlp	; jmp/ no string - why delete?
	XCHG
	DCX	H
	SHLD	bblb	; else delete char
	LHLD	bptr
	DCX	H
	SHLD	bptr	; dink buffer pointer
	JMP	Entlp	; and scram
	;
Entesc	LHLD	bbls	; get original bblb
	SHLD	bblb	; restore it
	LHLD	bptr
	DCX	H	; HL --> last char in buffer
	ANA	A	; clear carry
	RET		; and split.
	PAGE
;========================================================
;|		    Claim Buffer Space			|
;========================================================
;
;   Enter:  HL - # of contiguous bytes to claim
;	     A - <> 0 to init claimed space to zero
;		  = 0 to not init claimed space
;
;   Exit:    C set if no room -------------------
;	      else ----				|
;		      | 		    BC - length
;	    DE - first address		    DE - length
;	    HL - last address+1 	    HL - room-DE
;
Claim	PUSH	PSW	; save the init flag
	PUSH	H	; stack: length/
	LHLD	bblb	; how much space available?
	XCHG
	LHLD	bble
	CALL	Dsub
	INX	H	; # of bytes free in HL
	POP	D	; length desired (length) in DE
	PUSH	D	; stack: length/
	CALL	Dsub	; do we have enough room? (HL-DE)
	POP	B	; clean stack - length in BC
	JC	Cxit1	; sorry bub, no pub.
	;
	; ok. we can make the claim. first thing is to update
	; bfptr, the display start pointer, so things don't
	; go funnie.
	;
	LHLD	bfptr
	DAD	B
	SHLD	bfptr	; bfptr = bfptr+len
	;
	LHLD	bblb
	PUSH	H	; stack: old bblb/
	DAD	B	; compute new bblb
	SHLD	bblb	; bblb = bblb + length
	PUSH	H	; stack: bblb+length/ old bblb
	LHLD	ustrt	; get start of buffer address
	PUSH	H	; stack: buf start/bblb+length/old bblb
	DAD	B	; add length
	SHLD	ustrt	; new start of buffer address
	POP	H	; HL has old start address
	POP	B	; BC has destination address+1
	POP	D	; DE has from address+1
	;
	;   now we move until (from addr) = (old ustrt)
	;
Clp	CALL	Cmpdh	; are we done?
	JZ	Cdn	; yup/ scram.
	DCX	B	; else dink ptrs
	DCX	D
	LDAX	D
	STAX	B	; copy over
	JMP	Clp	; spin.
	;
Cdn	XCHG		; first addr of claimed space in DE
	LHLD	ustrt	; last addr + 1 in HL
	POP	PSW
	ORA	A
	JZ	Cxit	; if init flag zero, we split.
	;
	PUSH	D	; else save regs
	PUSH	H	
	XCHG		; and set up to do init
Cinit	MVI	M,0
	INX	H
	CALL	Cmpdh
	JNZ	Cinit	; zap till done.
	;
	POP	H	; last addr+1 of claimed space in HL
	POP	D	; first addr of claimed space in DE
	PUSH	PSW	; this saves a jump.
	;
Cxit1	POP	PSW	; clean stack.
Cxit	RET		; and split.
	PAGE
;========================================================
;|			  Gnomus!			|
;========================================================
;
GnInit	MVI	A,1
	STA	GnVFlag	; set validation flag for ^[^h hack
	LXI	H,256	; # of bytes to claim (RAT size)
	MVI	A,1	; flag to zero claimed space
	CALL	Claim	; get space for RAT (reserved addr tab)
	JC	Gnono	; oops - no room for RAT
	SHLD	Gnose	; end of RAT for start of strings
	SHLD	Gtail	; mark string space empty
	SHLD	Gtop	; and no space available (Gtail=Gtop)
	XCHG
	SHLD	Grat	; set address of RAT
	XRA	A
	STA	Gnlvl	; and reset recursion stack
	;
	;  Now the dirty stuff... store reserved ESC sequences!
	;  (this is done so we know which keys are already used
	;  for commands).
	;
	LXI	H,esctb	; where we start
Gshufle MOV	A,M	; pickup char index
	INX	H	; point past char
	CPI	0FFH
	RZ		; ret/ end of table
	RLC		; index times two...
	MOV	C,A
	MVI	B,0	; ...in BC
	XCHG		; save HL
	LHLD	Grat	; pickup address of RAT
	DAD	B	; compute slot address
	XCHG		; DE = slot addr, HL = table ptr
	MOV	A,M	; get routine address
	STAX	D
	INX	H
	INX	D
	MOV	A,M
	STAX	D
	INX	H	; fill in slot
	JMP	Gshufle	; and go shuffle some more...
	PAGE
;********************************************************
;*							*
;*	 Gtext, Ptext, FillIn & FlshOut Primitives	*
;*							*
;********************************************************
;
;  Gtext is the 'Append' service routine. It determines how
;  many characters are to be operated on, and calls TxtIn to
;  perform the actual I/O. Then it does cleanup of the text
;  (removing nulls and markers) before returning.
;
Gtext	CALL	Edmeb	; end of bubble, please
	;
	LHLD	bblb	; compute # of chars to read
	XCHG		; DE = bblb
	LHLD	bble	; HL = bble
	CALL	Dsub	; number of free bytes
	MOV	A,H
	ORA	L
	RZ		; ret/ no room in bubble
	MOV	A,H
	RAR
	MOV	B,A
	MOV	A,L
	RAR
	MOV	C,A
	INX	B	; BC = free bytes/2+1
	;
	JMP	GetInp	; well, go do the read and return.
;
;****************
;
;  Ptext is the 'Output' service routine. It determines how
;  many characters are to be operated on, passes over the
;  text removing markers, and then calls TxtOut to perform
;  the actual I/O.
;
Ptext	CALL	Edtb	; top of bubble, please
	LHLD	bble
	XCHG
	LHLD	utop
	DCX	H	; point to char before etx marker
	CALL	Dsub	; compute length of text
	MOV	A,H
	ORA	L
	RZ		; ret/ no text in buff
	MOV	A,H
	RAR
	MOV	B,A
	MOV	A,L
	RAR
	MOV	C,A
	INX	B	; BC = length of text/2+1
	;
	JMP	PutOut	; jmp/ well, do the write and return
;
;****************
;
;  FillIn is an Exit service routine. It fills the entire
;  free RAM with the input file (as opposed to Gtext which
;  inputs only half of free RAM).
;
;	CARRY is returned on an EOF condition.
;
FillIn	CALL	Edmeb	; end of bubble, please
	;
	LHLD	bblb	; compute # of chars to read
	XCHG		; DE = bblb
	LHLD	bble	; HL = bble
	CALL	Dsub	; number of free bytes
	MOV	A,H
	ORA	L
	RZ		; ret/ no room in bubble
	;
	MOV	B,H
	MOV	C,L	; BC = # of bytes to read
	;
	JMP	GetInp	; well, go do the read and return.
;
;****************
;
;  FlshOut is an Exit service routine. It flushes the entire
;  free RAM to the output file (as opposed to Ptext which
;  outputs only half of free RAM).
;
FlshOut CALL	Edtb	; top of bubble, please
	LHLD	bble
	XCHG
	LHLD	utop
	DCX	H	; point to char before etx marker
	CALL	Dsub	; compute length of text
	MOV	A,H
	ORA	L
	RZ		; ret/ no room in bubble
	;
	MOV	B,H
	MOV	C,L	; BC = # of bytes to output
	;
	JMP	PutOut	; jmp/ well, do the write and return
	PAGE
;********************************************************
;*							*
;*		  GetInp & PutOut Services		*
;*							*
;********************************************************
;
; Get BC chars into memory starting at the bubble position,
; then scan over the text deleting markers and nulls. Finally
; update the bubble position and return.
;
GetInp	LHLD	bblb
	PUSH	H	; save old bblb for later
	CALL	Txtin	; read BC chars into (bblb)
	XCHG		; mem ptr in HL
	;
	; now read char by char until eof or carriage return.
	;
	JC	Gtxt2
Gtxt1	PUSH	D
	XCHG
	LHLD	bble
	CALL	Cmpdh
	XCHG
	POP	D
	JZ	Gtxt2	; bubble full, quit.
	RNC		; this should be impossible.
	;
	DCX	H
	MOV	A,M	; get last char read
	INX	H
	CPI	cr	; is it a cr?
	JZ	Gtxt2	; jmp/ yassuh
	LXI	B,1	; else setup to read one char
	CALL	Txtin	; do read
	XCHG		; mem ptr in HL
	JNC	Gtxt1	; and spin if no eof
	;
	; now we scan from the old bblb to the new bblb,
	; eating bad chars (nulls & markers). The final
	; value of BC is the new bblb pointer.
	;
Gtxt2	POP	D	; old bblb in DE (from-ptr)
	PUSH	PSW	; save flags
	MOV	B,H
	MOV	C,L	; BC is new bblb
	MOV	H,D
	MOV	L,E	; HL is to-ptr
Gtlp	MOV	A,D
	CMP	B
	JNZ	Gtlp1	; jmp/ not done yet
	MOV	A,E
	CMP	C
	JZ	Gtlpdn	; jmp/ done
Gtlp1	LDAX	D
	INX	D	; get a char
	ANI	7FH	; strip 80 bit
	MOV	M,A
	INX	H	; stash it
	JNZ	Gtlp	; jmp/ not a null
	DCX	H	; dink B to trash char
	JMP	Gtlp
Gtlpdn	SHLD	bblb	; update bblb
	POP	PSW
	RET
;
;****************
;
; Scan over BC chars starting after the bubble deleting
; markers, then put BC chars to disk starting at the end of
; the bubble, Finally update the bubble position and return.
;
PutOut	LHLD	bble
	INX	H	; --> char @ (bble+1)
	;
	; first we pass over text and convert all
	; markers to nulls.
	;
	PUSH	H
	PUSH	B	; we need these later
Ptxt1	MOV	A,B
	ORA	C
	JZ	Ptxt3	; jmp/ all done
	MOV	A,M
	ORA	A
	JP	Ptxt2	; jmp/ not marker
	MVI	M,0	; zap!
Ptxt2	INX	H
	DCX	B
	JMP	Ptxt1	; bump, dink, and spin.
Ptxt3	POP	B
	POP	H	; restore these
	CALL	Txtout	; output BC chars from (bble)+1
	XCHG		; HL --> next char to output
	;
	; now output char by char until etx or cr
	;
Ptxt4	MOV	A,M	; peek ahead
	CPI	etx
	DCX	H	; dink for bble
	JZ	Ptxt5	; jmp/ next char is etx
	MOV	A,M
	CPI	cr
	JZ	Ptxt5	; jmp/ last char read was cr
	INX	H
	LXI	B,1
	PUSH	H	; --> char to output
	MOV	A,M
	ORA	A	; is it a marker?
	CP	Txtout	; output 1 char if not
	POP	H
	INX	H	; HL --> next char to output
	JMP	Ptxt4	; and spin
	;
Ptxt5	SHLD	bble	; set bble
	ORI	1
	RET		; and split non-zero
	PAGE
;********************************************************
;*							*
;*		   File System I/O drivers		*
;*							*
;********************************************************
;
;  Txtin - read BC chars into memory
;	   starting at (HL).
;
;    entry:  HL --> buffer area
;	     BC: # of bytes
;
;    exit:   DE --> next byte
;	     BC: # of bytes remaining to read
;
;    flags:  C set if eof encountered during read
;	     Z set if operation completed (BC=0)
;
Txtin	LDA	inflg	; file open flag
	ORA	A
	JZ	Gteof	; jmp/ give eof
	;
	PUSH	H	; Save pointer
	DAD	B	; This is last address to be stored in
	XCHG		; Put in DE
	LHLD	utop	; Load dymanic top of RAM
	CALL	Cmpdh	; Check the two
	POP	D	; DE = mem ptr
	JC	Txtin1	; If OK continue processing
	ORI	80H	; Unset zero flag
	RET		; Split reporting error
	;
Txtin1	LHLD	iptr	; get buffer ptr
	;
Inp1	INR	L
	JZ	Inp2	; jmp/ buffer empty
	;
Inp1a	MOV	A,B
	ORA	C
	RZ		; ret/ all done
	;
	SHLD	iptr	; update buffer ptr
	MOV	A,M
	STAX	D	; install char
	INX	D
	DCX	B	; dink byte counter
	JMP	Inp1	; and spin.
	;
Gteof	MVI	A,1
	STA	eoflg
	STC
	RET		; set eoflg, carry, and split.
	;
Inp2	MOV	A,B
	ORA	C	; do we want any more chars?
	RZ		; my, my, what a coincidence!
	;
	LHLD	inbs	; get # sectors remaining
	MOV	A,H
	ORA	L
	JZ	Gteof	; jmp/ end of file
	;
	PUSH	D	; save mem ptr
	MOV	E,B	; save B
	MVI	D,0	; DE = B
	CALL	Cmpdh	; compare (DE-HL)
	MOV	B,D	; B = 0
	JC	Inp3	; jmp/ B < inbs
	XCHG
	DCX	D	; DE = inbs-1
	LXI	B,100H	; BC = 100H to read last sector
	;
Inp3	POP	H	; get mem ptr in HL
	PUSH	B	; save BC
	PUSH	H	; and memptr
	;
	MOV	A,D
	ORA	E
	JZ	Inp4	; jmp/ DE = 0 (no sectors)
	LHLD	ifda	; get disk address
	LDA	ifde	; disk #
	ANI	DEVMASK
	MOV	C,A	; in C
	MVI	B,1	; read!
	MOV	A,E	; # sectors
	POP	D	; mem addr
	PUSH	D
	PUSH	PSW	; stash nsect
	CALL	Dio	; do io!
	JC	Hrderr	; jmp/ AAUGH!!!
	;
	POP	PSW	; A = nsect
	POP	B	; BC = mem addr
	MOV	E,A
	MVI	D,0	; DE = nsect
	ADD	B
	MOV	B,A	; BC = mem addr+(256*nsect)
	PUSH	B	; stash
	LHLD	ifda	; get disk addr
	DAD	D
	SHLD	ifda	; ifda = ifda + nsect
	LHLD	inbs
	CALL	Dsub
	SHLD	inbs	; inbs = inbs - nsect
	;
Inp4	LHLD	ifda	; disk addr
	LDA	ifde	; disk #
	ANI	DEVMASK
	MOV	C,A	; in C
	MVI	B,1	; read!
	MOV	A,B	; 1 sector
	LXI	D,dibuf
	CALL	Dio	; do io!
	JC	Hrderr	; jmp/ AAUGH!!!
	;
	LHLD	inbs
	DCX	H
	SHLD	inbs	; inbs = inbs - 1
	LHLD	ifda
	INX	H
	SHLD	ifda	; ifda = ifda + 1
	LXI	H,dibuf
	SHLD	iptr	; init buffer ptr
	;
	POP	D	; DE = mem ptr
	POP	B	; BC = chars remaining
	JMP	Inp1a	; finish up last bytes
;
;****************
;
;	--------   Output text	--------
;
;  Txtout - write BC chars into output file from
;	    memory starting at (HL).
;
;    entry:  HL --> buffer area
;	     BC: # of bytes
;
;    exit:   DE --> next byte
;	     BC: # of bytes remaining to read
;
;    flags:  C set if unrecoverable error encountered
;	     during write (error code in DE)
;
Txtout	LDA	otflg	; get output file opened flag
	ORA	A
	LXI	D,-1	; error code for system error
	STC		; set carry
	RZ		; ret/ no output file open
	;
	XCHG		; DE = mem ptr
Out1	LHLD	optr	; get buffer pointer
	;
Out1a	MOV	A,B
	ORA	C
	RZ		; ret/ all done
	;
	LDAX	D	; get char
	MOV	M,A	; stash in buffer
	INX	D	; bump mem ptr
	DCX	B	; dink char counter
	INR	L	; bump buff ptr
	SHLD	optr	; stash buffer ptr
	JNZ	Out1a	; jmp/ buffer not full
	;
Out2	PUSH	D	; save mem ptr
	LHLD	smax	; and maximum sector count
	XCHG
	LHLD	onda	; get current disk address
	CALL	Cmpdh	; compare (smax-onda)
	POP	D	; restore mem ptr
	CC	Otopn	; call/ open new output file
	JC	Out2	; and try again	
	;
	PUSH	D	; save mem ptr
	PUSH	B	; save character count
	;
	LHLD	onda	; output disk addr
	LDA	ofde
	ANI	DEVMASK	; get unit #
	MOV	C,A	; in C
	LXI	D,dobuf	; buffer to flush
	MVI	A,1	; 1 sector, please
	MVI	B,0	; write!
	CALL	Dio
	JC	Hrderr	; jmp/ AAUGH!
	;
	LHLD	onda
	INX	H
	SHLD	onda	; bump output disk addr
	LHLD	onbs
	INX	H
	SHLD	onbs	; bump sector count, too
	LXI	H,dobuf
	SHLD	optr	; reset buffer ptr
	;
	POP	B	; restore char count
	POP	D	; and mem ptr
	;
	MOV	A,B
	ORA	A
	JZ	Out1	; jmp/ < 256 chars left
	;
Out3	PUSH	B	; save char count
	PUSH	D	; then mem ptr
	MOV	E,B
	MVI	D,0	; DE = # of sectors desired (nsect)
	PUSH	D	; stash nsect
	LHLD	onda	; get current sector #
	DAD	D
	DCX	H
	XCHG		; DE = (onda+nsect-1)
	LHLD	smax	
	XCHG		; DE = smax, HL = (onda+nsect-1)
	CALL	Cmpdh	; compare (DE-HL)
	JNC	Out4	; jmp/ we can do entire write
	;
	POP	H	; trash old nsect
	LHLD	onda	; get current sector number again
	XCHG		; DE = sector, HL = smax
	CALL	Dsub	; HL = smax-sector		
	INX	H	; bump
	XCHG		; DE = (smax-onda)+1
	PUSH	D	; stash this nsect
;
; right now the stack looks like:
;
;	(top)	# sectors to write
;		mem ptr
;		char count
;
;
Out4	POP	D	; DE = # of sectors to write
	XRA	A
	ORA	D
	JNZ	Otimp	; jmp/ nsect > 255 - somebody goofed
	ORA	E
	JNZ	Out5	; jmp/ nsect <> 0 - ok
;
; oh wow. somebody goofed. either the disk we want to
; write on is exactly full, or else somebody pissed on
; the directory and we are pointed way past the end of
; the damn disk. All we do is bail out and keep asking
; for a new output file until we get one we like.
;
Otimp	POP	D	; mem ptr
	POP	B	; char count
	CALL	Otopn	; open a new output file
	JMP	Out3	; and try again.
	;
Out5	LHLD	onda	; output disk addr
	LDA	ofde
	ANI	DEVMASK
	MOV	C,A	; unit # in C
	MOV	A,E	; # of sectors in A
	POP	D	; mem ptr in DE
	PUSH	D	; keep on stack
	MVI	B,0	; write!
	PUSH	PSW	; save sector count
	CALL	Dio	; do it!
	JC	Err	; jmp/ AAUGH!!!
	POP	PSW	; get nsect
	MOV	E,A
	MVI	D,0	; in DE
	LHLD	onda
	DAD	D
	SHLD	onda	; update output disk addr
	LHLD	onbs
	DAD	D
	SHLD	onbs	; and sector count
	;
	POP	H	; mem ptr in HL
	POP	B	; char count in BC
	ADD	H
	MOV	H,A	; HL = mem ptr+(256*nsect)
	MOV	A,B
	SUB	E	
	MOV	B,A	; BC = BC-(256*nsect)
	XCHG		; DE = mem ptr
	JZ	Out1	; jmp/ completed write - go finish up
	;
	CALL	Otopn	; well, go open another file
	JMP	Out3	; and try another block write
	PAGE
;
;	***** UTILITIES *****
;
Cmpdh	MOV	A,D
	CMP	H
	RNZ
	MOV	A,E
	CMP	L
	RET
;
Dsub	MOV	A,L
	SUB	E
	MOV	L,A
	MOV	A,H
	SBB	D
	MOV	H,A
	RET
;
	IF	USERS=1
OVGFID	CALL	Ovrto		; PVEC isn't pointing into 
	DB	'Gfid'		; overlay area.
	RET
	ENDIF
	PAGE
;
;	***** CONSTANT AREA *****
;
;-----Tables-----
;
MsgTab	DW	WrnMsg0 	; 0
	DW	WrnMsg1 	; 1
	DW	PrtMsg2 	; 2
	DW	WrnMsg3 	; 3
	DW	PrtMsg4 	; 4
	DW	PrtMsg5 	; 5
	DW	PrtMsg6 	; 6
LastMsg EQU	($-MsgTab)/2
;
;-----Messages-----
;

	IF	GERMAN
Anymsg	 db	cr,'Bitte eine Taste druecken...',cr,0
ReentMsg db	'PROGRAM hier nicht erlaubt!',cr,0
;
WrnMsg0  db	cr,'WARNUNG: Das ganze File passt nicht'
	 db	' auf die',cr,tab
	 db	' Ausgabe-Disk.  (Control-Y fuer Abort)',cr,0

WrnMsg1  db cr,cr,tab,'WARNUNG:  Das derzeitige Eingabe-File'
	 db ' wurde nicht',cr,tab,tab,' ganz in den Speicher'
	 db ' gelesen! (Control-Y fuer Abort)',cr,0

PrtMsg2  db	cr,cr,cr,tab,tab,'Das Eingabe-File ist leer!'
	 db	cr,cr,0

WrnMsg3	 db	cr,cr,tab,'WARNUNG: Sie koennen nicht weiter,'
	 db	' haben aber noch',cr,tab,tab,'mehr Text von'
	 db	' der Disk zu lesen.  Bitte',cr,tab,tab
	 db	'Control-A druecken um noch weiter'
	 db	' einzulesen.',cr,0

PrtMsg4	 db cr,'Bei Fragezeichen bitte eine der folgenden'
	 db ' Angaben machen:',cr
	 db tab,'Ctrl-Y',tab,tab,'Ende Suchen und Austausch',cr
	 db tab,'ESC',tab,tab,'ueberspringt und verarbeitet'
	 db ' weiter',cr
	 db tab,'RETURN',tab,tab,'vollendet das restliche'
	 db ' Austauschen',cr
	 db tab,'<Taste>',tab,tab,'aendert und verarbeitet'
	 db ' weiter',cr,0

PrtMsg5  db	cr,'RETURN fuer Abort.'
	 db ' (andere Taste zur Weiterverarbeitung)...',0

PrtMsg6  db	cr,cr,'Edit abgebrochen---KEIN UPDATE',cr
	 db	'(Eingabe PROGRAM zur Weiterverarbeitung)',cr,0

	ELSEIF	DANISH		;Danish option
Anymsg	 db	cr,'Tryk RETURN...',cr,0
ReentMsg db	'REENTER ikke tilladt..',cr,0
;
WrnMsg0  db	cr,'BEM[RK: Input fil kan ikke v{re p} output'
	 db	cr,tab,'disk. (CTRL-Y)',cr,0

WrnMsg1  db	cr,cr,tab,'BEM[RK:  Input fil blev ikke helt'
	 db	' indl{st',cr,tab,tab,'  i hukommelsen.'
	 db	' (CTRL-Y)',cr,0

PrtMsg2  db	cr,cr,cr,tab,tab,'Input fil er tom..'
	 db	cr,cr,0

WrnMsg3  db	cr,cr,tab,'BEM[RK:  Ikke mere tekst i'
	 db	' hukommelsen, men',cr,tab,tab,'  mere'
	 db	' tekst p} disk. Brug CTRL-A for at',cr,tab
	 db	tab,'  indl{se mere tekst.',cr,0 

PrtMsg4  db	cr,'Ved ? tast en af f|lgende:'
	 db	cr,tab,'    Ctrl-Y',tab
	 db	'standser global s|ge/{ndre,',cr,tab,'    ESC'
	 db	tab,tab,'{ndrer ikke denne,',cr
	 db	tab,'    RETURN	{ndrer resten,',cr
	 db	tab,'    <~>   {ndrer denne.'
	 db	cr,0

PrtMsg5  db	cr,'Tryk RETURN for at slutte..'
	 db	' (Tryk ~ for at komme videre)...',0

PrtMsg6  db	cr,cr,'EDIT afbrudt - FIL IKKE [NDRET',cr
	 db	'(Tast REENTER for at forts{atte)',cr,0


	ELSE		;default to English
Anymsg	 db	cr,'Hit any key to continue...',cr,0
ReentMsg db	'REENTER not allowed now!',cr,0
;
WrnMsg0  db	cr,'WARNING: The entire input file will'
	 db	' not fit on the',cr,tab
	 db	' output drive.  (control-Y to abort)',cr,0

WrnMsg1  db	cr,cr,tab,'WARNING:  The current input file'
	 db	' was not entirely',cr,tab,tab,'  read into'
	 db	' memory! (control-Y to abort)',cr,0

PrtMsg2  db	cr,cr,cr,tab,tab,'The current input'
	 db	' file is empty!',cr,cr,0

WrnMsg3  db	cr,cr,tab,'WARNING:  You can''t move any'
	 db	' further, but you still',cr,tab,tab,'  have'
	 db	' more text to read from the disk. Use',cr,tab
	 db	tab,'  control-A to read some of it into'
	 db	' memory.',cr,0 

PrtMsg4  db	cr,'When the question mark appears type one'
	 db	' of the following: ',cr,tab,'    Ctrl-Y',tab
	 db	'quits global search/replace,',cr,tab,'    ESC'
	 db	tab,tab,'skips this change and continues,',cr
	 db	tab,'    RETURN	makes remaining changes,',cr
	 db	tab,'    <any key>   does this change and'
	 db	' continues.',cr,0

PrtMsg5  db	cr,'Hit RETURN to abort'
	 db	' (any other key to continue)...',0

PrtMsg6  db	cr,cr,'Edit aborted---FILE NOT UPDATED',cr
	 db	'(Type REENTER to resume editing)',cr,0

	ENDIF		;end language options
;
;********
;
; Eftail is the "tail" and the checksum of the overlay.
;
Efcksm	SET	(NOT %) AND 0FFH	; Overlay checksum
Eftail	DB	Efcksm,-Efcksm
;
	END
