	PAGE
;********************************************************
;*							*
;*	 	----  Command Driver  ----		*
;*							*
;********************************************************

GnChew	CALL	ONdui		; turn off display update

GnClp	LDA	GnLF
	ORA	A		; library load in progress?
	JZ	GnChx		; nope, all done!
	CALL	Edeat		; yep/ go eat a char
	JMP	GnClp		; don't bother to update

GnChx	CALL	OFFdui		; turn on display update
	RET

;********

Edlp
	ctly	OFF		; disable control-Y
	CALL	Update		; do update if wanted
	CALL	Edeat		; and eat a char

	; now check for needed flushing
	CPI	Eups
	JZ	EtFlsh		; jmp/ up a line
	CPI	Edns
	JZ	EtFlsh		; jmp/ down a line
	CPI	Ergt
	JZ	EtFlsh		; jmp/ right a char
	CPI	Elft
	JZ	EtFlsh		; jmp/ left a char
	JMP	Edlp		; and spin

EtFlsh	CALL	Flush		; flush typeahead
	EI
	JMP	Edlp		; and spin...

;********

;  get a character from either an escape library file
;  or the keyboard, and do what it says to do...

Edeat
	CALL	WH0		; get KBD character
	ANI	7FH		; strip marker bit

	PUSH	PSW		; save the char
	LXI	H,edctb		; --> command table
	CALL	Lookup		; look this up
	JNC	Edlok		; jmp/ got a command
	LXI	H,Edchr		; i guess we eat this char...

Edlok	POP	PSW		; get char back from stack
	PUSH	PSW		; but keep it there, too
	PUSH	H
	LXI	H,RetHere	; ret addr goes on stack
	XTHL			; and HL gets routine address
	PCHL			; call handler (come back here)
RetHere	POP	PSW		; get char back from stack
	RET			; and return w/ character

;****************

Escape	
	CALL	WH0		; get KBD character
	ANI	7FH		; strip marker bit
	CPI	esc		; is it another escape?
	JZ	Escape		; yep/ ingore it...
	PUSH	PSW		; save register name
	LXI	H,esctb		; --> escape command table
	CALL	Lookup		; look this up
	JNC	EscGo		; jmp/ escape sequence!
	
	POP	PSW		; restore register name in A
	MOV	B,A		;     (pass to Efn1 in B)
	CALL	GnEat		; and feed it to the Gnome
	RET			; return

EscGo	POP	PSW		; clean stack
	PCHL			; and jump to handler

;****************


	PAGE
;********************************************************
;*							*
;*		     Command Decoder			*
;*							*
;********************************************************

;  Get a command from the keyboard and return a token which
;  which points to the routine that handles the command.

;	on entry:
;		 A = command character
;		HL --> command table

;	on return:
;		HL = token

Lookup
	LXI	D,3		; # bytes per entry
	MOV	B,A
LScan	MOV	A,M		; get command char
	CPI	0FFH
	JZ	Lnfg		; end of table/ unrecognizable!
	DAD	D		; --> next entry
	CMP	B		; does command char match key?
	JNZ	LScan		; jmp/ nope
	;
	DCX	H		; else backup and pickup token
	MOV	D,M
	DCX	H
	MOV	E,M
	XCHG			; token in HL
	RET			; scram.

Lnfg	STC
	RET			; set carry and split.

;****************

	PAGE
;********************************************************
;*							*
;*			Command Tables			*
;*							*
;********************************************************

op	MACRO
	DB	#1
	DW	#2
	IF	NOT NULL[#L]
#L	EQU	#1
	ENDIF
	ENDM

cop	MACRO
	DB	'#1'-40H
	DW	#2
	IF	NOT NULL[#L]
#L	EQU	'#1'-40H
	ENDIF
	ENDM

	MACLIST	0

;********************************

;  The regular command table

edctb:
	op	esc,Escape	; ESC prefix for esc commands

	cop	A,Append	; Append text from disk
	cop	B,Edtb		; Beginning of buffer
	cop	C,Edcon 	; Continue search
	cop	D,Eddel 	; Delete a char
	cop	E,Edeb		; End of buffer
	cop	F,Edfnd		; Find string
				; ******************
	cop	G,Edchr 	; **   bell	  **
	cop	H,Edchr		; **   backspace  **
	cop	I,Edchr		; **   tab	  **
	cop	J,Edchr		; **   line feed  **
	cop	K,Edchr		; **   vert tab	  **
	cop	L,Edchr		; **   form feed  **
	cop	M,Edchr		; **   cr	  **
				; ******************
	cop	N,Edgnp		; Next page
	cop	O,Output	; Output text to disk
	cop	P,Edgpp		; Previous page
Eups	cop	Q,Edups		; Straight up (up arrow)
Edns	cop	R,Eddns		; Straight down (down arrow)
Ergt	cop	S,Edgr		; Right (right arrow)
Elft	cop	T,Edgl		; Left (left arrow)
	cop	U,Edundl	; Undelete a char!
	cop	V,Edswap	; Swap upper/lower case
	cop	W,Eddw		; Delete word
	cop	X,Edkl		; Kill line
	DW	-1

;********************************

;  The ESC sequence command table.

esctb:
	op	colon,GnColon	; repeat!
	op	equals,GnDef	; define string
	op	quest,GnDsp	; display strings
cmtchr	op	quote,Edcmt	; comment for escape libraries
	op	ctldel,Eddb	; delete block (shift delete)
	op	del,Eddm	; delete markers (delete)

	cop	C,Edcb		; copy block
	cop	D,Edotb		; dump block
	cop	E,Exit		; exit editor
	cop	G,GRepl		; global change/replace
	cop	H,Edwt		; halt function
	cop	I,Reinp		; reopen input
	cop	K,Xcopy		; copy rest of input to output
	cop	L,GnLoad	; load escape library
	cop	M,Eddm		; delete markers
	cop	O,Reout		; reopen output
	cop	P,Edptb		; print block
	cop	Q,Edupl		; up a line (up arrow)
	cop	R,Eddnl		; down a line (down arrow)
	cop	S,Edmh		; mark head (right arrow)
	cop	T,Edmt		; mark tail (left arrow)
	cop	U,Updatf	; force update
	cop	V,EdTgl		; toggle message flag
	cop	W,GnWrite	; write escape library
	cop	X,EdAbt		; abort editor (no write)

	DW	-1		; end of table

	MACLIST	2

;****************
