	PAGE
;********************************************************
;*							*
;*		      Editor Services			*
;*							*
;********************************************************
;
;  The look functions below are called by different editor
;  functions for the purpose of finding certain types of
;  markers:  cursor, block head, and block tail.
;
Lkcr	LXI	H,crmkr	; find cursor marker
	JMP	Lkmkr
	;
Lkhd	LXI	H,hdmkr	; find head of block marker
	JMP	Lkmkr
	;
Lktl	LXI	H,tlmkr	; find tail of block marker
	;
Lkmkr	SHLD	sbufst	; set start of search string
	SHLD	sbufnd	; set end of search string
	JMP	Edlkup	; do lookup of string and split
;
;**********
;
;  ON & OFF are called to tweak the display inhibit flags
;
ONdui	MVI	A,dui	; display update inhibit
	JMP	ONbit
ONdin	MVI	A,din	; internal display inhibit (!?)
	JMP	ONbit
ONdai	MVI	A,dai	; all display activity inhibit
	;
ONbit	PUSH	H
	LXI	H,inflgs
	ORA	M
	MOV	M,A
	POP	H
	RET		; turn on the bit and scram.
;
;**********
;
OFFdui	MVI	A,dui	; display update inhibit
	JMP	OFFbit
OFFdin	MVI	A,din	; internal display inhibit (!?)
	JMP	OFFbit
OFFdai	MVI	A,dai	; all display activity inhibit
	;
OFFbit	PUSH	H
	LXI	H,inflgs
	CMA
	ANA	M
	MOV	M,A
	POP	H
	RET		; turn off the bit and scram.
;
;****************
;
Warning LDA	WMsgFlg
	ORA	A
	RZ		; ret/ no messages today
	PUSH	H
	CALL	Hailer	; clear screen and put hailing message
	POP	H
	CALL	PrtMsg	; print message for code HL.
	JMP	Anykey	; and return after getting any key
	PAGE
;========================================================
;|		Cnote & Crenew (Cursor services)	|
;========================================================
;
;  Cnote remembers the cursor position. Crenew tries to
;  put the bubble back. We use the cursor marker char and
;  use Edtb and Lkcr to do the dirty work. We also set
;  the display update inhibit flag and reset it in Crenew.
;
Cnote	LDA	inflgs
	ANI	dai
	RNZ		; ret/ no display activity
	;
	LDA	crmkr
	CALL	Edchr	; put marker in buffer
	RZ		; split if no room
	;
	CALL	ONdin
	RET		; turn on display inhibit and scram.
	;
Crenew	LDA	inflgs
	ANI	dai
	RNZ		; ret/ no display activity
	;
	CALL	Edtb	; fly to the top
	CALL	Lkcr	; look for marker
	JC	Cren1	; ret/ it disappeared!
	CALL	Eddel	; delete marker
	;
Cren1	CALL	OFFdin
	RET		; clear display inhibit and split.
	PAGE
;********************************************************
;*							*
;*	 	       Display Driver			*
;*							*
;********************************************************
;
; Initialize screens.
;
Upint	CALL	Uprst		; reset display parameters
Upclr	LHLD	fkscn		; --> fkscn
	LXI	D,400H
Up1	MVI	M,0
	INX	H
	DCX	D
	MOV	A,D
	ORA	E
	JNZ	Up1		; clear whole screen
	RET			; and scram.
;
;	Reset the display buffer pointer and cursor position
;  to the very beginning of the bubble.
;
Uprst	LXI	H,0
	SHLD	curse		; reset cursor
	LHLD	bblb
	SHLD	bfptr		; init buffer pointer
	RET			; and scram.
;
;****************
;
;  Update the screen, locking out slicing.
;
Update	
	LDA	inflgs
	ANI	dai+dui+din
	RNZ		; return/ no update wanted
	;
	; Updatf is the entry point for the ESC-U function!
	;
Updatf	LHLD	bblb
	MOV	A,M		; get char @ (bblb)
	PUSH	PSW		; and save for restore on exit
	MVI	M,dsmkr		; put display marker at (bblb)
	lock			; lock against slicing
	CALL	Updat		; update the screen
	unlock			; allow slicing
	LHLD	bblb		; --> bblb
	POP	PSW		; get char we removed
	MOV	M,A		; and restore for undelete hack
	RET			; and return!
;
;  Update the screen.
;	For speed we use in-line macro coding and retain the
;	following in registers in the update loop:
;		HL = fake screen cursor position
;		DE = buffer pointer
;		BC = relative cursor position
;
Upscn	MACRO		; macro for in-line coding
#L	CMP	M
	JZ	$+11	; jmp/ already there
	MOV	M,A
	PUSH	H
	LHLD	rlscn
	DAD	B
	MOV	M,A	; else put char on both screens
	POP	H
	INX	H
	INX	B	; bump both pointers
	ENDM
;
Updat	LHLD	bfptr	; init buffer pointer
	XCHG
	LHLD	fkscn	; and fake screen
	LXI	B,0	; screen index = 0
Uplp	LDAX	D	; get char
	ORA	A
	JP	Upnchr	; jmp/ not marker char
	CPI	dsmkr
	JZ	Upbbl	; display marker/ we hit bubble
	CPI	etx
	JZ	Upend	; end of text/ weez done!
	;
Upnchr	CPI	ctlgrp
	JP	Upnctl	; jmp/ not control char
	CPI	cr
	JZ	Upcr	; cr/ skip line
	CPI	tab
	JZ	Uptab	; tab/ skip to tab stop
	;
Upnctl	ORI	80H	; make a displayable char
UpCurs	Upscn		; put it on screen
	MVI	A,3
	CMP	B	; set carry is screen full
	RC
	INX	D	; else bump bfptr
	JMP	Uplp	; fly
	;
Upbbl	XCHG		; swap bfptr, fake screen cursor
	MOV	H,B
	MOV	L,C
	SHLD	ptsav	; save pos of brick
	LHLD	bble	; end of bubble is new bufptr
	XCHG		; swap 'em back
	;
	LDA	crmkr	; cursor marker
	JMP	UpCurs	; use common code
	;
Upcr	PUSH	D
	LXI	D,3F00H+grbk	; do cr w/ graphics blank
	JMP	Upspc		; do it
	;
Uptab	PUSH	D
	LXI	D,700H+grbk	; do tab w/ graphics blank
	;
Upspc	MOV	A,E	; put pad char out
	Upscn		; put it on screen (can't overflow)
	MOV	A,C
	ANA	D	; check tab/cr boundry
	JNZ	Upspc
	POP	D
	MVI	A,3
	CMP	B	; now check for screen overflow!
	INX	D
	JNC	Uplp	; jmp/ not done, spin
	RET
	;
Upend	MVI	A,grbk
	Upscn		; put blank on the screen
	MVI	A,3
	CMP	B	; set carry if screen full
	JNC	Upend	; spin till full
	RET
	PAGE
;********************************************************
;*							*
;*	 	       Cursor Driver			*
;*							*
;********************************************************
;
Cursr	LDA	inflgs
	ANI	dai
	JNZ	Curlv	; split if no update wanted
	;
	LHLD	bblb	; get char @ (bblb-1)
	DCX	H
	MOV	A,M
	LHLD	curse	; HL = current cursor position (CPOS)
	CPI	ctlgrp
	JP	Cursr1	; jmp/ not control char
	CPI	tab
	JZ	Curstb	; jmp/ tab char
	CPI	cr
	JZ	Curscr	; jmp/ cr char
Cursr1	INX	H	; normal char - bump CPOS
	JMP	Cursr3	; and go store it
	;
Curscr	LXI	D,64	; setup to move to start of next line
	MVI	A,0C0H
	JMP	Cursr2	; go do it
	;
Curstb	LXI	D,8	; setup to move to tab stop
	MVI	A,0F8H
	;
Cursr2	DAD	D	; update cursor
	ANA	L	; mask to cr/tab stop
	MOV	L,A	; HL = new CPOS
Cursr3	SHLD	curse	; store it
	MOV	A,H	; get high byte of curse
	CPI	4
	JC	Curlv	; no overflow/ split
	;
	; scroll up one display line, if possible.
	;
	IF	USERS=2
	giveup		; be polite!
	ENDIF
	MVI	B,64	; set limit
Covfl	LHLD	bfptr	; get DSA (display start)
	MOV	A,M	; get char @ (DSA)
	INX	H
	SHLD	bfptr	; and bump DSA.
	DCR	B	; dink counter
	JZ	Covldn	; jmp/ hit limit - weez done!
	CPI	ctlgrp
	JP	Covfl	; jmp/ not control char
	CPI	cr
	JZ	Covldn	; jmp/ cr - weez done...
	CPI	tab
	JNZ	Covfl	; jmp/ not tab
	MOV	A,B
	ANI	0F8H	; set count to tab stop
	MOV	B,A
	JNZ	Covfl	; spin/ didn't hit limit yet.
	;
Covldn	LXI	H,960
	SHLD	curse	; set cursor position to last line
	JMP	Curlv	; jmp/ split.
;
;****************
;
Cursl	LDA	inflgs
	ANI	dai
	JNZ	Curlv	; split if no cursor update wanted
	;
	LHLD	bble
	INX	H
	MOV	A,M	; get char @ (bble+1)
Cursl1	PUSH	PSW	; save char
	;
	LHLD	bblb	; cursor off screen if (bblb < DSA)
	XCHG
	LHLD	bfptr
	CALL	Cmpdh	; comp. bblb-bfptr
	POP	B	; restore char to B (flags to C)
	JC	Cundr	; jmp/ cursor off screen
	;
	MOV	A,B	; char in A
	CPI	ctlgrp
	JM	Cursl3	; jmp/ control char
Cursl2	LHLD	curse	; else normal char
	DCX	H
	SHLD	curse	; dink cursor position
	JMP	Curlv	; and split.
	;
Cursl3	CPI	tab
	JZ	Cursl4	; jmp/ tab
	CPI	cr
	JNZ	Cursl2	; jmp not cr
Cursl4	;
	;  all cr and tab mean is that we have to haul ass and
	;  find the location of the cursor relative to DSA.
	;
	IF	USERS=2
	giveup		; be polite!
	ENDIF
	LHLD	bblb
	MOV	B,H
	MOV	C,L
	LHLD	bfptr	; HL = DSA, BC = bblb
	LXI	D,0	; DE = CPOS
Cursl5	MOV	A,L	; compare DSA & bblb
	CMP	C
	JNZ	Cursl6
	MOV	A,H
	CMP	B
	JZ	Cursdn	; if equal, scram.
Cursl6	MOV	A,M	; get char @ ptr
	INX	H	; bump ptr
	CPI	ctlgrp
	JM	Cursl8	; jmp/ control char
Cursl7	INX	D	; bump CPOS
	JMP	Cursl5	; and loop.
Cursl8	CPI	tab
	JZ	Curltb	; jmp/ tab
	CPI	cr
	JNZ	Cursl7	; jmp/ not cr
	;
	PUSH	H	; save ptr
	MVI	A,0C0H	; mask
	LXI	H,64	; index to start of next line
	JMP	Cursl9
Curltb	PUSH	H	; save ptr
	MVI	A,0F8H	; mask
	LXI	H,8	; index to tab stop
Cursl9	DAD	D
	ANA	L	; mask to cr/tab stop
	MOV	L,A	; HL = new CPOS
	XCHG		; CPOS back in DE
	POP	H	; restore HL
	JMP	Cursl5	; and spin...
	;
Cursdn	XCHG		; CPOS in HL
	SHLD	curse	; stash.
	JMP	Curlv	; and scram.
;
;  Cursor underflow is a damn nightmare... we must bring in
;  a full DISPLAY line (not necessarily a TEXT line), which
;  means hassle. We put the cursor at the home position and
;  then move display start until we get the cursor as near to
;  the end of the line without overflowing. TABS give us a
;  pain. This is the algorithim I finally developed for tabs:
;
;	1> get Char; if CR: exit
;	2> if TAB: add 8 to Ptr; set Flag; clear Ctr, goto 1
;	3> if Flag not set: add 1 to Ptr; goto 1
;	4> Add 1 to Ctr; if Ctr=8: clear Ctr; add 8 to Ptr
;	5> goto 1
;
;  'Ptr' represents the computed cursor position; 'get Char'
;  represents decrementing DSA and getting a char.
;
Cundr	LHLD	bblb	; set display start to bblb
	LXI	D,0	; set cursor position to zero
	LXI	B,0	; B is flag/ C is counter
	;
Cundlp1 SHLD	bfptr	; stash display start address (DSA)
	XCHG
	SHLD	curse	; stash CPOS as current cursor position
	XCHG		; HL = DSA, DE = cursor position (CPOS)
	DCX	H	; dink DSA
	MOV	A,M	; get char @ (DSA-1)
	;
	CPI	stx
	JZ	Curlv	; jmp/ start of text - split.
	CPI	cr
	JZ	Curlv	; jmp/ cr - split.
	CPI	tab
	JNZ	Cundnt	; jmp/ no tab
	;
Cunac	MOV	A,E	; else process tab
	ADI	8
	MOV	E,A	; add 8 to CPOS
	LXI	B,100H	; set flag and clear counter
	JMP	Cundlp2	; and go test for limit
	;
Cundnt	MOV	A,B
	ORA	A
	JZ	Cunfns	; jmp/ tab flag not set
	INR	C	; bump counter	
	MOV	A,C
	CPI	8	; is ctr = 8?
	JZ	Cunac	; jmp/ same as tab - bump
	JMP	Cundlp1	; else spin.
	;
Cunfns	INX	D	; bump CPOS
	JMP	Cundlp2	; and spin.
	;
Cundlp2 MOV	A,E	; is curse < 64?
	CPI	64
	JM	Cundlp1	; jmp/ yep.
	;
Curlv	ORI	1
	RET		; split non-zero
