	PAGE
;********************************************************
;*							*
;*		    U T I L I T I E S			*
;*							*
;********************************************************
;
; FIELD advances HL to point to the next field in a
; buffer. First we skip to the end of the current field,
; then we skip delimiters until the next field is hit. If
; we hit a cr we stop and point to it.
;
Field	CALL	Nxtc
	JNZ	Field	; skip to end of last field
 	DCX	H	; point to delimiter
Flp1	CALL	Nxtc
	JNZ	Flp2	; jmp/ not delim
	CPI	cr
	JNZ	Flp1	; eat all delimiters except cr
Flp2	DCX	H	; point at char of next field
	RET		; scram
;
;****************
;
; NXTC returns the next character in the buffer pointed
; to by (HL), incrementing (HL). Z flag set if the char
; is a delimiter - cr, space, or comma.
;
Nxtc	MOV	A,M
	INX	H
	CPI	cr
	RZ
	CPI	space
	RZ
	CPI	comma
	RET
;
;****************
;
; A hard disk error causes an error message and jump to Exec.
;
Hrderr	MVI	A,ff
	CALL	WH1	; clear screen
	JMP	Err	; and report error
;
;****************
;
; Cmpdh compares DE & HL by subtraction. (DE-HL)
;
Cmpdh	MOV	A,D
	CMP	H
	RNZ
	MOV	A,E
	CMP	L
	RET
;
;****************
;
; Dsub performs the double subtraction HL = HL-DE
;
Dsub	MOV	A,L
	SUB	E
	MOV	L,A
	MOV	A,H
	SBB	D
	MOV	H,A
	RET
	PAGE
;********************************************************
;*							*
;*		   Bubble primitives			*
;*							*
;********************************************************
;
bblsiz	LHLD	bblb	; get bubble start
	XCHG
	LHLD	bble	; get bubble end
	JMP	Cmpdh	; compare
;
;****************
;
; Mvbl will move the bubble buffer one location to the
; left. This is accomplished by copying the character at
; (bblb-1) to (bble), and then decrementing bblb and bble.
;
Mvbl	LHLD	bblb
	DCX	H	; get bblb-1
	MOV	A,M	; get char
	CPI	stx
	RZ		; split/ top of buffer
	SHLD	bblb	; bblb = bblb-1
	LHLD	bble	; get bble
	MOV	M,A	; toss char to end of bubble
	DCX	H	; dink bble
	SHLD	bble
	RET		; ret non-zero
;
;****************
;
; Mvbr will move the bubble buffer one location to the
; right. This is accomplished by copying the character
; at (bble+1) to (bblb), and then incrementing bblb & bble.
;
Mvbr	LHLD	bble
	INX	H	; get bble+1
	MOV	A,M	; get char
	CPI	etx
	RZ		; split/ end of buffer
	SHLD	bble	; bble = bble+1
	LHLD	bblb	; get bblb
	MOV	M,A	; toss char to front of bubble
	INX	H	
	SHLD	bblb
	RET		; ret non-zero
	PAGE
;========================================================
;|		   String Lookup Service		|
;========================================================
;
;  Find a string, starting at the end of the bubble. The
;  search (pattern) string address is in sbufst.
;  If we find the string, the bubble is moved there, else
;  we set CARRY and sbuflg non-zero to indicate a search
;  failure.  We also set sdone to indicate that Edlkup was
;  called (for the escape colon processor).
;
Edlkup	LHLD	bble
	XCHG
	INX	D	; start of text to search in DE.
	;
Edlk1	LHLD	sbufst	; HL - start of search string
Edlklp	LDAX	D
	CPI	etx
	JZ	Edfdn	; jmp/ end of buffer hit
	CMP	M	; compare characters
	JZ	Edlkit	; jmp/ match!
	INX	D	; else bump text ptr
	JMP	Edlklp	; and spin.
	;
Edlkit	PUSH	D	; save copy for failure recovery
Edlk2	PUSH	D	; save text ptr
	XCHG
	LHLD	sbufnd	; tail of search buffer
	CALL	Cmpdh	; compare string ptr for end
	XCHG
	POP	D	; get back text & search pointers
	JZ	Edfun	; jmp/ found string!
	INX	D	
	INX	H	; bump string & text pointers
	LDAX	D
	CMP	M	; compare chars
	JZ	Edlk2	; jmp/ match!
	POP	D	; restore ptr to start of text matched
	INX	D
	JMP	Edlk1	; jmp/ nfg... go spin.
	;
Edfun	POP	H	; clean stack
	LHLD	bble	; get bubble end
	XCHG
	CALL	Dsub	; compute length of move
	SHLD	bbls	; save it
	CALL	Update	; show the find text gone
Edfld	CALL	Edgr	; bubble down
	CALL	Flash	; flash it
	LHLD	bbls
	DCX	H	; decrement length counter
	SHLD	bbls
	MOV	A,H
	ORA	L	
	JNZ	Edfld	; loop if not there yet
	CALL	Update	; else update screen
	ANA	A	; clear carry
	JMP	Edfxit	; and go bye-bye
	;
Edfdn	MVI	A,0FFH
	STA	sbuflg	; set search failure flag
	CALL	Update	; and update screen
	STC		; set error flag
	;
Edfxit	LXI	H,sdone
	MVI	M,1
	RET		; and return
	PAGE
;========================================================
;|		   More utilities			|
;========================================================
;
Flash	PUSH	H	; save regs
	PUSH	D
	PUSH	PSW
	LDA	inflgs
	ANI	dai
	JNZ	Flshx	; jump/ no display activity wanted
	;
	LHLD	ptsav	; loc to flash on rlscn
	XCHG
	LHLD	fkscn
	DAD	D	; compute addr on fkscn
	LDA	TIMER	; get fast byte
	ANI	8	; get 8 bit
	MOV	A,M	; get char
	JZ	Flshn	; if timer bit zero, unblink
	ANI	7FH	; else blink it
	JMP	Flshv
Flshn	ORI	80H	; sign on to make normal
	;
Flshv	CMP	M	; is it there already?
	JZ	Flshx	; yup/ scram
	MOV	M,A	; nope/ stuff it
	LHLD	rlscn
	DAD	D	; compute address on rlscn
	MOV	M,A	; stuff it, too
	;
Flshx	POP	PSW
	POP	D
	POP	H	; restore regs
	RET
;
;****************
;
; Hailer puts up the Edit/nn message.
;
Hailer	LXI	H,hail
	CALL	MESSAGE	; put up the hailing
	CALL	Upclr	; force screen update
	RET		; and scram.
;
;****************
;
Ederr	CALL	Ovrto	; report error message and ret carry
	DB	'Emsg'
	MVI	A,cr
	CALL	WH1	; put up a cr
	STC		; set error flag
	RET		; scram
;
;****************
;
MESSAGE LDA	inflgs
	ANI	dui+dai+din
	RNZ		; return/ display activity inhibited
	JMP	Msg	; else pump message and scram
;
;****************
;
PCHAR	PUSH	B
	MOV	B,A
	LDA	inflgs
	ANI	dui+dai+din
	MOV	A,B
	POP	B
	RNZ		; return/ display activity inhibited
	JMP	WH1	; else put the char and scram
;
;****************
;
;	Copy BC bytes RIGHT from HL@ to DE@.  Unrolled once 
;  for some speed.  Returns with HL & DE pointing past the 
;  last bytes moved.
;
COPYR	ORA	A		; Clear CARRY
	MOV	A,B
	RAR
	MOV	B,A
	MOV	A,C
	RAR
	MOV	C,A		; BC = BC/2, rem in CARRY
	JNC	COPYR1		; Jmp/ no remainder
	MOV	A,M
	STAX	D
	DCX	H
	DCX	D		; else move one byte
COPYR1	MOV	A,B
	ORA	C
	RZ			; Ret/ handle zero case
COPYR2	MOV	A,M
	STAX	D
	DCX	H
	DCX	D		; Move one
	MOV	A,M
	STAX	D
	DCX	H
	DCX	D		; and another.
	DCX	B		; Decrement count.
	MOV	A,B
	ORA	C
	JNZ	COPYR2		; Jmp/ move two more.
	RET
;
;	Copy BC bytes LEFT from HL@ to DE@.  Unrolled once 
;  for some speed.  Returns with HL & DE pointing past the 
;  last bytes moved.
;
COPYL	ORA	A
	MOV	A,B
	RAR
	MOV	B,A
	MOV	A,C
	RAR
	MOV	C,A		; BC = BC/2, rem in CARRY
	JNC	COPYL1		; Jmp/ no remainder
	MOV	A,M
	STAX	D
	INX	H
	INX	D		; else move one byte
COPYL1	MOV	A,B
	ORA	C
	RZ			; Ret/ handle zero case
COPYL2	MOV	A,M
	STAX	D
	INX	H
	INX	D		; Move one
	MOV	A,M
	STAX	D
	INX	H
	INX	D		; and another.
	DCX	B		; Decrement count.
	MOV	A,B
	ORA	C
	JNZ	COPYL2		; Jmp/ move two more.
	RET
