;********************************************************
;*							*
;*			 Gfid.OV 			*
;*							*
;*	Version 3.6			R.T. Martin	*
;*	02/08/81			F.E. Anderson	*
;*							*
;*	Get file identifier service.  (The heart of	*
;*  the crude System 88 file system).			*
;*  							*
;*	Copyright (C), 1980 Interactive Products Corp-	*
;*  oration dba PolyMorphic Systems, 460 Ward Drive,	*
;*  Santa Barbara  CA 93111  (805) 967-2351		*
;*							*
;********************************************************
;
; Get File Identifier overlay. April 1977, R. T. Martin
; Modified for directory tree structure (FEA)
;
; Last updated:
;	4/18/79 FEA	Support of tree structured directories
;	7/19/79 RTM	Disable setting of "new" bit in FDE
;	8/14/79 BFS	Fixed extension match bug in replace
;			function.
;	9/28/79 BFS	Enter function sets "new" bit again.
;	10/18/79 RTM	Modify for two user system
;	10/24/79 RTM	Add GLOCK semaphore...
;	10/29/79 RTM	Hacks.
;	11/16/79 RTM	Move to EE00
;	12/11/79 RTM	Problems with ^Y and semaphores...
;	12/31/79 RTM	Disaster recovery, move to EF00
;	1/4/80   RTM	Another disaster, interlocks.
;	2/26/80  RTM	Interlocks/devlock support
;	3/28/80  RTM	Make update directory a service
;	5/2/80	 RTM	Bandaid over DIRADDR/blized disk?!?
;	5/27/80  BFS	Bandaid on the bandaid.  Move more
;			stuff into per user area.
;	5/28/80  BFS	The real source of the interlock
;			problem is the self modifying code
;			in the Gfid load routine.  Changing
;			start address to byte after GFLOCK
;			will make the interlock appear to work.
;			I removed the bandaids.
;	5/29/80	 BFS	Moved DIRADDR again.
;	5/29/80  Staff	Fixed GFLOCK double-unlock problem,
;			Control-Y disabled w/ EIC, and GfExit
;			added for increased reliability.
;			Also GFLOCK moved out of Gfid into
;			vector area, and Gfid lock cleaned up
;			for when Gfid loads for the first time.
;	7/22/80	RTM	Set EIC for real, don't lock up system
;			while reading from the user!
;	8/13/80 BFS	Sprinkled ANI 31's throughout to mask
;			off flag bits when we want length.
;	10/03/80 BFS	Changed EFID to avoid setting flag bits
;			in last flename length byte of lookup
;			block.
;	10/20/80 Staff	Fixed lookup of old type looks passed
;			externally by PUSH H - POP H in look.
;			Uses NFDIR & PATH more intelligently 
;			now.  Still more work on new bit.
;			(Called it version 3.0)
;	10/21/80 LSA	Fixed extension match problem created
;			by yesterday's changes.
;	10/23/80 Staff	PATH refers to correct disk?
;	10/30/80 Staff	Fixed 2U invalidation of NFDIR and
;			merged SU w/ 2U source.
;	11/12/80 BFS	Put in update of DIRADDR from DirAddr
;			if look finds that current PATH is
;			correct. (Other user may have stomped
;			DIRADDR between calls by this user.)
;			Also added validity check on NFDA
;			in EFID function.  Before entering
;			new file checks that actual end of
;			directory entries is same as SBUF1+11.
;			As long as I was here I cleaned up the
;			SCAN routine .
;	11/15/80 Staff	Save HL around above DIRADDR fix.
;	12/16/80 BFS	Fix SEARCH.
;	1/12/81  FEA	Changed assembly address for the twin
;			version to F200H to accomidate volume
;			manager and hard disk driver.
;	01/26/81 BFS	Changed PFC so it doesn't write
;			into RAM at ZERO if it exists.
;	02/07/81 BFS	Added 40H bit function to clear
;			new bit on replace function.
;			See definition.
;	02/08/81 BFS	Fixed unit parse so <x<, x<, or even
;			< followed by the filename will all
;			work.
;	04/02/81 BFS	SEARCH sets string back to ? if file
;			not found.
;
;
; This overlay performs common file system functions.
; The function performed is determined by the lower
; four bits of A to select:
;
;	A and 0FH	Function
;	   0		Get file identifier
;	   1		Enter/replace FDE
;	   2		Look up a file
;	   3		Update directory on disk (2U-only)
;
; Get File Identifier: The registers on entry contain:
;	HL:	Points to either the prompt string to
;		use in reading from the user (Via RLWE)
;		or the address of the text buffer to
;		examine
;	DE:	Points to the buffer that is used to
;		return the file specification in form
;		suitable for feeding to LOOK
;	BC:	May contain the default extension
;		to use if the user does not give one
;	A: 80H	if set, read (RLWE) from the user, 
;		prompting with the MSG -> by HL
;	   40H	if set, LOOK the file up. If it exists,
;		copy the FDE to the buffer -> (DE)+1,
;		thus returning FDA, NSCTR, LA, SA. If 
;		an 0300 error is returned from LOOK,
;		return NFA from directory in the
;		FDE slot in the buffer -> (DE)+1
;	   20H	If set, use the extension in BC if
;		the user does not specify one
;	Return:
;	HL:	Points to the ending delimiter in the 
;		text buffer
;	DE:	If C set in PSW, error code. If not, FDE
;		address if LOOK was requested, else ??
;	BC:	crap. 
;	A (PSW):C set if error, clear if not. A ??
;
;	Errors:
;	05/00	Invalid disk #
;	05/01	Name longer than 31 characters
;	05/02	Extension longer than 2 characters
;	05/03	Name zero length
;
; Enter FDE:
;	HL:	Points to a file block as built by get
;		file function. First byte has disk #,
;		next byte has flags and name length, etc
;	A: 80H	If set, replace an existing FDE with 
;		the one pointed to by HL, else enter a
;		new one at the end of the disk
;	   40H  If set, and if 80H set, replace and clear
;		new bit.  If 40H not set, and 80H set
;		replace and set new bit.
;	Returns:
;	A (PSW) If C set, then DE has error code/subcode
;		If C not set, then registers scrambled.
;	Errors:
;	05/04	Directory is full (you lose.)
;
; Update directory:
;	Added for 2 User to allow invalidation of NFDIR hack.
;	uses NFDIR, DIRADDR, SBUF1 information
;
	MACLIST	0
	REFS	SYSTEM
	REF

DirEnd	EQU	SBUF1+400H	; End of directory in SBUF.
MAXDRIVE EQU	9		; Largest legal drive #.

;
;	Macros...
;
rORG	MACRO
rlc	SET	#1
	ENDM
;
rDS	MACRO
#L	EQU	rlc
rlc	SET	rlc+#1
	ENDM
;
	IF	USERS=1
	ORG	OVRLY
	IDNT	$,$
CODE	DB	'Gfid'
	JMP	Start
	ELSE		; 2-User system
CODE	EQU	0F200H
	ORG	CODE
	IDNT	$,$
	ENDIF
;
; Character codes that are delimiters for file names
;
CR	EQU	0DH	; carriage return
ESC	EQU	1BH	; escape
SPACE	EQU	20H	; space
TAB	EQU	09H	; tab
COMMA	EQU	2CH	; comma
PLUS	EQU	2BH	; plus sign
;
	IF	USERS=1
;
;	4.0 routines used.
;
MOVE	EQU	0100H
;
;	2-User synonyms.
;
Moven	EQU	MOVE
gLook	EQU	Look
	ENDIF		; End of USERS
	PAGE
	IF	USERS=1
;
;	Entry and exit code for single user Gfid.
;
RETURN	XTHL
	SHLD	PVEC
	POP	H
	RET
;
;	CONTY handler for read buffer logic.
;
CONTY	LHLD	ESP
	SPHL
	LXI	D,503H	; Name null or weird.
	STC
	EI
	RET
;
;	Enter here.
;
Start	PUSH	H	; Save HL.
	LHLD	PVEC	; Get PVEC.
	PUSH	H	; Save it.
	LXI	H,Ioret	; Point PVEC at Ioret.
	SHLD	PVEC
	POP	H	; Get old PVEC.
	XTHL
	PUSH	H	; Slide HL under PVEC
	LXI	H,RETURN ; Common return address.
	XTHL		; Put common return address on stack
			; and restore entry HL.
	PUSH	H	; Save HL again.
	LXI	H,2	; offset for saved HL.
	DAD	SP	; Phony entry SP.
	SHLD	ESP	; Save it.
	POP	H	; Restore HL.
	STA	ACRUD
	SHLD	SPTR	; A and HL
	PUSH	PSW
	XRA	A
	STA	EFIDFLG	; In case we are reinvoked
	POP	PSW
	XCHG		; Save pointer to where file is built
	SHLD	FPTR	; (DE).
	PUSH	B	; Save default extension (BC)
	POP	H
	SHLD	DEXT
	ANI	0FH	; Mask out command (Gfid or Efid?)
	ADD	A	; Multiply by two
	MOV	E,A	; Offset to DE
	MVI	D,0
	LXI	H,DTBL	; Base of dispatch table to HL
	DAD	D	; address to go to
	MOV	E,M
	INX	H
	MOV	D,M
	PUSH	D	; stack the address...
	LHLD	FPTR
	XCHG
	LHLD	SPTR	; get regs back
	LDA	ACRUD
	RET		; Dive into the routine we stacked.
;
; Dispatch table
;
DTBL	DW	GFID	; 0
	DW	EFID	; 1
	DW	look	; 2
;
	ELSE			; 2-User system
;
; NOTE:  The semaphore has already been acquired by this time.
;	 It is very important that this code RETURN to its
;	 caller so the semaphore will clear. We disable CTL-Y
;	 via EIC to make sure that it does.
;
Start	PUSH	PSW	; save A
	ANI	0FH	; see if get file identifier code
	JNZ	Gf0	; jmp/nope.

; On Get file identifier calls, we trap out ones that
; specify a read so we don't lock up the file system for
; the duration of the read from the user, and also so we
; can allow ^Y from that read.

	POP	PSW
	PUSH	PSW	; get it back.
	ANI	80H	; does it say to read from user?
	JZ	Gf0	; jmp/nope.

; We're supposed to read from the user. Let's build up our
; environment to re-enter Gfid when the read completes.

	POP	PSW
	ANI	7FH	; turn off read bit
	PUSH	PSW	; stash PSW
	PUSH	B	; BC
	PUSH	D	; DE
	LXI	D,RBUF0	; assume a read for user 0
	LDA	BRG
	ANI	PHANTOM	; see if it's user 00
	JZ	Grd0	; jmp/yup
	LXI	D,RBUF1	; use user 1 buffer
Grd0	PUSH	D	; "HL" now is buffer address

; Give up the interlock

	leave	GFLOCK	; we are supposedly now reentrant
	XCHG		; put prompt in DE, buf in HL
	LXI	B,34	; length of buf
	CALL	Rlwe	; read from the user

; Now unstack our environment and split!

	POP	H
	POP	D
	POP	B
	POP	PSW
	JMP	Gfid

; The main path starts here again

Gf0	POP	PSW
	STA	ACRUD
	SHLD	SPTR	; A and HL
	PUSH	PSW

	PUSH	H
	LXI	H,EFLG1
	MOV	A,M
	STA	OldEFLG1 ; Save old EIC bit.
	ORI	EIC
	MOV	M,A	; This disables CTL-Y.
	POP	H
	XRA	A
	STA	EFIDFLG	; Initialize flag.

	EI		; Now enable interrupts.
	POP	PSW
	XCHG		; Save pointer to where file is built
	SHLD	FPTR	; (DE).
	MOV	H,B	; Save default extension (BC)
	MOV	L,C
	SHLD	DEXT
	ANI	0FH	; Mask out command (Gfid or Efid?)
	ADD	A	; Multiply by two
	MOV	E,A	; Offset to DE
	MVI	D,0
	LXI	H,GfExit ; Setup return to GfExit so we clear
	PUSH	H	;  the semaphore and restore EFLG1.
	LXI	H,DTBL	; Base of dispatch table to HL
	DAD	D	; address to go to
	MOV	E,M
	INX	H
	MOV	D,M
	PUSH	D	; stack the address...
	LHLD	FPTR
	XCHG
	LHLD	SPTR	; get regs back
	LDA	ACRUD
	RET		; Dive into the routine we stacked.

; Dispatch table

DTBL	DW	GFID		; 0
	DW	EFID		; 1
	DW	look		; 2
	DW	Updir		; 3
;
;	Here is where we return from any of the above Gfid
;  services.
;
GfExit	DI
	PUSH	PSW
	LDA	OldEFLG1
	STA	EFLG1
	POP	PSW
	RET		; And split
	ENDIF		; End of USERS
	PAGE
; Get me a string of characters from (SPTR) until a delimiter
; is seen.  Put in (FPTR).  Return count in B.

GSTRING	MVI	B,0	; Initialize count
GS1	CALL	GFC	; Get a char
	RZ		; If a delimiter split
	INR	B	; Increment count
	PUSH	B	; Save count
	CALL	PFC	; Put character in (FPTR)
	POP	B	; Restore count
	RC		; Error control for PFC
	JMP	GS1	; Get the next one

;
;
;
CKUNIT	CPI	'#'	; Is it #?
	RZ		; Yup, return zero.
	CPI	'?'	; Is it ?
	RZ		; Yup, return zero.
	SUI	'0'	; subtract ASCII.
	CPI	10	; Is it 10 or greater?
	JNC	CKU1	; Yup, return non-zero below.
	XRA	A	; It's 1-9, set zero flag.
	RET
CKU1	MVI	A,0FFH
	ORA	A
	RET
;
;	Check character in A for some combination of digit
;  ? or # plus HL pointing to delimiter ('<' or '>').  If not
;  found return ZERO flag set.  Don't change SPTR.
;
UNIT	CALL	CKUNIT
	RNZ		; It's not a unit id.
	MOV	A,M
	CALL	CHKDLM
	RET		; If it's a delimiter, it's an ID.
;
;	Scan buffer at SPTR for commas, spaces and tabs bumping
;  pointer past them.  Leave pointer at first non-match.
;
BREAK	LHLD	SPTR
BRK1	MOV	A,M
	INX	H
	CPI	' '
	JZ	BRK1
	CPI	TAB
	JZ	BRK1
	CPI	','
	JZ	BRK1
	DCX	H
	SHLD	SPTR
	RET
	
;
;	GFC returns the next chr from the buffer pointed to
;  by SPTR, and has Z set if the chr is a delimiter.
;
GFC	LHLD	SPTR
	MOV	A,M
	INX	H
	SHLD	SPTR	; advance scan pointer too.
	CPI	CR
	RZ
	CPI	TAB
	RZ
	CPI	COMMA
	RZ
	CPI	SPACE
	RZ
	CPI	PLUS
	RZ
	CPI	'.'
	RZ
	CALL	CHKDLM
	RZ		; simple.
	ORA	A	; Null?
	RZ
	CPI	80H	; Termination for #?
	RNZ		; If not just split
	LHLD	SPTR1	; Tell guy to take from
	SHLD	SPTR	; old scan pointer now.
	JMP	GFC

CHKDLM	CPI	'<'
	RZ
	CPI	'>'
	RET

PFC	LHLD	FPTR	; Where to put character
	MOV	B,A	; Save Acc here.
	XRA	A	; Make a zero.
	CMP	H	; Is H zero?
	JZ	PFC1	; Yes, don't write.
	MOV	M,B	; Put it
PFC1	INX	H	; Increment pointer
	SHLD	FPTR	; and update

; Find out if we're too long

	XCHG		; Where we are to DE
	LHLD	FPTR1	; Beginning of our area
	LXI	B,35	; # of bytes allowed
	DAD	B	; This is top address alowed
	CALL	DCMP	; Compare values
	RET		; Return with flags set
;
; Error bail-outs
;
E300	MVI	E,0
	DB	21H
E3FF	MVI	E,0FFH
E3XX	MVI	D,3
	JMP	EXXX
E500	MVI	E,0	; Invalid drive #
	DB	21H
E501	MVI	E,1	; Name too long
	DB	21H
E502	MVI	E,2	; Extension length wierd
	DB	21H
E503	MVI	E,3	; Name null or wierd
	DB	21H
E504	MVI	E,4	; Directory full
	DB	21H
E505	MVI	E,5	; file already exists error
	DB	21H
E509	MVI	E,9	; <?> not valid here error
	DB	21H
E50A	MVI	E,0AH	; Directory not found
	DB	21H
E50B	MVI	E,0BH	; Directory specification wierd
E5XX	MVI	D,5	; error code
EXXX	XRA	A
	LHLD	SPTR	; load scan pointer
	DCX	H	; Point at last char
	STC		; set carry
	RET		; and split.
	PAGE
; Start the thing. Stash away stuff, and go to it.

GFID	LHLD	FPTR		; FPTR to HL
	SHLD	FPTR1
	MVI	E,43
GFIDZ	INX	H		; Zero file buffer
	MVI	M,0
	DCR	E
	JNZ	GFIDZ
	;
	IF	USERS=1
	LDA	ACRUD
	ANI	80H
	JZ	GFID1
	;
	; Read buffer option specified.
	;
	LXI	H,CONTY ; Setup control-Y path.
	SHLD	PVEC
	LHLD	SPTR	; Prompt string to DE
	XCHG
	LXI	H,RBUF	; our buffer
	SHLD	SPTR	; set that up
	LXI	B,34	; length
	CALL	Rlwe	; go do the read
	PUSH	H
	LXI	H,Ioret
	SHLD	PVEC
	POP	H
	ENDIF		; End of USERS
;
; Start the scan
; Skip over spaces and tabs till we get to text
;
GFID1	CALL	BREAK	; skip spaces and tabs.
	CALL	GFC	; get the next character.
	CALL	CHKDLM  ; Is it < or >?
	CZ	GFC	; If first was delimiter get next.
	MOV	B,A	; Save character as it may get stomped.
	CALL	UNIT	; See if digit or ? or # followed by <.
	MOV	A,B	; Restore the character.
	JNZ	DEFAULT	; If not zero, it wasn't.

; See where to vector to

GFID0	CPI	'?'	; If ? search all drives
	JZ	SEARCH
	CPI	'#'	; If # check # drive
	JZ	SETDRIV
	CPI	'1'
	JC	E500	; If < 1 it's not legal
	CPI	MAXDRIVE+'0'+1 ; Ascii code for maxdrive + 1
	JNC	E500	; If not between 1 and maxdrive
			; it's illegal
	SUI	'0'	; Decode ASCII character
	CALL	PFC	; And stick it where it belongs
	LHLD	FPTR	; Point past length
	INX	H
	SHLD	FPTR
	CALL	GFC	; Get next character
	CALL	CHKDLM	; Should be a delimiter
	JNZ	E500
	JMP	GFID2a	; Find out what to do next

SETDRIV	LHLD	SPTR	; Stow current scan pointer under
	SHLD	SPTR1	; a rock
	LXI	H,DEFPATH ; Where we info on where to go
	SHLD	SPTR	; Scan that turkey!
	CALL	GFC
	JMP	GFID0

DEFAULT	CALL	CHKDLM
	JZ	DF1	; It was leading delimiter.
	LHLD	SPTR	
	DCX	H
	SHLD	SPTR	; Put it back to non-delimiter.
DF1	LDA	SYSRES	; Set up default drive
	CALL	PFC	; Plotz
	LHLD	FPTR	; Save room for length
	INX	H
	SHLD	FPTR
	JMP	GFID2a	; Jump to name parse.

; Start building directory name.

GFID2	LHLD	FPTR	; Get address to put string into
	PUSH	H	; Length goes here later
	INX	H	; Build name after length
	SHLD	FPTR
	CALL	GSTRING	; Get first directory name
	POP	H	; Get where to put length
	JC	E501	; If we got a carry we exceded max len
	MOV	C,A	; Save termination character
	MOV	A,B	; Get length
	ORA	A	; Is length 0?
	JZ	E503
	CPI	31	; Or longer than 31 chars
	JNC	E50A
	MOV	M,A	; If OK insert it.
	MOV	A,C	; Get delimiter
	CPI	'.'	; Did he specify an extension
	JNZ	GFID21	; If not put in .DX extension
	CALL	GSTRING	; Else use the one he specified
	JC	E501	; Oops/ went over limit
	MOV	C,A	; Save termination character
	MOV	A,B	; Get length
	CPI	2	; Extension must be two bytes in length
	JNZ	E502
	MOV	A,C	; Get back termination character
	JMP	GFID2a	; Continue processing
;
; Put in .DX extension
;
GFID21	PUSH	B	; Save termination char
	MVI	A,'D'
	CALL	PFC	; Put the character into the string
	POP	B	; Balance stack in case of error
	JC	E501	; String got too long error
	PUSH	B	; Save termination char again
	MVI	A,'X'
	CALL	PFC
	POP	B	; Get back termination character
	JC	E501
	MOV	A,C	; Termination char to Acum.
GFID2a	LHLD	SPTR	; Tuck FPTR and SPTR away
	PUSH	H
	LHLD	FPTR
	PUSH	H
	LXI	H,0	; Point FPTR where it wont do any
	SHLD	FPTR	; damage ( HA, see fix of PFC)
GFID2b	CALL	GSTRING	; Look past next string
	JNZ	GFID2b
	CPI	'.'	; Skip '.'
	JZ	GFID2b
	CALL	CHKDLM	; Check for '<' or '>'
	POP	H	; Restore pointers
	SHLD	FPTR
	POP	H
	SHLD	SPTR
	JZ	GFID2	; If more indirections go parse them

; If we fall through we are ready to parse file name.

GFID3	LHLD	FPTR	; Get address of end of directory spec
	DCX	H
	XCHG		; Save it in DE
	LHLD	FPTR1	; Get original start of file spec
	XCHG
	CALL	DSUB	; Length to HL
	LXI	D,43	; Max length
	CALL	DCMP	; Are we over?
	JNC	E50A	; If so dive out with directory
			; specification too long error
	LHLD	FPTR
	PUSH	H	; Save address to put length in
	INX	H	; And point past it.
	SHLD	FPTR
	CALL	GSTRING	; Build name
	POP	H	; Get where to put length
	JC	E501	; Too long error
	MOV	C,A	; Save termination character
	MOV	A,B	; Get length
	ORA	A	; 0 length?
	JZ	E503
	CPI	31	; Too big?
	JNC	E501
	MOV	M,A	; If OK insert.
	LHLD	FPTR1	; Find out length of overall FDE
	XCHG
	LHLD	FPTR	
	CALL	DSUB
	DCX	H	; Because FPTR is pointing at one
			; past where we want it
	DCX	H	; Overhead for drive # byte
	MOV	A,L
	LHLD	FPTR1	; And stick it in the length spec
	INX	H
	MOV	M,A
	MOV	A,C	; Get back termination character
	CPI	'.'	; Extension?
	JNZ	GFID6	; No, use default extension
	CALL	GSTRING	; Get extension
	JC	E501	; jmp/went over char limit
	MOV	A,B	; How long was it
	CPI	2	; If not 2 characters he blew it
	JNZ	E502	; Extension too long?
GFID4	LHLD	FPTR1
	MVI	A,7FH	; Tell look to make extension match
	ANA	M
	MOV	M,A
GFID5	XCHG		; FPTR1 to DE
	LHLD	SPTR
	DCX	H
	LDA	ACRUD	; Find out if look was specified
	ANI	40H
	RZ		; If we don't look it up we're done
	CALL	LOOK	; Otherwise look it up
	JC	GFID52	; If error handle it
	LHLD	FPTR1	; Else copy the FDE into (->FPTR)+1
	INX	H
	XCHG		; FDE -> (HL), where it goes -> (DE)
	PUSH	H	; Save pointer to FDE
	CALL	EMOV	; Move it
	LHLD	SPTR	; Set up exit regs.
	DCX	H
	POP	D	; Retrieve old FDE address
	ORA	A	; Undo error flag
	RET		; And split

GFID52	PUSH	PSW	; Save flags
	LXI	H,300H	; If error make sure it was a 300
	CALL	DCMP
	JNZ	GFID53	; If not bomb out
	PUSH	D
	LHLD	FPTR1	; Else plop the NFA onto the FDE
	INX	H	; Get flag byte
	MOV	A,M
	ANI	31	; Mask off flag bits.
	ADI	3	; Point HL at where FDA goes
	MOV	E,A
	MVI	D,0
	DAD	D
	XCHG
	LHLD	NFA	; Get what to put there
	XCHG
	MOV	M,E	; So put it
	INX	H
	MOV	M,D
	POP	D	; Get back error code
GFID53	POP	PSW	; And split reporting error
	LHLD	SPTR
	DCX	H
	RET
GFID6	LDA	ACRUD	; Do we care about the extension?
	ANI	20H
	LHLD	FPTR1
	JNZ	GFID7	; Yes, put in default
	MVI	A,80H	; No, tell look a match is not reqd.
	ORA	M
	MOV	M,A
	JMP	GFID5

GFID7	LDA	DEXT+1	; Get first byte of default extension
	CALL	PFC	; Put it in
	JC	E501	; Error control
	LDA	DEXT	; Second byte
	CALL	PFC
	JC	E501
	JMP	GFID4	; And go back to what we were doing
	PAGE
; LOOK is the generalized file lookup critter as needed by
; Gfid.  It figures out which directory to parse then looks
; up the file.  It expects a file spec as generated by the
; Gfid function to be pointed to by FPTR1.  It returns the
; FDE address in DE.

look	SHLD	FPTR1	; Entry point for external call

LOOK	LHLD	FPTR1	; Point at parsed FDE
	MOV	A,M
	STA	LACRUD
LOOKa	MVI	B,0	; Initialize # of indirections
	INX	H	; Point past drive designation
	MOV	A,M	; Get overall length
	ANI	31
	MOV	C,A
	DCR	C	; Subtract overhead
	INX	H	; Point at first name length
LOOK00	MOV	A,C	; Check it against overall length
	SUB	M
	DCX	H	; Point at beginning of name in case we
			; don't jump.
	JNC	LOOK000	; If we have a carry we were looking at
			; an old type look passed externally.
			; This fails miserably on filenames
			; with the first character a control
			; character.
	SHLD	FPTR
	XRA	A	; So zero out NDIRS
	STA	NDIRS
	PUSH	H
	LXI	H,1	; And update SPECLEN
	SHLD	SPECLEN
	POP	H
	LDA	EFIDFLG	; Were we invoked by EFID?
	ORA	A
	JZ	LOOK010 ; Jmp/ nope, just look up
	XRA	A
	STA	NFDIR
	LHLD	DirAddr	; Else figure out where to get dir.
	LXI	D,SBUF1	; Load it
	LDA	LACRUD
	ANI	DEVMASK
	MOV	C,A
	MVI	B,1
	MVI	A,4
	SHLD	DIRADDR	; Update this before we lose it
	CALL	Dio
	JC	EXXX	; Error control
	CALL	Ckdr	; Checksum it
	MOV	B,A
	LDA	SBUF1
	CMP	B	; Check checksum
	JNZ	E3FF	; If not ok bomb out
	STA	NFCK
	LDA	LACRUD	; Update directory address
	PUSH	PSW	; Save for use in SCAN call.
	ANI	DEVMASK
	ORI	80H
	STA	NFDIR
	LHLD	FPTR	; Point at FDE to match
	POP	PSW	; Restore LACRUD.
	CALL	SCAN
	JC	E300	; If no find scream
	JMP	LOOKOUT	; Else done, split

LOOK000	INX	H	; jumped so reset pointer
	JZ	LOOK001	; If it's the same, we're done
	SUI	3	; Subtract file overhead
	MOV	C,A	; Else update total count
	MOV	A,M	; And point to the next name length
	ANI	31	; Mask off flag bits.
	ADI	3
	MOV	E,A
	MVI	D,0
	DAD	D
	INR	B	; Increment # of indirections count
	JMP	LOOK00	; And try again
LOOK001	MOV	A,B	; Get # of indirections
	STA	NDIRS	; Set up number of indirections counter
	XCHG		; Address of last length spec to DE
	LHLD	FPTR1	; Get beginning of lookup block
	XCHG
	CALL	DSUB	; HL now contains path spec. length
	SHLD	SPECLEN	; Save this for use by EFID
	MOV	B,L	; Length of directory spec to B
	LDA	NFDIR	; Get current directory in RAM
	ANI	80H	; Is it an extension?
	JZ	LOOK01	; If not let Look take care of things
	LDA	NDIRS	; See if we have to worry about 
	ORA	A	; extensions	
	JZ	LOOK01	; If not just look the junk up
	LDA	LACRUD	; Get drive number requested.
	ANI	DEVMASK
	MOV	D,A
	LDA	NFDIR	; Get current drive number.
	ANI	DEVMASK
	CMP	D	; Is PATH on correct disk?
	JNZ	LOOK01	; Jump, it isn't.
	LHLD	FPTR1	; Directory spec to HL
	INX	H	; Point past drive designation
	INX	H	;  and overall length
	DCR	B
	DCR	B	; Compensate length for two INX's.
	LXI	D,PATH	; Current directory spec to DE
LOOK0	LDAX	D	; Get char of current dir spec
	CMP	M	; Is it the same as we have
	JNZ	LOOK01	; If not do directory fetch
	INX	H	; Point to next byte
	INX	D
	DCR	B	; # of chars done?
	JNZ	LOOK0	; If not check some more
	LDAX	D	; Make sure we don't have any more
	ORA	A	;  chars in our current spec else we
	JNZ	LOOK01	;  only had a partial match.
	PUSH	H	; Save pointer to file block.
	LHLD	DirAddr	; Path is correct so DirAddr is too.
	SHLD	DIRADDR	; Update this so it will be correct.
	POP	H	; Restore pointer.
	XRA	A	; If we had the whole thing in, scan
	STA	NDIRS	;  the directory and give the guy the
	JMP	LOOK31	;  results.
LOOK01	LHLD	FPTR1	; File text to look up
	INX	H
	INX	H
LOOK010	SHLD	FPTR	; Update pointer
	LDA	LACRUD	; Get which drive to look on
	MOV	B,A
	LDA	NDIRS	; Is this a directory look
	ORA	A	; If so we need extension match
	MOV	A,B
	JZ	LOOK1	; jmp/ not a directory look
	ANI	7FH	; add in extension match flag
LOOK1	CALL	gLook	; Look in base directory
	LXI	H,0	; Update address where current dir
	SHLD	DIRADDR	; goes
	LHLD	SBUF1+13
	SHLD	NFA	; Update next free address
	MVI	A,0	; Use MVI A,0 to retain flags
	STA	PATH
	JC	LOOK51	; Error control for above Look
LOOK3	LDA	NDIRS
	ORA	A
	JZ	LOOKOUT	; If no extension directories
			; specified, we're done so split
	DCR	A	; If not try again
	STA	NDIRS
	LDAX	D	; Get Name size
	ANI	31	; Mask off length
	ADI	3	; Overhead to get to FILE Address
	MOV	L,A
	MVI	H,0
	DAD	D	; HL now points to address of next
			; directory entry on disk we want
	MOV	E,M	; Put address into HL
	INX	H
	MOV	D,M
	XCHG		; Directory addr into HL.
	SHLD	DIRADDR	; Update directory address
	LXI	D,SBUF1 ; Where to put the junk	
	LDA	LACRUD	; Get drive #	
	ANI	DEVMASK ;  mask off flag
	MOV	C,A	; Stash here for dio
	XRA	A
	STA	NFDIR	; In case of error
	MVI	A,4
	MVI	B,1
	CALL	Dio	; Read 4 sectors
	JC	EXXX	; Error control
	LDA	LACRUD	; Get drive #
	ANI	DEVMASK
	ORI	80H	; Says this is an extension directory
	STA	NFDIR	; Store in current directory
	CALL	Ckdr	; Checksum directory in memory
	MOV	B,A	; Save this
	LDA	SBUF1	; See what this thing thinks its
			; checksum is.
	CMP	B	; Do we agree
	JNZ	E3FF	; If not bomb out
	STA	NFCK	; Update checksum in RAM
	;
	; Update path name
	;
	LHLD	FPTR	; Point to beginning of next name
	MOV	A,M	; Get length
	ANI	31	; Mask off flag bits.
	ADI	3	; Add overhead (length byte +extension)
	MOV	E,A	; Offset to DE
	MVI	D,0
	DAD	D	; Add to HL
	SHLD	FPTR	; Update pointer
	XCHG		; Free up HL
	LHLD	FPTR1	; Find out where we started
	CALL	DSUB	; Should get a negative number suitable
			; for MOVE
	MOV	B,H	; # of bytes to move to BC
	MOV	C,L
	INX	B	; Don't move in the first two bytes
	INX	B
	LHLD	FPTR1	; Where to move form
	INX	H	; Point past drive designation
	INX	H	;  and overall length
	LXI	D,PATH	; Where to move to
	CALL	Moven	; Path name now updated
	XCHG		; Path pointer to HL
	MVI	M,0	; Make sure we're terminated with a 0
	LHLD	FPTR	; Point at text to look up
	LDA	NDIRS	; Get counter
	ORA	A	; Is this a file or a directory?
	JNZ	LOOK4	; It's a directory, so it requires
			; an extension match
LOOK31	LDA	LACRUD	; Get ext match flag
	JMP	LOOK5
LOOK4	XRA	A	; Says make it an exact match
LOOK5	CALL	SCAN	; Scan the directory
	JNC	LOOK3	; If OK continue
	LXI	D,300H	; Error code
LOOK51	LDA	NDIRS	; So we can set zero flag to tell
			; whoever called us whether error
			; 300 was on dir of file
	CALL	LOOKOUT	; File not found
	STC
	RET		; scram.

LOOKOUT LHLD	DIRADDR	; Set up pointer for Dfns
	SHLD	DirAddr
	LHLD	SPTR
	DCX	H
	ORA	A	; Fix error flag
	RET		; And split
	PAGE
; Look up a <?> file....
;	 We will search drives 1-Maxdrive for a match

SEARCH	LDA	ACRUD	; Is look specified?
	ANI	40H
	JZ	E509	; If not no can do.
	LDA	SYSRES	; First drive to check on
	LHLD	SPTR	; Get scan pointer
	DCX	H	; Point to where drive # goes
SEARCH1	PUSH	PSW	; Save drive #
	PUSH	H	; Save scan pointer
	ANI	DEVMASK	; Drives are modulo DEVMASK
	JZ	SEARCH3	; Don't do drive 0.
	ADI	'0'	; Make it ASCII
	MOV	M,A	; Insert drive # into lookup block
	SHLD	SPTR	; Scan starts at it.
	LHLD	FPTR1	; Reset file pointer
	SHLD	FPTR
	CALL	GFID	; Try that drive
	JNC	SEARCH5	; We found it so split		
SEARCH3 POP	H	; Get back beginning of scan pointer
	POP	PSW	; Get back drive # we were checking
	INR	A	; Point to next one
	MOV	C,A	; See if it's the SYSRES drive
	LDA	SYSRES
	CMP	C
	JZ	SEARCH6	; If so didn't find it.
	MOV	A,C	; Else check it
	CPI	MAXDRIVE+1
	JC	SEARCH1	; jump, still < MAXDRIVE+1
	XRA	A	; Start over at 0.
	JMP	SEARCH1
	;
	; Found it routine
	;
SEARCH5	POP	PSW	; junk
	POP	PSW	; Get back drive #
	ANI	DEVMASK	; We're modulo DEVMASK remember?
	ADI	'0'	; Make ASCII
	STA	DEFPATH	; Update default path
	MVI	A,80H	; (Ensure base directory)
	STA	DEFPATH+1
	XRA	A	; Zero carry flag
	RET		; And split
	;
	;	Didn't find it routine.
	;
SEARCH6	MVI	A,'?'
	MOV	M,A	; Put question mark back in string.
	JMP	E300	; Return error 300.
	PAGE
; This routine scans the directory.  It expects to see the 
; lookup block of the file it's looking for pointed to by HL.
; A has 80H bit set if exact match required.
;
; This code is stolen (mostly) from the listing of Root/75, 
; and is attributed to RTM.
;
SCAN	LXI	D,SBUF1+15 ; Start of directory.
SCAN1	PUSH	H
	PUSH	PSW	; And restack it.
	LDAX	D	; Get length of this FDE
	MOV	C,A	; save it here.
	ANI	31	; get just length
	MOV	B,A	; length in B.
	MOV	A,C	; get back the virgin copy.
	ORA	A	; set flags.
	JZ	SCAN7	; End hit/scram
	JM	SCAN3	; Jmp/ deleted.  Skip this one.
	;
	; Compare this file with what we wanted
	;
SCAN2	MOV	A,M	; Get length of wanted file
	ANI	31
	CMP	B	; Same length?
	JZ	SCAN4	; Jmp/ yup
	;
	; Lengths don't agree or deleted file.  Skip to next
	; FDE in the list
	;
SCAN3	MOV	A,B	; Length we hid away
	ADI	11	; + overhead
	MVI	H,0	; Msb of offset to H
	MOV	L,A	; Lsb of offset to L
	DAD	D	; Point at next file's flag byte
	XCHG		; Put it in DE
	POP	PSW
	POP	H
	JMP	SCAN1	; And go check the next FDE
	;
	; Lengths agreed, check names
	;
SCAN4	MOV	C,A	; Number of chars to check
	POP	PSW	; Entry PSW on stack.
	PUSH	PSW	; Get extension check flag back
	PUSH	D	; Save rescan pointer
	ANI	80H	; Check extension check flag
	JNZ	SCAN5	; If no extension match reqd, jmp
	INR	C
	INR	C	; Else check extension as well
	;
	; Check loop
	;
SCAN5	INX	H
	INX	D
	LDAX	D
	CMP	M
	JNZ	SCAN6	; Jmp/ not the one
	DCR	C
	JNZ	SCAN5
	;
	; We found it, restore stuff and split
	;
	POP	D	; -> FDE
	POP	PSW	; junk
	POP	H	; junk
	ORA	A	; clear carry
	RET		; and split.
	;
	; Not the one. Get pointer to FDE start and skip it.
	;
SCAN6	POP	D
	JMP	SCAN3	; jmp/ scan next entry
	;
	; SCAN7 says we didn't find it
	;
SCAN7	POP	PSW	; Restore stuff
	POP	H
	STC		; Set error flag
	RET		; And split
	PAGE
;
; Enter or replace an FDE. HL points to the block to
; deal with.
;
EFID	MVI	A,7FH	; Tell look we need an exact match
	STA	EFIDFLG	; Update this
	ANA	M
	MOV	M,A
	;
	IF	USERS=2
	;
	; Get MUNG access to the device so we can futz with 
	; directory
	;
	PUSH	B
	MOV	C,A		; device #
	MVI	B,0A5H		; temp mung
	CALL	Devlock		; go get it
	POP	B
	JC	EXXX		; scram if errors reported
	ENDIF
	;
	; Force the proper directory in by LOOKing up the file.
	;
	CALL	look	; do it.
	JNC	EFID1	; jmp/looks like it exists.
	;
	; Error 0300 is the only error we allow here.
	; Anything else croaks out.
	;
	LXI	H,300H	; Was it a 300 error (Can't find
	CALL	DCMP	;  that file)
	JNZ	EXXX	; if not 0300, scram.
	LDA	ACRUD
	ORA	A	; see if this is an update
	JM	EXXX	; and croak out if it is...
	JMP	EFID2
;
; The file exists. If this is not update, crap out with
; error 505-file exists.
;
EFID1	LDA	ACRUD
	ANI	80H
	JZ	E505	; oops...file already exists.
	;
	; Replace the FDE
	;
	PUSH	D	; Save directory entry address
	LHLD	SPTR	; Get file spec
	XCHG
	LHLD	SPECLEN
	DAD	D
	POP	D
	PUSH	D
	CALL	EMOV	; copy in new thing
	POP	D
	LHLD	FPTR1
	INX	H
	MOV	A,M
	ANI	0C0H	; Leave system and deleted bits.
	XCHG
	ORA	M	; And OR them into the flag byte
	PUSH	B	; Save BC
	MOV	C,A	; keep flag bits and length in C.
	LDA	ACRUD	; Get original spec.
	ANI	40H	; Clear new specified?
	MVI	A,20H	; Get a new bit.
	JZ	SNEW	; No, jump.
	CMA		; complement new bit
	ANA	C	; Clear new bit and put back in A.
	JMP	SNEW1
SNEW	ORA	C	; set it with other stuff.
SNEW1	POP	B
	MOV	M,A	; Put them into memory.
	JMP	UPDIR	; and update the directory
;
; Enter a new file on the end of the directory. Watch
; out for overflow.
;
; First verify that complete path has been gone
; through and that the last directory is in memory.
;
EFID2	LXI	H,0	; NFA's accurate.
	SHLD	NFAoff
EFIDD	LDA	SPECLEN	; Get the number of chars in the spec.
	CPI	3
	JC	EFID31	; Jmp/ no indirections (or old type)
	SUI	2
	MOV	B,A
	LHLD	FPTR1	; Point at the parsed file name
	INX	H
	INX	H	; (Past drive no and total length)
	XCHG
	LXI	H,PATH	; And at the current in RAM dir path
EFID3	LDAX	D	; Compare the two
	CMP	M	; If not the same we need to create
	JNZ	NEWDIR	; a new directory.
	INX	H
	INX	D
	DCR	B
	JNZ	EFID3
	MOV	A,M	; Insurance, make sure what we want is
	ORA	A	; all we have
	JNZ	E50B	; Jmp/ something is funny, BAIL OUT!
EFID31	LHLD	SBUF1+11 ; end of directory FDE list
	PUSH	H	; that's copy to (save it for now)
	LXI	H,Gfile	; Impossible file name.
	CALL	SCAN	; find it. (returns DE pointing to
			; next available byte in directory.
	POP	H	; Get directory's version.
	CALL	DCMP	; Are they the same?
	JNZ	E3FF	; No, that's a garbage directory.
	PUSH	H	; Save copy to again.
	LHLD	SPTR	; copy from
	XCHG		; Point past all the path stuff
	LHLD	SPECLEN
	DAD	D
	POP	D	; Get back copy to addr.
	PUSH	D
	CALL	EMOV	; so copy the thing
	XCHG		; directory addr back to HL.
	POP	D	; Start of FDE in directory.
	PUSH	H	; Save pointer.
	LHLD	SPTR
	INX	H
	MOV	A,M	; Get any flag bits
	ANI	0C0H	; Leave system and deleted bits.
	XCHG		; And put them into the flag byte
	ORA	M	; OR with length.
	ORI	20H	; Set "new" bit.
	MOV	M,A	; Move flag byte into memory.
	POP	H	; New end of directory.
	;
	;	We return a 504 error when the directory is
	;  ALMOST full.  We leave enough room for a 4-char
	;  directory name.
	;
	LXI	D,DirEnd-15
	CALL	DCMP
	JC	EFID32	; Jmp/ directory's NOT full
	XRA	A
	STA	NFDIR	; Mark directory bad!
	JMP	E504	; oops...directory full!
EFID32	SHLD	SBUF1+11 ; new end point
	MVI	M,0	; pow.
	XCHG		; Save spot in dir in DE again
	LHLD	SBUF1+9
	INX	H
	SHLD	SBUF1+9	; update file count
	;
	; Now fix up NFA pointer. Note DE points past FDE
	;
	LXI	H,-6
	DAD	D	; point at NSCTR in the FDE
	MOV	E,M
	INX	H
	MOV	D,M	; that should be the size
	LHLD	NFAoff
	DAD	D
	SHLD	NFAoff	; update amount NFA's off by.
	LDA	NFDIR
	ANI	80H	; If we're on the base directory
	JZ	EFID33	; we simply have to update NFA.
	;
	CALL	UPDIR	; Write out subdirectory.
	JC	EXXX	; Didn't make it.
	LHLD	DIRADDR	; Update the pointer for the
	SHLD	DirAddr	; disk functions overlays
	MVI	A,0
	STA	PATH	; Kill PATH.
	LHLD	SPTR
	MOV	A,M
	ANI	DEVMASK ; Get device in A
	LXI	H,Gfile
	CALL	gLook	; Force in base directory.
	JNC	EFID33	; Somehow it worked.
	LXI	H,300H
	CALL	DCMP
	JNZ	EXXX	; Jmp/ not 300 error.
	;
EFID33	LHLD	SBUF1+13
	XCHG
	LHLD	NFA	; Check NFA in base directory against
	CALL	DCMP	; what we think it's supposed to be
	JNZ	E3FF	; Jmp/ different, return error.
	XCHG		; NFA to DE
	LHLD	NFAoff	; Amount NFA is off.
	DAD	D	; Correct it.
	SHLD	NFA	; Update NFA and stick in directory.
	SHLD	SBUF1+13
	LXI	H,0
	SHLD	DIRADDR ; Ensure we write at addr 0.
	JMP	UPDIR	; Update it.
	PAGE
;
; Updir is the service provided on the 2-User system.
; (and the source of many problems).
;
	IF	USERS=2
Updir	LHLD	DirAddr
	SHLD	DIRADDR ; use right address, please!
	ENDIF
;
;	UPDIR checksums and writes out the directory in SBUF1 
;  to the NFDIR disk.
;
UPDIR	CALL	Ckdr	; checksum the thing
	STA	SBUF1	; save
	STA	NFCK	; and over here too.
	LDA	NFDIR
	ANI	DEVMASK ; Mask off extension flag
	MOV	C,A	; set device
	LHLD	DIRADDR	; disk address
	LXI	D,SBUF1	; memory address
	MVI	B,0	; write
	MVI	A,4	; 4 sectors
	CALL	Dio
	JC	EXXX	; oops....very bad news!
	;
	IF	USERS=2
	;
	; See if we updated the other guy's NFDIR disk and
	; invalidate it if we did!
	;
	DI
	LDA	NFDIR	; Drive number.
	ANI	7FH	; Mask off subdir indicator.
	MOV	B,A	; Save it.
	LDA	PMASK
	ORA	A
	RZ		; jmp/no other guy around!
	MOV	C,A
	LDA	BRG
	XRA	C
	OUT	BRGEN	; ***** SWAP *****
	LDA	NFDIR	; Other guy's drive number.
	ANI	7FH	; Mask it.
	CMP	B	; Same drive?
	JNZ	Shf1	; jmp/nope, that's not it.
	XRA	A
	STA	NFDIR	; invalidate his directory!
Shf1	LDA	BRG
	OUT	BRGEN	; ***** SWAP ***** back!
	XRA	A
	ENDIF		; End of USERS=2
	;
	RET		; split (NO CARRY).
	PAGE
;
; NEWDIR is the critter which creates a new directory
; When the guy needs one.  It is entered with DE
; pointing at the flag byte and name for the directory
; FDE.  The directory into which this goes is in RAM.
;
NEWDIR	LHLD	SBUF1+11 ; Where to copy to
	XCHG
	MOV	A,M	; Get length byte
	ANI	31	; Mask off flag bits.
	ADI	3	; Figure out how much to move
	PUSH	D	; Save start of this directory FDE
	CALL	EMOV0	; Move that many bytes
	PUSH	D	; Save spot in entry
	LHLD	SPTR	; First find out where on the
	INX	H	; disk to put the new directory
	MOV	A,M
	ANI	31	; Mask off flag bits
	ADI	5	; Point to DNS in FDE
	MOV	E,A	
	MVI	D,0
	DAD	D
	MOV	E,M	; Put DNS in DE
	INX	H
	MOV	D,M
	LHLD	NFAoff	; The place to put the directory is
	DAD	D	; at DNS+NFAoff+NFA!
	XCHG		; DNS+NFAoff
	LHLD	NFA
	DAD	D	; DNS+NFAoff+NFA in HL.
	;
	; This gives us the next free address where a new
	; directory can go.  So now insert the entry for that
	; directory in the current one.
	;
	POP	D	; Get back address where DFA goes
	PUSH	D
	XCHG
	MOV	M,E	; Insert DFA
	INX	H
	MOV	M,D
	INX	H
	MVI	M,4	; Insert DNS
	INX	H
	MVI	M,0
	INX	H	; Point to LA
	MVI	B,4
NEWDIRa MVI	M,01	; Squirt 0101 into LA and SA
	INX	H
	DCR	B
	JNZ	NEWDIRa
	SHLD	SBUF1+11 ; Update NFDA
	MVI	M,0	; Insurance
	LXI	D,DirEnd
	CALL	DCMP	; Check for overflow
	JNC	E504
	LHLD	SBUF1+9	; Get # of files
	INX	H	; Update
	SHLD	SBUF1+9
	CALL	UPDIR	; Go write that one out
	POP	H	; Clean stack in case of error
	POP	B	; Get back beginning of FDE entry too
	JC	EXXX	; Error control
	PUSH	B	; Save this again
	MOV	E,M	; Put address where new directory
	INX	H	; is to go on disk in DE
	MOV	D,M
	XCHG
	SHLD	DIRADDR	; Store that addr in our update addr
	LXI	H,PATH-1
NEWDIR0 INX	H	; Move along entry until we reach
	MOV	A,M	; The end of where we are
	ORA	A
	JNZ	NEWDIR0
	XCHG		; Put that address in DE
	POP	H	; Get beginning of new dir name
	MOV	A,M	; Get name length
	ANI	31	; Mask off flag bits.
	ADI	3	; Add overhead
	CALL	EMOV0	; Update this
	XRA	A
	STAX	D	; Make sure a 0 terminates it
	;
	; Clear the directory area in preparation to making 
	; a new one
	;
	LXI	H,SBUF1 	; Zero starting here
NEWDIR1 MVI	M,0		;ZAP!
	INX	H
	MOV	A,H
	CPI	DirEnd/256	; Are we done?
	JNZ	NEWDIR1 	; Keep spinning till we are
	LXI	H,SBUF1+15	; Where our first entry goes
	SHLD	SBUF1+11	; Shovel it into here
	CALL	Ckdr		; Checksum this mess
	STA	SBUF1		; And update directory checksum
	STA	NFCK		; Stow it in current chksm. too
	;
	; Update NFAoff so that the new directory we just
	; created won't get written upon.
	;
	LHLD	NFAoff
	LXI	D,4	; Directory size.
	DAD	D
	SHLD	NFAoff	; Update amount of error in NFA
	LDA	NFDIR	; Tell system base directory is no
	ORI	80H	;  longer in RAM
	STA	NFDIR
	JMP	EFIDD	; And try again
;
; Copy the FDE -> by (HL) to the spot -> (DE)
;
EMOV	MOV	A,M
	ANI	31	; strip to length
	ADI	11	; copy whole FDE
EMOV0	MOV	B,A
EMOV1	MOV	A,M
	STAX	D
	INX	H
	INX	D
	DCR	B
	JNZ	EMOV1
	RET		; and spit when done.
;
DCMP	MOV	A,H
	CMP	D
	RNZ
	MOV	A,L
	CMP	E
	RET
;
DSUB	MOV	A,D
	CMA
	MOV	D,A
	MOV	A,E
	CMA
	MOV	E,A
	INX	D
	DAD	D
	RET
	PAGE
;
;	TEXT and CONSTANTS.
;
Gfile	DB	1,19H,0 ; Impossible file name
;
	IF	USERS=1
ESP	DS	2	; entry stack pointer
RBUF	DS	35
	ELSE
RBUF0	DS	35	; input buffer user 0
RBUF1	DS	35	; input buffer user 1
OldEFLG1 DS	1
	ENDIF
;
DEXT	DS	2	; extension
SPTR	DS	2	; scan/spec pointer
SPTR1	DS	2
FPTR	DS	2	; file block pointer
FPTR1	DS	2
DIRADDR DS	2
ACRUD	DS	2	; A contents.
LACRUD	DS	2
NDIRS	DS	2
SPECLEN DS	2
EFIDFLG DS	2
NFAoff	DS	2
;
;	All done.
;
	ORG	CODE+800H	; Check size.
	END
