;********************************************************
;*							*
;*			Pack.OV				*
;*							*
;*	Version 2.0			Frank Anderson	*
;*	04/28/81					*
;*							*
;*	Copyright (c) 1981, Interactive Products	*
;*  Corporation dba PolyMorphic Systems 460 Ward Dr.	*
;*  Santa Barbara, CA  93111				*
;*							*
;********************************************************
;
; PACK code for directory tree structure.
;
; Last updated April 10, 1980 by Frank E. Anderson	
;
; 8/20/80  1.2	BFS	Change number of sectors calculation
; 			so that a strange MEMTOP (not xxFF)
;			will still work and not be written
;			past to a page boundary.
; 01/29/81 1.3	BFS	No devmask strikes again, changed
;			ANI 7 to ANI DEVMASK.
; 04/17/81 1.4  BFS	Check for table overflow on too
;			many files. The bug that ate Chicago.
; 04/24/81 2.0	FEA	Rewrote the pack algorhythm in order
;			to allow it to handle more files.  In
;			addition the new program will do a
;			"Super Sniff" on all areas of data it
;			is going to move before it alters
;			anything on the disk.  This should
;			allow easier recovery from errors
;			encountered during packing.
; 05/18/81 2.0  DKM	Add German Option

TEST	EQU	0
;
; Equate table
;
	REFS	SYSTEM
;
	REF	DONT	; Shuts off ^Y etc.
	REF	Warm	; Warm start executive
	REF	CMPTR	; Exec command line scan pointer
	REF	NFDIR	; Current directory # in RAM
	REF	OVRLY	; Where overlay area exists
	REF	Ovrto	; Fetch me an overlay
	REF	MEMTOP	; Top of usable RAM
	REF	SYSRES	; Where # of system drive is stored
	REF	Dio	; Generalized Disk I/O routine
	REF	DEFPATH	; What "#" means
	REF	DEVMASK ; Number legal devs.
	REF	CMDF	; Command file in progress?
	REF	CMDD	; Command file drive #
	REF	CMDA	; Command file disk address
	REF	WH1	; Print a character
	REF	Ioret	; Interupt return
	REF	PVEC	; Control Y vector
	REF	UVEC	; User interupt
	REF	SCHR	; Front panel character
	REF	USERS	; For single or two user system.
	REF	SBUF1	; Directory area.
	REF	Ckdr	; Checsum the directory area function
	REF	USER	; Start of USER RAM
	REF	Msg	; Print message to screen
	REF	Deout	; Print DE in hex to screen

	IF	USERS=2
	REF	Devlock	; Device lock routine.
	ENDIF

TAB	EQU	9	; ASCII horizontal tab char
CR	EQU	0DH	; ASCII carriage return
;
	IF	TEST
	ORG	USER
	ELSE
	ORG	OVRLY		; Where we live and run
	ENDIF

	IDNT	$,$		; Load and start addresses
;
	IF	NOT TEST
	DB	'Pack'		;Name (overlays gotta have one)
	ENDIF
;
PACK	LHLD	PVEC		; Save old PVEC
	PUSH	H
	LHLD	UVEC		; Same with old UVEC
	PUSH	H
	LDA	SCHR		; Front panel character
	PUSH	PSW
	IF	NOT TEST
	LXI	H,Ioret		; Point vectors here
	SHLD	PVEC		; Take that control Y
	SHLD	UVEC		; You too user interupt
	XRA	A		; Front panel character = 0
	STA	SCHR		; User can't get to me now
	ENDIF
	LXI	H,0		; Save stack pointer on entry
	DAD	SP
	SHLD	ESP
	XRA	A		; Invalidate directory in RAM
	STA	NFDIR
	CMA
	IF	NOT TEST
	STA	DONT		; Don't bug me
	ENDIF
	CALL	GDN		; Get drive number
	LXI	H,FILTAB	; Initialize file table
	SHLD	FILTAB0
	LHLD	MEMTOP
	SHLD	DIRTAB0		; Initialize dir addr table
	LXI	D,0		; First dir address on disk
	MOV	M,D
	DCX	H
	MOV	M,E
	DCX	H
	SHLD	DIRTAB1		; Save as extent of dir addr
				; table
	CALL	GETDIR		; Bring a directory into memory
	LHLD	SBUF1+0DH	; Get old NFA
	SHLD	OLDNFA		; Save it for later
LOADIR	LXI	H,SBUF1+15	; Point to start of files
	;
	; See if file is deleted
	;
LOADIR0	MOV	A,M		; Get file flag byte
	ORA	A		; End hit?
	JZ	LOADIR2		; Jmp/ end of this directory
	ANI	80H		; Deleted?
	JNZ	LOADIR1		; jmp/ deleted
	PUSH	H		; Save FDE address
	MOV	A,M
	ANI	31		; Get file name length
	ADI	3		; Add offset to point to FDA
	MOV	E,A
	MVI	D,0
	DAD	D
	XCHG			; Point to non-deleted file
	LHLD	FILTAB0		;  table with HL.  DE now
				;  points to FDA.
	LXI	B,4		; # of bytes to transfer
	XCHG			; format to move
	CALL	MOVE		; Move in the FDA and the FNS
	XCHG			; New table address to HL
	SHLD	FILTAB0		; Save here for future refr.
	PUSH	D		; Save pointer to FDE
	XCHG			; Non-deleted tab pointer to DE
	LHLD	DIRTAB1		; Compare to extent of dir
	CALL	DCMP		;  table.
	XCHG			; Non-deleted tab pointer to HL
	LXI	D,0206H		; Error code (assume error)
	JC	OHNO		; jmp/ not enough memory error
	POP	D		; Get back in use FDE pointer 
	LXI	H,-6		; Point to extension
	DAD	D
	XCHG			; Extension pointer to DE
	POP	H		; Get back pointer to FDE
	LDAX	D		; Get first char of extension
	CPI	'D'		; Was it a D?
	JNZ	LOADIR1		; jmp/ can't be a .DX file
	INX	D
	LDAX	D		; Check second character
	CPI	'X'		; Was it an X?
	JNZ	LOADIR1		; jmp/ not a .DX file
	;
	; Else this was a directory, so enter its address
	; into the directory address table.
	;
	INX	D
	INX	D
	PUSH	H		; Save start of FDE pointer
	LHLD	DIRTAB1		; Get extent pointer
	LDAX	D		; Enter address in
	MOV	M,A
	DCX	D
	DCX	H
	LDAX	D
	MOV	M,A
	DCX	H
	SHLD	DIRTAB1		; Save new extent pointer
	POP	H		; Get back FDE pointer
	;
	; Entry point for a non-deleted or a non-.DX file
	;
LOADIR1	MOV	A,M		; Bump to next FDE
	ANI	31
	ADI	11		; FDE overhead
	MOV	E,A
	MVI	D,0
	DAD	D		; HL -> next FDE
	JMP	LOADIR0		; jmp, check next FDE
	;
	; Entry point for end of current directory scan
	;
LOADIR2	LHLD	DIRTAB1		; Get directory table extent
	XCHG			; Put it into DE
	LHLD	DIRTAB0		; Get current dir table base 
	CALL	DCMP		; Are they the same?
	JZ	LOADIR3		; jmp/ yup, all directorys
				;  have been scanned.
	CALL	GETDIR		; Get next directory
	JMP	LOADIR		; jmp/ go do next directory
	;
	; Entry point when all directories have been scanned
	; and the table of non-deleted files has been built.
	;
LOADIR3	MVI	A,0FFH		; Termination character
	LHLD	FILTAB0		; Where to put it
	MOV	M,A		; Put termination character
	INX	H		;  onto the end of the
	MOV	M,A		;  undeleted file table
	INX	H
	SHLD	DELTAB0		; Where to put the deleted
				;  sectors table
	SHLD	DELTAB1		; Scan pointer for above table
	;
	; Build table of deleted sectors from table of
	; undeleted sectors.
	;
BLDDEL	LXI	D,4		; Check to see that the first
	JMP	BLDDEL1		; sector isn't deleted
BLDDEL0	PUSH	H		; Save scan pointer
	INX	H		; Point at size
	INX	H
	MOV	A,M		; 0 sectors?
	INX	H
	ORA	M
	POP	H		; Restore pointer
	JZ	BLDEL2a		; jmp/ skip 0 sector files
	MOV	E,M		; Load address into DE
	INX	H
	MOV	D,M
	INX	H
	MOV	C,M		; Load file size into BC
	INX	H
	MOV	B,M
	INX	H
	XCHG			; Address to HL
	DAD	B		;Addr of next file if it exists
	XCHG			; New compare address to DE
BLDDEL1	LXI	H,FILTAB
BLDDEL2	MOV	C,M		; Search table for match
	INX	H
	MOV	B,M
	CALL	CmpBCDE
	DCX	H		; Point to beginning of entry
				;  in case
	JZ	BLDDEL0		; jmp/ it matches, use as
				;  next base
	PUSH	D		; Check for end hit
	LXI	D,-1
	CALL	CmpBCDE
	POP	D
	JZ	BLDDEL3		; If end hit it's not in table
BLDEL2a	INX	H		; Else check next entry
	INX	H
	INX	H
	INX	H
	JMP	BLDDEL2
	;
	; Entry point for a deleted file
	;
BLDDEL3	LHLD	DELTAB1		; Enter into del sects table
	MOV	M,E
	INX	H
	MOV	M,D
	INX	H
	PUSH	H		; Save while we find out how
				; many sectors are deleted
	;
	; Find the next lowest undeleted entry
	;
	LXI	H,-1		;Set up initial next lowest reg
	SHLD	NLOWST
	LXI	H,FILTAB 	; Point at where to start scan
BLDDEL4	MOV	C,M		; Get an entry and put it in BC
	INX	H
	MOV	B,M
	INX	H		; Point at next entry
	INX	H
	INX	H
	CALL	CmpBCDE		; Compare to beginning of
				;  deleted file
	JC	BLDDEL4		; If less than or equal to old
	JZ	BLDDEL4		; address we aren't interested
				; in it
	PUSH	D		; See if it's an end hit
	LXI	D,-1
	CALL	CmpBCDE
	POP	D
	JZ	BLDDEL6		; If an end hit clean up
				;  and leave
	PUSH	H		; Compare to our current
	PUSH	D		; next lowest
	LHLD	NLOWST		; undeleted sector
	MOV	D,B
	MOV	E,C
	CALL	DCMP
	JC	BLDDEL5		; jmp/ it's bigger than the
	XCHG			;  one we've already got
	SHLD	NLOWST		; Else update
BLDDEL5	POP	D		; Restore address of first
	POP	H		;  deleted sector and pointer
				;  into table
	JMP	BLDDEL4		; Check the next entry
BLDDEL6	LHLD	NLOWST		; Get the next lowest undeleted
				;  sector
	PUSH	H		; Save for later
	PUSH	D
	LXI	D,-1		; If NLOWST is a -1 we are done
	CALL	DCMP		; building table,
	POP	D
	JZ	BLDDEL7		; so terminate it.
	CALL	DSUB		; Subtract from first del addr 
	MOV	B,H		; Size to BC
	MOV	C,L
	POP	D		; Get back NLOWST
	POP	H		; Get back table pointer
	MOV	M,C		; Make size entry into table
	INX	H
	MOV	M,B
	INX	H
	SHLD	DELTAB1		; And update pointer into
				;  deleted table
	PUSH	D		; Save new sector to check
	XCHG			; Deleted table pointer to DE
	LHLD	DIRTAB1		; Compare to directory table
	CALL	DCMP		; Overflow?
	LXI	D,0206H		; error code (assume error)
	JC	OHNO		; jmp/ ran out of memory
	POP	D		; Get back sector to check
	JMP	BLDDEL1		; And return to table building
				;  routine
BLDDEL7	POP	H		; Throw-away to clean stack
	POP	H		; Get back table pointer
BLDDEL8	MVI	M,0FFH		; Terminate table
	INX	H
	MVI	M,0FFH
	;
	; Update NFA
	;
	DCX	H
	DCX	H		; Get last addr entry in table
	MOV	D,M
	DCX	H
	MOV	E,M
	DCX	H
	PUSH	D		; Save for subtraction
	LXI	B,-1		; What to compare to for
				;  end of table
NFA	MOV	D,M		; Load # deleted sects in this
	DCX	H		; block into DE
	MOV	E,M
	DCX	H		; Point to next # deleted
	DCX	H		;  sects entry in deleted
	DCX	H		;  table (scanning down).
	CALL	CmpBCDE		; Check for end of table
	JZ	NFA1		; If done bail out
	XTHL			; Else switch scan pointer with
				;rnd sect. # to update on stack
	CALL	DSUB		; Subtract off # deleted sects.
	XTHL			; Switch result with scan ptr
	JMP	NFA		; And go scan next entry
NFA1	POP	H		; Get updated NFA
	SHLD	NEWNFA		; Save it
	;
	; Move the deleted table down over the old non-
	; deleted table to free up space for a buffer.
	;
	LHLD	DELTAB0		; First find the extent of the
	PUSH	H		; deleted sectors table
	LXI	B,0FFFFH
EXTENT	MOV	E,M
	INX	H
	MOV	D,M
	INX	H
	CALL	CmpBCDE		; End hit
	JNZ	EXTENT		; jmp/ nope, not at the end yet
	POP	D		; Put beginning into DE
	CALL	DSUB		; HL contains length
	MOV	B,H		; Put length into BC
	MOV	C,L
	PUSH	B		; Save # bytes
	XCHG			; Copy "from" to HL
	LXI	D,BUFFER	; Copy "to" to DE 
	CALL	MOVE		; Shovel the stuff down
	POP	B		; Figure out where the new
	LXI	H,BUFFER	; end of the deleted sectors
	DAD	B		; table is.
	SHLD	BUFST		; Save as buffer start
	;
	; Do a "super sniff" do find out if disaster would
	; have struck.
	;
	XCHG
	LXI	H,BUFFER	; Fix the pointer to the start
	SHLD	DELTAB0		; First figure out how big the
				; of the deleted sectors table.
	LHLD	DIRTAB1		; buffer is that we can read
	CALL	DSUB		; into.
	MOV	A,H
	STA	BUFSECS		; Save over here
	LHLD	BUFFER+2	; Get first size entry in
				; deleted sectors table
	LXI	D,-1		; Is it a termination?
	CALL	DCMP
	JZ	UPDIR		; jmp/ no data shuffling to do
	LHLD	BUFFER		; Else get base of data area
				; where shuffling will go on
	SHLD	GETPTR		; Save as first data area to do
	XCHG
	LHLD	OLDNFA		; Find out how much data I
	CALL	DSUB		; have to look at.
	SHLD	OLDNFA		; We'll use this area to store
				; remaining sectors to check.
	;
	; Find out if the remaining stuff to check is bigger
	; than our buffer area
	;
SSniff	LDA	BUFSECS		; Get the number of sects we
	MOV	E,A		; can do and put it as a 16
	MVI	D,0		; bit quantity in DE
	CALL	DCMP		; Compare to remaining stuff
	JC	LstSnif		; jmp/ last one to do
	JZ	LstSnif		; jmp/ same deal
	LDA	BUFSECS		; Else grab this much
	PUSH	D		; save this offset quantity
	CALL	SniffIO		; Try to read.  If we return
				; all is OK.
	POP	D		; Get back 16 bit BUFSECS
	LHLD	GETPTR		; Update disk address to start
	DAD	D		; sniff I/O
	SHLD	GETPTR
	LHLD	OLDNFA		; Subtract # of sectors we just
	CALL	DSUB		; checked from what remains to
	SHLD	OLDNFA		; be checked
	JMP	SSniff		; Loop, Sniff another block
	;
LstSnif	MOV	A,L		; Remaining # of sects to do
	CALL	SniffIO		; Try them
	JMP	UPDIR		; Done sniffing, do next thing
	;
	; This code does the actual I/O.  It is entered with
	; with the number of sectors to do in the Accum.
	;
SniffIO	LHLD	BUFST		; Memory area to read into
	XCHG			; Put mem pointer into DE
	LHLD	GETPTR		; and disk addr pointer into HL
	PUSH	PSW		; Save number of sectors
	LDA	DDR		; Get drive number
	ANI	DEVMASK
	MOV	C,A
	MVI	B,1		; Read command
	POP	PSW		; Get back number of sectors
	CALL	Dio		; Try to read it
	RNC			; If OK split
	;
	; If we fall through we have to warn the poor slob
	; that disaster is imminent if he packs, so we STOP!
	;
	LXI	H,NoWorky	; Tell the guy what happened
	CALL	Msg
	JMP	SPLIT		; Split unlocking world
	;
	; Now update all the directories 
	;
UPDIR	LHLD	MEMTOP		; Reset directory get pointer
	SHLD	DIRTAB0
	;
	; Load the next directory in the table into memory
	; and checksums the directory.
	;
UPDIR0	LHLD	DIRTAB0		; Dir addr get pointer
	MOV	D,M		; Get the address
	DCX	H		; Put addr into DE
	MOV	E,M
	XCHG			; Directory addr to HL
	LXI	D,SBUF1		; Where to put it
	LDA	DDR		; From which drive
	ANI	DEVMASK
	MOV	C,A
	MVI	A,4		; Get 4 sectors
	MVI	B,1		; Read command
	PUSH	H		; Save Disk address
	CALL	Dio		; Get it with error control
	POP	H		; Retrieve disk address
	MVI	A,4		; We were doing 4 sectors
	JC	GETDIR1		; Error control for Dio
	PUSH	H		; Save disk address
	CALL	Ckdr		; Checksum the directory
	LXI	H,SBUF1
	CMP	M		; Does it agree
	LXI	D,03FFH		; error code (assume error)
	POP	H		; Get back disk address
	JZ	GETDIR2		; All's cool, split 
GETDIR1	MOV	A,H		; Find out if we blew it on the
	ORA	L		; first directory
	JZ	DIRERR		; jmp/ report this error since
				; no data has been altered yet
	MVI	A,4		; Say we were doing 4 sectors
	JMP	FATAL		; jmp/ disaster time
	;
	; Now update the entries in the directory
	;	
GETDIR2	LXI	H,SBUF1+15	; Point at next FDE
UPDIR1	PUSH	H		; Save beginning address of FDE
	MOV	A,M		; Get flag byte
	ORA	A		; End hit?
	JZ	UPDIR5		; jmp / end hit
	ANI	80H		; Deleted?
	JNZ	UPDIR4		; jmp/ deleted
	MOV	A,M		; Point at FDA
	ANI	31
	ADI	3
	MOV	E,A
	MVI	D,0
	DAD	D
	MOV	E,M		; Put FDA in DE
	INX	H
	MOV	D,M
	PUSH	H		;Save where we will be updating
	LXI	H,0		; Initialize # of deleted sects
	SHLD	GUNGE		;  under me to 0
	LHLD	DELTAB0		; Point HL at deleted table
UPDIR2	MOV	C,M		; Pick up address in table
	INX	H
	MOV	B,M
	INX	H
	CALL	CmpBCDE		; Is del addr less than FDA?
	JNC	UPDIR3		; jmp/ NOT less
	MOV	C,M		; Size to BC
	INX	H
	MOV	B,M
	INX	H		; Point to next entry in table
	PUSH	H		; Save pointer
	LHLD	GUNGE		; Update # sects under me
	DAD	B
	SHLD	GUNGE
	POP	H		; Restore pointer
	JMP	UPDIR2		; Check next entry
UPDIR3	LHLD	GUNGE		; # del sectors under me to HL
	XCHG			; # sects to DE, FDA to HL
	CALL	DSUB		; HL now contains new FDA
	XCHG			; DE now has it
	POP	H		; Get back pointer to FDA
	MOV	M,D		; Update FDA
	DCX	H
	MOV	M,E
UPDIR4	POP	H		; Get back pointer to FDE
	MOV	A,M		; Get flag byte
	ANI	31		; Mask off size
	ADI	11		; Add total overhead
	MOV	E,A		; Offset to next FDE to DE
	MVI	D,0
	DAD	D
	JMP	UPDIR1		; Look at next FDE
	;
	; Entry point for end of directory
	;
UPDIR5	POP	H		; Clean stack
	;
	; Pack deleted entrys out of directories
	;
	LXI	H,SBUF1+15	; Point at first entry
DIRPAK2	MOV	A,M		; Get flag byte
	ORA	A		; End hit?
	JZ	DIRPAK4
	ANI	80H		; Deleted
	JZ	DIRPAK3		; jmp/ nope point to next entry
	MOV	A,M
	ANI	31		; Get size of deleted FDE
	ADI	11
	PUSH	H		; Save move to address
	PUSH	PSW		; Save offset
	MOV	E,A
	MVI	D,0
	DAD	D		; Calculate move from addr
	PUSH	H		; Save that
		;
		; Calculate how many bytes to move down
		;
	LXI	H,SBUF1+0BH	; Where NFDA is
	MOV	E,M		; Get NFDA
	INX	H
	MOV	D,M
	POP	B		; Retrieve offset #
	POP	PSW		; While saving move from addr
	PUSH	B
	MOV	C,A		; Make it a 16 bit # in BC
	MVI	B,0
	PUSH	D		; Save old NFDA
	XCHG			; Old NFDA to HL
	MOV	D,B		; Put offset into DE
	MOV	E,C
	CALL	DSUB		; HL now contains new NFDA
	XCHG			; Pointer to NFDA to HL
	LXI	H,SBUF1+0CH	; Point at NFDA
	MOV	M,D		; Update NFDA
	DCX	H
	MOV	M,E
	DCX	H		; Update number of files on dir
	MOV	D,M
	DCX	H
	MOV	E,M
	DCX	D		; NF = NF-1
	MOV	M,E		; Update
	INX	H
	MOV	M,D
	POP	H		; Get back old NFDA
	LXI	D,-2800H	; Subrtact off offset
	DAD	D
	POP	D		; Retrieve move from address
	PUSH	D		; And restack both it and
	PUSH	H		;  stripped end address
	LXI	H,SBUF1
	XCHG			;  directory address
	CALL	DSUB		; Subtract off offset
	POP	D		; Get back stripped NFDA
	XCHG
	CALL	DSUB		; HL now contains #
				;  bytes to move
	INX	H		; Move one extra byte so we
				;  don't hang if we're at the
				;  end of a directory
	MOV	B,H		; # bytes to BC
	MOV	C,L
	POP	H		; Get from addr.
	POP	D		; Get to addr.
	PUSH	D		; We'll need this later
	CALL	MOVE		; Shovel it
	POP	H		; Get back scan addr.
	JMP	DIRPAK2		; And continue
	;
	; Entry point for undeleted file
	;
DIRPAK3	MOV	A,M		; Get flag byte
	ANI	31
	ADI	11		; Add total overhead
	MOV	E,A
	MVI	D,0
	DAD	D
	JMP	DIRPAK2
	;
	; Entry point for end of directory
	;
DIRPAK4	LHLD	SBUF1+0BH	; Starting addr to zero to HL
	MVI	A,(SBUF1+1024)/256	; Ending address for
					; comparison
DIRPAK5	MVI	M,0		; Zero the junk
	INX	H
	CMP	H		; End?
	JNZ	DIRPAK5		; Nope/spin
	;
	; Checksum directory
	;
	CALL	Ckdr		; Checksum it
	STA	SBUF1		; Save it
	;
	; Now write the directory out to the disk
	;
	LHLD	DIRTAB0
	MOV	D,M		; Disk address to DE
	DCX	H
	MOV	E,M
	DCX	H
	SHLD	DIRTAB0
	XCHG			; Disk addr to HL,
	MOV	A,H		; Base directory?
	ORA	L
	JNZ	WRITDIR		; jmp/ nope
	PUSH	H		; Save disk addr
	LHLD	NEWNFA		; Else insert new NFA
	SHLD	SBUF1+0DH
	CALL	Ckdr		; Re-checksum the thing
	STA	SBUF1
	POP	H		; Get back disk addr and contin
WRITDIR	LXI	D,SBUF1		;  mem addr to DE
	LDA	DDR		; Get device #
	MOV	C,A		; Device # to C
	MVI	B,0		; Write command
	MVI	A,4		; 4 sectors
	PUSH	H		; Save disk addr
	CALL	Dio		; Do dio 
	POP	H		; Get back disk address
	MVI	A,4		; Say we were doing 4 sectors
	JC	FATAL		; Error control for Dio
	LHLD	DIRTAB0
	XCHG
	LHLD	DIRTAB1	
	CALL	DCMP		; End hit?
	JNZ	UPDIR0		; jmp/no, do the next directory
	;	
	; Directories rebuilt and rewritten to the disk.
	; Now comes the time to shovel disk data
	;
SHOVEL	LHLD	BUFST		; Get start of buffer area
	XCHG			; Put into DE
	LHLD	MEMTOP		; Put memtop into HL
	CALL	DSUB		; HL = Buffer memory available
	MOV	A,H
	STA	BUFSECS		; Save number of sectors.
	LHLD	DELTAB0		; Initialize put pointer
	MOV	E,M
	INX	H
	MOV	D,M
	INX	H
	XCHG
	SHLD	PUTPTR
	XCHG			; Initialize get pointer
	MOV	C,M
	INX	H
	MOV	B,M
	INX	H
	PUSH	D		; Check to see if we have any
	LXI	D,-1		; stuff to shovel
	CALL	CmpBCDE
	POP	D
	JZ	EXIT		; If not just split
	SHLD	SPTR		; Save scan pointer
	XCHG
	DAD	B
	SHLD	GETPTR
SHOVEL1	MVI	B,0		; Initialize # sectors counter
	LDA	BUFSECS
	MOV	C,A
	LHLD	BUFST		; Initialize mem addr.
	SHLD	MPTR
SHOVEL2	LHLD	SPTR		; Deleted sectors scan pointer
	MOV	E,M		; Get next deleted sector addr
	INX	H
	MOV	D,M
	LHLD	GETPTR
	XCHG
	CALL	DSUB
	MOV	A,C		; How many sectors left?
	SUB	B
	PUSH	PSW		; Save this
	MOV	E,A		; Make result into 16 bit
	MVI	D,0		;  word in DE
	CALL	DCMP		; Compare with number in this
				;  undeleted block.
	JZ	SHOVEL4		; jmp/ # sectors in buffer = #
				; to next deleted area 
	JNC	SHOVEL5		; jmp/ # sectors in buffer < #
				; to next deleted area, ELSE
				; fall through >
	POP	PSW		; Throw away
	PUSH	B		; Save # sects remaining
	LDA	DDR		; Set up registers for Dio
	MOV	C,A		; Drive # to C
	MOV	A,L		; # of sectors to A
	LHLD	MPTR		; Memory pointer to DE
	XCHG
	LHLD	GETPTR		; Get pointer to disk data
	PUSH	D
	PUSH	H		; Save disk address
	MOV	B,A		; # sectors
	PUSH	B		; Save for error control
	MVI	B,1		; Read command
	CALL	Dio		; Do Dio 
	POP	B		; Get back # sectors
	MOV	A,B		; Put in accum
	POP	H		; Get back disk address
	JC	FATAL		; Error control for Dio
	MOV	D,A		; Multiply by 256 and put in DE
	MVI	E,0
	POP	H		; Update memory pointer
	DAD	D
	SHLD	MPTR
	POP	B		; Update # sects read
	ADD	B
	MOV	B,A
	CALL	SHOVEL3		; Get next Undeleted sect ptr
	JNZ	SHOVEL2		; If no termination continue
	MOV	A,B		; Figure out how much to write
	CALL	DIOW		; Write the junk out
EXIT	LDA	CMDF		; Command file in progress?
	ORA	A
	JZ	EXIT1		; jmp/ nope just split
	LXI	H,CMDD		; Else is it on this drive?
	LDA	DDR
	CMP	M
	JNZ	EXIT1		; jmp/ nope just split
	;
	; If we fell through we need to adjust the disk address
	; pointer for the command file.
	;
	LHLD	CMDA		; Get old disk address
	XCHG
	LXI	H,0		; Initialize # of deleted sects
	SHLD	GUNGE		;  under me to 0
	LHLD	DELTAB0		; Point HL at deleted table
UPCMD2	MOV	C,M		; Pick up address in table
	INX	H
	MOV	B,M
	INX	H
	CALL	CmpBCDE		; Is del addr less than FDA?
	JNC	UPCMD3		; jmp/ NOT less
	MOV	C,M		; Size to BC
	INX	H
	MOV	B,M
	INX	H		; Point to next entry in table
	PUSH	H		; Save pointer
	LHLD	GUNGE		; Update # sects under me
	DAD	B
	SHLD	GUNGE
	POP	H		; Restore pointer
	JMP	UPCMD2		; Check next entry
UPCMD3	LHLD	GUNGE		; # del sectors under me to HL
	XCHG			; # sects to DE, FDA to HL
	CALL	DSUB		; HL now contains new FDA
	SHLD	CMDA		; Save as command file address
;
EXIT1	POP	PSW		; Get back front panel char
	STA	SCHR		; Restore it
	POP	H		; Get back user vector
	SHLD	UVEC		; Restore it
	POP	H		; Get back control Y vector
	SHLD	PVEC		; Restore it
	MVI	A,0		; Update don't flag
	STA	DONT
	RET			; And return to Exec!
	;
SHOVEL3	LHLD	SPTR		; Update fetch pointer by
	MOV	E,M		; finding next undeleted sector
	INX	H
	MOV	D,M		; deleted sector # to DE
	INX	H
	PUSH	D		; Save
	MOV	E,M		; get # deleted sectors
	INX	H
	MOV	D,M
	INX	H
	PUSH	B		; Check for table termination
	LXI	B,-1
	CALL	CmpBCDE
	POP	B
	SHLD	SPTR		; New scan pointer
	POP	H		; Retrieve deleted sector #
	DAD	D		; add to # deleted sectors
	SHLD	GETPTR		; = next undeleted sector
	RET
	;
	; Routine for # sects remaining in buffer = to # of
	; sects till next deleted area
	;
SHOVEL4	LDA	DDR		; Set up regs for Dio read 
	MOV	C,A		; Drive # to C
	POP	PSW		; Get back # sects
	LHLD	MPTR		; Memory address to DE
	XCHG
	LHLD	GETPTR		; Disk address to HL
	MOV	B,A		; Save # of sectors
	PUSH	B
	PUSH	H		; Save disk address
	MVI	B,1		; Read command
	CALL	Dio
	POP	H		; Retrieve disk address 
	POP	B		;  and # sectors
	MOV	A,B
	JC	FATAL
	LDA	BUFSECS
	CALL	DIOW		; Write it back out
	CALL	SHOVEL3		; Update scan pointer and get
				; pointer and return
	JZ	EXIT		; If termination return to Exec
	JMP	SHOVEL1		; Else pack some more
	;
	; Routine for # sects in undeleted area > # sects 
	; in buffer ram
	;
SHOVEL5	LDA	DDR
	MOV	C,A		; Drive # to C
	POP	PSW		; Restore # sects
	PUSH	PSW		; And save it again
	LHLD	MPTR		; Memory pointer to DE
	XCHG
	LHLD	GETPTR		; Fetch data addr on disk
	MOV	B,A
	PUSH	B		; Save # sectors
	PUSH	H		; Save for later
	MVI	B,1		; Read command
	CALL	Dio
	POP	H		; Get back disk address
	POP	B		;  and # sectors
	MOV	A,B
	JC	FATAL		; Error control for Dio
	PUSH	H
	LDA	BUFSECS		; Write whole buffer out
	CALL	DIOW
	POP	H		; Update fetch pointer
	POP	PSW		; Retrieve # sects
	MOV	C,A
	MVI	B,0		; Make a 16 bit number
	DAD	B
	SHLD	GETPTR	
	JMP	SHOVEL1		; Continue
;
; DIOW writes the number of sectors passed in the accum from
; BUFFER to the disk address pointed to by the PUTPTR and
; then updates the put pointer.
;
DIOW	MOV	B,A		; Save # sectors
	PUSH	B
	LHLD	PUTPTR		; Where to write to
	PUSH	H		; Save for now
	MOV	C,A		; Update pointer
	MVI	B,0		; MSB for DAD ( also write
				; command for later)
	DAD	B		; HL now contains new PUTPTR
	SHLD	PUTPTR		; Update
	XCHG
	LHLD	BUFST		; Get memory address
	XCHG			; Memory address to DE
	PUSH	PSW		; Save # sects
	LDA	DDR		; Get drive #
	MOV	C,A		; Drive # to C
	POP	PSW		; # sects
	POP	H		; Disk address
	PUSH	H
	CALL	Dio		; Do it.
	POP	H		; Get back disk address
	POP	B		; Get back # sectors
	MOV	A,B		; Put in accum
	JC	FATAL
	RET
;
; Routine for picking up drive number to pack.
;
GDN	LHLD	CMPTR		; Get command pointer
GDN1	MOV	A,M		; Get next char.
	LXI	D,500H		; Error code
	INX	H
	CPI	TAB		; Skip tabs
	JZ	GDN1
	CPI	' '		; and spaces
	JZ	GDN1
	CPI	CR		; If no # use default
	JNZ	GDN2		; Else jump to decode
	LDA	SYSRES		; Use sysres drive as default
	JMP	GDN4

GDN2	CPI	'#'
	JNZ	GDN3
	MOV	A,M
	LXI	D,0207H		; Error code
	CPI	CR
	JNZ	OHNO
	LDA	DEFPATH
GDN3	SUI	'0'		; Make into ASCII
	JZ	OHNO
	JC	OHNO		; If < 0 error
	CPI	10		; See if too big
	JNC	OHNO
GDN4	STA	DDR		; Save
	IF	USERS=2
	MOV	C,A
	MVI	B,0A7H
	CALL	Devlock
	JC	OHNO
	ENDIF
	RET
;
; GETDIR loads the next directory in the table into memory,
; bumps the pointer into the table, and checksums the
; directory.
;
GETDIR	LHLD	DIRTAB0		; Dir addr get pointer
	MOV	D,M		; Get the address
	DCX	H		; Put addr into DE
	MOV	E,M
	DCX	H
	SHLD	DIRTAB0		; Update dir addr get pointer
	XCHG			; Directory addr to HL
	LXI	D,SBUF1		; Where to put it
	LDA	DDR		; From which drive
	ANI	DEVMASK
	MOV	C,A
	MVI	B,1		; Read command
	MVI	A,4		; Get 4 sectors
	CALL	Dio		; Get it with error control
	JC	DIRERR
	CALL	Ckdr		; Checksum the directory
	LXI	H,SBUF1
	CMP	M		; Does it agree
	LXI	D,03FFH		; error code (assume error)
	JNZ	DIRERR
	RET			; Else all's cool, split 
;
; MOVE moves BC bytes from HL to DE
;
MOVE	MOV	A,B		; Are we done?
	ORA	C
	RZ			; If so split.
	MOV	A,M		; Else get a byte
	STAX	D		; Put a byte
	INX	H		; Bump get pointer
	INX	D		; Bump put pointer
	DCX	B		; Decrement count
	JMP	MOVE		; Do some more
;
; 16 bit compare of HL to DE.  Sets all flags
; Accum only messed up.
;
DCMP	MOV	A,H
	CMP	D
	RNZ
	MOV	A,L
	CMP	E
	RET
;
; Double subtract (sets all flags).  HL=HL-DE
;
DSUB	MOV	A,L
	SUB	E
	MOV	L,A
	MOV	A,H
	SBB	D
	MOV	H,A
	RET
;
; Double compare of BC to DE.  Sets all flags.
; Accum. only messed up
;
CmpBCDE	MOV	A,B
	CMP	D
	RNZ
	MOV	A,C
	CMP	E
	RET
;
; DIRERR is the routine which reports funny directories before
; we shuffle any data on the disk
;
DIRERR	LXI	H,DirErr	; Print first part of message
	CALL	Msg
	INX	H		; Bump to next part of message
	PUSH	H
	CALL	Deout		; Print error code
	POP	H
	CALL	Msg		; Print last half of message
	JMP	SPLIT		; Split
;
; FATAL is where we report errors encountered while doing the
; data shuffling.
; 
FATAL	PUSH	H		; Save bad disk address
	PUSH	PSW		; And number of sectors
	LXI	H,Fatal		; Point to error message
	CALL	Msg		; Display first part
	INX	H		; Bump to next part of message
	PUSH	H		; Save message pointer
	CALL	Deout		; Print Error Code
	POP	H
	CALL	Msg		; Display second part
	INX	H		; Bump to next part of message
	POP	PSW		; Get # sectors
	PUSH	H		; Save message pointer
	MOV	E,A		; Print bad sector #
	MVI	D,0
	CALL	Deout
	POP	H
	CALL	Msg		; Print third part of Msg
	INX	H		; Bump to next part of message
	POP	D		; Get error address
	PUSH	H		; Save message pointer
	CALL	Deout		; Print error address
	POP	H
	CALL	Msg		; Print period and CR
SPLIT	LHLD	ESP		; Restore stack pointer
	SPHL
	JMP	EXIT1		; And split
;
; Error bail out.  Sets error flags and bails out through
; executive.  Can be entered from any stacked environment.
;
OHNO	LHLD	ESP		; Restore stack pointer
	SPHL
	CALL	Ovrto		; Invoke Emsg
	DB	'Emsg'
	MVI	A,CR		; Terminate message
	CALL	WH1
	JMP	EXIT1		; Split back to calling routine
;
; Message storage
;
	IF	GERMAN
Fatal	DB 'Toedlicher Fehler ',0,' ist waehrend PACK '
	DB 'gefunden.',CR
	DB 'Der Fehler ist stattgefunden, als ',0
	DB ' Sektoren, vom Sektor ',0,' hinaus, uebertragen ' 
	DB 'wurden.',CR,0
DirErr	DB 'Fehler ',0,' ist beim Lesen der Verzeichnisse '
	DB 'gefunden.',CR
	DB 'Keine Daten sind uebertragen worden.',CR,0
NoWorky	DB 'Ihre Disk hat schadhafte Sektoren!  PACK ist '
	DB 'nicht moeglich.',CR,0

	ELSEIF	DANISH
Fatal	DB 'Fatal error ',0,' encountered while packing disk.'
	DB CR,'Error occured while doing ',0,' sectors'
	DB ' starting at ',0,'.',CR,0
DirErr	DB 'Error ',0,' encountered while reading directories.'
	DB CR,'No data on the disk has been altered.',CR,0
NoWorky	DB 'There is a bad sector of data on your disk.  You'
	DB ' must fix',CR,'this sector before I can pack.',CR,0	

	ELSE	;default to English
Fatal	DB 'Fatal error ',0,' encountered while packing disk.'
	DB CR,'Error occured while doing ',0,' sectors'
	DB ' starting at ',0,'.',CR,0
DirErr	DB 'Error ',0,' encountered while reading directories.'
	DB CR,'No data on the disk has been altered.',CR,0
NoWorky	DB 'There is a bad sector of data on your disk.  You'
	DB ' must fix',CR,'this sector before I can pack.',CR,0	
	ENDIF
;
; Data storage
;
ESP	DS	2	; Rock under which we save the SP
FILTAB0	DS	2	; Current undeleted file table pointer
DDR	DS	1	; Drive number
DELTAB0	DS	2	; Deleted table base
DELTAB1	DS	2	; Deleted table pointer
DIRTAB0	DS	2	; Get pointer to table of directory
			; addrs.
			; (this table grows down in memory)
DIRTAB1	DS	2	; End of table of directory addrs.
;
GUNGE	DS	2	; Used for various things
GETPTR	DS	2	; Get pointer (where data comes from)
PUTPTR	DS	2	; Put pointer (where data goes to)
MPTR	DS	2	; Memory pointer (Where to read into)
SPTR	DS	2	; Scan pointer (Where pack address are)
BUFSECS	DS	1	; How many sectors we can read at once
NLOWST	DS	2	; Next lowest entry
OLDNFA	DS	2	; Temporary storage of old NFA
NEWNFA	DS	2	; Temporary storage of new NFA
BUFST	DS	2	; Start of data buffer
TAIL			; Program end

	IF	TEST
BUFFER	EQU	TAIL
	ELSE
BUFFER	EQU	USER
	ENDIF
;
FILTAB	EQU	BUFFER	; Where to start building undeleted
			; stuff table.

	END
