;************************************************************
;*		Volume Manager Overlay			    *
;*		     version 2.5			    *
;*		      1/21/81				    *
;*							    *
;*	Original implementation by Robert T. Martin	    *
;*	Redesigned and rewritten by Frank E. Anderson	    *
;*							    *
;************************************************************
;
; Last updated:
;	12/28/79	RTM	Disaster recovery
;	1/9/80		RTM	ROM Dio support
;	10/30/80	FEA	Added support for offset, size
;				and parameter functions.  Also
;				added functions 11, 12 and 13.
;	11/11/80	FEA	Made into an overlay
;	1/12/81		FEA	Made hookup function work by
;				being passed an absolute device
;				number, rather than parameters.
;	1/19/81		FEA	Added support for checksumed
;				device definition blocks
;	1/21/81		FEA	Added relocating loader and
;				single user support
;	02/03/81	BFS	Changed error code in first
;				section of Connect to 1D0H.
;				Also changed terminology so
;				so that what used to be drives
;				are not called volumes but
;				units.  Sections of a device
;				that are connected to a unit
;				are called volumes.
;	02/24/81	BFS	Added check in Twin for other
;				user access before disconnect.
;				Returns 1C2H error if other
;				user has access.  Also added
;				check for SYSRES before
;				disconnect.
;	03/16/81	BFS	Rip out all busy checks in 
;				single user version.
;
;-------------------------------------------------------------
; Each unit control block is 13 bytes long and is defined as
; follows:
;
;	Offset	Size	Use
;	  0	  2	Busy semaphore address
;	  2	  2	Device driver address
;	  4	  2	Parm Storage Addr (0 if driver can't
;			handle dsize > 0FFFFH.
;	  6	  2	Unit size (0 if size changes on fly)
;	  8	  3	Offset from physical 00 on device
;	  B	  1	Physical drive number for I/O driver
;	  C	  1	Unit status (0 = free, FF= connected)
;
;------------------------------------------------------------
; All physical devices handled by a device driver which can
; understand disk sizes larger than 0FFFFH have a device
; definition block stored on the last sector of the physical
; device.  This block contains the following information:
;
;	Size	Value			Definition
;	 1	1 - 19		Number of logical devices
;				contained on this physical
;				device.
;
; 	* The folowing crud is repeated 1 - # devices times
;
;	 2	1 - dsize-1	Size of logical device
;	 3	0 - dsize-1	Offset for logical device
;	 8	ASCII chars	Volume name
;
;	 1	0-FFH		ones complement checksum
;				(adds to 0FFH)
;
;-------------------------------------------------------------
; The commands to the volume manager and their register setups
; are as follows:
;
;	C	unit number (1-9)
;	A	command:
;		10H - Return ucb address in HL
;
;		11H - Connect logical device
;			B - Physical Device Number
;			HL-> volume name (00 if not desired)
;
;
;		12H - Get device definition block
;			B - Pysical Device Number
;			HL-> memory address to read stuff into
;			( no volume number required )
;
;		13H - Diconnect logical unit #
;
;-------------------------------------------------------------

	REFS	SYSTEM
	REF	OVRLY
	REF	Look
	REF	Runr
	REF	Ovrto
	REF	USERS
	REF	SYSRES
	REF	DRVADTAB
	REF	VCBTAB
	REF	POS
	REF	WH1
	REF	Msg
	REF	Dio
	REF	MEMTOP
	IF	USERS=2
	REF	Mtos
	REF	excl
	REF	sett
	REF	Devlock
	ENDIF

CR	EQU	0DH

; Macros

db	MACRO
#L	DB	#A
	ENDM	db

uncursor MACRO			; Poke out brick
#L	LHLD	POS
	MVI	M,' '+80H
	ENDM	uncursor

prints	MACRO	text
#L	JMP	$+LEN[#A]+3+1
	db	#A,0
	LXI	H,$-(LEN[#A]+1)
	CALL	Msg
	uncursor
	ENDM	prints

printl	MACRO	text
#L	prints	#A,CR
	ENDM	printl

voldef	MACRO	device number, name
#L	DB	#1
	IF	NOT NULL[#2]
	DB	#2
	ENDIF
	ORG	#L+9
	ENDM	voldef

	IF	USERS=1
UNITS	EQU	7
	ELSE
UNITS	EQU	9
	ENDIF

DEVICES	EQU	20

	MACLIST	0

	PAGE
	ORG	OVRLY		; Run at this address
	IDNT	OVRLY,Vmgr

	db	'Vmgr'		; Overlays needa a name

Vmgr	LXI	D,204H		; Error code "What?" in case
	CPI	13H		; Disconnect?
	JZ	Unhook		; jmp/ yup
	CPI	12H		; Was it an Get Block?
	JZ	GetDef		; jmp /yup
	CPI	11H		; Was it a Connect?
	JZ	Connect		; jmp/ yup, doit
	CPI	10H		; Was it an Index?
	JZ	Index		; jmp/ yup, doit
	ORA	A		; Was it an init?
	STC			; Set error flag in case
	RNZ			; split if not 10H!

	PAGE
;---------------------------------------------------------
; Initialize volume manager
;
Init	LXI	H,DrvrNam	; Try to find Driver code
	LDA	SYSRES
	CALL	Look		; Does it exist?
	IF	USERS=1
	CMC			; Reverse error flag polarity
	RNC			; Split, can't find it
	ELSE
	JC	HookVCB
	ENDIF
	;
	; file exists check size and load address
	;	
	IF	USERS=1		; Save FDE
	PUSH	D
	ENDIF
	LDAX	D		; Get flag byte
	ANI	31		; Mask off flags
	ADI	5		; Add in offset for DNS	
	MOV	L,A		; Index HL to DNS
	MVI	H,0
	DAD	D
	MOV	A,M		; Get # sectors
	IF	USERS=2
	CPI	3		; more than 2 sectors?
	JNC	BadDrvr		; if so gripe.
	ELSE
	MOV	B,A
	ENDIF
	INX	H		; Point to load address 
	INX	H
	IF	USERS=1
	INX	H		; For single user try start add
	INX	H
	ENDIF
	MOV	E,M
	INX	H
	MOV	D,M
	IF	USERS=2
	LXI	H,0FA00H	; Here's where it's realy goes
	ELSE
	LXI	H,0		; Start address 0 for reloc. 
	ENDIF
	CALL	DCMP		; are they the same?
	IF	USERS=1
	POP	D		; Get back FDE pointer
	ENDIF
	JNZ	BadDrvr		; jmp/ nope it won't work
 	;
	; Load in driver file
	;
	IF	USERS=2
	LDA	SYSRES
	LXI	H,DrvrNam
	CALL	Runr		; So get it
	RC			; Report if error
	ELSE
	LHLD	MEMTOP		; Load in below MEMTOP
	MOV	A,H
	SUB	B		; by the # sectors
	MOV	H,A
	INX	H		; Put on even boundary
	XCHG			; Align registers
	PUSH	D		; Save load address
	CALL	Loader		; Load me in
	POP	H		; Retrieve load address
	RC			; Report any errors
	ENDIF
	;
	; Execute driver initialization
	;
	LXI	D,HookVCB	; Push return address on stack
	PUSH	D
	PCHL			; And execute driver
	;
	; Hook up initial unit control tables
	;
	; Start by saying that none of the units are in use
	;

HookVCB	MVI	C,1		; Squirt out vol 1's flag first
	CALL	Index		; Point me at unit 1
	LXI	D,12		; Offset to connected flag
	DAD	D		; Point at first connected flag
	INX	D		; Make so I can indx to nxt vol
ZeroVCB	MVI	M,0		; Squirt, not connected
	DAD	D		; Index to next unit
	INR	C		; Bump unit count
	MOV	A,C
	CPI	UNITS+1	; Done?
	JNZ	ZeroVCB		; jmp/ nope do next one
	;
	; Put in default values
	;
	LXI	H,VolDefTab+1	; Point to first unit
	MVI	C,1		; Tell it to hook up unit 1
VCBLoop	PUSH	B		; Save important stuff
	PUSH	H	
	MOV	B,M		; Put physical device in B
	INX	H		; Point HL at name
	MOV	A,M		; Name null?
	ORA	A
	JNZ	VCBLop1		; jmp/nope
	LXI	H,0
VCBLop1	CALL	Connect		; Call hookup function
	POP	H		; Balance stack
	POP	B
	RC			; return in case of error
	LXI	D,9		; Index to next unit
	DAD	D
	INR	C		; Bump unit #
	MOV	A,C		; Done?
	CPI	UNITS+1
	JC	VCBLoop		; if not, do some more
	;
	; Hook up initial boot unit
	;
	LDA	VolDefTab	; Get default unit #
	ORA	A		; Is it zero
	RZ			; split, don't touch
	STA	SYSRES		; Else save
	;
	; Hack to hook up sysres on other user
	;
	IF	USERS=2
	LXI	B,1		; Copy 1 byte
	LXI	H,SYSRES	;   from this users space
	MOV	D,H		;   to the same spot in the
	MOV	E,L		;   other users space
	CALL	Mtos		;   to squirt int his SYSRES
	ENDIF
	;
	RET			; Leave this mess

BadDrvr	LXI	D,01D6H		; Error code to DE
	STC
	RET

DrvrNam	db	6,'DriverDD'

	PAGE
;---------------------------------------------------------
; Return ucb address in HL. C has unit #.
;
Index	MOV	A,C		; Check to see that the 
	ORA	A		; unit number is legal
	JZ	BadVol
	CPI	UNITS+1
	JNC	BadVol
	PUSH	B		; Save these
	PUSH	D
	LXI	H,0
	MVI	B,0		; unit in C, 00 in B
	DCR	C		; unit - 1, index from 0.
	DAD	B		; start with something.
	DAD	H		; unit *2
	DAD	H		; *4
	PUSH	H
	DAD	H		; *8
	POP	D
	DAD	D		; *8 + *4 = *12
	DAD	B		; gives *13 (finally!)
	XCHG			; put into DE
	LHLD	VCBTAB
	DAD	D		; HL now points to VCB start
	POP	D		; Restore these
	POP	B
	ORA	A		; Clear carry flag
	RET			; split with stuff in HL.

BadVol	LXI	D,0101H		; Error code
	STC			; Error flag
	RET			; Split

	PAGE
;------------------------------------------------------------
; Connect is what we use to hook up a device driver to the
; volume manager.
	;
	; First we see if we already have something hooked up
	; to this logical device.  If so split with error 01D0.
	;
Connect	PUSH	H	; Save pointer to possible name
	CALL	Index	; Get me to the right block
	PUSH	D	; Save possible error code
	PUSH	PSW	; Save Error status
	LXI	D,12	; Offset to unit status
	DAD	D	; Index me
	MOV	E,M	; Get unit status
	POP	PSW	; Get back error status from call
	MOV	A,E	; Put unit status here
	POP	D	; Get back possible error
	POP	H	; HL now points to possible name
	RC		; ret/ Index returned an error
	LXI	D,01D0H	; Put error code into DE just in case
	ORA	A	; Set 0 flag according to status
	STC		; Assume error will occur so set flag
	RNZ		; Split with error if already connected
	;
	; Next check to see if he has given us a legal physical
	; device number, and if that device number has a
	; driver associated with it.
	;
	LXI	D,01D5H		; Put error code in DE in case
	MOV	A,B		; Get physical device number
	CPI	DEVICES+1	; Only 20 devices defined
	CMC			; Toggle flag for error return
	RC			; ret/ oops, device # > 20
	ORA	A		; Was device number 0?
	JZ	NoDev		; Use side door to Unhook
	PUSH	H		; Save pointer to possible name
	CALL	IndxDvr		; Point to Driver Parameters
	MOV	A,M		; Get exist flag
	ORA	A		; Device driver exist?
	POP	H		; Balance stack in case error		LXI	D,01D5H		; Put error code in DE in case
	STC			; Set error flag in case error
	RZ			; Split/ error, no driver
	;
	; Next copy in "busy" semaphore address, device driver
	; address, parameter area address and physical drive #
	;
	PUSH	H		; Save name pointer again
	CALL	IndxDvr		; Point to driver params
	XCHG			; Put into DE
	INX	D		; Point past exist flag
	CALL	Index		; Point HL at VCB entry
	PUSH	B		; Save unit and phys device
	MVI	C,6		; Copy in 6 bytes
Conn1	LDAX	D		; Get byte
	MOV	M,A		; Put byte
	INX	D		; Bump get pointer
	INX	H		; Bump put pointer
	DCR	C		; Done?
	JNZ	Conn1		; jmp/ nope, loop
	LXI	B,5		; Offset from here to drive #
	DAD	B		; Index to drive #
	POP	B		; Get back drive number
	MOV	M,B		; Save it
	LXI	B,-0BH
	DAD	B		; Get back pointer to block
	;
	; Next check parameter area address to see if we need
	; hook up offsets and volume size.
	;
	PUSH	H		; Save pointer to block
	CALL	DrvrTyp
	POP	H		; Get back pointer to block
	JNZ	Hookem		; jmp/ need to be connected
	;
	; Else offset = 0 and we let drivers handle dsize
	;
	POP	D		; Throw away possible name
	PUSH	H		; Save ucb pointer
	LXI	D,6		; Offset for volume size
	DAD	D		; Index to volume size
	MVI	A,5		; Squirt 5 zeros in
Conn2	MVI	M,0
	INX	H
	DCR	A
	JNZ	Conn2
	POP	H		; Get back ucb pointer
	JMP	Duped		; Check for duplication and go
	;
	; Else need to connect a large disk driver
	; Read disk definition block into memory
	; (Entered with HL pointing at beginning of ucb.)
	;
Hookem	PUSH	H		; Save ucb pointer
	LXI	D,ddbStor	; Where I can read this into
	CALL	GetDef1		; Read it into memory
	POP	H		; Retrieve ucb pointer
	POP	B		; and name pointer
	RC			; If error split
	PUSH	B		; name back on stack
	;
	; Check to see if named volume desired.  If so fall
	; through, if not jmp to Hookem1 and connect first
	; available volume.
	;
	POP	D		; Get pointer to name
	MOV	A,E
	ORA	D		; Zero?
	JZ	NxtFree		; jmp/ no name
	;
	; Look for named block then check to see that it hasn't
	; already been connected.  If not hook it up.  If in
	; use return with error code 01D1.  If it can't be
	; found return with error 01D2.
	;
	PUSH	H		; Save ucb pointer
	LXI	H,ddbStor	; Point HL at ddb we just got
	MOV	B,M		; # names to check
	MVI	A,6		; Offset to first name
FndName	ADD	L		; Index to next name
	MOV	L,A
	MOV	A,H
	ACI	0
	MOV	H,A
	PUSH	D		; Save name pointer
	PUSH	H		; Save ddb pointer
	MVI	C,8		; Name length to check
FndNam1	LDAX	D		; Get a character
	CMP	M		; Compare it
	JNZ	FndNam2		; jmp/ no match
	INX	H		; bump
	INX	D		; bump
	DCR	C		; Any more characters? 
	JNZ	FndNam1		; jmp/ yup, check next char 
	JMP	FndNam3		; jmp/ match!	
FndNam2	POP	H		; Retrieve ddb pointer
	POP	D		; Retrieve name pointer
	MVI	A,13		; Index to next name in ddb
	DCR	B		; Decrement # names to check
	JNZ	FndName		; jmp/ found a match
	;
	LXI	D,01D2H		; Else can't find name error
	POP	H		; Throw away ucb pointer
	STC			; Flag error
	RET			; And split
	;
	; Copy the volume size and the offset into the ucb
	;
FndNam3	POP	H		; ddb pointer
	LXI	B,-5		; offset to start of device def
	DAD	B		; Point me there
	XCHG			; Stick into DE
	POP	H		; Name pointer (throw away)
	POP	H		; Get back VCB pointer
	PUSH	H		; And save it again
	LXI	B,6		; Offset to Volume size
	DAD	B		; Index to volume size
	MVI	C,5		; # bytes to copy
FndNam4	LDAX	D		; Get byte
	MOV	M,A		; Put byte
	INX	H		; Bump
	INX	D		; Bump
	DCR	C		; Done?
	JNZ	FndNam4		; jmp/ nope, do more
	;
	; Check to see that this isn't already connected
	;
	POP	H		; New ucb to HL
	JMP	Duped		; If not duped hook up else
				; return to caller with error
	;
	; Parse definition block and compare entries to what
	; we already have defined in the volume manager.  We
	; are looking for a free logical device to hook in.
	; If there are none return with error 01D3H.
	;
NxtFree	LXI	D,ddbStor	; Point at ddb
	LDAX	D		; get # volumes
	INX	D		; point at first volume def.
	MOV	B,A		; # volumes to B
NxtFre1	PUSH	B		; Save that
	PUSH	D		; Save pointer to volume
				; definition in ddb
	PUSH	H		; Save ucb address again
	LXI	B,6		; Offset for volume size
	DAD	B		; Index ucb to volume size
	MVI	C,5		; # bytes to copy
NxtFre2	LDAX	D		; Get byte
	MOV	M,A		; Put byte
	INX	D		; Bump
	INX	H		; Bump
	DCR	C		; Done?
	JNZ	NxtFre2		; jmp/ nope, copy more
	POP	H		; Get back pointer to ucb
	PUSH	H		; and save it again
	CALL	Duped		; Is this a duplicate entry?
	POP	H		; Get back ucb pointer
	POP	D		; Get back next volume pointer 
	POP	B		; Get back # volumes count
	RNC			; If Duped connected us split
	DCR	B		; Else this is a duplicate,
				; see if any more volumes
				; available
	JZ	NonFree		; jmp/ no more volumes
				; available on this device
	PUSH	H		; Else index DE to next volume
	LXI	H,13		; definition
	DAD	D
	XCHG
	POP	H
	JMP	NxtFre1		; Try next one
NonFree	LXI	D,01D3H		; Error code for none available
	STC			; Set error flag
	RET			; Split
	;
	; Duped checks to see if we have made a duplicate
	; entry.  If we have it return with error 01D1H and
	; carry set.  If the entry is not a duplicate it
	; turns status on and returns with the carry not set.
	;
Duped	XCHG
	LHLD	VCBTAB		; Point at first ucb
	MVI	B,UNITS		; # ucbs to check
Dup	PUSH	D		; Re-save new ucb
	PUSH	H		; Save ucb being checked
	CALL	DCMP		; Are we looking at ourself
	JZ	Dup2		; jmp/ yup, no match
	MVI	C,12		; # bytes in ucb to check
Dup1	LDAX	D		; Get a byte 
	CMP	M		; Compare it
	JNZ	Dup2		; jmp/ no match
	INX	D		; Else test next byte
	INX	H
	DCR	C		; Are there any more?
	JNZ	Dup1		; jmp/ yup, check next byte
	POP	H		; DUPLICATE, error time
	POP	H		; Clean stack
	LXI	D,01D1H		; Error code
	STC			; Error flag
	RET			; Split
	;
	; Test next ucb entry
	;
Dup2	POP	H		; ucb being checked
	LXI	D,13		; Bump to next ucb
	DAD	D
	POP	D		; Get back ucb to compare to
	DCR	B		; Any more?
	JNZ	Dup		; jmp/ yup check next entry
	;
	; Else all's cool, turn on unit status and split
	;
	LXI	H,12		; Offset to unit status
				; (DE points at ucb)
	DAD	D		; Index
	MVI	M,0FFH		; Set flag, connected
	ORA	A		; Clear carry
	RET			; Split

IndxDvr	PUSH	B		; Save world
	PUSH	D
	MOV	A,B		; Get Physical device number
	DCR	A		; Make 0-19 rather than 1-20
	CPI	DEVICES		; If we were fed too much
	JNC	InxDvr1		;   reject it
	MOV	C,A		; Multiply by 7
	ADD	A		; * 2
	MOV	B,A
	ADD	A		; *4
	ADD	B		; *6
	ADD	C		; *7
	LHLD	DRVADTAB	; Where to index from
	ADD	L		; Index HL by A
	MOV	L,A
	MOV	A,H
	ACI	0
	MOV	H,A
	POP	D		; Restore world
	POP	B
	ORA	A		; Clear error flag
	RET			; Split
	;
InxDvr1	POP	D
	LXI	D,01D5H		; Error code
	POP	B
	STC			; Set error flag
	RET


NoDev	CALL	Index		; Point me in the right place
	JMP	Unhook1		; And dive right in

	PAGE	
; -------------------------------------------------------------
; Get definition block
;
	; First parse to device definition table
	;
GetDef	PUSH	H		; Save address to put data
	PUSH	B
	CALL	IndxDvr		; Point driver definition
	POP	B
	MOV	A,B
	POP	B		; Put destination addr into BC
	RC
	MOV	E,C		; Destination addr to DE
	MOV	D,B
	MOV	B,A		; Device # to B
	JMP	GetDef0
	;
	; Next see if this device has a definition block
	; (Parameter area address non-zero).  If not split
	; with error 01D4H.
	;
	; NOTE: This is also the internal entry point with
	; HL pointing at the VCB and DE pointing at the
	; destination address in memory.
	;
GetDef1	PUSH	D		; Save memory address
	LXI	D,11		; Offset to device
	DAD	D
	MOV	B,M		; Get phyical device
	PUSH	B
	CALL	IndxDvr		; Point HL at device def
	POP	B
	POP	D
	;
GetDef0	MOV	A,M		; Device exist
	ORA	A
	JZ	Err01D5
	PUSH	D		; Save destination address
	INX	H
	PUSH	H		; Save device def address
	CALL	DrvrTyp
	POP	H		; Restore def address
	LXI	D,01D4H		; Assume error so set up error
	STC			;    code and flag
	MOV	A,B
	POP	B		; Clean stack in case of exit
	RZ			; Split if error
	PUSH	B		; re-save memory address
	;
	; Get physical device #
	;
	MOV	C,A
	;
	; Find Dsize of that device
	;
	PUSH	H		; Save def address
	MVI	B,5		; Dsize command
	PUSH	B
	CALL	DoDrivr
	SHLD	AddrBuf
	POP	B
	POP	H		; def address to HL
	MOV	A,C		; Device #
	POP	B		; Memory address
	RC			; RET/ Argh, it's broken
	;
	; Read in last sector
	;
	PUSH	B		; Save memory address again
	PUSH	PSW		; Save device number
	PUSH	H		; Save dev drv pointer
	LXI	D,4		; Offset to parm storage area
	DAD	D		; Index me to parm storage area
	MOV	E,M		; Parm address to DE
	INX	H
	MOV	D,M
	LHLD	AddrBuf		; Copy from here
	PUSH	D		; Save parm addr for later
	MVI	B,3		; Copy three bytes
GetDef2	MOV	A,M		;  from the table
	STAX	D		;  into the parameter area
	INX	H
	INX	D
	DCR	B
	JNZ	GetDef2
	POP	H		; Get pointer to parm addr
GetDef3	DCR	M		; Decrement dsize
	INX	H
	JC	GetDef3		; If no carry we're done
	POP	H		; HL-> dev def
	POP	PSW
	MOV	C,A		; Put into C
	POP	D		; Get memory address back
	MVI	A,1		; Only do 1 sector
	MVI	B,1		; Read command
	PUSH	D		; Save memory address
	CALL	DoDrivr		; Get the sector
	POP	H		; Retrieve memory address
	RC			; If dio error split
	;
	; Verify that what we got was a ddb
	;
	XRA	A		; Initialize checksum
	MOV	B,A		; And loop count
	PUSH	H		; Save memory address
Cksm	ADD	M		; Calculate checksum
	INX	H		; Bump pointer
	DCR	B		; decrement 256 byte counter
	JNZ	Cksm		; jmp/ not done yet
	INR	A		; Result 0FFH?
	POP	H		; Get back ddb start
	LXI	D,01D6H		; Error code in case bomb-out
	STC			; Error flag "   "    "
	RNZ
	MOV	A,M		; Check # entries
	ORA	A		; 0?
	STC			; Set error flag in case
	RZ			; Can't have ddb with 0 entries
	CPI	20		;  or more than 19 entries
	CMC			; Make error flag polarity OK
	RET			; Split (error flag set)

Err01D5	LXI	D,01D5H
	STC
	RET

	PAGE	
;--------------------------------------------------------------
; Disconnect a logical unit number.
;
	; First index to correct VCB flag
	;
Unhook	CALL	Index		; Point to correct block
	RC			; Error control
	;
	; Check to see if device is connected
	;
	PUSH	H		; Save pointer
	LXI	D,13		; Offset to Status
	DAD	D		; Index me
	MOV	A,M		; Get Status
	ORA	A		; Already off?
	POP	H		; Get back pointer
	JZ	Unhook1		; If off, skip busy check
	LDA	SYSRES
	CMP	C
	JNZ	Unh1
	LXI	D,111H		; Not to SYSRES you don't.
	STC
	RET

Unh1
	IF	USERS=2
	;
	;	Now see if other user has any access to device.
	;  If not, clear all our allocations.  Device number is
	;  still in C.
	;
	sett	excl
	RC			; Error control.
	PUSH	B
	PUSH	H
	MVI	L,6
	MVI	B,0C7H		; Clear permanent allocation.
Clear1	CALL	Devlock
	DCR	B		; Do next bit.
	DCR	L
	JNZ	Clear1
	POP	H
	POP	B

	;
	; Check busy semaphore to see if device is being used.
	; If so return with error 01C2H.
	;
	MOV	E,M		; Address of busy flag to DE
	INX	H
	MOV	D,M
	DCX	H		; Point back at beginning
	LDAX	D		; Get busy flag
	ORA	A		; Test it
	LXI	D,01C2H		; Error code in case
	STC			; Error flag in case
	RNZ			; Ret/ busy right now
	ENDIF

	;
	; Else disconnect by squirting 00 into unit status
	; and split.
	;
Unhook1	MVI	A,13		; Offset from here to status
Unhook2	MVI	M,0		; Squirt, I'm off
	INX	H		; Bump
	DCR	A		; Done?
	JNZ	Unhook2		; jmp/ nope
	ORA	A		; Clear error flag (carry)
	RET			; Split

	PAGE
	IF	USERS=1
;********************************************************
;*							*
;*		   Relocating Loader			*
;*							*
;*	Version 0.12			Lennie Araki	*
;*	01/15/80			Brian Smith	*
;*							*
;*	Copyright (C), 1979 Interactive Products	*
;*  Corporation dba PolyMorphic Systems, 460 Ward Dr.	*
;*  Santa Barbara CA 93111 (805) 967-2351.		*
;*							*
;********************************************************
;
;	Loader loads file whose description block is pointed
;  to by HL at address pointed to by DE.
;
; NOTE: This is a special version of the loader which only
; loads from the base directory on SYSRES.  It is also
; passed the pointer to the FDE entry in the dirctory in
; the DE registers.
;
Loader	XCHG			; Save memory address
	SHLD	Prog_Start
	XCHG
	CALL	GetVals
	MOV	A,B
	ORA	A
	JNZ	TooBig
	SHLD	DiskAdd
	XCHG
	SHLD	FilLA
	MOV	A,C
	STA	Nsect
NEWENT	LHLD	Prog_Start
	XCHG			; Memory address in DE.
	LDA	SYSRES
	MOV	C,A		; Drive number
	LHLD	DiskAdd 	; Disk address
	LDA	Nsect		; Number of sect.
	MVI	B,1		; Read
	CALL	Dio
	JC	LdErr
	;
	;	Relocate the file
	;
	LHLD	Prog_Start	; Assume it's assembled
	MOV	B,H		; at address 0
	MOV	C,L
	XCHG
	LHLD	FilLA		; Load address points
	DAD	D		; at relocation block.
	CALL	RELOCR		; Do it!
	ORA	A
	RET
;
; RELOCR
;	Relocate a file loaded into memory at DE
;  with HL pointing at the beginning of the relocation
;  block.  Offset is in BC.
;
RELOCR	SHLD	RELBLK	 	;Save HL in RELBLK.
	MOV	H,B
	MOV	L,C
	SHLD	RELOFF
	LHLD	RELBLK
	XCHG
	JMP	Reloc3		;Jmp get into the loop!
	;
Reloc1	INX	H		;Main loop
	INX	H		;Bump HL two.
	PUSH	D
	XCHG
	LHLD	RELBLK		; Get end of code pointer.
	XCHG
	CALL	DCMP		; Present pointer - end.
	POP	D
	RNC			;Ret/ all done, no carry.
	;
Reloc2	DCR	C		;Dink bit count
	MOV	A,B		;Get bit pattern in A
	JNZ	Reloc4		;Jmp/ some bits left
Reloc3	LDAX	D		; else get a new byte
	INX	D
	MVI	C,8		; and reset bit count
Reloc4	RAL
	MOV	B,A		;Put bits back into B.
	JNC	Reloc1		;Jmp/ 0: just skip two
	;
	DCR	C		;Dink bit count
	MOV	A,B		;Get bit pattern in A
	JNZ	Reloc5		;Jmp/ some bits left
	LDAX	D		; else get a new byte
	INX	D
	MVI	C,8		; and reset bit count
Reloc5	RAL
	MOV	B,A		;Put bits back into B.
	JNC	Reloc6		;Jmp/ 0: relocate even
	INX	H		; else relocate odd.
Reloc6	PUSH	B
	PUSH	D
	MOV	C,M
	INX	H
	MOV	B,M		;Get old addr in BC
	XCHG
	LHLD	RELOFF		;Offset to add.
	DAD	B		;Add the two together.
	XCHG
	MOV	M,D
	DCX	H
	MOV	M,E		;Stick back in memory.
	POP	D
	POP	B
	JMP	Reloc1		;Back to main loop.
;
;	Error bailout.  Returns CARRY.
;
LdErr	CALL	Ovrto
	db	'Emsg'
	MVI	A,CR
	CALL	WH1
	STC
	RET
;
	IF	GERMAN
TooBig	printl	'File passt nicht in den Speicher!'
	ELSE
TooBig	printl	'File won''t fit in memory!'
	ENDIF
	STC
	RET
;
;	Extract disk address, file length and load address
;  from FDE pointed to by HL.  Return HL = disk address, DE =
;  load address, and BC = file length.
;
GetVals	MOV	A,M
	ANI	31
	ADI	3
	MOV	E,A
	MVI	D,0
	DAD	D
	MOV	E,M
	INX	H
	MOV	D,M
	INX	H
	PUSH	D	 	; Disk addr on stack.
	MOV	E,M
	INX	H
	MOV	D,M
	INX	H
	PUSH	D		; File length on stack.
	MOV	E,M
	INX	H
	MOV	D,M		; Load address in DE.
	POP	B		; File length in BC.
	POP	H		; Disk address in HL.
	RET
;
FILL	PUSH	D
	MOV	E,A
FILL0	MOV	A,B
	ORA	C
	JZ	FILL2
FILL1	MOV	M,E
	INX	H
	DCX	B
	JMP	FILL0
FILL2	MOV	A,E
	POP	D
	RET
;
;	Data section
;
Prog_Start	DS	2
TXPTR		DS	2
DiskAdd		DS	2
Nsect		DS	1
FilLA		DS	2
RELBLK		DS	2
RELOFF		DS	2
BLOCK		DS	44

	ENDIF
	PAGE
;------------------------------------------------------------
;			UTILITIES
;------------------------------------------------------------
; DrvrTyp returns with zero flag set if the device driver can
; only handle dsize of 0-0FFFFH or with the zero flag not set
; if it can handle bigger ones.  It expects HL to point at the
; beginning of the ucb and returns with it intact.
;
DrvrTyp	LXI	D,4		; Offset for parm area addr
	DAD	D		; Index
	MOV	A,M		; Get first byte
	INX	H		; Point to second byte
	ORA	M		; Test, are they both 0?
	RET
;
;-------------------------------------------------------------
; I/O driver.  This code executes and I/O using the parameters
; stored in the ucb pointed to by HL.  Disk address is pointed
; to by Parm Storage Addr in the ucb.  The physical device (not
; the logical device) is in C.  All other registers correspond
; to standart DIO setup.
;
DoDrivr	PUSH	D		; Save memory addr
	INX	H		; Index to Device driver addr.
	INX	H
	MOV	E,M		; Load driver address into DE
	INX	H
	MOV	D,M
	XCHG			; HL -> driver address
	XTHL			; Save on stack
	PUSH	H		; Save memory addr again
	XCHG
	INX	H		; Point HL at parm storage addr
	MOV	E,M		; Get LSB of Parm Storage addr.
	INX	H
	MOV	D,M		; Get MSB of Parm Storage addr.
	XCHG			; Move to HL
	POP	D		; Put memory address back in DE
	RET			; Execute driver

	IF	USERS=1
;
; Subtracts DE from HL, destroys accum, zero flag not set.
;
DSUB	MOV	A,L
	SUB	E
	MOV	L,A
	MOV	A,H
	SBB	D
	MOV	H,A
	RET

	ENDIF
;
; DCMP is a double compare of HL to DE with HL acting as the
; accumulator.  All registers preserved except A.
;
DCMP	MOV	A,H
	CMP	D
	RNZ
	MOV	A,L
	CMP	E
	RET

	IF	USERS=1
;
;	Move BC bytes from HL to DE.
;
MOVE	MOV	A,B		; Check count
	ORA	C		; Zero?
	RZ			; ret/ done
	MOV	A,M		; Else get byte
	STAX	D		; Put byte
	INX	H		; Bump get ptr
	INX	D		; Bump put ptr
	DCX	B		; Bump count
	JMP	MOVE		; and loop

	ENDIF
;
AddrBuf	DS	2

	PAGE
;-----------------------------------------------------------
; Here are the default unit hookups which are used by the
; initialize function to set up the VCB table.
;
; NOTE: This area must be aligned on a 256 byte boundry so
; that the Unit Setup program can easily modify it.
;
; The table is organized as follows:
;
;	Size	Offset	Definition
;	  1	  0	Boot Unit (if 0 use SYSRES)
;	  1	  1	Unit 1 Physical Device Number
;	  8	  2	Unit 1 Name
;	  1	  10	Unit 2 Physical Device Number
;	  8	  11	Unit 2 Name
;	....
;	  1	  73	Unit 9 Physical Device Number
;	  8	  74	Unit 9 Name	

	ORG	(($+256)/256)*256  ; Put me on an even boundry

; Here are the default value we ship the system with
;
VolDefTab

BootVol	db	0

Vol1Def	voldef	1	; Units 1-3 are DD
Vol2Def	voldef	2
Vol3Def	voldef	3
Vol4Def	voldef	4	; Units 4-7 are MS
Vol5Def	voldef	5
Vol6Def	voldef	6
Vol7Def	voldef	7

	IF	USERS=2

Vol8Def	voldef	0	; Units 8-9 are undefined for now
Vol9Def	voldef	0

	ENDIF

; This is where we store the device definition block when we
; are doing a connect device function.
;
ddbStor	

	END
